{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpE00.pas. The Original Code is available through the  }
{ home page of DSpatial at http://dspatial.sourceforge.net.                    }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit imports spatial data from ArcInfo E00 format files. The unit       }
{ contains both the driver and the broker. Currently only grid coverages are   }
{ supported, in a single file. By default, the raster bands are created with   }
{ the dspFloat32 or dspInt32 data type, depending on the grid coverage data    }
{ type. Only GRID and INFO sections are currently processed, although the unit }
{ contains stubs for LOG and PRJ sections as well.                             }
{                                                                              }
{ The E00 format is a very inefficient format to read, because it is plain     }
{ ASCII. Due to this and other peculiarities of the format, the file is read in}
{ its entirity when the driver is created, the sections parsed into a binary   }
{ representation, and the file closed. This creates data duplication between   }
{ the driver and the data storage at the kernel level, so you are advised to   }
{ use this driver exclusively for importing data and immediately saving it in a}
{ more appropriate format.                                                     }
{                                                                              }
{ E00 files can contain things other than a supported format (grids only at the}
{ moment). These could be INFO files, a 1039 drawing, etc. Currently you will  }
{ get an empty raster object with no bands or other information. Since this is }
{ a read-only format the solution is adequate, but an end-user may raise his/  }
{ her eyebrows when the empty raster object is returned.                       }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 03                                             }
{                                                                              }
{ Dependencies  : DSpPFA                                                       }
{                 DSpUtils (FloatsSame)                                        }
{                 DSpGlobals                                                   }
{                 DB (TFieldDef)                                               }
{                 DBClient (TClientDataSet)                                    }
{                 SysUtils (Exception, LastDelimiter)                          }
{                                                                              }
{ Version       : 0.4  03Dec03 Support for saving data added. SupportsSource   }
{                              and CanSave added to the broker.                }
{                 0.3  03Nov25 Parsing by section and internal storage.        }
{                 0.2  03Nov07 PFA overhaul                                    }
{                 0.1  03Jul20 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpE00;
{$I DSpatial.inc}
interface
uses DBClient, DSpGlobals, DSpPFA;
type
  TDSpE00Grid = record
                  Rows, Cols : integer;
                  Pixel      : TDSpFloat64;
                  Data       : pointer;
                end;

  TDSpE00Info = record
                  Names      : array of string;
                  Data       : array of TClientDataSet;
                end;

  TDSpE00Driver = class(TDSpBaseDriver)
  private
    FFile: TextFile;
    FFileName: string;
    FLoc: string;
    FPrecision: TDSpDataType;
    FNull: TDSpFloat64;
    FNorth, FEast, FSouth, FWest: TDSpFloat64;
    FGrid: ^TDSpE00Grid;
    FInfo: TDSpE00Info;

    procedure   ReadGRD;
    procedure   ReadLOG;
    procedure   ReadIFO;
    procedure   ReadPRJ;
  protected
    property    FileName: string read FFileName write FFileName;
    function    Description: string; override;
    function    DataSets: integer; override;

    function    DataSetID(i: integer): integer; override;
    function    DataSetDescription(i: integer): string; override;
    function    DataSetCoordSys(i: integer): string; override;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); override;
    function    DataSetType(i: integer): TDSpPFAFormat; override;
  public
    constructor Open(const fn: string);
    destructor  Destroy; override;

    function    CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer; override;
    function    AddDataSet(const ds: TDSpSourceDataSet): integer; override;
    function    CloneDataSet(const ds: TDSpSourceDataSet): integer; override;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   AddRasterBand(i: integer; info: PDSpRasterbandInfo); override;
    procedure   DeleteRasterBand(i: integer); override;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); override;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); override;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); override;
  end;

  TDSpE00Broker = class(TDSpBaseBroker)
  public
    constructor Create;
    function    CreateSource(const fn: string): TDSpSourceHandler; override;
    function    OpenSource(const fn: string): TDSpSourceHandler; override;
    function    SupportsSource(const source: string): boolean; override;
    function    CanSave(const ds: TDSpSourceDataSet): string; override;
  end;

implementation
uses SysUtils, DSpUtils, DB;

////////////////////////////////////////////////////////////////////////////////
// TDSpE00Reader
constructor TDSpE00Driver.Open(const fn: string);
var num: integer;
    section: string[3];
begin
  inherited Create;

  // Initializing variables
  FGrid := nil;

  // Read the file
  FFileName := fn;
  AssignFile(FFile, FFileName);
  try
    Reset(FFile);
    ReadLn(FFile, section, num, FLoc);
    if (section <> 'EXP') or (num <> 0) then raise Exception.Create('');

    // Read individual sections
    repeat
      ReadLn(FFile, section);
      if section = 'GRD' then ReadGRD
      else if section = 'LOG' then ReadLOG  // not actually read
      else if section = 'IFO' then ReadIFO
      else if section = 'PRJ' then ReadPRJ  // not actually read
      else if section = 'EOS' then Break;   // end-of-file marker
    until Eof(FFile);
  finally
    CloseFile(FFile);
  end;
end;

destructor TDSpE00Driver.Destroy;
var i: integer;
begin
  if FGrid <> nil then begin
    if FGrid^.Data <> nil then FreeMem(FGrid^.Data);
    FreeMem(FGrid);
  end;
  if Length(FInfo.Names) > 0 then begin
    for i := 0 to High(FInfo.Names) do
      FInfo.Data[i].Free;
    Finalize(FInfo.Names);
    Finalize(FInfo.Data);
  end;
  inherited;
end;

procedure TDSpE00Driver.ReadGRD;
var linesPerRow, leftOver, row, col, num: integer;
    flt: array [1..5] of TDSpFloat32;
    int: array [1..5] of TDSpInt32;
    p: pointer;
    dx, xmax, ymax: TDSpFloat64;
begin
  New(FGrid);
  FGrid^.Data := nil;

  ReadLn(FFile, FGrid^.Cols, FGrid^.Rows, num, FNull);
  case num of
    1: FPrecision := dspInt32;
    2: FPrecision := dspFloat32;
  else raise Exception.Create('E00 file has an invalid GRID data type');
  end;

  if (FGrid^.Cols = 0) or (FGrid^.Rows = 0) then
    raise Exception.Create('E00 file has an invalid GRID dimension');
  ReadLn(FFile, dx, FGrid^.Pixel);
  if dx <> FGrid^.Pixel then
    raise Exception.Create('GRID must have equal x and y resolutions.');
  ReadLn(FFile, FWest, FSouth);
  ReadLn(FFile, xmax, ymax);
  if not (FloatsSame(FWest + dx * FGrid^.Cols, xmax) and FloatsSame(FSouth + dx * FGrid^.Rows, ymax)) then
    raise Exception.Create('Invalid GRID extent.');
  FEast := FWest + dx * FGrid^.Cols;
  FNorth := FSouth + dx * FGrid^.Rows;

  FGrid^.Data := AllocMem(FGrid^.Cols * FGrid^.Rows * 4);
  linesPerRow := FGrid^.Cols div 5;
  leftOver := FGrid^.Cols mod 5;
  p := FGrid^.Data;
  for row := 1 to FGrid^.Rows do begin
    for col := 1 to linesPerRow do begin
      case FPrecision of
        dspFloat32: begin
          ReadLn(FFile, flt[1], flt[2], flt[3], flt[4], flt[5]);
          Move(flt, p^, 20);
        end;
        dspInt32: begin
          ReadLn(FFile, int[1], int[2], int[3], int[4], int[5]);
          Move(int, p^, 20);
        end;
      end;
      p := pointer(cardinal(p) + 20);
    end;
    if leftOver > 0 then begin
      case FPrecision of
        dspFloat32: begin
          ReadLn(FFile, flt[1], flt[2], flt[3], flt[4], flt[5]);
          Move(flt, p^, 4 * leftOver);
        end;
        dspInt32: begin
          ReadLn(FFile, int[1], int[2], int[3], int[4], int[5]);
          Move(int, p^, 4 * leftOver);
        end;
      end;
      p := pointer(integer(p) + 4 * leftOver);
    end;
  end;
  ReadLn(FFile); // skip the EOG line
end;

procedure TDSpE00Driver.ReadLOG;
var section: string[3];
begin
  // Sink the info for now
  repeat
    ReadLn(FFile, section);
  until section = 'EOL';
end;

procedure TDSpE00Driver.ReadIFO;
type
  FileInfo = packed record
               Name   : packed array [0..31] of char;
               IntExt : packed array [0..1] of char;
               Items  : packed array [0..3] of char;
               Again  : packed array [0..3] of char;
               RecLen : packed array [0..3] of char;
               Recs   : packed array [0..9] of char;
             end;
  ItemInfo = packed record
               Name   : packed array [0..15] of char;
               Len    : packed array [0..2] of char;
               Min1   : packed array [0..1] of char;
               Start  : packed array [0..3] of char;
               Four1  : packed array [0..2] of char;
               Num    : packed array [0..3] of char;
               Dec    : packed array [0..1] of char;
               Space  : char;
               DType  : packed array [0..1] of char;
             end;
  ItemDesc = record
               Name : string;
               DType: TFieldType;
               Len  : integer;
             end;

var line, recLine: string;
    i, r, ndx, num, numItems, recLen: integer;
    infoFile: FileInfo;
    infoItem: ItemInfo;
    fd: TFieldDef;
    items: array of ItemDesc;
begin
  repeat
    // Read the starting line of the INFO file, update structures
    ReadLn(FFile, line);
    if line = 'EOI' then Break;
    ndx := Length(FInfo.Names);
    SetLength(FInfo.Names, ndx + 1);
    SetLength(FInfo.Data, ndx + 1);
    FInfo.Data[ndx] := TClientDataSet.Create(nil);
    Move(line[1], infoFile, SizeOf(FileInfo));
    FInfo.Names[ndx] := Trim(infoFile.Name);

    // Read the item definitions, populate the client data set with fields
    recLen := 0;
    numItems := StrToInt(infoFile.Items);
    SetLength(items, numItems);
    if numItems > 0 then
      for i := 0 to numItems - 1 do begin
        ReadLn(FFile, line);
        Move(line[1], infoItem, SizeOf(ItemInfo));
        fd := FInfo.Data[ndx].FieldDefs.AddFieldDef;
        items[i].Name := Trim(infoItem.Name);
        fd.Name := items[i].Name;
        num := StrToInt(infoItem.DType);
        items[i].Len := StrToInt(infoItem.Len);
        case num of
          20: begin
                fd.DataType := ftString;
                fd.Size := items[i].Len;
                Inc(recLen, items[i].Len);
              end;
          50: begin
                fd.DataType := ftInteger;
                Inc(reclen, 11);
              end;
          60: begin
                fd.DataType := ftFloat;
                if items[i].Len = 4 then Inc(recLen, 14) else Inc(recLen, 24);
              end;
        end;
        items[i].DType := fd.DataType;
      end;
    FInfo.Data[ndx].CreateDataSet;
    FInfo.Data[ndx].Open;

    // Read the data and put it in the client data set
    for r := 0 to StrToInt(infoFile.Recs) - 1 do begin
      // First read in the entire record in ASCII
      recLine := '';
      i := 0;
      repeat
        ReadLn(FFile, line);        // A line may be shorter than 80 chars when
        Inc(i, 80);                 // trailing blanks are removed, but those
        recLine := recLine + line;  // deleted spaces still need to be counted
        if Length(line) < 80 then   // and they must be reinserted in recLine.
          recLine := recLine + StringOfChar(' ', 80 - Length(line));
      until i >= recLen;

      // Convert items into their binary representation
      FInfo.Data[ndx].Append; // Add a new record, which becomes active
      for i := 0 to numItems - 1 do begin
        case items[i].DType of
          ftString : begin
                       FInfo.Data[ndx].FieldValues[items[i].Name] := Copy(recLine, 1, items[i].Len);
                       recLine := Copy(recLine, items[i].Len+1, 100000);
                     end;
          ftInteger: begin
                       FInfo.Data[ndx].FieldValues[items[i].Name] := StrToInt(Copy(recLine, 1, 11));
                       recLine := Copy(recLine, 12, 100000);
                     end;
          ftFloat:   begin
                       if items[i].Len = 4 then num := 14 else num := 24;
                       FInfo.Data[ndx].FieldValues[items[i].Name] := StrToFloat(Copy(recLine, 1, num));
                       recLine := Copy(recLine, num+1, 100000);
                     end;
        end;
      end;
      FInfo.Data[ndx].Post;
    end;
    FInfo.Data[ndx].Close;
  until false;
end;

procedure TDSpE00Driver.ReadPRJ;
var section: string[3];
begin
  // Sink the info for now
  repeat
    ReadLn(FFile, section);
  until section = 'EOP';
end;

function TDSpE00Driver.CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer;
begin
  Assert(false, 'E00 driver does not support writing new data sets');
  Result := -1;
end;

function TDSpE00Driver.AddDataSet(const ds: TDSpSourceDataSet): integer;
begin
  Assert(false, 'E00 driver does not support writing new data sets');
  Result := -1;
end;

function TDSpE00Driver.CloneDataSet(const ds: TDSpSourceDataSet): integer;
begin
  Assert(false, 'E00 driver does not support writing new data sets');
  Result := -1;
end;

procedure TDSpE00Driver.AddRasterBand(i: integer; info: PDSpRasterbandInfo);
begin
  info^.BandIndex := -1;
end;

procedure TDSpE00Driver.DeleteRasterBand(i: integer);
begin
  Assert(false, 'E00 driver does not support writing data sets');
end;

function TDSpE00Driver.Description: string;
begin
  Result := 'ArcInfo export file: ' + FFileName;
end;

function TDSpE00Driver.DataSets: integer;
begin
  Result := 1;
end;

function TDSpE00Driver.DataSetID(i: integer): integer;
begin
  Result := 0;
end;

function TDSpE00Driver.DataSetDescription(i: integer): string;
begin
  Result := Trim(Copy(FLoc, LastDelimiter('/\', FLoc) + 1, 250));
end;

function TDSpE00Driver.DataSetCoordSys(i: integer): string;
begin
  Result := 'LOCAL_CS["Local coordinate system (meter)",LOCAL_DATUM["Local datum",0],UNIT["Meter",1.0],AXIS["X",EAST],AXIS["Y",NORTH]]';  // <<<<<<<<<<<< projection not yet supported
end;

procedure TDSpE00Driver.DataSetExtent(i: integer; const ext: TDSpExtent);
begin
  ext.North := FNorth;
  ext.East := FEast;
  ext.South := FSouth;
  ext.West := FWest;
end;

function TDSpE00Driver.DataSetType(i: integer): TDSpPFAFormat;
begin
  Result := dspPFARaster;
end;

procedure TDSpE00Driver.GetRasterInfo(i: integer; info: PDSpRasterInfo);
begin
  if FGrid = nil then info^.Bands := 0
  else begin
    info^.BandFormat := dspRasterFlat;
    info^.Bands := 1;
    info^.Rows := FGrid^.Rows;
    info^.Columns := FGrid^.Cols;
    info^.PixelSize := FGrid^.Pixel;
  end;
end;

procedure TDSpE00Driver.GetRasterBandInfo(i: integer; info: PDSpRasterInfo);
var ndx: integer;
begin
  Assert(FGrid <> nil);
  with info^.Band[0] do begin
    BandIndex := 0;
    DataType := FPrecision;
    BandType := dspRasterBandSurface;
    MissingValue := FNull;
    Min := 0;       // Default values if no STA file found
    Max := -1;
    if Length(FInfo.Names) > 0 then
      for ndx := 0 to High(FInfo.Names) do
        if Copy(FInfo.Names[ndx], Pos('.', FInfo.Names[ndx])+1, 100) = 'STA' then begin
          FInfo.Data[ndx].Open;
          FInfo.Data[ndx].First;
          Min := FInfo.Data[ndx].FieldValues['MIN'];
          Max := FInfo.Data[ndx].FieldValues['MAX'];
          FInfo.Data[ndx].Close;
        end;
  end;
end;

procedure TDSpE00Driver.ReadRasterBand(dataSet, band: integer; const into: pointer);
begin
  Assert((dataSet = 0) and (band = 0) and (FGrid <> nil) and (FGrid^.Data <> nil));
  Move(FGrid^.Data^, into^, FGrid^.Rows * FGrid^.Cols * 4);
end;

procedure TDSpE00Driver.ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer);
begin
  Assert(FGrid <> nil);
  Assert(false, 'not yet implemented');
end;

procedure TDSpE00Driver.ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer);
begin
  Assert(FGrid <> nil);
  Assert(false, 'not yet implemented');
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpE00Broker
constructor TDSpE00Broker.Create;
begin
  inherited;
  FDescription := 'ArcInfo export files';
  FFilter := '*.E00';
  FCaps := [dspReadRaster];
end;

function TDSpE00Broker.CreateSource(const fn: string): TDSpSourceHandler;
begin
  Result := nil;
  Assert(false, 'E00 is a read-only format');
end;

function TDSpE00Broker.OpenSource(const fn: string): TDSpSourceHandler;
var r: TDSpE00Driver;
begin
  r := TDSpE00Driver.Open(fn);
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

function TDSpE00Broker.SupportsSource(const source: string): boolean;
begin
  Result := UpperCase(ExtractFileExt(source)) = '.E00';
end;

function TDSpE00Broker.CanSave(const ds: TDSpSourceDataSet): string;
begin
  Result := '';
end;

initialization
{ Enable this method for static loading of the broker. Please note that the    }
{ DSpPFA unit MUST be loaded before this unit. If the broker is to be loaded   }
{ dynamically, you will have to provide a DLL with this unit.                  }
  PFA.RegisterBroker(TDSpE00Broker.Create);
end.
