{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpHDFDriver.pas. The Original Code is available through}
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit provides a driver and reader for the HDF format for use with the   }
{ DSpatial PFA. Only SD data sets can be accessed at this time. In the default }
{ implementation up to three dimensions of data can be used, with the highest  }
{ dimension being mapped to the x (east-west) coordinate, the second highest to}
{ the y (north-south) coordinate, and the third highest to the raster band.    }
{ Other dimensions are ignored and the lowest index is used. Specific layouts  }
{ of data (for instance, with time as a fourth dimension) require a specific   }
{ driver (for which this driver can be used as a template).                    }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 03                                             }
{                                                                              }
{ Dependencies  : DSpHDFObject                                                 }
{                 DSpHDFTypes                                                  }
{                 DSpPFA (register driver with the PFA)                        }
{                 DSpGlobals                                                   }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.2  03Dec03 Support for saving data added. SupportsSource   }
{                              and CanSave added to the broker.                }
{                 0.1  03Jul21 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpHDFDriver;
{$I DSpatial.inc}
interface
uses Classes, DSpGlobals, DSpPFA, DSpHDFObject;
type
  TDSpHDFSDDriver = class(TDSpBaseDriver)
  private
    FFile: THDFFile;
    FSD  : THDF_SD;
  protected
    function    GetFileName: string;

    property    FileName: string read GetFileName;
    function    Description: string; override;
    function    DataSets: integer; override;

    function    DataSetID(i: integer): integer; override;
    function    DataSetDescription(i: integer): string; override;
    function    DataSetCoordSys(i: integer): string; override;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); override;
    function    DataSetType(i: integer): TDSpPFAFormat; override;
  public
    constructor Open(const fn: string);
    destructor  Destroy; override;

    function    CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer; override;
    function    AddDataSet(const ds: TDSpSourceDataSet): integer; override;
    function    CloneDataSet(const ds: TDSpSourceDataSet): integer; override;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   AddRasterBand(i: integer; info: PDSpRasterbandInfo); override;
    procedure   DeleteRasterBand(i: integer); override;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); override;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); override;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); override;
  end;

  TDSpHDFSDBroker = class(TDSpBaseBroker)
  public
    constructor Create;
    destructor  Destroy; override;

    function    CreateSource(const fn: string): TDSpSourceHandler; override;
    function    OpenSource(const fn: string): TDSpSourceHandler; override;
    function    SupportsSource(const source: string): boolean; override;
    function    CanSave(const ds: TDSpSourceDataSet): string; override;
  end;

implementation
uses SysUtils, DSpUtils, DSpHDFTypes;

////////////////////////////////////////////////////////////////////////////////
// TDSpHDFSDReader
constructor TDSpHDFSDDriver.Open(const fn: string);
begin
  inherited Create;
  FFile := THDFFile.Create(fn);
  FSD := FFile.OpenSD;
end;

destructor TDSpHDFSDDriver.Destroy;
begin
  FSD.Free;
  FFile.Free;
  inherited;
end;

function TDSpHDFSDDriver.GetFileName: string;
begin
  if Assigned(FFile) then Result := FFile.FileName
  else Result := ''
end;

function TDSpHDFSDDriver.Description: string;
begin
  Result := Format('HDF file %s', [FFile.FileName]);
end;

function TDSpHDFSDDriver.DataSets: integer;
begin
  Result := FSD.NumDataSets;
end;

function TDSpHDFSDDriver.DataSetID(i: integer): integer;
begin
  Result := FSD.SDS[i].ID;
end;

function TDSpHDFSDDriver.DataSetDescription(i: integer): string;
begin
  Result := FSD.SDS[i].Name;
end;

function TDSpHDFSDDriver.DataSetCoordSys(i: integer): string;
begin
  Result := 'LOCAL_CS["Local coordinate system (meter)",LOCAL_DATUM["Local datum",0],UNIT["Meter",1.0],AXIS["X",EAST],AXIS["Y",NORTH]]';  // <<<<<<<<<<<< projection not yet supported
end;

procedure TDSpHDFSDDriver.DataSetExtent(i: integer; const ext: TDSpExtent);
var sds: THDF_SDS;
    rank: integer;
begin
  Assert(Assigned(FSD));
  sds := FSD.SDS[i];
  rank := sds.Rank;
  ext.West := 0;                         // work in pixel space
  ext.East := sds.Dimension[rank] - 1;
  ext.South := 0;
  if rank > 1 then ext.North := sds.Dimension[rank-1] - 1
  else ext.North := 0;
end;

function TDSpHDFSDDriver.DataSetType(i: integer): TDSpPFAFormat;
begin
  Result := dspPFARaster;
end;

function TDSpHDFSDDriver.CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer;
begin
  Assert(false, 'HDF driver does not support writing new data sets');
  Result := -1;
end;

function TDSpHDFSDDriver.AddDataSet(const ds: TDSpSourceDataSet): integer;
begin
  Assert(false, 'Not yet implemented.');
  Result := -1;
end;

function TDSpHDFSDDriver.CloneDataSet(const ds: TDSpSourceDataSet): integer;
begin
  Assert(false, 'Not yet implemented.');
  Result := -1;
end;

procedure TDSpHDFSDDriver.AddRasterBand(i: integer; info: PDSpRasterbandInfo);
begin
  info^.BandIndex := -1;
end;

procedure TDSpHDFSDDriver.DeleteRasterBand(i: integer);
begin
  Assert(false, 'HDF driver does not support writing data sets');
end;

procedure TDSpHDFSDDriver.GetRasterInfo(i: integer; info: PDSpRasterInfo);
var sds: THDF_SDS;
    rank: integer;
begin
  Assert(Assigned(FSD));
  sds := FSD.SDS[i];
  rank := sds.Rank;

  info^.BandFormat := dspRasterFlat;
  if rank > 2 then info^.Bands := sds.Dimension[rank-2]
  else info^.Bands := 1;
  if rank > 1 then info^.Rows := sds.Dimension[rank-1]
  else info^.Rows := 1;
  info^.Columns := sds.Dimension[rank];
  info^.PixelSize := 1.0;
end;

procedure TDSpHDFSDDriver.GetRasterBandInfo(i: integer; info: PDSpRasterInfo);
var b:  integer;
    sds: THDF_SDS;
begin
  Assert(Assigned(FSD));
  sds := FSD.SDS[i];
  for b := 0 to High(info^.Band) do
    with info^.Band[b] do begin
      BandIndex := sds.ID;
      case sds.DataType of
         4: DataType := dspInt8;     // NOTE: this is HDF char data type
         5: DataType := dspFloat32;
         6: DataType := dspFloat64;
        20: DataType := dspInt8;
        21: DataType := dspUInt8;
        22: DataType := dspInt16;
        23: DataType := dspUInt16;
        24: DataType := dspInt32;
        25: DataType := dspUInt32;
      else Assert(false, 'Need code to act on non-numeric HDF SD data');
      end;
      BandType := dspRasterBandOrdinal;
      MissingValue := sds.MissingValue;
      Min := 0;
      Max := -1;  
    end;
end;

procedure TDSpHDFSDDriver.ReadRasterBand(dataSet, band: integer; const into: pointer);
var start, len: TDimsArray;
    sds: THDF_SDS;
    i, rank: integer;
begin
  Assert(Assigned(FSD));
  sds := FSD.SDS[dataSet];
  rank := sds.Rank;

  for i := 1 to 6 do begin
    start[i] := 0;
    len[i] := 1;
  end;

  len[rank] := sds.Dimension[rank];
  if rank > 1 then len[rank-1] := sds.Dimension[rank-1];
  if rank > 2 then start[rank-2] := band;

  sds.ReadData(start, len, into);
end;

procedure TDSpHDFSDDriver.ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer);
begin
end;

procedure TDSpHDFSDDriver.ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer);
begin
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpHDFSDBroker
constructor TDSpHDFSDBroker.Create;
begin
  inherited;
  FDescription := 'HDF files';
  FFilter := '*.HDF';
  FCaps := [dspReadRaster];
end;

destructor TDSpHDFSDBroker.Destroy;
begin
  inherited;
end;

function TDSpHDFSDBroker.CreateSource(const fn: string): TDSpSourceHandler;
begin
  Result := nil;
  Assert(false, 'HDF is a read-only format');
end;

function TDSpHDFSDBroker.OpenSource(const fn: string): TDSpSourceHandler;
var r   : TDSpHDFSDDriver;
begin
  r := TDSpHDFSDDriver.Open(fn);
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

function TDSpHDFSDBroker.SupportsSource(const source: string): boolean;
begin
  Result := UpperCase(ExtractFileExt(source)) = '.HDF';
end;

function TDSpHDFSDBroker.CanSave(const ds: TDSpSourceDataSet): string;
begin
  Result := '';
end;

initialization
{ Enable this method for static loading of the driver. Please note that the    }
{ DSpPFA unit MUST be loaded before this unit. If the driver is to be loaded   }
{ dynamically, you will have to provide a DLL with this unit.                  }
  PFA.RegisterBroker(TDSpHDFSDBroker.Create);
end.
