{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpHDFInterface.pas. The Original Code is available     }
{ through the home page of DSpatial at http://dspatial.sourceforge.net.        }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit provides an interface to the HDF library, version 41r5, contained  }
{ in the DLL for Windows (HDF41r5.dll). Currently only the SD interface is     }
{ implemented, which will allow you to read many science data sets from        }
{ assorted NASA satellite platforms.                                           }
{                                                                              }
{ HDF is developed by the National Center for Supercomputing Applications at   }
{ the University of Illinois at Urbana-Champaign (NCSA). Documentation, latest }
{ versions, and legalia regarding HDF can be found on their website at         }
{ http://hdf.ncsa.uiuc.edu/.                                                   }
{                                                                              }
{ This unit is a complete translation of the SD interface, which is nearly     }
{ identical to the call interface of the original HDF library. Use the HDF     }
{ documentation for information on using the functions. This code is specific  }
{ to the version 4.1r5 library, it is incompatible with HDF5.                  }
{                                                                              }
{ Start using the library by simply calling SDstart. This will automagically   }
{ load the DLL and set all the function pointers to the right address. If the  }
{ HDF DLL is not in some convenient location you will have to call             }
{ SetHDFLibPath first. When you're done call UnloadHDF, which will release the }
{ DLL. This nifty trick allows you to save some resources that you might       }
{ require elsewhere. Please note that the DSpHDFObject unit contains an        }
{ object-oriented wrapper for this unit, which makes it much easier to include }
{ HDF data in your applications.                                               }
{                                                                              }
{ This is a product in development. Reading from existing SD data sets has been}
{ tested and is considered to be reliable. You can read both SD data sets and  }
{ the older, deprecated, DFSD data sets. Writing new data sets has not been    }
{ tested. Any results, bug reports, and comments are greatly appreciated.      }
{ There are plans to work on the GR and Vdata interfaces and the H calls, but  }
{ this will not be undertaken until someone can demonstrate the need to do so  }
{ (all HDF data that the developer is aware of consists of SD data sets).      }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 July 21                                                 }
{                                                                              }
{ Dependencies  : DSpHDFTypes                                                  }
{                 Windows (LoadLibrary, FreeLibrary)                           }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.1  03Jul21 Initial coding. Port from an earlier code base  }
{                              developed by the Initial Developer.             }
{                                                                              }
{******************************************************************************}
unit DSpHDFInterface;
interface
uses Windows, DSpHDFTypes;
const
  // Access modes
  DFACC_READ   : cardinal = 1;
  DFACC_WRITE  : cardinal = 2;
  DFACC_CREATE : cardinal = 4; // will also enable read and write
  DFACC_ALL    : cardinal = 7;
  DFACC_RDONLY : cardinal = 1;
  DFACC_RDWR   : cardinal = 3;

type
  TCompInfo = integer; // Lazy implementation

  THDFChunkDef = record
    chunk_lengths : array [1..2] of integer;
    case integer of
    0: ( comp_type : integer;
         cinfo     : TCompInfo;);
    1: ( start_bit,
         bit_len,
         sign_ext,
         fill_one  : integer;);
  end;

var
  // SD function prototypes
  SDattrinfo        : function(obj_id, attr_index: integer; attr_name: PChar; var data_type, count: integer): integer; cdecl;
  SDcheckempty      : function(sds_id: integer; var emptySDS: integer): integer; cdecl;
  SDcreate          : function(sd_id: integer; name: PChar; data_type, rank: integer; dimsizes: TDimsArray): integer; cdecl;
  SDdiminfo         : function(dim_id: integer; name: PChar; var size, data_type, num_attrs: integer): integer; cdecl;
  SDend             : function(sd_id: integer): integer; cdecl;
  SDendaccess       : function(sds_id: integer): integer; cdecl;
  SDfileinfo        : function(sd_id: integer; var num_datasets, num_global_attrs: integer): integer; cdecl;
  SDfindattr        : function(obj_id: integer; attr_name: PChar): integer; cdecl;
  SDgetcal          : function(sds_id: integer; var cal, cal_err, offset, offset_err: double; var data_type: integer): integer; cdecl;
  SDgetchunkinfo    : function(sds_id: integer; var cdef: THDFChunkDef; var flag: integer): integer; cdecl;
  SDgetcompress     : function(sds_id: integer; var comp_type: integer; var c_info: TCompInfo): integer; cdecl;
  SDgetdatastrs     : function(sds_id: integer; lab, unt, format, coordsys : PChar; length: integer): integer; cdecl;
  SDgetdimid        : function(sds_id, dim_index: integer): integer; cdecl;
  SDgetdimscale     : function(dim_id: integer; scale_buf: pointer): integer; cdecl;
  SDgetdimstrs      : function(dim_id: integer; lab, unt, format: PChar; length: integer): integer; cdecl;
  SDgetfillvalue    : function(sds_id: integer; fill_value: pointer): integer; cdecl;
  SDgetinfo         : function(sds_id : integer; sds_name : PChar; var rank : integer; dimsizes : TDimsArray; var data_type, num_attrs : integer) : integer; cdecl;
  SDgetrange        : function(sds_id: integer; max, min: pointer): integer; cdecl;
  SDidtoref         : function(sds_id: integer): integer; cdecl;
  SDiscoordvar      : function(sds_id: integer): boolean; cdecl;
  SDisdimval_bwcomp : function(dim_id: integer): integer; cdecl;
  SDisrecord        : function(sds_id: integer): boolean; cdecl;
  SDnametoindex     : function(sd_id: integer; sds_name: PChar): integer; cdecl;
  SDreadattr        : function(obj_id, attr_index: integer; attr_buf: pointer): integer; cdecl;
  SDreadchunk       : function(sds_id: integer; var origin: integer; datap: pointer): integer; cdecl;
  SDreaddata        : function(sds_id: integer; start, stride, edge: PDimsArray; buffer: pointer): integer; cdecl;
  SDreftoindex      : function(sds_id, sds_ref: integer): integer; cdecl;
  SDselect          : function(sd_id, sds_index: integer) : integer; cdecl;
  SDsetattr         : function(obj_id: integer; attr_name: PChar; data_type, count: integer; values: pointer): integer; cdecl;
  SDsetblocksize    : function(sd_id, block_size: integer): integer; cdecl;
  SDsetcal          : function(sds_id: integer; cal, cal_err, offset, offset_err: double; data_type: integer): integer; cdecl;
  SDsetchunk        : function(sds_id: integer; cdef: THDFChunkDef; flag: integer): integer; cdecl;
  SDsetchunkcache   : function(sds_id, maxcache, flag: integer): integer; cdecl;
  SDsetcompress     : function(sds_id, comp_type: integer; var c_info: TCompInfo): integer; cdecl;
  SDsetdatastrs     : function(sds_id: integer; lab, unt, format, coordsys: PChar): integer; cdecl;
  SDsetdimname      : function(dim_id: integer; dim_name: PChar): integer; cdecl;
  SDsetdimscale     : function(dim_id, count, data_type: integer; data: pointer): integer; cdecl;
  SDsetdimstrs      : function(dim_id: integer; lab, unt, format: PChar): integer; cdecl;
  SDsetdimval_comp  : function(dim_id, comp_mode: integer): integer; cdecl;
  SDsetexternalfile : function(sds_id: integer; filename: PChar; offset: integer): integer; cdecl;
  SDsetfillmode     : function(sd_id, fill_mode: integer): integer; cdecl;
  SDsetfillvalue    : function(sds_id: integer; fill_value: pointer): integer; cdecl;
  SDsetnbitdataset  : function(sds_id, start_bit, bit_len, sign_ext, fill_one: integer): integer; cdecl;
  SDsetrange        : function(sds_id: integer; max, min: pointer): integer; cdecl;
  // SDstart is hidden
  SDwritechunk      : function(sds_id: integer; var origin: integer; datap: pointer): integer; cdecl;
  SDwritedata       : function(sds_id: integer; start, stride, edge: TDimsArray; buffer: pointer): integer; cdecl;

procedure SetHDFLibPath(libPath : string);
procedure UnloadHDF;
function IsLoadedHDF : boolean;

// Declarations of interfaces to hidden prototypes
function SDstart(const filename : PChar; access_mode : integer) : integer;

implementation
uses SysUtils;
var hDLLm, hDLLd : THandle;
    IsLoadedSD : boolean;
    hmName, hdName : string;

    // Hidden prototypes
    _SDstart : function(const filename : PChar; access_mode : integer) : integer; cdecl;

procedure SetHDFLibPath(libPath : string);
begin
  if (libPath <> '') and (libPath[Length(libPath)] <> '\') then libPath := libPath + '\';
  hmName := libPath + 'hm415m.dll';
  hdName := libPath + 'hd415m.dll';
end;

procedure UnloadHDF;
begin
  if hDLLm <> 0 then FreeLibrary(hDLLm);
  hDLLm := 0;
  if hDLLd <> 0 then FreeLibrary(hDLLd);
  hDLLd := 0;
end;

function IsLoadedHDF : boolean;
begin
  Result := (hDLLm <> 0) or (hDLLd <> 0);
end;

procedure UseSD;
begin
  if hDLLm = 0 then hDLLm := LoadLibrary(PChar(hmName));
  if hDLLm = 0 then raise Exception.CreateFmt('DLL %s could not be loaded', [hmName]);
  _SDstart          := GetProcAddress(hDLLm, 'SDstart');
  SDattrinfo        := GetProcAddress(hDLLm, 'SDattrinfo');
  SDcheckempty      := GetProcAddress(hDLLm, 'SDcheckempty');
  SDcreate          := GetProcAddress(hDLLm, 'SDcreate');
  SDdiminfo         := GetProcAddress(hDLLm, 'SDdiminfo');
  SDend             := GetProcAddress(hDLLm, 'SDend');
  SDendaccess       := GetProcAddress(hDLLm, 'SDendaccess');
  SDfileinfo        := GetProcAddress(hDLLm, 'SDfileinfo');
  SDfindattr        := GetProcAddress(hDLLm, 'SDfindattr');
  SDgetcal          := GetProcAddress(hDLLm, 'SDgetcal');
  SDgetchunkinfo    := GetProcAddress(hDLLm, 'SDgetchunkinfo');
  SDgetcompress     := GetProcAddress(hDLLm, 'SDgetcompress');
  SDgetdatastrs     := GetProcAddress(hDLLm, 'SDgetdatastrs');
  SDgetdimid        := GetProcAddress(hDLLm, 'SDgetdimid');
  SDgetdimscale     := GetProcAddress(hDLLm, 'SDgetdimscale');
  SDgetdimstrs      := GetProcAddress(hDLLm, 'SDgetdimstrs');
  SDgetfillvalue    := GetProcAddress(hDLLm, 'SDgetfillvalue');
  SDgetinfo         := GetProcAddress(hDLLm, 'SDgetinfo');
  SDgetrange        := GetProcAddress(hDLLm, 'SDgetrange');
  SDidtoref         := GetProcAddress(hDLLm, 'SDidtoref');
  SDiscoordvar      := GetProcAddress(hDLLm, 'SDiscoordvar');
  SDisdimval_bwcomp := GetProcAddress(hDLLm, 'SDisdimval_bwcomp');
  SDisrecord        := GetProcAddress(hDLLm, 'SDisrecord');
  SDnametoindex     := GetProcAddress(hDLLm, 'SDnametoindex');
  SDreadattr        := GetProcAddress(hDLLm, 'SDreadattr');
  SDreadchunk       := GetProcAddress(hDLLm, 'SDreadchunk');
  SDreaddata        := GetProcAddress(hDLLm, 'SDreaddata');
  SDreftoindex      := GetProcAddress(hDLLm, 'SDreftoindex');
  SDselect          := GetProcAddress(hDLLm, 'SDselect');
  SDsetattr         := GetProcAddress(hDLLm, 'SDsetattr');
  SDsetblocksize    := GetProcAddress(hDLLm, 'SDsetblocksize');
  SDsetcal          := GetProcAddress(hDLLm, 'SDsetcal');
  SDsetchunk        := GetProcAddress(hDLLm, 'SDsetchunk');
  SDsetchunkcache   := GetProcAddress(hDLLm, 'SDsetchunkcache');
  SDsetcompress     := GetProcAddress(hDLLm, 'SDsetcompress');
  SDsetdatastrs     := GetProcAddress(hDLLm, 'SDsetdatastrs');
  SDsetdimname      := GetProcAddress(hDLLm, 'SDsetdimname');
  SDsetdimscale     := GetProcAddress(hDLLm, 'SDsetdimscale');
  SDsetdimstrs      := GetProcAddress(hDLLm, 'SDsetdimstrs');
  SDsetdimval_comp  := GetProcAddress(hDLLm, 'SDsetdimval_comp');
  SDsetexternalfile := GetProcAddress(hDLLm, 'SDsetexternalfile');
  SDsetfillmode     := GetProcAddress(hDLLm, 'SDsetfillmode');
  SDsetfillvalue    := GetProcAddress(hDLLm, 'SDsetfillvalue');
  SDsetnbitdataset  := GetProcAddress(hDLLm, 'SDsetnbitdataset');
  SDsetrange        := GetProcAddress(hDLLm, 'SDsetrange');
  SDwritechunk      := GetProcAddress(hDLLm, 'SDwritechunk');
  SDwritedata       := GetProcAddress(hDLLm, 'SDwritedata');
  IsLoadedSD  := Assigned(_SDstart) and Assigned(SDend) and Assigned(SDselect) and
                 Assigned(SDendaccess) and Assigned(SDfileinfo) and Assigned(SDgetinfo) and
                 Assigned(SDcheckempty) and Assigned(SDcreate) and Assigned(SDdiminfo) and
                 Assigned(SDfindattr) and Assigned(SDgetcal) and
                 Assigned(SDgetchunkinfo) and Assigned(SDgetcompress) and Assigned(SDgetdatastrs) and
                 Assigned(SDgetdimid) and Assigned(SDgetdimscale) and Assigned(SDgetdimstrs) and
                 Assigned(SDgetfillvalue) and Assigned(SDgetrange) and Assigned(SDidtoref) and
                 Assigned(SDiscoordvar) and Assigned(SDisdimval_bwcomp) and Assigned(SDisrecord) and
                 Assigned(SDnametoindex) and Assigned(SDreadchunk) and Assigned(SDreftoindex) and
                 Assigned(SDsetattr) and Assigned(SDsetblocksize) and Assigned(SDsetcal) and
                 Assigned(SDsetchunk) and Assigned(SDsetchunkcache) and Assigned(SDsetcompress) and
                 Assigned(SDsetdatastrs) and Assigned(SDsetdimname) and Assigned(SDsetdimscale) and
                 Assigned(SDsetdimstrs) and Assigned(SDsetdimval_comp) and Assigned(SDsetexternalfile) and
                 Assigned(SDsetfillmode) and Assigned(SDsetfillvalue) and Assigned(SDsetnbitdataset) and
                 Assigned(SDsetrange) and Assigned(SDwritechunk) and Assigned(SDwritedata) and
                 Assigned(SDreaddata) and Assigned(SDreadattr) and Assigned(SDattrinfo);
  if not IsLoadedSD then raise Exception.CreateFmt('Could not grab all SD functions in %s.', [hmName]);
end;

function SDstart(const filename : PChar; access_mode : integer) : integer;
begin
  access_mode := access_mode and 7;  // mask out any unwanted bits
  if not IsLoadedSD then UseSD;
  Result := _SDStart(filename, access_mode);
end;

initialization
  hDLLm := 0;
  hDLLd := 0;
  hmName := 'hm415m.dll';
  hdName := 'hd415m.dll';
  IsLoadedSD := false;
finalization
  UnloadHDF;
end.
