{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpHDFObject.pas. The Original Code is available through}
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit provides an object-oriented wrapper around the HDF interface in the}
{ DSpHDFInterface unit. It supports the reading of and writing to SDS data sets}
{ in an HDF file.                                                              }
{                                                                              }
{ Note that this unit does not follow the standard DSpatial naming of classes  }
{ and class members. This is done to maintain the names given to parameters in }
{ the native HDF calls, making the use of HDF documentation easier.            }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 18                                             }
{                                                                              }
{ Dependencies  : DSpHDFInterface                                              }
{                 DSpHDFTypes                                                  }
{                 Classes (TStringList)                                        }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.2  03Dec18 Writing of SDSs enabled.                        }
{                 0.1  03Jul21 Initial coding. Port from an earlier code base  }
{                              developed by the Initial Developer.             }
{                                                                              }
{******************************************************************************}
unit DSpHDFObject;
{$I DSpatial.inc}
interface
uses Classes, SysUtils, DSpHDFTypes, DSpHDFInterface;
type
  THDF_SDS = class(TObject)
  private
    sds_id : integer;
    sds_name : PChar;
    rnk, data_type, num_attrs: integer;
    dims : TDimsArray;
  protected
    function GetName: string;
    function GetDimension(d: integer): integer;
    function GetAttributeString(a: integer): string;
    function GetMissingValue: double;
  public
    constructor Create(sds: integer);
    destructor  Destroy; override;

    function  AttributeStrings: TStringList;
    procedure ReadData(start, len: TDimsArray; buf: pointer);

    property  ID: integer read sds_id;
    property  Name: string read GetName;
    property  Rank: integer read rnk;
    property  Dimension[d: integer]: integer read GetDimension;
    property  DataType: integer read data_type;
    property  NumAttributes: integer read num_attrs;
    property  AttributeAsString[a: integer]: string read GetAttributeString;
    property  MissingValue: double read GetMissingValue;
  end;

  THDF_SD = class(TObject)
  private
    sd_id : integer;
    ndatasets, numattrs: integer;
    SDSs  : array of THDF_SDS;
  protected
    function GetSDS(s: integer) : THDF_SDS;
    function GetGlobAttr: TStringList;
  public
    constructor Open(sd: integer);
    destructor  Destroy; override;

    function  SDSNamesAndObjects: TStringList;
    procedure CloseSDS(var sds: THDF_SDS);
    property  NumDataSets: integer read ndatasets;
    property  NumGlobalAttributes: integer read numattrs;
    property  SDS[s: integer]: THDF_SDS read GetSDS;
    property  GlobalAttributes: TStringList read GetGlobAttr;
  end;

  THDFFile = class(TObject)
  private
    fname    : string;
    sd_id    : integer;
  public
    constructor Create(HDFname : string);
    destructor  Destroy; override;
    function    OpenSD: THDF_SD;

    property    FileName : string read fname;
  end;

procedure SetHDFDLLPath(p : string);
procedure ReleaseHDF;

implementation
///////////////////////////////////////////////////////////////////////////////
// generic functions
function ReadAttributeString(obj_id, ndx: integer): string;
// Read an attribute as a string. Used by SD and SDS.
var name : string;
    dtype, values, v, len : integer;
    p : pointer;
begin
  SetLength(name, 256);
  if SDattrinfo(obj_id, ndx, PChar(name), dtype, values) = HDF_FAIL then
    raise Exception.CreateFmt('Could not read attribute info for %d on %d', [ndx, obj_id]);
  len := Pos(#0, name) - 1;
  SetLength(name, len);
  name := name + ' = '; 
  p := Allocmem(values * HDF_DATATYPE_SIZE[dtype]);
  try
    if SDreadattr(obj_id, ndx, p) = HDF_FAIL then
      raise Exception.CreateFmt('Could not read attribute %s (%d) on %d', [name, ndx, obj_id]);
    case dtype of
       4: name := AdjustLineBreaks(name + PChar(p));
       5: begin // single
            name := Format('%s %10.5f', [name, SArray(p^)[0]]);
            if values > 1 then
              for v := 1 to values - 1 do name := Format('%s, %10.5f', [name, SArray(p^)[v]]);
          end;
       6: begin // double
            name := Format('%s %10.5f', [name, DArray(p^)[0]]);
            if values > 1 then
              for v := 1 to values - 1 do name := Format('%s, %10.5f', [name, DArray(p^)[v]]);
          end;
      20: begin // short
            name := Format('%s %d', [name, SHArray(p^)[0]]);
            if values > 1 then
              for v := 1 to values - 1 do name := Format('%s, %d', [name, SHArray(p^)[v]]);
          end;
      21: begin // byte
            name := Format('%s %d', [name, BArray(p^)[0]]);
            if values > 1 then
              for v := 1 to values - 1 do name := Format('%s, %d', [name, BArray(p^)[v]]);
          end;
      22: begin // smallint
            name := Format('%s %d', [name, SMArray(p^)[0]]);
            if values > 1 then
              for v := 1 to values - 1 do name := Format('%s, %d', [name, SMArray(p^)[v]]);
          end;
      23: begin // word
            name := Format('%s %d', [name, WArray(p^)[0]]);
            if values > 1 then
              for v := 1 to values - 1 do name := Format('%s, %d', [name, WArray(p^)[v]]);
          end;
      24: begin // integer
            name := Format('%s %d', [name, IArray(p^)[0]]);
            if values > 1 then
              for v := 1 to values - 1 do name := Format('%s, %d', [name, IArray(p^)[v]]);
          end;
      25: begin // cardinal
            name := Format('%s %d', [name, CArray(p^)[0]]);
            if values > 1 then
              for v := 1 to values - 1 do name := Format('%s, %d', [name, CArray(p^)[v]]);
          end;
      else name := Format('Unhandled data_type for %s%d', [name, dtype]);
    end;
    Result := name;
  finally
    FreeMem(p);
  end;
end;

/////////////////////////////////////////////////////////////////////////////
// THDF_SDS
constructor THDF_SDS.Create(sds: integer);
begin
  inherited Create;
  sds_id := sds;
  if sds_id = HDF_FAIL then raise Exception.Create('SDS not found');
  sds_name := AllocMem(64);
  if SDgetinfo(sds_id, sds_name, rnk, dims, data_type, num_attrs) <> 0 then
    raise Exception.Create(Format('Could not read SDS with id %d', [sds_id]));
end;

destructor THDF_SDS.Destroy;
begin
//  SDendaccess(sds_id);   // <<<<<<<<<<<<<<<<<<< THIS THROWS AN EXCEPTION
  FreeMem(sds_name);
  inherited;
end;

function THDF_SDS.GetName: string;
begin
  Result := sds_name;
end;

function THDF_SDS.GetDimension(d: integer): integer;
begin
  if d in [1..rnk] then Result := dims[d]
  else raise Exception.Create('Dimension out of bounds');
end;

function THDF_SDS.GetAttributeString(a: integer): string;
begin
  if not a in [0..num_attrs-1] then
    raise Exception.Create('Attribute index out of bounds');
  Result := ReadAttributeString(sds_id, a);
end;

function THDF_SDS.GetMissingValue: double;
var buf: array[0..7] of byte;
    s: single;
    sm: smallint;
    w: word;
    i: integer;
    c: cardinal;
begin
  if SDgetfillvalue(sds_id, @buf) = HDF_FAIL then Result := 0 // <<<<<<<<<<<<<<<<<
  else begin
    case data_type of
       4: Result := 0; // <<<<<<<<<<<<<<<<<<<<<<<< char data
       5: begin
            Move(buf[0], s, 4);
            Result := s;
          end;
       6: Move(buf[0], Result, 8);
      20: Result := shortint(buf[0]);
      21: Result := buf[0];
      22: begin
            Move(buf[0], sm, 2);
            Result := sm;
          end;
      23: begin
            Move(buf[0], w, 2);
            Result := w;
          end;
      24: begin
            Move(buf[0], i, 4);
            Result := i;
          end;
      25: begin
            Move(buf[0], c, 4);
            Result := c;
          end;
    end;
  end;
end;

function THDF_SDS.AttributeStrings: TStringList;
var n: integer;
begin
  Result := TStringList.Create;
  Result.Capacity := num_attrs;
  for n := 0 to num_attrs - 1 do Result.Add(GetAttributeString(n));
end;

procedure THDF_SDS.ReadData(start, len: TDimsArray; buf: pointer);
begin
  SDreaddata(sds_id, @start, nil, @len, buf);
end;

/////////////////////////////////////////////////////////////////////////////
// THDF_SD
constructor THDF_SD.Open(sd : integer);
begin
  inherited Create;
  sd_id := sd;
  if SDfileinfo(sd_id, ndatasets, numattrs) <> 0 then
    raise Exception.Create('Could not retrieve SD file info');
  SetLength(SDSs, ndatasets);
end;

destructor THDF_SD.Destroy;
var n : integer;
begin
  if ndatasets > 0 then 
    for n := 0 to ndatasets - 1 do
      if SDSs[n] <> nil then SDSs[n].Free;
  Finalize(SDSs);
//  if sd_id >= 0 then SDend(sd_id);  // <<<<<<<<<<<<<<<<<<< THIS THROWS AN EXCEPTION
  inherited;
end;

function THDF_SD.GetSDS(s : integer) : THDF_SDS;
begin
  Assert((s >= 0) and (s < ndatasets));
  if SDSs[s] = nil then SDSs[s] := THDF_SDS.Create(SDselect(sd_id, s));
  Result := SDSs[s];
end;

procedure THDF_SD.CloseSDS(var sds : THDF_SDS);
var n : integer;
begin
  for n := 0 to ndatasets - 1 do
    if SDSs[n] = sds then begin
      sds.Free;                      // free object when no longer referenced
      SDSs[n] := nil;               // empty the container slot
      Break;
    end;
  sds := nil;                        // set to nil no matter if found or not
end;

function THDF_SD.SDSNamesAndObjects: TStringList;
var n: integer;
begin
  Result := TStringList.Create;
  if ndatasets = 0 then Exit;
  Result.Capacity := ndatasets;
  for n := 0 to ndatasets - 1 do
    Result.AddObject(SDSs[n].Name, SDSs[n]);
end;

function THDF_SD.GetGlobAttr: TStringList;
var n : integer;
begin
  Result := TStringList.Create;
  Result.Capacity := numattrs;
  for n := 0 to numattrs - 1 do
    Result.Add(ReadAttributeString(sd_id, n));
end;

/////////////////////////////////////////////////////////////////////////////
// THDFfile
constructor THDFfile.Create(HDFname : string);
begin
  inherited Create;
  fname := HDFname;
end;

destructor THDFfile.Destroy;
begin
  inherited;
end;

function THDFfile.OpenSD: THDF_SD;
begin
  if FileExists(fname) then sd_id := SDstart(PChar(fname), DFACC_WRITE)
  else sd_id := SDstart(PChar(fname), DFACC_CREATE);

  if sd_id >= 0 then Result := THDF_SD.Open(sd_id) else Result := nil;
end;

procedure SetHDFDLLPath(p : string);
begin
  SetHDFLibPath(p);
end;

procedure ReleaseHDF;
begin
  UnloadHDF;
end;

end.
