{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpISO8211.pas. The Original Code is available through  }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit reads ISO8211 formatted files. Currently only level 1 and 2 files  }
{ are supported. This format is best known for its use with SDTS data          }
{ transfers.                                                                   }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 April 17                                                }
{                                                                              }
{ Dependencies  : SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.1  03Apr17 Initial coding. Supports only one level of      }
{                              nesting in formats strings.                     }
{                                                                              }
{******************************************************************************}
unit DSpISO8211;
interface
type
  TDirEntryMap = record
                   Tag : integer;
                   Pos : integer;
                   Len : integer;
                 end;

  TDDRLeader = record
                 RecordLength  : integer;
                 Interchange   : char;
                 LeaderID      : char;
                 CodeExtension : char;
                 FieldControl  : integer;
                 DataAddress   : integer;
                 CharacterSet  : array[0..2] of char;
                 DirMap        : TDirEntryMap;
               end;

  TDRLeader = record
                RecordLength  : integer;
                LeaderID      : char;
                DataAddress   : integer;
                DirMap        : TDirEntryMap;
              end;

  TDirEntry   = record
                  Tag    : string[7];
                  Length : integer;
                  Data   : PChar;   // not owned, do not Dispose!!
                end;
  TDirEntries = array of TDirEntry;

  // Data Descriptive Area
  TDDAField = record
                Name     : string;    // field name (ISO8211: label)
                DataType : char;      // type of data
                Length   : integer;   // length of field; 0 if terminated
                Offset   : integer;   // offset into user structure, filled by caller
                Format   : string;    // format string for data, filled by caller
              end;
  TDDAFieldArray = array of TDDAField;
  
  TDDAEntry = record
                Tag          : string[7];
                FieldControl : string;
                Name         : string;
                Fields       : TDDAFieldArray;
              end;
  TDDAEntries = array of TDDAEntry;

  TISO8211 = class
  private
    FFileName   : string;
    FData       : PChar;            // the entire file contents
    FCur        : PChar;            // current position in the stream
    FEnd        : PChar;            // points at the #0 byte at the end of FData
    FDDRLeader  : TDDRLeader;
    FDDREntries : TDirEntries;
    FDDA        : TDDAEntries;      // Data Descriptive Area
    FRLeader    : TDirEntries;      // directory of repetitive DRs

    procedure ParseDirEntries(const dir: TDirEntries; const d: TDirEntryMap);
    function  GetDirEntry(e: integer): TDirEntry;
    function  FindDirEntry(tag: string): TDirEntry;
    function  GetDDREntries: integer;
    function  GetDDAEntries: integer;
    function  GetDREntries: integer;
    procedure ParseFormat(ent: TDDAEntry; fmt: PChar);

  public
    constructor Create(fn: string);
    destructor  Destroy; override;

    function  ReadNextRecord(const rec: TDirEntries): boolean;

    property  DDAEntries: integer read GetDDAEntries;
    property  DDREntry[e: integer]: TDirEntry read GetDirEntry;
    property  DDREntryByTag[tag: string]: TDirEntry read FindDirEntry;
    property  DDA: TDDAEntries read FDDA;
    property  DREntries: integer read GetDREntries;
  end;

implementation
uses SysUtils;

///////////////////////////////////////////////////////////////////////////////
// TISO8211: parsing of ISO 8211 files
constructor TISO8211.Create(fn: string);
var f: file;
    s, res, dda: integer;
    str : string[5];
    definition : TDirEntry;
    fieldName, sep, fldEnd : PChar;
begin
  FFileName := fn;
  AssignFile(f, FFileName);
  try
    Reset(f, 1);
    s := FileSize(f);
    FData := StrAlloc(s + 1);
    BlockRead(f, FData[0], s, res);     // what about HUGE files???
    if res <> s then
      raise Exception.Create('Corrupt file');
    FData[s] := #0;
    FCur := FData;
    FEnd := @FData[s];

    // read the DDR leader
    with FDDRLeader do begin
      SetLength(str, 5);
      Move(FCur[0], str[1], 5);
      RecordLength := StrToInt(str);
      InterChange := FCur[5];
      LeaderID := FCur[6];
      CodeExtension := FCur[7];
      SetLength(str, 2);
      Move(FCur[10], str[1], 2);
      FieldControl := StrToInt(str);
      SetLength(str, 5);
      Move(FCur[12], str[1], 5);
      DataAddress := StrToInt(str);
      Move(FCur[17], CharacterSet[0], 3);
      DirMap.Len := StrToInt(FCur[20]);
      DirMap.Pos := StrToInt(FCur[21]);
      DirMap.Tag := StrToInt(FCur[23]);
      Inc(FCur, 24);           // FCur now points to the first directory entry

      SetLength(FDDREntries, (DataAddress - 25) div (DirMap.Pos + DirMap.Len + DirMap.Tag));
      ParseDirEntries(FDDREntries, DirMap);
      FCur := PChar(integer(FData) + RecordLength); // FCur now points to the first DR
    end;

    // Parse Data Descriptive Area
    SetLength(FDDA, GetDDREntries - 2);
    for dda := 0 to GetDDREntries - 3 do begin  // skip external file name and RIF
      definition := DDREntry[dda+2];

      FDDA[dda].Tag := FDDREntries[dda+2].Tag;
      FDDA[dda].FieldControl := '    '; // 4 blanks
      Move(definition.Data[0], FDDA[dda].FieldControl[1], 4);

      fieldName := StrScan(definition.Data, #$1F);
      s := integer(fieldName) - integer(definition.Data) - FDDRLeader.FieldControl;
      SetLength(FDDA[dda].Name, s);
      Move(PChar(integer(definition.Data) + FDDRLeader.FieldControl)^, FDDA[dda].Name[1], s);

      Inc(fieldName);            // fieldName now points to the first field name
      if fieldName[0] = '*' then Inc(fieldName); // jump over array hint
      fldEnd := StrScan(fieldName, #$1F);
      res := 1;
      repeat
        SetLength(FDDA[dda].Fields, res);
        sep := StrScan(fieldName, '!');
        if (sep = nil) or (sep > fldEnd) then sep := fldEnd;
        s := integer(sep) - integer(fieldName);
        SetLength(FDDA[dda].Fields[res-1].Name, s);
        Move(fieldName[0], FDDA[dda].Fields[res-1].Name[1], s);
        FDDA[dda].Fields[res-1].Format := '';
        Inc(fieldName, s);
        if fieldName[0] = #$1F then Break;
        Inc(fieldName);
        Inc(res);
      until false;

      // Parse the format
      Inc(fieldName);    // fieldName now points at the format
      ParseFormat(FDDA[dda], fieldName);
    end;
  finally
    CloseFile(f);
  end;
end;

destructor TISO8211.Destroy;
begin
  StrDispose(FData);
  Finalize(FDDREntries);
  Finalize(FRLeader);
  Finalize(FDDA);
end;

procedure TISO8211.ParseDirEntries(const dir: TDirEntries; const d: TDirEntryMap);
var str    : string[9];
    i, off : integer;
    base   : integer;
begin
  // FCur points to the beginning of the directory
  if High(dir) = -1 then Exit;
  base := integer(FCur) + Length(dir) * (d.Tag + d.Pos + d.Len) + 1;
  for i := 0 to High(dir) do begin
    SetLength(dir[i].Tag, d.Tag);
    Move(FCur^, dir[i].Tag[1], d.Tag);
    Inc(FCur, d.Tag);

    SetLength(str, d.Len);
    Move(FCur^, str[1], d.Len);
    dir[i].Length := StrToInt(str);
    Inc(FCur, d.Len);

    SetLength(str, d.Pos);
    Move(FCur^, str[1], d.Pos);
    off := StrToInt(str);
    Inc(FCur, d.Pos);

    dir[i].Data := PChar(base + off);
  end;
  Inc(FCur); // skip the FT
end;

function TISO8211.GetDirEntry(e : integer): TDirEntry;
begin
  if (e < Low(FDDREntries)) or (e > High(FDDREntries)) then Result.Length := 0
  else Result := FDDREntries[e];
end;

function TISO8211.FindDirEntry(tag: string): TDirEntry;
var i : integer;
begin
  Result.Length := 0;
  for i := Low(FDDREntries) to High(FDDREntries) do
    if FDDREntries[i].Tag = tag then begin
      Result := FDDREntries[i];
      Break;
    end;
end;

function TISO8211.GetDDREntries: integer;
begin
  Result := High(FDDREntries) - Low(FDDREntries) + 1;
end;

function TISO8211.GetDDAEntries: integer;
begin
  Result := High(FDDA) - Low(FDDA) + 1;
end;

function TISO8211.GetDREntries: integer;
begin
  Result := High(FDDREntries) - Low(FDDREntries); // no external file title
end;

function TISO8211.ReadNextRecord(const rec: TDirEntries): boolean;
var str    : string[5];
    i, recSize, base : integer;
    map    : TDirEntryMap;
    leader : char;
begin
  // FCur points at the DR
  Result := FCur < FEnd;
  if Result then try
    if FRLeader = nil then begin
      // Parse the DR leader
      SetLength(str, 5);
      Move(FCur[0], str[1], 5);
      recSize := StrToInt(str);
      leader := FCur[6];
      Move(FCur[12], str[1], 5);
      base := StrToInt(str);
      map.len := StrToInt(FCur[20]);
      map.pos := StrToInt(FCur[21]);
      map.tag := StrToInt(FCur[23]);
      Inc(FCur, 24);

      ParseDirEntries(rec, map);          // FCur now points at the user data

      // Initialize FRLeader if required
      if UpCase(leader) = 'R' then begin
        SetLength(FRLeader, DREntries);
        for i := Low(FRLeader) to High(FRLeader) do begin
          FRLeader[i].Tag := rec[i].Tag;
          FRLeader[i].Length := rec[i].Length;
        end;
      end;

      Inc(FCur, recSize - base);         // FCur now points to the next DR
    end else begin
      // Copy the values over from the FRLeader to the rec parameter
      for i := Low(FRLeader) to High(FRLeader) do begin
        rec[i].Tag    := FRLeader[i].Tag;
        rec[i].Length := FRLeader[i].Length;
        rec[i].Data   := FCur;
        Inc(FCur, FRLeader[i].Length);
      end;
    end;
  except
    Result := false;                     // probably EOF
  end;
end;

procedure TISO8211.ParseFormat(ent: TDDAEntry; fmt: PChar);
var curField, numSubs, rep, i, j, width: integer;

  function Reps: integer;
  begin
    Result := 0;
    while fmt[0] in ['0'..'9'] do begin
      Result := 10 * Result + Ord(fmt[0]) - 48;
      Inc(fmt);
    end;
    if Result = 0 then Result := 1;
  end;

  procedure ParseSubFormat;
  var subRep, i: integer;
  begin
    numSubs := 0;
    while fmt[0] <> ')' do begin
      subRep := Reps;
      if fmt[0] in ['A', 'B', 'C', 'I', 'R', 'S', 'X'] then begin
        for i := 0 to subRep - 1 do
          ent.Fields[curField+numSubs+i].DataType := fmt[0];
        Inc(fmt);
        if fmt[0] in [',', ')'] then
          for i := 0 to subRep - 1 do
            ent.Fields[curField+numSubs+i].Length := 0
        else if fmt[0] = '(' then begin  // accept only a width, NOT a delimiter
          width := Reps;
          if fmt[0] <> ')' then
            raise Exception.Create('Invalid format string (delimiters not accepted)');
          for i := 0 to subRep - 1 do
            ent.Fields[curField+numSubs+i].Length := width;
          Inc(fmt);
        end;
        if fmt[0] = ',' then Inc(numSubs, subRep);
      end else raise Exception.Create('Invalid format string');
    end;
  end;

begin
  curField := 0;
  while fmt[0] <> ')' do begin
    Inc(fmt);
    rep := Reps;
    if fmt[0] = '(' then begin
      if rep = 1 then
        raise Exception.Create('Invalid format string');
      ParseSubFormat;
      for i := 2 to rep do   // the first one is already added by ParseSubFormat
        for j := 1 to numSubs do begin
          ent.Fields[curField+(i-1)*numSubs+j].DataType := ent.Fields[curField+j].DataType;
          ent.Fields[curField+(i-1)*numSubs+j].Length := ent.Fields[curField+j].Length;
        end;
      curField := curField + rep * numSubs;
    end else if fmt[0] in ['A', 'B', 'C', 'I', 'R', 'S', 'X'] then begin
      for i := 0 to rep - 1 do
        ent.Fields[curField+i].DataType := fmt[0];
      Inc(fmt);
      if fmt[0] in [',', ')'] then
        for i := 0 to rep - 1 do
          ent.Fields[curField].Length := 0
      else if fmt[0] = '(' then begin  // accept only a width, NOT a delimiter
        Inc(fmt);
        width := Reps;
        if fmt[0] <> ')' then
          raise Exception.Create('Invalid format string (delimiters not accepted)');
        for i := 0 to rep - 1 do
          ent.Fields[curField+i].Length := width;
        Inc(fmt);
      end;
      if fmt[0] = ',' then Inc(curField, rep);
    end else raise Exception.Create('Invalid format string');
  end;
end;

end.
