{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpIdrisi.pas. The Original Code is available through   }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit imports spatial data from Idrisi raster files. A header file with  }
{ the same name as the data file but with an .rdc extension must be present as }
{ well. (Currently coordinate system information is not parsed and a local     }
{ coordinate system is assigned.)                                              }
{ This driver currently only supports non-compressed binary data, RGB8 and     }
{ RGB24. Other formats (ASCII and compressed data) will be supported in a later}
{ version of this driver. RGB8 and RGB24 data are transformed into rasters of  }
{ TDSpUInt8 data type with 1 or 3 bands, respectively.                         }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 06                                             }
{                                                                              }
{ Dependencies  : DSpPFA                                                       }
{                 DSpRasterSource                                              }
{                 DSpUtils (SwapXBytes)                                        }
{                 DSpGlobals                                                   }
{                 Classes (TFileStream)                                        }
{                 SysUtils (Exception, ChangeFileExt, ExtractFileExt)          }
{                                                                              }
{ Version       : 0.2  03Dec06 Support for saving data added. SupportsSource   }
{                              and CanSave added to the broker.                }
{                 0.1  03Nov24 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpIdrisi;
{$I DSpatial.inc}
interface
uses DSpGlobals, DSpPFA;
type
  TDSpIdrisiDriver = class(TDSpBaseDriver)
  private
    FFileName : string;
    FTitle    : string;
    FBands, FCols, FRows: integer;
    FDataType : TDSpDataType;
    FBandType : TDSpRasterBandType;
    FPixelSize: TDSpFloat64;
    FNull     : TDSpFloat64;
    FULX, FULY: TDSpFloat64;
    FMin, FMax: string; // there could be one or three values
  protected
    procedure   ReadRDC;
    function    Description: string; override;
    function    DataSets: integer; override;

    function    DataSetID(i: integer): integer; override;
    function    DataSetDescription(i: integer): string; override;
    function    DataSetCoordSys(i: integer): string; override;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); override;
    function    DataSetType(i: integer): TDSpPFAFormat; override;
  public
    constructor Create(const fn: string);
    constructor Open(const fn: string);

    function    CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer; override;
    function    AddDataSet(const ds: TDSpSourceDataSet): integer; override;
    function    CloneDataSet(const ds: TDSpSourceDataSet): integer; override;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   AddRasterBand(i: integer; info: PDSpRasterbandInfo); override;
    procedure   DeleteRasterBand(i: integer); override;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); override;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); override;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); override;
  end;

  TDSpIdrisiBroker = class(TDSpBaseBroker)
  public
    constructor Create;
    function    CreateSource(const fn: string): TDSpSourceHandler; override;
    function    OpenSource(const fn: string): TDSpSourceHandler; override;
    function    SupportsSource(const source: string): boolean; override;
    function    CanSave(const ds: TDSpSourceDataSet): string; override;
  end;

implementation
uses SysUtils, Classes, DSpUtils, DSpRasterSource;

////////////////////////////////////////////////////////////////////////////////
// TDSpIdrisiDriver
constructor TDSpIdrisiDriver.Create(const fn: string);
begin
  inherited Create;
  FBands := 0;
  FFileName := fn;
  FTitle := 'New Idrisi raster';
end;

constructor TDSpIdrisiDriver.Open(const fn: string);
begin
  inherited Create;
  FBands := 1;
  FBandType := dspRasterBandOrdinal;
  FPixelSize := 1.0;
  FNull := -9E200;             // just an improbable value

  // Read the header file
  FFileName := fn;
  ReadRDC;

  // Read the projection file

end;

procedure TDSpIdrisiDriver.ReadRDC;
var hdr: TextFile;
    line, token, value: string;
    white: integer;
begin
  AssignFile(hdr, ChangeFileExt(FFileName, '.RDC'));
  try
    Reset(hdr);
    while not Eof(hdr) do begin
      ReadLn(hdr, line);
      white := Pos(':', line);
      token := UpperCase(Trim(Copy(line, 1, white-1)));
      value := UpperCase(Trim(Copy(line, white+1, 1000)));
      if token = 'COLUMNS' then FCols := StrToInt(value)
      else if token = 'ROWS' then FRows := StrToInt(value)
      else if token = 'FILE TITLE' then FTitle := value
      else if token = 'MIN. X' then FULX := StrToFloat(value)
      else if token = 'MAX. Y' then FULY := STrToFloat(value)
      else if token = 'RESOLUTION' then FPixelSize := StrToFloat(value)
      else if token = 'FLAG VALUE' then FNull := StrToFloat(value)
      else if token = 'MIN. VALUE' then FMin := value
      else if token = 'MAX. VALUE' then FMax := value
      else if token = 'DATA TYPE' then begin
        if value = 'REAL' then FDataType := dspFloat32
        else if value = 'BYTE' then FDataType := dspUInt8
        else if value = 'INTEGER' then FDataType := dspInt16
        else if value = 'RGB8' then begin
          FDataType := dspUInt8;
          FBandType := dspRasterBandImage;
        end else if value = 'RGB24' then begin
          FDataType := dspUInt8;
          FBandType := dspRasterBandImage;
          FBands := 3;
        end;
      end;
    end;
  finally
    CloseFile(hdr);
  end;

  // Check for fatal combinations and omissions
  if (FCols = 0) or (FRows = 0) or FloatsSame(FPixelSize, 0) then begin
    ; // show parameter dialog
  end;
end;

function TDSpIdrisiDriver.Description: string;
begin
  if FTitle <> '' then Result := FTitle
  else Result := 'Idrisi raster file: ' + FFileName;
end;

function TDSpIdrisiDriver.DataSets: integer;
begin
  Result := 1;
end;

function TDSpIdrisiDriver.DataSetID(i: integer): integer;
begin
  Result := 0;
end;

function TDSpIdrisiDriver.DataSetDescription(i: integer): string;
begin
  if FTitle <> '' then Result := FTitle
  else Result := ExtractFileName(FFileName);
end;

function TDSpIdrisiDriver.DataSetCoordSys(i: integer): string;
begin
  Result := 'LOCAL_CS["Local coordinate system (meter)",LOCAL_DATUM["Local datum",0],UNIT["Meter",1.0],AXIS["X",EAST],AXIS["Y",NORTH]]';  // <<<<<<<<<<<< projection not yet supported
end;

procedure TDSpIdrisiDriver.DataSetExtent(i: integer; const ext: TDSpExtent);
begin
  ext.North := FULY;
  ext.West  := FULX;
  ext.East  := ext.West + FCols * FPixelSize;
  ext.South := ext.North - FRows * FPixelSize;
end;

function TDSpIdrisiDriver.DataSetType(i: integer): TDSpPFAFormat;
begin
  Result := dspPFARaster;
end;

function TDSpIdrisiDriver.CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer;
begin
  Assert(false, 'Idrisi driver does not support writing new data sets');
  Result := -1;
end;

function TDSpIdrisiDriver.AddDataSet(const ds: TDSpSourceDataSet): integer;
var fs: TFileStream;
    rdc: TextFile;
    data: pointer;
    dtype: string;
begin
  Assert(ds is TDSpRasterSource);
  Result := CloneDataSet(ds);
  if (Result = -1) or (FBands = 0) then Exit;
  Result := 0;

  try
    // Create the data file
    fs := TFileStream.Create(FFileName, fmCreate or fmShareExclusive);
    try
      data := DSpKernelReadRasterBand(ds, 0);
      fs.Write(data^, FRows * FCols * DataSize(FDataType));
      FreeMem(data);
    finally
      fs.Free;
    end;

    // Create the RDC
    AssignFile(rdc, ChangeFileExt(FFileName, '.rdc'));
    try
      Rewrite(rdc);
      WriteLn(rdc, 'file format : IDRISI Raster A.1');
      WriteLn(rdc, Format('file title  : %s', [FTitle]));

      if FDataType = dspFloat32 then dtype := 'real'
      else if FDataType = dspUInt8 then dtype := 'byte'
      else if FDataType = dspInt16 then dtype := 'integer'
      else raise Exception.Create('Wrong data type for Idrisi raster');
      WriteLn(rdc, Format('data type   : %s', [dtype]));

      WriteLn(rdc, 'file type   : binary');
      WriteLn(rdc, Format('columns     : %d', [FCols]));
      WriteLn(rdc, Format('rows        : %d', [FRows]));
      WriteLn(rdc, 'ref. system : plane');
      WriteLn(rdc, 'ref. units  : meters');
      WriteLn(rdc, 'unit dist.  : 1.0000000');
      WriteLn(rdc, Format('min. X      : %.7f', [FULX]));
      WriteLn(rdc, Format('max. X      : %.7f', [FULX + FCols * FPixelSize]));
      WriteLn(rdc, Format('min. Y      : %.7f', [FULY - FRows * FPixelSize]));
      WriteLn(rdc, Format('max. Y      : %.7f', [FULY]));
      WriteLn(rdc, 'pos''n error : unknown');
      WriteLn(rdc, Format('resolution  : %.7f', [FPixelSize]));
      WriteLn(rdc, Format('min. value  : %s', [FMin]));
      WriteLn(rdc, Format('max. value  : %s', [FMax]));
      WriteLn(rdc, Format('display min : %s', [FMin]));
      WriteLn(rdc, Format('display max : %s', [FMax]));
      WriteLn(rdc, 'value units : unspecified');
      WriteLn(rdc, 'value error : unknown');
      WriteLn(rdc, Format('flag value  : %.7f', [FNull]));
      WriteLn(rdc, 'flag def''n  : Background');
    finally
      CloseFile(rdc);
    end;
  except
    Result := -1; // Flag failure on any error
  end;
end;

function TDSpIdrisiDriver.CloneDataSet(const ds: TDSpSourceDataSet): integer;
begin
  Assert(ds is TDSpRasterSource);

  // An Idrisi raster can have only 1 data set, having only 1 band (RGB24 not
  // supported for writing), so fail when the source already has a data set, or
  // when ds has multiple bands.
  if (FBands > 0) or (TDSpRasterSource(ds).Bands > 1) then begin
    Result := -1;
    Exit;
  end;

  with ds as TDSpRasterSource do begin
    FTitle := Description;
    FCols := Columns;
    FRows := Rows;
    FPixelSize := PixelSize;
    FULY := Extent.North;
    FULX := Extent.West;
    if Bands = 1 then begin
      FBands := 1;
      FBandType := Band[0].BandType;
      FDataType := Band[0].DataType;
      FNull := MissingValue[0];
      FMin := FloatToStrF(MinValue[0], ffFixed, 18, 7); // Should check if values are set
      FMax := FloatToStrF(MaxValue[0], ffFixed, 18, 7);
    end;
  end;
  Result := 0;
end;

procedure TDSpIdrisiDriver.AddRasterBand(i: integer; info: PDSpRasterbandInfo);
begin
  info^.BandIndex := -1;
end;

procedure TDSpIdrisiDriver.DeleteRasterBand(i: integer);
begin
  // Just do nothing
end;

procedure TDSpIdrisiDriver.GetRasterInfo(i: integer; info: PDSpRasterInfo);
begin
  if FBands = 1 then info^.BandFormat := dspRasterFlat
  else info^.BandFormat := dspRasterInterleavePixel;
  info^.Bands := FBands;
  info^.Rows := FRows;
  info^.Columns := FCols;
  info^.PixelSize := FPixelSize;
end;

procedure TDSpIdrisiDriver.GetRasterBandInfo(i: integer; info: PDSpRasterInfo);
var b, c: integer;
    minimum, maximum: array [0..2] of TDSpFloat64;
begin
  if FBands = 0 then Exit;

  // Estiblish band minimum and maximum values
  if FBands = 1 then begin
    minimum[0] := StrToFloat(FMin);
    maximum[0] := StrToFloat(FMax);
  end else begin
    // RGB24, FMin and FMax hold three values each
    b := Pos(' ', FMin);
    c := Pos(' ', Copy(FMin, b+1, 80));
    minimum[0] := STrToFloat(Copy(FMin, 1, b-1));
    minimum[1] := STrToFloat(Copy(FMin, b+1, c-1));
    minimum[2] := STrToFloat(Copy(FMin, c+1, 80));
    b := Pos(' ', FMax);
    c := Pos(' ', Copy(FMax, b+1, 80));
    maximum[0] := STrToFloat(Copy(FMax, 1, b-1));
    maximum[1] := STrToFloat(Copy(FMax, b+1, c-1));
    maximum[2] := STrToFloat(Copy(FMax, c+1, 80));
  end;

  for b := 0 to High(info^.Band) do
    with info^.Band[b] do begin
      BandIndex := b;
      BandType := FBandType;
      DataType := FDataType;
      MissingValue := FNull;
      Min := minimum[b];
      Max := maximum[b];
    end;
end;

procedure TDSpIdrisiDriver.ReadRasterBand(dataSet, band: integer; const into: pointer);
// For now only reads flat binary and RGB8/24 files
var r, c: integer;
    data: TFileStream;
    p, pix, row: PDSpUInt8;
begin
  if FBands = 0 then Exit;
  Assert((dataSet = 0) and (band < FBands));

  data := TFileStream.Create(FFileName, fmOpenRead or fmShareDenyWrite);
  try
    if FBands = 1 then data.Read(into^, FRows * FCols * DataSize(FDataType))
    else begin
      // Read RGB24 data
      GetMem(row, FCols * 3);
      try
        p := into;
        for r := 0 to FRows - 1 do begin
          data.Read(row^, FCols * 3);
          pix := row;
          Inc(pix, band);
          for c := 0 to FCols - 1 do begin
            p^ := pix^;
            Inc(p);
            Inc(pix, 3);
          end;
        end;
      finally
        FreeMem(row);
      end;
    end;
  finally
    data.Free;
  end;
end;

procedure TDSpIdrisiDriver.ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

procedure TDSpIdrisiDriver.ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpIdrisiBroker
constructor TDSpIdrisiBroker.Create;
begin
  inherited;
  FDescription := 'Idrisi raster files';
  FFilter := '*.RST';
  FCaps := [dspReadRaster, dspWriteRaster];
end;

function TDSpIdrisiBroker.CreateSource(const fn: string): TDSpSourceHandler;
var r: TDSpIdrisiDriver;
begin
  r := TDSpIdrisiDriver.Create(fn);
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

function TDSpIdrisiBroker.OpenSource(const fn: string): TDSpSourceHandler;
var r: TDSpIdrisiDriver;
begin
  r := TDSpIdrisiDriver.Open(fn);
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

function TDSpIdrisiBroker.SupportsSource(const source: string): boolean;
begin
  Result := UpperCase(ExtractFileExt(source)) = '.RST';
end;

function TDSpIdrisiBroker.CanSave(const ds: TDSpSourceDataSet): string;
begin
  if (ds is TDSpRasterSource) and
     (TDSpRasterSource(ds).Bands = 1) and
     (TDSpRasterSource(ds).Band[0].DataType in [dspUInt8, dspInt16, dspFloat32]) then
    Result := 'Idrisi raster|*.RST' else Result := '';
end;

initialization
{ Enable this method for static loading of the broker. Please note that the    }
{ DSpPFA unit MUST be loaded before this unit. If the broker is to be loaded   }
{ dynamically, you will have to provide a DLL with this unit.                  }
  PFA.RegisterBroker(TDSpIdrisiBroker.Create);
end.
