{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRawBinary.pas. The Original Code is available through}
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit imports spatial data from raw binary files. Data files can be in   }
{ BIL, BIP, or BSQ format, or an ESRI binary raster file. DEM files are also   }
{ supported and considered to contain a single band with elevation data. A     }
{ header file with the same name as the data file but with an .hdr extension   }
{ must be present as well. The driver will look for an optional third file with}
{ a .prj extension as well, containing coordinate system information in either }
{ WKT or ArcInfo format. In the absence of this latter file a local coordinate }
{ system is assigned to the data set. (Currently coordinate system information }
{ is not parsed and a local coordinate system is assigned.)                    }
{ This driver supports 8, 16 and 32 bit data. Eight bit data is interpreted as }
{ dspUInt8, 16 and 32 bit as dspInt16/32. ESRI rasters are always single band  }
{ dspFloat32.                                                                  }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 06                                             }
{                                                                              }
{ Dependencies  : DSpPFA                                                       }
{                 DSpUtils (SwapXBytes)                                        }
{                 DSpGlobals                                                   }
{                 Classes (TFileStream)                                        }
{                 SysUtils (Exception, ChangeFileExt, ExtractFileExt)          }
{                                                                              }
{ Version       : 0.4  03Dec06 Support for saving data added. SupportsSource   }
{                              and CanSave added to the broker.                }
{                 0.3  03Nov24 Added support for ESRI binary rasters.          }
{                 0.2  03Nov07 PFA overhaul                                    }
{                 0.1  03Aug24 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpRawBinary;
{$I DSpatial.inc}
interface
uses DSpGlobals, DSpPFA;
type
  TDSpRawDriver = class(TDSpBaseDriver)
  private
    FFileName: string;
    FOrder: char; // I(ntel) or M(otorola)
    FLayout: string[3]; // BIL, BIP, BSQ
    FBands, FCols, FRows: integer;
    FBits, FSkip: integer;
    FDataType: TDSpDataType;
    FBandRow, FTotalRow, FBandGap: integer;
    FPixelSize, FDY: TDSpFloat64;
    FNull: TDSpFloat64;
    FULX, FULY: TDSpFloat64;
  protected
    procedure   ReadHDR;
    procedure   ReadFLTHDR;
    procedure   WriteHDR;
    procedure   WriteFLTHDR;

    function    Description: string; override;
    function    DataSets: integer; override;

    function    DataSetID(i: integer): integer; override;
    function    DataSetDescription(i: integer): string; override;
    function    DataSetCoordSys(i: integer): string; override;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); override;
    function    DataSetType(i: integer): TDSpPFAFormat; override;
  public
    constructor Create(const fn: string);
    constructor Open(const fn: string);

    function    CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer; override;
    function    AddDataSet(const ds: TDSpSourceDataSet): integer; override;
    function    CloneDataSet(const ds: TDSpSourceDataSet): integer; override;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   AddRasterBand(i: integer; info: PDSpRasterbandInfo); override;
    procedure   DeleteRasterBand(i: integer); override;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); override;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); override;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); override;
  end;

  TDSpRawBroker = class(TDSpBaseBroker)
  public
    constructor Create;
    function    CreateSource(const fn: string): TDSpSourceHandler; override;
    function    OpenSource(const fn: string): TDSpSourceHandler; override;
    function    SupportsSource(const source: string): boolean; override;
    function    CanSave(const ds: TDSpSourceDataSet): string; override;
  end;

implementation
uses SysUtils, Classes, DSpUtils, DSpRasterSource;

////////////////////////////////////////////////////////////////////////////////
// TDSpRawDriver
constructor TDSpRawDriver.Create(const fn: string);
begin
  inherited Create;
  FOrder := 'I';                // set defaults
  FLayout := 'BIL';
  FBands := 0;
  FBits := 8;
  FPixelSize := 1.0;
  FNull := -9E200;             // just an improbable value
  FFileName := fn;
end;

constructor TDSpRawDriver.Open(const fn: string);
var ext: string;
begin
  Create(fn);

  // Read the header
  ext := UpperCase(ExtractFileExt(FFileName));
  if ext = '.FLT' then ReadFLTHDR else ReadHDR;

  // Read the projection file

end;

procedure TDSpRawDriver.ReadHDR;
var hdr: TextFile;
    line, token, value: string;
    white: integer;
begin
  AssignFile(hdr, ChangeFileExt(FFileName, '.hdr'));
  try
    Reset(hdr);
    while not Eof(hdr) do begin
      ReadLn(hdr, line);
      white := Pos(' ', line);
      if white = 0 then white := Pos(#9, line);
      token := UpperCase(Copy(line, 1, white-1));
      value := UpperCase(Trim(Copy(line, white+1, 1000)));

      if token = 'NROWS' then FRows := StrToInt(value)
      else if token = 'NCOLS' then FCols := StrToInt(value)
      else if token = 'NBANDS' then FBands := StrToInt(value)
      else if token = 'NBITS' then FBits := StrToInt(value)
      else if token = 'BYTEORDER' then FOrder := value[1]
      else if token = 'LAYOUT' then FLayout := value
      else if token = 'SKIPBYTES' then FSkip := StrToInt(value)
      else if token = 'ULXMAP' then FULX := StrToFloat(value)
      else if token = 'ULYMAP' then FULY := StrToFloat(value)
      else if token = 'XDIM' then FPixelSize := StrToFloat(value)
      else if token = 'YDIM' then FDY := StrToFloat(value)
      else if token = 'BANDROWBYTES' then FBandRow := StrToInt(value)
      else if token = 'TOTALROWBYTES' then FTotalRow := StrToInt(value)
      else if token = 'BANDGAPBYTES' then FBandGap := StrToInt(value)
      else if token = 'NODATA' then FNull := StrToFloat(value);
    end;
  finally
    CloseFile(hdr);
  end;

  // Check for fatal combinations??? FPixelSize <> FDY???

  // Some derived and updated parameters
  if FBits <= 8 then FDataType := dspUInt8
  else if FBits = 16 then FDataType := dspInt16 else FDataType := dspInt32;
  FULY := FULY + 0.5 * FPixelSize;     // correct to edge of pixel
  FULX := FULX - 0.5 * FPixelSize;
end;

procedure TDSpRawDriver.ReadFLTHDR;
var hdr: TextFile;
    line, token, value: string;
    white: integer;
begin
  AssignFile(hdr, ChangeFileExt(FFileName, '.hdr'));
  try
    Reset(hdr);
    while not Eof(hdr) do begin
      ReadLn(hdr, line);
      white := Pos(' ', line);
      if white = 0 then white := Pos(#9, line);
      token := UpperCase(Copy(line, 1, white-1));
      value := UpperCase(Trim(Copy(line, white+1, 1000)));
      if token = 'NCOLS' then FCols := StrToInt(value)
      else if token = 'NROWS' then FRows := StrToInt(value)
      else if token = 'XLLCORNER' then FULX := StrToFloat(value)
      else if token = 'YLLCORNER' then FULY := STrToFloat(value)  // must correct
      else if token = 'CELLSIZE' then FPixelSize := StrToFloat(value)
      else if token = 'NODATA_VALUE' then FNull := StrToFloat(value)
      else if token = 'BYTEORDER' then
        if UpperCase(value) = 'LSBFIRST' then FOrder := 'I' else FOrder := 'M';
    end;
    FBands := 1;
  finally
    CloseFile(hdr);
  end;

  // Check for fatal combinations and omissions
  if (FCols = 0) or (FRows = 0) or FloatsSame(FPixelSize, 0) then begin
    ; // show parameter dialog
  end else FULY := FULY + FCols * FPixelSize;

  // Some fixed parameters
  FBits := 32;
  FDataType := dspFloat32;
end;

function TDSpRawDriver.Description: string;
begin
  Result := 'Raw binary file: ' + FFileName;
end;

function TDSpRawDriver.DataSets: integer;
begin
  Result := 1;
end;

function TDSpRawDriver.DataSetID(i: integer): integer;
begin
  Result := 0;
end;

function TDSpRawDriver.DataSetDescription(i: integer): string;
begin
  Result := ExtractFileName(FFileName);
end;

function TDSpRawDriver.DataSetCoordSys(i: integer): string;
begin
  Result := 'LOCAL_CS["Local coordinate system (meter)",LOCAL_DATUM["Local datum",0],UNIT["Meter",1.0],AXIS["X",EAST],AXIS["Y",NORTH]]';  // <<<<<<<<<<<< projection not yet supported
end;

procedure TDSpRawDriver.DataSetExtent(i: integer; const ext: TDSpExtent);
begin
  ext.North := FULY;
  ext.West  := FULX;
  ext.East  := ext.West + FCols * FPixelSize;
  ext.South := ext.North - FRows * FPixelSize;
end;

function TDSpRawDriver.DataSetType(i: integer): TDSpPFAFormat;
begin
  Result := dspPFARaster;
end;

function TDSpRawDriver.CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer;
begin
  Assert(false, 'Idrisi driver does not support writing new data sets');
  Result := -1;
end;

function TDSpRawDriver.AddDataSet(const ds: TDSpSourceDataSet): integer;
var fs: TFileStream;
    data: pointer;
    size: integer;
begin
  Assert(ds is TDSpRasterSource);
  Result := CloneDataSet(ds);
  if (Result = -1) or (FBands = 0) then Exit;
  Result := 0;

  try
    // Create the data file: simply write all bands sequentially
    size := FRows * FCols * DataSize(FDataType);
    fs := TFileStream.Create(FFileName, fmCreate or fmShareExclusive);
    try
      data := DSpKernelReadRasterBand(ds, 0);
      fs.Write(data^, size);
      FreeMem(data);
    finally
      fs.Free;
    end;

    if UpperCase(ExtractFileExt(FFileName)) = '.FLT' then WriteFLTHDR
    else WriteHDR;
  except
    Result := -1; // Flag failure on any error
  end;
end;

procedure TDSpRawDriver.WriteHDR;
var hdr: TextFile;
begin
  AssignFile(hdr, ChangeFileExt(FFileName, '.hdr'));
  try
    Rewrite(hdr);
    WriteLn(hdr, 'BYTEORDER         I');
    WriteLn(hdr, 'LAYOUT            ' + FLayout);
    WriteLn(hdr, Format('NROWS             %d', [FRows]));
    WriteLn(hdr, Format('NCOLS             %d', [FCols]));
    WriteLn(hdr, Format('NBANDS            %d', [FBands]));
    WriteLn(hdr, Format('NBITS             %d', [FBits]));
    WriteLn(hdr, Format('BANDROWBYTES      %d', [FCols * FBits div 8]));
    WriteLn(hdr, Format('TOTALROWBYTES     %d', [FCols * FBits div 8]));
    WriteLn(hdr, 'BANDGAPBYTES      0');
    WriteLn(hdr, Format('XDIM              %.8f', [FPixelSize]));
    WriteLn(hdr, Format('YDIM              %.8f', [FPixelSize]));
    WriteLn(hdr, Format('ULXMAP            %.8f', [FULX]));
    WriteLn(hdr, Format('ULYMAP            %.8f', [FULY]));
    WriteLn(hdr, Format('NODATA            %d', [Round(FNull)]));
  finally
    CloseFile(hdr);
  end;
end;

procedure TDSpRawDriver.WriteFLTHDR;
var hdr: TextFile;
begin
  AssignFile(hdr, ChangeFileExt(FFileName, '.hdr'));
  try
    Rewrite(hdr);
    WriteLn(hdr, Format('ncols         %d', [FCols]));
    WriteLn(hdr, Format('nrows         %d', [FRows]));
    WriteLn(hdr, Format('xllcorner     %.8f', [FULX]));
    WriteLn(hdr, Format('yllcorner     %.8f', [FULY - FCols * FPixelSize]));
    WriteLn(hdr, Format('cellsize      %.8f', [FPixelSize]));
    WriteLn(hdr, Format('NODATA_value  %.8f', [FNull]));
    WriteLn(hdr, 'byteorder     LSBFIRST');
  finally
    CloseFile(hdr);
  end;
end;

function TDSpRawDriver.CloneDataSet(const ds: TDSpSourceDataSet): integer;
begin
  Assert(ds is TDSpRasterSource);

  // An raw binary raster can have only 1 data set, having 1 or more bands, each
  // having the same data type, so fail when the source already has a data set.
  // Assume that all bands have the same data type.
  if FBands > 0 then begin
    Result := -1;
    Exit;
  end else Result := 0;

  with ds as TDSpRasterSource do begin
    FCols := Columns;
    FRows := Rows;
    FPixelSize := PixelSize;
    FULY := Extent.North;
    FULX := Extent.West;
    FBands := Bands;
    if FBands = 1 then FLayout := 'BIL' else FLayout := 'BSQ';
    FDataType := Band[0].DataType;
    FNull := MissingValue[0];

    if FDataType = dspFloat32 then FBits := 32
    else if FDataType = dspUInt8 then FBits := 8
    else if FDataType = dspInt16 then FBits := 16
    else Result := -1;
  end;
end;

procedure TDSpRawDriver.AddRasterBand(i: integer; info: PDSpRasterbandInfo);
begin
  info^.BandIndex := -1;
end;

procedure TDSpRawDriver.DeleteRasterBand(i: integer);
begin
  // Just do nothing
end;

procedure TDSpRawDriver.GetRasterInfo(i: integer; info: PDSpRasterInfo);
begin
  if (FBands = 1) or (FLayout = 'BSQ') then info^.BandFormat := dspRasterFlat
  else if FLayout = 'BIL' then info^.BandFormat := dspRasterInterleaveLine
  else info^.BandFormat := dsprasterInterleavePixel;
  info^.Bands := FBands;
  info^.Rows := FRows;
  info^.Columns := FCols;
  info^.PixelSize := FPixelSize;
end;

procedure TDSpRawDriver.GetRasterBandInfo(i: integer; info: PDSpRasterInfo);
var b: integer;
begin
  for b := 0 to High(info^.Band) do
    with info^.Band[b] do begin
      BandIndex := b;
      if UpperCase(ExtractFileExt(FFileName)) = '.DEM' then BandType := dspRasterBandSurface
      else BandType := dspRasterBandOrdinal;
      DataType := FDataType;
      MissingValue := FNull;
      Min := 0;
      Max := -1;
    end;
end;

procedure TDSpRawDriver.ReadRasterBand(dataSet, band: integer; const into: pointer);
// For now only reads multiple byte sized data: 8, 16, 32 bit
var r, c, rowSize, len, skip: integer;
    data: TFileStream;
    p: pointer;
begin
  Assert((dataSet = 0) and (band < FBands));
  Assert(FBits >= 8);

  data := TFileStream.Create(FFileName, fmOpenRead or fmShareDenyWrite);
  len := FBits shr 3; // len is length in bytes of pixel DOES NOT WORK FOR BINARY OR 4-BIT DATA
  rowSize := FCols * len;
  p := into;
  try
    data.Seek(FSkip, soFromBeginning);
    if (FBands = 1) or (FLayout = 'BSQ') then begin
      // Flat file, just move the block
      if band > 0 then data.Seek(band * FRows * rowSize + FBandGap, soFromCurrent);
      data.Read(p^, FRows * rowSize);
    end else if FLayout = 'BIL' then begin
      // Read BIL, DEM, or FLT data
      if FBandRow = 0 then FBandRow := rowSize;
      if FTotalRow > 0 then skip := FTotalRow - (FBands - 1) * FBandRow else skip := 0;
      if band > 0 then data.Seek(band * FBandRow, soFromCurrent);
      for r := 0 to FRows - 1 do begin
        if data.Read(p^, rowSize) <> rowSize then
          raise Exception.Create('Unexpected end-of-file encountered');
        data.Seek(skip, soFromCurrent);    // skip over the other bands
        p := pointer(integer(p) + rowSize);
      end;
    end else begin
      // Read BIP data
      if FTotalRow > 0 then skip := FTotalRow - FBands * FCols * (FBits shr 3) else skip := 0;
      rowSize := (FBands - 1) * len;  // rowSize is now size of interleaved pixel spacing
      if band > 0 then data.Seek(band * len, soFromCurrent);
      for r := 0 to FRows - 1 do begin
        for c := 0 to FCols - 1 do begin
          if data.Read(p^, len) <> len then
            raise Exception.Create('Unexpected end-of-file encountered');
          data.Seek(rowSize, soFromCurrent);
          p := pointer(integer(p) + len);
        end;
        data.Seek(skip, soFromCurrent);
      end;
    end;
  finally
    data.Free;
  end;

  if FOrder = 'M' then begin
    if FBits = 16 then Swap2Bytes(into^, FRows * FCols)
    else if FBits = 32 then Swap4Bytes(into^, FRows * FCols);
  end;
end;

procedure TDSpRawDriver.ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

procedure TDSpRawDriver.ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRawBroker
constructor TDSpRawBroker.Create;
begin
  inherited;
  FDescription := 'Raw binary files';
  FFilter := '*.BIL;*.BIP;*.BSQ;*.DEM;*.FLT';
  FCaps := [dspReadRaster, dspWriteRaster];
end;

function TDSpRawBroker.CreateSource(const fn: string): TDSpSourceHandler;
var r: TDSpRawDriver;
begin
  r := TDSpRawDriver.Create(fn);
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

function TDSpRawBroker.OpenSource(const fn: string): TDSpSourceHandler;
var r: TDSpRawDriver;
begin
  r := TDSpRawDriver.Open(fn);
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

function TDSpRawBroker.SupportsSource(const source: string): boolean;
var ext: string;
begin
  ext := UpperCase(ExtractFileExt(source));
  Result := (ext = '.BIL') or (ext = '.BIP') or (ext = '.BSQ') or (ext = '.DEM') or (ext = '.FLT');
end;

function TDSpRawBroker.CanSave(const ds: TDSpSourceDataSet): string;
var i, bands: integer;
    dt: TDSpDataType;
begin
  Result := '';    // Always be negative
  if (ds is TDSpRasterSource) then begin
    bands := TDSpRasterSource(ds).Bands;
    if bands = 0 then Exit;
    if bands = 1 then begin
      // Single band formats
      if TDSpRasterSource(ds).Band[0].DataType in [dspUInt8, dspInt16, dspInt32] then
        Result := 'Raw binary (single band, BIL)|*.bil'
      else if TDSpRasterSource(ds).Band[0].DataType = dspFloat32 then
        Result := 'ESRI binary grid|*.flt';
    end else begin
      // Multi band formats: all bands must have same data type
      dt := TDSpRasterSource(ds).Band[0].DataType;
      if dt in [dspUInt8, dspInt16, dspInt32] then
        for i := 1 to bands - 1 do
          if TDSpRasterSource(ds).Band[i].DataType <> dt then Exit;
      Result := 'Raw binary (multiple bands, BSQ)|*.bsq';
    end;
  end;
end;

initialization
{ Enable this method for static loading of the broker. Please note that the    }
{ DSpPFA unit MUST be loaded before this unit. If the broker is to be loaded   }
{ dynamically, you will have to provide a DLL with this unit.                  }
  PFA.RegisterBroker(TDSpRawBroker.Create);
end.
