{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpSDTS.pas. The Original Code is available through the }
{ home page of DSpatial at http://dspatial.sourceforge.net.                    }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements a reader for the Spatial Data Transfer Standard data    }
{ transfers. The reader does not provide any specific logic to interpret the   }
{ contents of the transfer. This logic is implemented through profiles, which  }
{ should be implemented as classes inheriting from TSDTStransfer. This reader  }
{ is for sequential record access only. SDTS is a transfer specification, and  }
{ it is not suitable as a format for analysis or random access. This is a      }
{ read-only format.                                                            }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 July 18                                                 }
{                                                                              }
{ Dependencies  : * DSpISO8211 (essential)                                     }
{                 * DSpUtils (byte swapping)                                   }
{                 * SysUtils (Exception)                                       }
{                                                                              }
{ Version       : 0.4  03Jul18 Added interface calls                           }
{                 0.3          RDEF/LDEF/CELL dependencies now properly        }
{                              implemented. CELL record parsing deferred to    }
{                              request from implementor is made.               }
{                 0.2  03Apr20 All modules are now parsed based on DDA info.   }
{                              Data Quality modules not yet supported.         }
{                 0.1  03Apr17 Initial coding. Does not yet support all modules}
{                              nor are all modules properly implemented.       }
{                                                                              }
{ Limitations   : * Reads RPE profiles only                                    }
{                 * Default values not yet used                                }
{                 * Data quality modules not yet supported                     }
{                 * Attributes not yet supported                               }
{                 * Non-geospatial raster dimensions not supported             }
{                 * Compression not supported                                  }
{                                                                              }
{******************************************************************************}
unit DSpSDTS;
interface
uses SysUtils, DSpISO8211;

type
  // TSDTS module records. These records are filled by the Read.... methods.
  // Changing any of these records will likely break the methods, since these
  // rely on fixed offsets into the records.
  TSDTS_EREF = record                       // external reference
                 MODN : string;   //  0        used in several other structures
                 RCID : integer;  //  4
               end;

  TSDTS_SADR = record                       // spatial address
                 X    : double;   //  0        allow at most 6 dimensions
                 Y    : double;   //  8        used in several other structures
                 Z    : double;   // 16
                 DIM4 : double;   // 24     )
                 DIM5 : double;   // 32     ) These could be anything, depending
                 DIM6 : double;   // 40     )   on their format specifier
               end;

  TSDTS_IDEN = record
                 MODN : string;         //  0
                 RCID : integer;        //  4
                 STID : string;         //  8
                 STVS : string;         // 12
                 DOCU : string;         // 16
                 PRID : string;         // 20
                 PRVS : string;         // 24
                 PDOC : string;         // 28
                 TITL : string;         // 32
                 DAID : string;         // 36
                 DAST : string;         // 40
                 MPDT : string;         // 44
                 DCDT : string;         // 48
                 SCAL : integer;        // 52
                 COMT : string;         // 56
                 FFYN : string;         // 60       // CONF
                 VGYN : string;         // 64
                 GTYN : string;         // 68
                 RCYN : string;         // 72
                 EXSP : integer;        // 76
                 FTLV : integer;        // 80
                 CDLV : integer;        // 84
                 NGDM : string;         // 88
                 ATID : array of TSDTS_EREF;  // 92   ATID
               end;

  TSDTS_CATD = record             
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 NAME : string;   //  8
                 TYP_ : string;   // 12
                 VOLM : string;   // 16
                 FIL_ : string;   // 20
                 RECD : string;   // 24
                 EXTR : string;   // 28   
                 MVER : string;   // 32
                 COMT : string;   // 36
                 NREC : integer;  // 40 STAT.NREC
                 NSAD : integer;  // 44 STAT.NSAD
                 RECDtype : string[3]; // I|A
               end;

  TSDTS_CATX = record
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 NAM1 : string;   //  8
                 TYP1 : string;   // 12
                 NAM2 : string;   // 16
                 TYP2 : string;   // 20
                 COMT : string;   // 24
               end;

  TSDTS_CATS = record
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 NAME : string;   //  8
                 TYP_ : string;   // 12
                 DOMN : string;   // 16
                 MAP  : string;   // 20
                 THEM : string;   // 24
                 AGOB : string;   // 28
                 AGTP : string;   // 32
                 COMT : string;   // 36
               end;

  TSDTS_IREF = record
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 COMT : string;   //  8
                 SATP : string;   // 12
                 XLBL : string;   // 16
                 YLBL : string;   // 20
                 HFMT : string;   // 24
                 VFMT : string;   // 28
                 SFAX : double;   // 32
                 SFAY : double;   // 40
                 SFAZ : double;   // 48
                 XORG : double;   // 56
                 YORG : double;   // 64
                 ZORG : double;   // 72
                 XHRS : double;   // 80
                 YHRS : double;   // 88
                 VRES : double;   // 96
                 DMID : array of TSDTS_EREF; // 104
               end;

  TSDTS_VATT = record
                 VDAT : string;   //  0
                 VEM  : string;   //  4
                 ATLB : string;   //  8
                 AUTH : string;   // 12
               end;

  TSDTS_SATT = record
                 SDAT : string;   //  0
                 SEM  : string;   //  4
                 ATLB : string;   //  8
                 AUTH : string;   // 12
               end;

  TSDTS_XREF = record
                 MODN : string;              //  0
                 RCID : integer;             //  4
                 COMT : string;              //  8
                 RDOC : string;              // 12
                 RSNM : string;              // 16
                 VDAT : string;              // 20
                 SDAT : string;              // 24
                 HDAT : string;              // 28
                 ZONE : string;              // 32
                 PROJ : string;              // 36
                 ATID : array of TSDTS_EREF; // 40
                 VATT : array of TSDTS_VATT; // 44
                 SATT : array of TSDTS_SATT; // 48
               end;

  TSDTS_SPDM = record
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 DTYP : string;   //  8
                 DSTP : string;   // 12
                 COMT : string;   // 16
                 DMSA : array of TSDTS_SADR; // 20
               end;

  TSDTS_DDDF = record
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 EORA : string;   //  8
                 EALB : string;   // 12
                 SRCE : string;   // 16
                 DFIN : string;   // 20
                 AUTH : string;   // 24
                 ADSC : string;   // 28
               end;

  TSDTS_DDOM = record
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 ATLB : string;   //  8
                 AUTH : string;   // 12
                 ATYP : string;   // 16
                 ADVF : string;   // 20
                 ADMU : string;   // 24
                 RAVA : string;   // 28
                 DVAL : double;   // 32
                 DVDF : string;   // 40
               end;

  TSDTS_DDSH = record
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 NAME : string;   //  8
                 TYP_ : string;   // 12
                 ETLB : string;   // 16
                 EUTH : string;   // 20
                 ATLB : string;   // 24
                 AUTH : string;   // 28
                 FMT  : string;   // 32
                 UNI_ : string;   // 36
                 PREC : double;   // 40
                 MXLN : integer;  // 48
                 KEY  : string;   // 52
               end;

  TSDTS_RDXT = record
                 DEXT : integer;
                 DSCO : string;
               end;

  TSDTS_RSDF = record
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 OBRP : string;   //  8
                 CSCD : string;   // 12
                 AQMD : string;   // 16
                 AQDT : string;   // 20
                 COMT : string;   // 24
                 DEFI : string;   // 28
                 CMPR : string;   // 32
                 CMMD : string;   // 36
                 DCOM : string;   // 40    <<<<<<<<<<<<<<<<<<< WRONG
                 METH : string;   // 44
                 RWXT : integer;  // 48
                 CLXT : integer;  // 52
                 PLXT : integer;  // 56
                 SCOR : string;   // 60
                 SCPT : string;   // 64
                 TIDX : string;   // 68
                 TIFT : string;   // 72
                 TIDS : string;   // 76
                 ALTN : integer;  // 80
                 FSCN : string;   // 84
                 ASPR : double;   // 88
                 NLAY : integer;  // 96
                 ISID : TSDTS_EREF;          // 100   [M](N)
                 RDXT : array of TSDTS_RDXT; // 108
                 SADR : TSDTS_SADR;          // 112   [M](N)
                 XXLB : array of double;     // 160
                 YXLB : array of double;     // 164
                 ZXLB : array of double;     // 168
                 LYID : array of TSDTS_EREF; // 172
                 RATP : array of TSDTS_EREF; // 176
                 CPID : array of TSDTS_EREF; // 180
               end;

  TSDTS_LDXT = record
                 DEXT : integer;
                 SODM : integer;
                 DTOO : integer;
                 DINR : string;
               end;

  TSDTS_LDEF = record
                 MODN : string;   //  0
                 RCID : integer;  //  4
                 CMNM : string;   //  8
                 LLBL : string;   // 12
                 CODE : string;   // 16
                 BMSK : string;   // 20
                 NROW : integer;  // 24
                 NCOL : integer;  // 28
                 NPLA : integer;  // 32
                 SORI : integer;  // 36
                 SOCI : integer;  // 40
                 SOPI : integer;  // 44
                 RWOO : integer;  // 48
                 CLOO : integer;  // 52
                 PLOO : integer;  // 56
                 INTR : string;   // 60
                 COMT : string;   // 64
                 LDXT : array of TSDTS_LDXT; // 68
                 LATP : array of TSDTS_EREF; // 72
               end;

  TSDTS_CELL = record
                 MODN : string;              //  0
                 RCID : integer;             //  4
                 NCEL : integer;             //  8
                 ROWI : integer;             // 12
                 COLI : integer;             // 16
                 PLAI : integer;             // 20
                 TIND : double;              // 24 I|B -> RSDF.TIFT
                 DNDX : array of integer;    // 32
                 SADR : TSDTS_SADR;          // 36 (SizeOf(TSDTS_SADR) = 48)
                 ATID : array of TSDTS_EREF; // 84
                 CVLS : pointer;             // 88
                 CVLSsize : integer;         // 92
               end;

  TSDTS_Modules = (sdtsIDEN, sdtsCATS, sdtsCATX, sdtsIREF, sdtsXREF, sdtsSPDM,
                   sdtsDDDF, sdtsDDOM, sdtsDDSH, sdtsSTAT, sdtsRSDF, sdtsLDEF,
                   sdtsCELL);

  TSDTStransfer = class
  protected
    FRoot : string;
    FModules : set of TSDTS_Modules;
    FIden : TSDTS_IDEN;
    FCatd : array of TSDTS_CATD;
    FCats : array of TSDTS_CATS;
    FIref : array of TSDTS_IREF;
    FXref : array of TSDTS_XREF; // are multiple records allowed? Not in RPE anyway
    FSpdm : array of TSDTS_SPDM;
    FDddf : array of TSDTS_DDDF;
    FDdom : array of TSDTS_DDOM;
    FDdsh : array of TSDTS_DDSH;
    FRsdf : array of TSDTS_RSDF;
    FLdef : array of TSDTS_LDEF;

  private
    procedure ParseSADR(dda: TDDAEntry; base: integer = 0);
    function  GetNextDelimitedData(var data: PChar; var dataLen: integer): string;
    function  GetNextFixedData(var data: PChar; const len: integer): string;
    procedure ConvertFMT(const str: string; const fmt: string; const x: pointer);
    function  BitLength(const str: string): integer;
    function  NumMultiple(const data: TDirEntry; const fields: TDDAFieldArray): integer;
    procedure ParseSingle(var data: TDirEntry; fields: TDDAFieldArray; p: pointer; rec: integer = 0);
    procedure ParseMultiple(var data: TDirEntry; fields: TDDAFieldArray; const p: pointer; const n, s: integer);
  protected
    procedure ReadCATD;
    procedure ReadSTAT(const fn: string);
    procedure ReadIDEN(const fn: string);
    procedure ReadCATS(const fn: string; nrec: integer);
    procedure ReadIREF(const fn: string; nrec: integer);
    procedure ReadXREF(const fn: string; nrec: integer);
    procedure ReadSPDM(const fn: string; nrec: integer);
    procedure ReadDDDF(const fn: string; nrec: integer);
    procedure ReadDDOM(const fn: string; nrec: integer);
    procedure ReadDDSH(const fn: string; nrec: integer);
    procedure ReadRSDF;
    procedure ReadLDEF;
    function  ReadCELL(const fn: string): TISO8211;

    function  FindRCIDIndex(a: pointer; rcid: integer): integer;
    function  FindDDSHIndex(const nm: string): integer;
    function  FindDDOMIndexByDVDF(const nm: string): integer;
    function  FindDDOMIndexByRAVA(const nm: string): integer;
  public
    constructor Open(fn: string);
    destructor  Destroy; override;

    function  ReadCELLRecord(parser: TISO8211): TSDTS_CELL;
    function  IsSRPE: boolean;
  end;

implementation
uses DSpUtils;
///////////////////////////////////////////////////////////////////////////////
// TSDTStransfer
constructor TSDTStransfer.Open(fn: string);
var i: integer;
begin
  inherited Create;

  if UpperCase(ExtractFileExt(fn)) <> '.DDF' then
    raise Exception.Create('Not an SDTS transfer file.');

  FRoot := Copy(fn, 1, LastDelimiter('\', fn) + 4);    // include common 4 chars

  ReadCATD;
  FRoot := Copy(FRoot, 1, LastDelimiter('\', FRoot));  // chop off 4 chars
  FModules := [];

  // Read STAT module so I know how many records every module contains
  for i := Low(FCatd) to High(FCatd) do
    if (FCatd[i].NAME = 'STAT') and (FCatd[i].EXTR = 'N') then begin
      ReadSTAT(FRoot + FCatd[i].FIL_);
      Break;
    end;

  // Read other global information modules. Missing are CATX, SCUR, RGIS
  for i := Low(FCatd) to High(FCatd) do
    if      (FCatd[i].NAME = 'IDEN') and (FCatd[i].EXTR = 'N') then ReadIDEN(FRoot + FCatd[i].FIL_)
    else if (FCatd[i].NAME = 'CATS') and (FCatd[i].EXTR = 'N') then ReadCATS(FRoot + FCatd[i].FIL_, FCatd[i].NREC)
    else if (FCatd[i].NAME = 'IREF') and (FCatd[i].EXTR = 'N') then ReadIREF(FRoot + FCatd[i].FIL_, FCatd[i].NREC)
    else if (FCatd[i].NAME = 'XREF') and (FCatd[i].EXTR = 'N') then ReadXREF(FRoot + FCatd[i].FIL_, FCatd[i].NREC)
    else if (FCatd[i].NAME = 'SPDM') and (FCatd[i].EXTR = 'N') then ReadSPDM(FRoot + FCatd[i].FIL_, FCatd[i].NREC)
    else if (FCatd[i].NAME = 'DDDF') and (FCatd[i].EXTR = 'N') then ReadDDDF(FRoot + FCatd[i].FIL_, FCatd[i].NREC)
    else if (FCatd[i].NAME = 'DDOM') and (FCatd[i].EXTR = 'N') then ReadDDOM(FRoot + FCatd[i].FIL_, FCatd[i].NREC)
    else if (FCatd[i].NAME = 'DDSH') and (FCatd[i].EXTR = 'N') then ReadDDSH(FRoot + FCatd[i].FIL_, FCatd[i].NREC);
end;

destructor TSDTStransfer.Destroy;
var i: integer;
begin
  Finalize(FIden.ATID);
  Finalize(FCatd);
  Finalize(FCats);
  for i := 0 to High(FIref) do Finalize(FIref[i].DMID);
  Finalize(FIref);
  for i := 0 to High(FXref) do begin
    Finalize(FXref[i].ATID);
    Finalize(FXref[i].VATT);
    Finalize(FXref[i].SATT);
  end;
  Finalize(FXref);
  for i := 0 to High(FSpdm) do Finalize(FSpdm[i].DMSA);
  Finalize(FSpdm);
  Finalize(FDdom);
  Finalize(FDdsh);
  for i := 0 to High(FRsdf) do begin
    Finalize(FRsdf[i].RDXT);
    Finalize(FRsdf[i].XXLB);
    Finalize(FRsdf[i].YXLB);
    Finalize(FRsdf[i].ZXLB);
    Finalize(FRsdf[i].LYID);
    Finalize(FRsdf[i].RATP);
    Finalize(FRsdf[i].CPID);
  end;
  Finalize(FRsdf);
  for i := 0 to High(FLdef) do begin
    Finalize(FLdef[i].LDXT);
    Finalize(FLdef[i].LATP);
  end;
  Finalize(FLdef);
  inherited;
end;

procedure TSDTStransfer.ReadCATD;
var parser     : TISO8211;
    userData   : TDirEntries;
    j, rec     : integer;
    dda        : TDDAEntries;
    recd       : string[3];
begin
  parser := TISO8211.Create(FRoot + 'CATD.ddf');
  try
    dda := parser.DDA;
    if (Length(dda) <> 1) or (dda[0].Tag <> 'CATD') then
      raise Exception.Create('Invalid module order in CATD');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'NAME' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'TYPE' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'VOLM' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'FILE' then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'RECD' then begin
        dda[0].Fields[j].Offset := 24; dda[0].Fields[j].DataType := 'A'; recd := dda[0].Fields[j].DataType;
      end else if dda[0].Fields[j].Name = 'EXTR' then dda[0].Fields[j].Offset := 28
      else if dda[0].Fields[j].Name = 'MVER' then dda[0].Fields[j].Offset := 32
      else if dda[0].Fields[j].Name = 'COMT' then dda[0].Fields[j].Offset := 36;

    // Read out the data
    SetLength(userData, 2);                 // include RIF
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      SetLength(FCatd, rec+1);
      ParseSingle(userData[1], dda[0].Fields, @FCatd[rec]);
      FCatd[rec].RECDtype := recd; // I|A, for interpreting RECD
      Inc(rec);
    until false;

    // Check for mandatory fields

  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadSTAT(const fn: string);
type
  TempStat = record
               MNRF: string;  // 0
               NREC: integer; // 4
               NSAD: integer; // 8
             end;

var parser     : TISO8211;
    userData   : TDirEntries;
    j, rec, max: integer;
    dda        : TDDAEntries;
    temp       : array of TempStat;
begin
  parser := TISO8211.Create(fn);
  try
    dda := parser.DDA;
    if (Length(dda) <> 1) or (dda[0].Tag <> 'STAT') then
      raise Exception.Create('STAT module of this transfer is corrupt');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MNRF' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'NREC' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'NSAD' then dda[0].Fields[j].Offset := 8
      else dda[0].Fields[j].DataType := 'X'; // don't care about other fields

    // Read out the data
    max := Length(FCatd);
    SetLength(userData, 2);                 // include RIF
    SetLength(temp, max);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      ParseSingle(userData[1], dda[0].Fields, @temp[rec]);
      Inc(rec);
    until false;

    // Move the data over to FCatd
    for j := 0 to max - 1 do begin
      rec := 0;
      while (rec < max) and (temp[rec].MNRF <> FCatd[j].NAME) do Inc(rec);
      if temp[rec].MNRF <> FCatd[j].NAME then
        raise Exception.Create('Error in transfer statistics file');
      FCatd[j].NREC := temp[rec].NREC;
      FCatd[j].NSAD := temp[rec].NSAD;
    end;
    Finalize(temp);

    // Check for mandatory fields

    FModules := FModules + [sdtsSTAT];
  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadCATS(const fn: string; nrec: integer);
var parser     : TISO8211;
    userData   : TDirEntries;
    j, rec     : integer;
    dda        : TDDAEntries;
begin
  parser := TISO8211.Create(fn);
  try
    dda := parser.DDA;
    if (Length(dda) <> 1) or (dda[0].Tag <> 'CATS') then
      raise Exception.Create('Invalid module order in CATS');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'NAME' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'TYPE' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'DOMN' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'MAP'  then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'THEM' then dda[0].Fields[j].Offset := 24
      else if dda[0].Fields[j].Name = 'AGOB' then dda[0].Fields[j].Offset := 28
      else if dda[0].Fields[j].Name = 'AGTP' then dda[0].Fields[j].Offset := 32
      else if dda[0].Fields[j].Name = 'COMT' then dda[0].Fields[j].Offset := 36;

    // Read out the data
    SetLength(userData, 2);                 // include RIF
    SetLength(FCats, nrec);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      ParseSingle(userData[1], dda[0].Fields, @FCats[rec]);
      Inc(rec);
    until false;

    // Check for mandatory fields

    FModules := FModules + [sdtsCATS];
  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadIDEN(const fn: string);
var parser     : TISO8211;
    userData   : TDirEntries;
    j          : integer;
    dda        : TDDAEntries;
begin
  parser := TISO8211.Create(fn);
  try
    dda := parser.DDA;
    if dda[0].Tag <> 'IDEN' then
      raise Exception.Create('Invalid module order in IDEN');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'STID' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'STVS' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'DOCU' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'PRID' then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'PRVS' then dda[0].Fields[j].Offset := 24
      else if dda[0].Fields[j].Name = 'PDOC' then dda[0].Fields[j].Offset := 28
      else if dda[0].Fields[j].Name = 'TITL' then dda[0].Fields[j].Offset := 32
      else if dda[0].Fields[j].Name = 'DAID' then dda[0].Fields[j].Offset := 36
      else if dda[0].Fields[j].Name = 'DAST' then dda[0].Fields[j].Offset := 40
      else if dda[0].Fields[j].Name = 'MPDT' then dda[0].Fields[j].Offset := 44
      else if dda[0].Fields[j].Name = 'DCDT' then dda[0].Fields[j].Offset := 48
      else if dda[0].Fields[j].Name = 'SCAL' then dda[0].Fields[j].Offset := 52
      else if dda[0].Fields[j].Name = 'COMT' then dda[0].Fields[j].Offset := 56;

    if dda[1].Tag <> 'CONF' then
      raise Exception.Create('Invalid module order in IDEN');
    for j := 0 to High(dda[1].Fields) do
      if dda[1].Fields[j].Name = 'FFYN' then dda[1].Fields[j].Offset := 60
      else if dda[1].Fields[j].Name = 'VGYN' then dda[1].Fields[j].Offset := 64
      else if dda[1].Fields[j].Name = 'GTYN' then dda[1].Fields[j].Offset := 68
      else if dda[1].Fields[j].Name = 'RCYN' then dda[1].Fields[j].Offset := 72
      else if dda[1].Fields[j].Name = 'EXSP' then dda[1].Fields[j].Offset := 76
      else if dda[1].Fields[j].Name = 'FTLV' then dda[1].Fields[j].Offset := 80
      else if dda[1].Fields[j].Name = 'CDLV' then dda[1].Fields[j].Offset := 84
      else if dda[1].Fields[j].Name = 'NGDM' then dda[1].Fields[j].Offset := 88;
 
    if Length(dda) = 3 then
      if dda[2].Tag = 'ATID' then begin
        for j := 0 to High(dda[2].Fields) do
          if dda[2].Fields[j].Name = 'MODN' then dda[2].Fields[j].Offset := 0
          else if dda[2].Fields[j].Name = 'RCID' then dda[2].Fields[j].Offset := 4;
      end else raise Exception.Create('Invalid module in file IDEN');

    // Read out the data
    SetLength(userData, Length(dda)+1);     // include RIF
    if not parser.ReadNextRecord(userData) then
      raise Exception.Create('File IDEN contains no data');
    FIden.CDLV := 0;      // default values
    FIden.NGDM := 'N';    //
    ParseSingle(userData[1], dda[0].Fields, @FIden);
    ParseSingle(userData[2], dda[1].Fields, @FIden);
    if Length(dda) = 3 then begin
      SetLength(FIden.ATID, NumMultiple(userData[3], dda[2].Fields));
      ParseMultiple(userData[3], dda[2].Fields, @(FIden.ATID[0]), Length(FIden.ATID), SizeOf(TSDTS_EREF));
    end;

    // Check for mandatory fields

    FModules := FModules + [sdtsIDEN];
  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadIREF(const fn: string; nrec: integer);
var parser     : TISO8211;
    userData   : TDirEntries;
    j, rec     : integer;
    dda        : TDDAEntries;
begin
  parser := TISO8211.Create(fn);
  try
    dda := parser.DDA;
    if (Length(dda) < 1) or (dda[0].Tag <> 'IREF') then
      raise Exception.Create('Invalid module order in IREF');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'COMT' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'SATP' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'XLBL' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'YLBL' then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'HFMT' then dda[0].Fields[j].Offset := 24
      else if dda[0].Fields[j].Name = 'VFMT' then dda[0].Fields[j].Offset := 28
      else if dda[0].Fields[j].Name = 'SFAX' then dda[0].Fields[j].Offset := 32
      else if dda[0].Fields[j].Name = 'SFAY' then dda[0].Fields[j].Offset := 40
      else if dda[0].Fields[j].Name = 'SFAZ' then dda[0].Fields[j].Offset := 48
      else if dda[0].Fields[j].Name = 'XORG' then dda[0].Fields[j].Offset := 56
      else if dda[0].Fields[j].Name = 'YORG' then dda[0].Fields[j].Offset := 64
      else if dda[0].Fields[j].Name = 'ZORG' then dda[0].Fields[j].Offset := 72
      else if dda[0].Fields[j].Name = 'XHRS' then dda[0].Fields[j].Offset := 80
      else if dda[0].Fields[j].Name = 'YHRS' then dda[0].Fields[j].Offset := 88
      else if dda[0].Fields[j].Name = 'VRES' then dda[0].Fields[j].Offset := 96;

    if Length(dda) = 2 then
      if dda[1].Tag = 'DMID' then begin
        for j := 0 to High(dda[2].Fields) do
          if dda[1].Fields[j].Name = 'MODN' then dda[1].Fields[j].Offset := 0
          else if dda[1].Fields[j].Name = 'RCID' then dda[1].Fields[j].Offset := 4;
      end else raise Exception.Create('Invalid module in file IREF');

    // Read out the data
    SetLength(userData, Length(dda)+1);                 // include RIF
    SetLength(FIref, nrec);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      ParseSingle(userData[1], dda[0].Fields, @FIref[rec]);
      if Length(dda) = 2 then begin
        SetLength(FIref[rec].DMID, NumMultiple(userData[2], dda[1].Fields));
        ParseMultiple(userData[2], dda[1].Fields, @(FIref[rec].DMID[0]), Length(FIref[rec].DMID), SizeOf(TSDTS_EREF));
      end;
      Inc(rec);
    until false;

    // Check for mandatory fields

    if rec > 0 then FModules := FModules + [sdtsIREF]
    else raise Exception.Create('No data in module IREF');
  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadXREF(const fn: string; nrec: integer);
var parser     : TISO8211;
    userData   : TDirEntries;
    i, j, rec  : integer;
    dda        : TDDAEntries;
begin
  parser := TISO8211.Create(fn);
  try
    dda := parser.DDA;
    if (Length(dda) = 0) or (dda[0].Tag <> 'XREF') then
      raise Exception.Create('Invalid module order in XREF');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'COMT' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'RDOC' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'RSNM' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'VDAT' then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'SDAT' then dda[0].Fields[j].Offset := 24
      else if dda[0].Fields[j].Name = 'HDAT' then dda[0].Fields[j].Offset := 28
      else if dda[0].Fields[j].Name = 'ZONE' then dda[0].Fields[j].Offset := 32
      else if dda[0].Fields[j].Name = 'PROJ' then dda[0].Fields[j].Offset := 36;

    if Length(dda) > 1 then
      for i := 1 to High(dda) do
        if dda[i].Tag = 'ATID' then begin
          for j := 0 to High(dda[i].Fields) do
            if dda[i].Fields[j].Name = 'MODN' then dda[i].Fields[j].Offset := 0
            else if dda[i].Fields[j].Name = 'RCID' then dda[i].Fields[j].Offset := 4;
        end else if dda[i].Tag = 'VATT' then begin
          for j := 0 to High(dda[i].Fields) do
            if dda[i].Fields[j].Name = 'VDAT' then dda[i].Fields[j].Offset := 0
            else if dda[i].Fields[j].Name = 'VEM'  then dda[i].Fields[j].Offset := 4
            else if dda[i].Fields[j].Name = 'ATLB' then dda[i].Fields[j].Offset := 8
            else if dda[i].Fields[j].Name = 'AUTH' then dda[i].Fields[j].Offset := 12;
        end else if dda[i].Tag = 'SATT' then
          for j := 0 to High(dda[i].Fields) do
            if dda[i].Fields[j].Name = 'SDAT' then dda[i].Fields[j].Offset := 0
            else if dda[i].Fields[j].Name = 'SEM'  then dda[i].Fields[j].Offset := 4
            else if dda[i].Fields[j].Name = 'ATLB' then dda[i].Fields[j].Offset := 8
            else if dda[i].Fields[j].Name = 'AUTH' then dda[i].Fields[j].Offset := 12;

    // Read out the data
    SetLength(userData, Length(dda)+1);     // include RIF
    SetLength(FXref, nrec);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      ParseSingle(userData[1], dda[0].Fields, @FXref[rec]);
      if Length(dda) > 1 then
        for i := 1 to High(dda) do
          if dda[i].Tag = 'ATID' then begin
            SetLength(FXref[rec].ATID, NumMultiple(userData[i+1], dda[i].Fields));
            ParseMultiple(userData[i+1], dda[i].Fields, @(FXref[rec].ATID[0]), Length(FXref[rec].ATID), SizeOf(TSDTS_EREF));
          end else if dda[i].Tag = 'VATT' then begin
            SetLength(FXref[rec].VATT, NumMultiple(userData[i+1], dda[i].Fields));
            ParseMultiple(userData[i+1], dda[i].Fields, @(FXref[rec].VATT[0]), Length(FXref[rec].ATID), SizeOf(TSDTS_VATT));
          end else if dda[i].Tag = 'SATT' then begin
            SetLength(FXref[rec].SATT, NumMultiple(userData[i+1], dda[i].Fields));
            ParseMultiple(userData[i+1], dda[i].Fields, @(FXref[rec].SATT[0]), Length(FXref[rec].ATID), SizeOf(TSDTS_SATT));
          end;
      Inc(rec);
    until false;

    // Check for mandatory fields

    if rec > 0 then FModules := FModules + [sdtsXREF]
    else raise Exception.Create('No data in XREF module');
  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadSPDM(const fn: string; nrec: integer);
var parser   : TISO8211;
    userData : TDirEntries;
    j, rec   : integer;
    dda      : TDDAEntries;
begin
  parser := TISO8211.Create(fn);
  try
    dda := parser.DDA;
    if (Length(dda) = 0) or (dda[0].Tag <> 'SPDM') then
      raise Exception.Create('Invalid module order in SPDM');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'DTYP' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'DSTP' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'COMT' then dda[0].Fields[j].Offset := 16;

    if Length(dda) = 2 then
      for j := 0 to High(dda[1].Fields) do
        if dda[1].Fields[j].Name = 'X' then dda[1].Fields[j].Offset := 0
        else if dda[1].Fields[j].Name = 'Y' then dda[1].Fields[j].Offset := 8
        else if dda[1].Fields[j].Name = 'Z' then dda[1].Fields[j].Offset := 16;

    // Read out the data, just one record
    SetLength(userData, Length(dda)+1);     // include RIF
    SetLength(FSpdm, nrec);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break;
      ParseSingle(userData[1], dda[0].Fields, @FSpdm[rec]);
      if Length(dda) > 1 then begin
        SetLength(FSpdm[rec].DMSA, NumMultiple(userData[2], dda[1].Fields));
        ParseMultiple(userData[2], dda[1].Fields, @(FSpdm[rec].DMSA[0]), Length(FSpdm[rec].DMSA), SizeOf(TSDTS_SADR));
      end;
      Inc(rec);
    until false;

    // Check for mandatory fields

  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadDDDF(const fn: string; nrec: integer);
var parser     : TISO8211;
    userData   : TDirEntries;
    j, rec     : integer;
    dda        : TDDAEntries;
begin
  parser := TISO8211.Create(fn);
  try
    dda := parser.DDA;
    if (Length(dda) <> 1) or (dda[0].Tag <> 'DDDF') then
      raise Exception.Create('Invalid module order in DDDF');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'EORA' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'EALB' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'SRCE' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'DFIN' then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'AUTH' then dda[0].Fields[j].Offset := 24
      else if dda[0].Fields[j].Name = 'ADSC' then dda[0].Fields[j].Offset := 28;

    // Read out the data
    SetLength(userData, 2);                 // include RIF
    SetLength(FDddf, nrec);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      ParseSingle(userData[1], dda[0].Fields, @FDddf[rec]);
      Inc(rec);
    until false;

    // Check for mandatory fields

  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadDDOM(const fn: string; nrec: integer);
var parser     : TISO8211;
    userData   : TDirEntries;
    j, rec     : integer;
    dda        : TDDAEntries;
begin
  parser := TISO8211.Create(fn);
  try
    dda := parser.DDA;
    if (Length(dda) <> 1) or (dda[0].Tag <> 'DDOM') then
      raise Exception.Create('Invalid module order in DDOM');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'ATLB' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'AUTH' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'ATYP' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'ADVF' then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'ADMU' then dda[0].Fields[j].Offset := 24
      else if dda[0].Fields[j].Name = 'RAVA' then dda[0].Fields[j].Offset := 28
      else if dda[0].Fields[j].Name = 'DVAL' then begin
        dda[0].Fields[j].Offset := 32; dda[0].Fields[j].Format := 'R'; // force to double
      end else if dda[0].Fields[j].Name = 'DVDF' then dda[0].Fields[j].Offset := 40;

    // Read out the data
    SetLength(userData, 2);                 // include RIF
    SetLength(FDdom, nrec);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      ParseSingle(userData[1], dda[0].Fields, @FDdom[rec]);
      Inc(rec);
    until false;

    // Check for mandatory fields

    if rec <> nrec then
      raise Exception.Create('Invalid number of data records in module DDOM');
  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadDDSH(const fn: string; nrec: integer);
var parser     : TISO8211;
    userData   : TDirEntries;
    j, rec     : integer;
    dda        : TDDAEntries;
begin
  parser := TISO8211.Create(fn);
  try
    dda := parser.DDA;
    if (Length(dda) <> 1) or (dda[0].Tag <> 'DDSH') then
      raise Exception.Create('Invalid module order in DDSH');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'NAME' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'TYPE' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'ETLB' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'EUTH' then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'ATLB' then dda[0].Fields[j].Offset := 24
      else if dda[0].Fields[j].Name = 'AUTH' then dda[0].Fields[j].Offset := 28
      else if dda[0].Fields[j].Name = 'FMT'  then dda[0].Fields[j].Offset := 32
      else if dda[0].Fields[j].Name = 'UNIT' then dda[0].Fields[j].Offset := 36
      else if dda[0].Fields[j].Name = 'PREC' then dda[0].Fields[j].Offset := 40
      else if dda[0].Fields[j].Name = 'MXLN' then dda[0].Fields[j].Offset := 48
      else if dda[0].Fields[j].Name = 'KEY'  then dda[0].Fields[j].Offset := 52;

    // Read out the data
    SetLength(userData, 2);                 // include RIF
    SetLength(FDdsh, nrec);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      ParseSingle(userData[1], dda[0].Fields, @FDdsh[rec]);
      Inc(rec);
    until false;

    // Check for mandatory fields

    if rec <> nrec then
      raise Exception.Create('Invalid number of data records in module DDSH');
  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadRSDF;
var parser     : TISO8211;
    userData   : TDirEntries;
    i, j       : integer;
    rec, nrec  : integer;
    dda        : TDDAEntries;
begin
  // Find RSDF module
  for i := Low(FCatd) to High(FCatd) do
    if FCatd[i].NAME = 'RSDF' then Break; // RSDF module should not be external
  parser := TISO8211.Create(FRoot + FCatd[i].FIL_);
  nrec := FCatd[i].NREC;

  try
    dda := parser.DDA;
    if (Length(dda) < 4) or (dda[0].Tag <> 'RSDF') then
      raise Exception.Create('Invalid module order in RSDF');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'OBRP' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'CSCD' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'AQMD' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'AQDT' then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'COMT' then dda[0].Fields[j].Offset := 24
      else if dda[0].Fields[j].Name = 'DEFI' then dda[0].Fields[j].Offset := 28
      else if dda[0].Fields[j].Name = 'CMPR' then dda[0].Fields[j].Offset := 32
      else if dda[0].Fields[j].Name = 'CMMD' then dda[0].Fields[j].Offset := 36
      else if dda[0].Fields[j].Name = 'DCOM' then dda[0].Fields[j].Offset := 40
      else if dda[0].Fields[j].Name = 'METH' then dda[0].Fields[j].Offset := 44
      else if dda[0].Fields[j].Name = 'RWXT' then dda[0].Fields[j].Offset := 48
      else if dda[0].Fields[j].Name = 'CLXT' then dda[0].Fields[j].Offset := 52
      else if dda[0].Fields[j].Name = 'PLXT' then dda[0].Fields[j].Offset := 56
      else if dda[0].Fields[j].Name = 'SCOR' then dda[0].Fields[j].Offset := 60
      else if dda[0].Fields[j].Name = 'SCPT' then dda[0].Fields[j].Offset := 64
      else if dda[0].Fields[j].Name = 'TIDX' then dda[0].Fields[j].Offset := 68
      else if dda[0].Fields[j].Name = 'TIFT' then dda[0].Fields[j].Offset := 72
      else if dda[0].Fields[j].Name = 'TIDS' then dda[0].Fields[j].Offset := 76
      else if dda[0].Fields[j].Name = 'ALTN' then dda[0].Fields[j].Offset := 80
      else if dda[0].Fields[j].Name = 'FSCN' then dda[0].Fields[j].Offset := 84
      else if dda[0].Fields[j].Name = 'ASPR' then dda[0].Fields[j].Offset := 88
      else if dda[0].Fields[j].Name = 'NLAY' then dda[0].Fields[j].Offset := 96;

    for i := 1 to High(dda) do
      if (dda[i].Tag = 'ISID') then begin
        for j := 0 to High(dda[i].Fields) do
          if dda[i].Fields[j].Name = 'MODN' then dda[i].Fields[j].Offset := 100
          else if dda[i].Fields[j].Name = 'RCID' then dda[i].Fields[j].Offset := 104;
      end else if (dda[i].Tag = 'LYID') or (dda[i].Tag = 'RATP') or (dda[i].Tag = 'CPID') then begin
        for j := 0 to High(dda[i].Fields) do
          if dda[i].Fields[j].Name = 'MODN' then dda[i].Fields[j].Offset := 0
          else if dda[i].Fields[j].Name = 'RCID' then dda[i].Fields[j].Offset := 4;
      end else if dda[i].Tag = 'RDXT' then begin
        for j := 0 to High(dda[i].Fields) do
          if dda[i].Fields[j].Name = 'DEXT' then dda[i].Fields[j].Offset := 0
          else if dda[i].Fields[j].Name = 'DSCO' then dda[i].Fields[j].Offset := 4;
      end else if dda[i].Tag = 'SADR' then ParseSADR(dda[i], 112)
      else if (dda[i].Tag = 'XXLB') or (dda[i].Tag = 'YXLB') or (dda[i].Tag = 'ZXLB') then
        dda[i].Fields[0].Offset := 0;

    // Read out the data
    SetLength(userData, Length(dda)+1);     // include RIF
    SetLength(FRsdf, nrec);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      ParseSingle(userData[1], dda[0].Fields, @FRsdf[rec]);
      for i := 1 to High(dda) do
        if dda[i].Tag = 'ISID' then ParseSingle(userData[i+1], dda[i].Fields, @FRsdf[rec])
        else if dda[i].Tag = 'RDXT' then begin
          SetLength(FRsdf[rec].RDXT, NumMultiple(userData[i+1], dda[i].Fields));
          ParseMultiple(userData[i+1], dda[i].Fields, @(FRsdf[rec].RDXT[0]), Length(FRsdf[rec].RDXT), SizeOf(TSDTS_RDXT));
        end else if dda[i].Tag = 'SADR' then ParseSingle(userData[i+1], dda[i].Fields, @FRsdf[rec])
        else if dda[i].Tag = 'XXLB' then begin
          SetLength(FRsdf[rec].XXLB, NumMultiple(userData[i+1], dda[i].Fields));
          ParseMultiple(userData[i+1], dda[i].Fields, @(FRsdf[rec].XXLB[0]), Length(FRsdf[rec].XXLB), SizeOf(double));
        end else if dda[i].Tag = 'YXLB' then begin
          SetLength(FRsdf[rec].YXLB, NumMultiple(userData[i+1], dda[i].Fields));
          ParseMultiple(userData[i+1], dda[i].Fields, @(FRsdf[rec].YXLB[0]), Length(FRsdf[rec].YXLB), SizeOf(double));
        end else if dda[i].Tag = 'ZXLB' then begin
          SetLength(FRsdf[rec].ZXLB, NumMultiple(userData[i+1], dda[i].Fields));
          ParseMultiple(userData[i+1], dda[i].Fields, @(FRsdf[rec].ZXLB[0]), Length(FRsdf[rec].ZXLB), SizeOf(double));
        end else if dda[i].Tag = 'LYID' then begin
          SetLength(FRsdf[rec].LYID, NumMultiple(userData[i+1], dda[i].Fields));
          ParseMultiple(userData[i+1], dda[i].Fields, @(FRsdf[rec].LYID[0]), Length(FRsdf[rec].LYID), SizeOf(TSDTS_EREF));
        end else if dda[i].Tag = 'RATP' then begin
          SetLength(FRsdf[rec].RATP, NumMultiple(userData[i+1], dda[i].Fields));
          ParseMultiple(userData[i+1], dda[i].Fields, @(FRsdf[rec].RATP[0]), Length(FRsdf[rec].RATP), SizeOf(TSDTS_EREF));
        end else if dda[i].Tag = 'CPID' then begin
          SetLength(FRsdf[rec].CPID, NumMultiple(userData[i+1], dda[i].Fields));
          ParseMultiple(userData[i+1], dda[i].Fields, @(FRsdf[rec].CPID[0]), Length(FRsdf[rec].CPID), SizeOf(TSDTS_EREF));
        end;      
      Inc(rec);
    until false;

    // Check for mandatory fields

    if rec > 0 then FModules := FModules + [sdtsRSDF]
    else raise Exception.Create('No data in RSDF module');
  finally
    Finalize(userData);
    parser.Free;
  end;
end;

procedure TSDTStransfer.ReadLDEF;
var parser     : TISO8211;
    userData   : TDirEntries;
    i, j       : integer;
    rec, nrec  : integer;
    dda        : TDDAEntries;
begin
  // Find LDEF module
  for i := Low(FCatd) to High(FCatd) do
    if FCatd[i].NAME = 'LDEF' then Break; // LDEF module should not be external
  parser := TISO8211.Create(FRoot + FCatd[i].FIL_);
  nrec := FCatd[i].NREC;

  try
    dda := parser.DDA;
    if (Length(dda) = 0) or (dda[0].Tag <> 'LDEF') then
      raise Exception.Create('Invalid module order in LDEF');
    for j := 0 to High(dda[0].Fields) do
      if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
      else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
      else if dda[0].Fields[j].Name = 'CMNM' then dda[0].Fields[j].Offset := 8
      else if dda[0].Fields[j].Name = 'LLBL' then dda[0].Fields[j].Offset := 12
      else if dda[0].Fields[j].Name = 'CODE' then dda[0].Fields[j].Offset := 16
      else if dda[0].Fields[j].Name = 'BMSK' then dda[0].Fields[j].Offset := 20
      else if dda[0].Fields[j].Name = 'NROW' then dda[0].Fields[j].Offset := 24
      else if dda[0].Fields[j].Name = 'NCOL' then dda[0].Fields[j].Offset := 28
      else if dda[0].Fields[j].Name = 'NPLA' then dda[0].Fields[j].Offset := 32
      else if dda[0].Fields[j].Name = 'SORI' then dda[0].Fields[j].Offset := 36
      else if dda[0].Fields[j].Name = 'SOCI' then dda[0].Fields[j].Offset := 40
      else if dda[0].Fields[j].Name = 'SOPI' then dda[0].Fields[j].Offset := 44
      else if dda[0].Fields[j].Name = 'RWOO' then dda[0].Fields[j].Offset := 48
      else if dda[0].Fields[j].Name = 'CLOO' then dda[0].Fields[j].Offset := 52
      else if dda[0].Fields[j].Name = 'PLOO' then dda[0].Fields[j].Offset := 56
      else if dda[0].Fields[j].Name = 'INTR' then dda[0].Fields[j].Offset := 60
      else if dda[0].Fields[j].Name = 'COMT' then dda[0].Fields[j].Offset := 64;

    if Length(dda) > 1 then
      for i := 1 to High(dda) do
        if dda[i].Tag = 'LDXT' then begin
          for j := 0 to High(dda[i].Fields) do
            if dda[i].Fields[j].Name = 'DEXT' then dda[i].Fields[j].Offset := 0
            else if dda[i].Fields[j].Name = 'SODM' then dda[i].Fields[j].Offset := 4
            else if dda[i].Fields[j].Name = 'DTOO' then dda[i].Fields[j].Offset := 8
            else if dda[i].Fields[j].Name = 'DINR' then dda[i].Fields[j].Offset := 12;
        end else if dda[i].Tag = 'LATP' then
          for j := 0 to High(dda[i].Fields) do
            if dda[i].Fields[j].Name = 'MODN' then dda[i].Fields[j].Offset := 0
            else if dda[i].Fields[j].Name = 'RCID' then dda[i].Fields[j].Offset := 4;

    // Read out the data
    SetLength(userData, Length(dda)+1);     // include RIF
    SetLength(FLdef, nrec);
    rec := 0;
    repeat
      if not parser.ReadNextRecord(userData) then Break; // no more records
      ParseSingle(userData[1], dda[0].Fields, @FLdef[rec]);
      if Length(dda) > 1 then
        for i := 1 to High(dda) do
          if dda[i].Tag = 'LDXT' then begin
            SetLength(FLdef[rec].LDXT, NumMultiple(userData[i+1], dda[i].Fields));
            ParseMultiple(userData[i+1], dda[i].Fields, @(FLdef[rec].LDXT[0]), Length(FLdef[rec].LDXT), SizeOf(TSDTS_LDXT));
          end;
      Inc(rec);
    until false;

    // Check for mandatory fields

//    if rec = FRsdf.NLAY then
    FModules := FModules + [sdtsLDEF]
//    else raise Exception.Create('Invalid number of layers in LDEF module');
  finally
    Finalize(userData);
    parser.Free;
  end;
end;

// ReadCELL will only set up the DDA structure. Actual reading of records is
// done by ReadCELLRecord and the results are interpreted by the caller.
function TSDTStransfer.ReadCELL(const fn: string): TISO8211;
var i, j: integer;
    dda : TDDAEntries;
begin
  Result := TISO8211.Create(fn);
  dda := Result.DDA;
  if (Length(dda) < 2) or (dda[0].Tag <> 'CELL') then
    raise Exception.Create('CELL module corrupt');
  for j := 0 to High(dda[0].Fields) do begin
    if dda[0].Fields[j].Name = 'MODN' then dda[0].Fields[j].Offset := 0
    else if dda[0].Fields[j].Name = 'RCID' then dda[0].Fields[j].Offset := 4
    else if dda[0].Fields[j].Name = 'NCEL' then dda[0].Fields[j].Offset := 8
    else if dda[0].Fields[j].Name = 'ROWI' then dda[0].Fields[j].Offset := 12
    else if dda[0].Fields[j].Name = 'COLI' then dda[0].Fields[j].Offset := 16
    else if dda[0].Fields[j].Name = 'PLAI' then dda[0].Fields[j].Offset := 20
    else if dda[0].Fields[j].Name = 'TIND' then begin
      dda[0].Fields[j].Offset := 24;
      if dda[0].Fields[j].DataType = 'B' then
        dda[0].Fields[j].Format := FRsdf[0].TIFT;  ////// $$$$$$$$$$$$$$$$$$$$
    end;
  end;
  for i := 1 to High(dda) do begin
    if dda[i].Tag = 'DNDX' then dda[i].Fields[0].Offset := 0
    else if dda[i].Tag = 'SADR' then ParseSADR(dda[i])
    else if dda[i].Tag = 'ATID' then begin
      for j := 0 to High(dda[i].Fields) do
        if dda[i].Fields[j].Name = 'MODN' then dda[i].Fields[j].Offset := 0
        else if dda[i].Fields[j].Name = 'RCID' then dda[i].Fields[j].Offset := 4;
    end else if dda[i].Tag = 'CVLS' then
      dda[i].Fields[0].Offset := 0;   // data goes into user-supplied pointer
  end;

  FModules := FModules + [sdtsCELL];
end;

// The next record from the CELL module is read into Result. Result.CVLS is
// allocated here, and needs to be freed by the caller.
function TSDTStransfer.ReadCELLRecord(parser: TISO8211): TSDTS_CELL;
var userData     : TDirEntries;
    dda          : TDDAEntries;
    i, len, size : integer;
begin
  SetLength(userData, parser.DREntries);
  if not parser.ReadNextRecord(userData) then begin  // no more records
    Result.MODN := 'NULL';
    Result.RCID := 0;
    Result.ROWI := -1;
    Result.COLI := -1;
    Exit;
  end;

  // Parse the data into the fields
  dda := parser.DDA;
  if Length(dda) < 2 then
    raise Exception.Create('CELL record corrupt');
  ParseSingle(userData[1], dda[0].Fields, @Result);
  for i := 1 to High(dda) do
    if dda[i].Tag = 'DNDX' then begin
      SetLength(Result.DNDX, NumMultiple(userData[i+1], dda[i].Fields));
      ParseMultiple(userData[i+1], dda[i].Fields, @(Result.DNDX[0]), Length(Result.DNDX), SizeOf(integer));
    end else if dda[i].Tag = 'SADR' then begin
      ParseSingle(userData[i+1], dda[i].Fields, @Result);
    end else if dda[i].Tag = 'ATID' then begin
      SetLength(Result.ATID, NumMultiple(userData[i+1], dda[i].Fields));
      ParseMultiple(userData[i+1], dda[i].Fields, @(Result.ATID[0]), Length(Result.ATID), SizeOf(integer));
    end else if dda[i].Tag = 'CVLS' then begin
      // Read out data from the stream into Result.CVLS if format = 'B'
      // else parse multiple "records"
      if dda[i].Fields[0].DataType = 'B' then begin
        Result.CVLSsize := userData[i+1].Length - 1;
        GetMem(Result.CVLS, Result.CVLSsize);
        Move(userData[i+1].Data^, Result.CVLS^, Result.CVLSsize);
        case dda[i].Fields[0].Length of
          16: Swap2Bytes(Result.CVLS^, Result.CVLSsize div 2);
          32: Swap4Bytes(Result.CVLS^, Result.CVLSsize div 4);
        end;
      end else begin
        if dda[i].Fields[0].DataType = 'I' then size := 4
        else if dda[i].Fields[0].DataType in ['R', 'S'] then size := 8
        else raise Exception.Create('Unsupported data type in CVLS');
        len := NumMultiple(userData[i+1], dda[i].Fields);
        Result.CVLSsize := len * size;
        GetMem(Result.CVLS, Result.CVLSsize);
        ParseMultiple(userData[i+1], dda[i].Fields, Result.CVLS, len, size);
      end;
    end;
end;

procedure TSDTStransfer.ParseSADR(dda: TDDAEntry; base: integer = 0);
var j : integer;
begin
  if Length(dda.Fields) > 3 then
    raise Exception.Create('This reader supports only geospatial dimensions');
  for j := 0 to High(dda.Fields) do begin
    if dda.Fields[j].DataType = 'I' then dda.Fields[j].DataType := 'R';
    if dda.Fields[j].Name = 'X' then begin
      dda.Fields[j].Offset := base;
      if dda.Fields[j].DataType = 'B' then dda.Fields[j].DataType := 'H';
    end else if dda.Fields[j].Name = 'Y' then begin
      dda.Fields[j].Offset := base + 8;
      if dda.Fields[j].DataType = 'B' then dda.Fields[j].DataType := 'H';
    end else if dda.Fields[j].Name = 'Z' then begin
      dda.Fields[j].Offset := base + 16;
      if dda.Fields[j].DataType = 'B' then dda.Fields[j].DataType := 'V';
    end;
  end;
end;

function TSDTStransfer.GetNextDelimitedData(var data: PChar; var dataLen: integer): string;
var pos : PChar;
    len : integer;
begin
  // find the field length and update the pointers
  pos := StrScan(data, #$1F);
  if pos = nil then len := dataLen - 1
  else len := cardinal(pos) - cardinal(data);
  if len > dataLen then len := dataLen - 1;
  Dec(dataLen, len + 1);

  // copy the data over to a Pascal string
  Result := Copy(data, 0, len);
  Inc(data, len + 1);
end;

function TSDTStransfer.GetNextFixedData(var data: PChar; const len: integer): string;
begin
  Result := Copy(data, 0, len);
  Inc(data, len);
end;

// The x pointer is here assumed to be 8 bytes long -> double. All formats
// except C (character bit field) and A (alphanumeric) will fit, but they are
// not converted to a double. The string is converted to its binary
// representation in little-endian format.
procedure TSDTStransfer.ConvertFMT(const str: string; const fmt: string; const x: pointer);
var i32 : integer;
begin
  if fmt = 'I' then
    integer(x^) := StrToInt(str)
  else if (fmt = 'R') or (fmt = 'S') then
    double(x^) := StrToFloat(str)
  else if (fmt = 'BI8') or (fmt = 'BUI8') then
    Move(str[1], x^, 1)
  else if (fmt = 'BI16') or (fmt = 'BUI16') then
    word(x^) := EndianSwap(word(str[1]))
  else if (fmt = 'BI24') or (fmt = 'BUI24') then begin
    Move(str[1], i32, 3);
    integer(x^) := EndianSwap(i32 shr 8);
  end else if (fmt = 'BUI') or (fmt = 'BI32') or (fmt = 'BUI32') or (fmt = 'BFP32') then
    integer(x^) := EndianSwap(integer(str[1]))
  else if fmt = 'BFP64' then Move(str[1], x^, 8);   /// must swap!!!!!!!!!!!!!!!
end;

function TSDTSTransfer.BitLength(const str: string): integer;
begin
  if (str = 'BI8') or (str = 'BUI8') then Result := 8
  else if (str = 'BI16') or (str = 'BUI16') then Result := 16
  else if (str = 'BI24') or (str = 'BUI24') then Result := 24
  else if (str = 'B') or (str = 'BI32') or (str = 'BUI32') or (str = 'BFP32') then Result := 32
  else if (str = 'BFP64') then Result := 64
  else raise Exception.Create('Invalid binary format string');
end;

procedure TSDTStransfer.ParseSingle(var data: TDirEntry; fields: TDDAFieldArray; p: pointer; rec: integer = 0);
var i, len: integer;
    str: string;
begin
  for i := 0 to High(fields) do begin
    if fields[i].Length = 0 then str := GetNextDelimitedData(data.Data, data.Length)
    else begin
      if fields[i].DataType in ['H', 'V'] then begin   // deal with SADR format
        if fields[i].DataType = 'H' then fields[i].Format := FIref[rec].HFMT
        else fields[i].Format := FIref[rec].VFMT;
        fields[i].Length := BitLength(fields[i].Format);
        SetLength(fields[i].Format, 1);
      end;
      if fields[i].Format = 'B' then len := fields[i].Length div 8
      else len := fields[i].Length;
      str := GetNextFixedData(data.Data, len);
    end;
    if fields[i].Format <> '' then begin
      if fields[i].DataType in ['A', 'C'] then
        raise Exception.Create('A and C data types are not supported for formatted data');
      ConvertFMT(str, fields[i].Format, pointer(integer(p) + fields[i].Offset));
    end else
      case fields[i].DataType of
        'A', 'C': string(pointer(integer(p) + fields[i].Offset)^) := str;
        'B': Move(str[1], pointer(integer(p) + fields[i].Offset)^, Length(str));
        'I': integer(pointer(integer(p) + fields[i].Offset)^) := StrToInt(str);
        'R', 'S': double(pointer(integer(p) + fields[i].Offset)^) := StrToFloat(str);
        'X': ; // ignore 'X' data type, which is ISO8211 unused data
      else raise Exception.Create('Unsupported format encountered in parsing data');
      end;
  end;
end;

function TSDTStransfer.NumMultiple(const data: TDirEntry; const fields: TDDAFieldArray): integer;
var i: integer;
    p, pend: PChar;
begin
  if fields[0].Length = 0 then begin      // delimited data, count delimiters
    Result := 1;
    p := data.Data;
    pend := PChar(integer(p) + data.Length);
    repeat
      p := StrScan(p, #$1F);
      if (p = nil) or (p > pend) then Break;
      Inc(p);
      Inc(Result);
    until false;
    Result := Result div Length(fields);
  end else begin                               // fixed length data
    Result := 0;
    for i := 0 to High(fields) do
      Inc(Result, fields[i].Length);
    Result := data.Length div Result;
  end;
end;

procedure TSDTStransfer.ParseMultiple(var data: TDirEntry; fields: TDDAFieldArray; const p: pointer; const n, s: integer);
var j: integer;
begin
  for j := 0 to n - 1 do
    ParseSingle(data, fields, pointer(integer(p) + j * s), j);
end;

function TSDTStransfer.FindRCIDIndex(a: pointer; rcid: integer): integer;
type
  EREF = array of TSDTS_EREF;
var x: EREF;
    i: integer;
begin
  Result := -1;
  x := EREF(a);
  for i := 0 to High(x) do
    if rcid = x[i].RCID then begin
      Result := i;
      Break;
    end;
  if Result = -1 then
    raise Exception.Create('Invalid module cross-reference encountered');
end;

function TSDTStransfer.FindDDSHIndex(const nm: string): integer;
var i: integer;
begin
  Result := -1;
  for i := 0 to High(FDdsh) do
    if FDdsh[i].NAME = nm then begin
      Result := i;
      Break;
    end;
  if Result = -1 then
    raise Exception.Create('DDSH module has no record for cross-referenced module');
end;

function TSDTStransfer.FindDDOMIndexByDVDF(const nm: string): integer;
var i: integer;
begin
  Result := -1;
  for i := 0 to High(FDdom) do
    if Pos(nm, UpperCase(FDdom[i].DVDF)) > 0 then begin
      Result := i;
      Break;
    end;
  if Result = -1 then
    raise Exception.Create('DDOM module has no record for requested string');
end;

function TSDTStransfer.FindDDOMIndexByRAVA(const nm: string): integer;
var i: integer;
begin
  Result := -1;
  for i := 0 to High(FDdom) do
    if UpperCase(FDdom[i].RAVA) = nm then begin
      Result := i;
      Break;
    end;
  if Result = -1 then
    raise Exception.Create('DDOM module has no record for requested range or value');
end;

function TSDTStransfer.IsSRPE: boolean;
begin
  Result := (Copy(FIden.PRID, 1, 4) = 'SRPE') and (FIden.FFYN = 'N') and
            (FIden.VGYN = 'N') and (FIden.GTYN = 'N') and (FIden.RCYN = 'Y');
end;

end.
