{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpSDTSDriver.pas. The Original Code is available       }
{ through the home page of DSpatial at http://dspatial.sourceforge.net.        }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements the PFA driver for the SDTS format. Currently it only   }
{ supports the RPE extension, and only so in a limited fashion. See DSpSDTS_RPE}
{ for more details on completeness of implementation.                          }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 03                                             }
{                                                                              }
{ Dependencies  : DSpPFA                                                       }
{                 DSpSDTS_RPE (pervasive)                                      }
{                 DSpBlobMath (convert FILL to MISSING)                        }
{                 DSpGlobals                                                   }
{                 SysUtils (Exception)                                         }
{                 Classes (TThread)                                            }
{                                                                              }
{ Version       : 0.7  03Dec03 Added Create/OpenSource pairs to driver and     }
{                              broker.                                         }
{                 0.6  03Nov29 Support for saving data added. SupportsSource   }
{                              added to the broker.                            }
{                 0.5  03Nov24 Convert Fill value to Missing value             }
{                 0.4  03Nov07 PFA overhaul                                    }
{                 0.3  03Aug14 CloseReader method is removed from the driver.  }
{                              It now uses the ancestor's virtual CloseReader  }
{                              method.                                         }
{                              GetAllRasterInfo removed.                       }
{                 0.2  03Jul18 Minor bug fixes.                                }
{                 0.1  03Apr26 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpSDTSDriver;
interface
uses DSpGlobals, DSpSDTS_RPE, DSpPFA;
type
  TDSpSDTSRasterDriver = class(TDSpBaseDriver)
  private
    FFile: string;
    FSDTS: TSDTS_RPE;
  protected
    property    FileName: string read FFile write FFile;
    function    Description: string; override;
    function    DataSets: integer; override;

    function    DataSetID(i: integer): integer; override;
    function    DataSetDescription(i: integer): string; override;
    function    DataSetCoordSys(i: integer): string; override;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); override;
    function    DataSetType(i: integer): TDSpPFAFormat; override;
  public
    constructor Open(const fn: string);
    destructor  Destroy; override;

    function    CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer; override;
    function    AddDataSet(const ds: TDSpSourceDataSet): integer; override;
    function    CloneDataSet(const ds: TDSpSourceDataSet): integer; override;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   AddRasterBand(i: integer; info: PDSpRasterbandInfo); override;
    procedure   DeleteRasterBand(i: integer); override;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); override;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); override;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); override;
  end;

  TDSpSDTSBroker = class(TDSpBaseBroker)
  public
    constructor Create;
    destructor  Destroy; override;

    function    CreateSource(const fn: string): TDSpSourceHandler; override;
    function    OpenSource(const fn: string): TDSpSourceHandler; override;
    function    SupportsSource(const source: string): boolean; override;
    function    CanSave(const ds: TDSpSourceDataSet): string; override;
  end;

implementation
uses SysUtils, DSpUtils, DSpBlobMath;

////////////////////////////////////////////////////////////////////////////////
// TDSpSDTSReader
constructor TDSpSDTSRasterDriver.Open(const fn: string);
begin
  inherited Create;
  FFile := fn;
  FSDTS := TSDTS_RPE.Open(FFile);
end;

destructor TDSpSDTSRasterDriver.Destroy;
begin
  FSDTS.Free;
  inherited;
end;

function TDSpSDTSRasterDriver.Description: string;
begin
  Result := 'SDTS Raster source: ' + FSDTS.DescriptiveName;
end;

function TDSpSDTSRasterDriver.DataSets: integer;
begin
  Result := FSDTS.Rasters;
end;

function TDSpSDTSRasterDriver.DataSetID(i: integer): integer;
begin
  Result := i;
end;

function TDSpSDTSRasterDriver.DataSetDescription(i: integer): string;
begin
  Result := Format('%s (%d)', [FSDTS.DescriptiveName, i]);
end;

function TDSpSDTSRasterDriver.DataSetCoordSys(i: integer): string;
begin
  Result := FSDTS.ProjectionString;
end;

procedure TDSpSDTSRasterDriver.DataSetExtent(i: integer; const ext: TDSpExtent);
var north, east, south, west: TDSpFloat64;
begin
  FSDTS.Extent(i, north, east, south, west);
  ext.North := north;
  ext.East := east;
  ext.South := south;
  ext.West := west;
end;

function TDSpSDTSRasterDriver.DataSetType(i: integer): TDSpPFAFormat;
begin
  Result := dspPFARaster;
end;

function TDSpSDTSRasterDriver.CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer;
begin
  Assert(false, 'SDTS driver does not support writing new data sets');
  Result := -1;
end;

function TDSpSDTSRasterDriver.AddDataSet(const ds: TDSpSourceDataSet): integer;
begin
  Assert(false, 'Can not write SDTS raster data sets');
  Result := -1;
end;

function TDSpSDTSRasterDriver.CloneDataSet(const ds: TDSpSourceDataSet): integer;
begin
  Assert(false, 'Can not write SDTS raster data sets');
  Result := -1;
end;

procedure TDSpSDTSRasterDriver.AddRasterBand(i: integer; info: PDSpRasterbandInfo);
begin
  info^.BandIndex := -1;
end;

procedure TDSpSDTSRasterDriver.DeleteRasterBand(i: integer);
begin
  Assert(false, 'Can not modify SDTS raster data sets');
end;

procedure TDSpSDTSRasterDriver.GetRasterInfo(i: integer; info: PDSpRasterInfo);
begin
  info^.BandFormat := FSDTS.RasterSequencing[i];
  info^.Bands := FSDTS.Layers[i];
  info^.Rows := FSDTS.Rows[i];
  info^.Columns := FSDTS.Columns[i];
  info^.PixelSize := FSDTS.XSize[i];
end;

procedure TDSpSDTSRasterDriver.GetRasterBandInfo(i: integer; info: PDSpRasterInfo);
var b: integer;
begin
  for b := 0 to High(info^.Band) do
    with info^.Band[b] do begin
      BandIndex := b;
      DataType := FSDTS.DataType[i, b];
      BandType := FSDTS.LayerType[i, b];
      MissingValue := FSDTS.MissingValue[i];
      Min := FSDTS.MinValue[i];
      Max := FSDTS.MaxValue[i];
    end;
end;

procedure TDSpSDTSRasterDriver.ReadRasterBand(dataSet, band: integer; const into: pointer);
var miss, fill: TDSpFloat64;
    ui8a, ui8b: TDSpUInt8;
    i8a, i8b: TDSpInt8;
    ui16a, ui16b: TDSpUInt16;
    i16a, i16b: TDSpInt16;
    ui32a, ui32b: TDSpUInt32;
    i32a, i32b: TDSpInt32;
    f32a, f32b: TDSpFloat32;
begin
  FSDTS.ReadRaster(dataSet, band, into);

  // Change FILL to MISSING
  miss := FSDTS.MissingValue[dataSet];
  fill := FSDTS.NullValue[dataSet];
  if miss <> fill then // can do binary comparison here, otherwise BlobReplace* doesn't work anyways
    case FSDTS.DataType[dataSet, band] of
      dspUInt8: begin
        ui8a := Round(miss);
        ui8b := Round(fill);
        BlobReplaceValue8Bit(into, FSDTS.Rows[dataSet] * FSDTS.Columns[dataSet], @ui8b, @ui8a);
      end;
      dspInt8: begin
        i8a := Round(miss);
        i8b := Round(fill);
        BlobReplaceValue8Bit(into, FSDTS.Rows[dataSet] * FSDTS.Columns[dataSet], @i8b, @i8a);
      end;
      dspUInt16: begin
        ui16a := Round(miss);
        ui16b := Round(fill);
        BlobReplaceValue16Bit(into, FSDTS.Rows[dataSet] * FSDTS.Columns[dataSet], @ui16b, @ui16a);
      end;
      dspInt16: begin
        i16a := Round(miss);
        i16b := Round(fill);
        BlobReplaceValue16Bit(into, FSDTS.Rows[dataSet] * FSDTS.Columns[dataSet], @i16b, @i16a);
      end;
      dspUInt32: begin
        ui32a := Round(miss);
        ui32b := Round(fill);
        BlobReplaceValue32Bit(into, FSDTS.Rows[dataSet] * FSDTS.Columns[dataSet], @ui32b, @ui32a);
      end;
      dspInt32: begin
        i32a := Round(miss);
        i32b := Round(fill);
        BlobReplaceValue32Bit(into, FSDTS.Rows[dataSet] * FSDTS.Columns[dataSet], @i32b, @i32a);
      end;
      dspInt64: ; // no conversion yet
      dspFloat32: begin
        f32a := miss;
        f32b := fill;
        BlobReplaceValue32Bit(into, FSDTS.Rows[dataSet] * FSDTS.Columns[dataSet], @f32b, @f32a);
      end;
      dspFloat64: BlobReplaceValueFloat64(into, FSDTS.Rows[dataSet] * FSDTS.Columns[dataSet], fill, miss);
    end;
end;

procedure TDSpSDTSRasterDriver.ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

procedure TDSpSDTSRasterDriver.ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpSDTSBroker
constructor TDSpSDTSBroker.Create;
begin
  inherited;
  FDescription := 'SDTS raster files';
  FFilter := '*CATD.DDF';
  FCaps := [dspReadRaster];
end;

destructor TDSpSDTSBroker.Destroy;
begin
  inherited;
end;

function TDSpSDTSBroker.CreateSource(const fn: string): TDSpSourceHandler;
begin
  Result := nil;
  Assert(false, 'SDTS is a read-only format');
end;

function TDSpSDTSBroker.OpenSource(const fn: string): TDSpSourceHandler;
var r   : TDSpSDTSRasterDriver;
begin
  r := TDSpSDTSRasterDriver.Open(fn);
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

function TDSpSDTSBroker.SupportsSource(const source: string): boolean;
begin
  Result := UpperCase(ExtractFileExt(source)) = '.DDF';
end;

function TDSpSDTSBroker.CanSave(const ds: TDSpSourceDataSet): string;
begin
  Result := '';
end;

initialization
{ Enable this method for static loading of the broker. Please note that the    }
{ DSpPFA unit MUST be loaded before this unit. If the broker is to be loaded   }
{ dynamically, you will have to provide a DLL with this unit.                  }
  PFA.RegisterBroker(TDSpSDTSBroker.Create);
end.
