{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpSDTS_RPE.pas. The Original Code is available through }
{ thehome page of DSpatial at http://dspatial.sourceforge.net.                 }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements the SDTS Raster Profile and Extensions (RPE) for reading}
{ gridded data in an SDTS transfer.                                            }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 07                                             }
{                                                                              }
{ Dependencies  : DSpSDTS (essential)                                          }
{                 DSpUtils (ValidateIndex)                                     }
{                 DSpISO8211 (TISO8211)                                        }
{                 DSpGlobals (RasterBandType)                                  }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.4  03Nov07 PFA overhaul.                                   }
{                 0.3  03Jul23 Projection string now WKT compliant(except SPCS)}
{                 0.2  03Jul18 Now supports multiple layers per raster         }
{                 0.1  03Apr17 Initial coding. Does not yet support multiple   }
{                              rasters, or multiple layers per raster.         }
{                                                                              }
{******************************************************************************}
unit DSpSDTS_RPE;
interface
uses DSpGlobals, DSpUtils, DSpSDTS;
type
  TSDTSDataType = (sdtsByte, sdtsShort, sdtsWord, sdtsSmallInt, sdtsCardinal,
                   sdtsInteger, sdtsSingle, sdtsDouble);

  TSDTSSequencing = (sdtsBSQ, sdtsBIL, sdtsBIP);

  TSDTS_RPE = class(TSDTStransfer)
  private
    FNumRasters: integer;
    FRasters   : array of array of integer;    //rasters[].layers[]

    function  GetDescriptiveName: string;
    function  GetRasterComment(r: integer): string;
    function  GetProjectionString: string;
    function  GetRasterSequencing(r: integer): TDSpRasterBandFormat;
    function  GetNumLayers(r: integer): integer;
    function  GetOriginX(r: integer): TDSpFloat64;
    function  GetOriginY(r: integer): TDSpFloat64;
    function  GetNumRows(r: integer): integer;
    function  GetNumColumns(r: integer): integer;
    function  GetLayerType(r, l: integer): TDSpRasterBandType;
    function  GetDataType(r, l: integer): TDSpDataType;
    function  GetDataSize(r, l: integer): integer;
    function  GetXSize(r: integer): TDSpFloat64;
    function  GetYSize(r: integer): TDSpFloat64;
    function  GetNullValue(r: integer): TDSpFloat64;
    function  GetMissingValue(r: integer): TDSpFloat64;
    function  GetMinValue(r: integer): TDSpFloat64;
    function  GetMaxValue(r: integer): TDSpFloat64;
  public
    constructor Open(fn: string);
    destructor  Destroy; override;

    function  ReadRaster(r, l: integer; p: pointer): boolean;
    procedure Extent(r: integer; var n, e, s, w: TDSpFloat64);

    property  Descriptivename: string read GetDescriptiveName;
    property  Rasters: integer read FNumRasters;
    property  RasterComment[r: integer]: string read GetRasterComment;
    property  ProjectionString: string read GetProjectionString;
    property  RasterSequencing[r: integer]: TDSpRasterBandFormat read GetRasterSequencing;
    property  Layers[r: integer]: integer read GetNumLayers;
    property  OriginX[r: integer]: TDSpFloat64 read GetOriginX;
    property  OriginY[r: integer]: TDSpFloat64 read GetOriginY;
    property  Rows[r: integer]: integer read GetNumRows;
    property  Columns[r: integer]: integer read GetNumColumns;
    property  LayerType[r, l: integer]: TDSpRasterBandType read GetLayerType;
    property  DataType[r, l: integer]: TDSpDataType read GetDataType;
    property  DataSize[r, l: integer]: integer read GetDataSize;
    property  XSize[r: integer]: TDSpFloat64 read GetXSize;
    property  YSize[r: integer]: TDSpFloat64 read GetYSize;
    property  NullValue[r: integer]: TDSpFloat64 read GetNullValue;
    property  MissingValue[r: integer]: TDSpFloat64 read GetMissingValue;
    property  MinValue[r: integer]: TDSpFloat64 read GetMinValue;
    property  MaxValue[r: integer]: TDSpFloat64 read GetMaxValue;
  end;

implementation
uses SysUtils, DSpISO8211;

constructor TSDTS_RPE.Open(fn: string);
var r, l: integer;
begin
  inherited;
  if not IsSRPE then
    raise Exception.Create('This file does not contain a valid raster profile');
  ReadRSDF;
  ReadLDEF;
  FNumRasters := Length(FRsdf);
  SetLength(FRasters, FNumRasters);
  for r := 0 to FNumRasters - 1 do begin
    SetLength(FRasters[r], FRsdf[r].NLAY);
    for l := 0 to FRsdf[r].NLAY - 1 do begin
      FRasters[r,l] := FindRCIDIndex(FLdef, FRsdf[r].LYID[l].RCID);
    end;
  end;
end;

destructor TSDTS_RPE.Destroy;
var r: integer;
begin
  for r := 0 to FNumRasters - 1 do Finalize(FRasters[r]);
  Finalize(FRasters);
  inherited;
end;

function TSDTS_RPE.GetDescriptiveName: string;
begin
  Result := FIden.TITL;
end;

function TSDTS_RPE.GetRasterComment(r: integer): string;
begin
  ValidateIndex(0, FNumRasters-1, r);
  Result := FRsdf[r].COMT;
end;

function TSDTS_RPE.GetProjectionString: string;
const GREENWICH_DEGREE: string = 'PRIMEM["Greenwich",0],UNIT["Degree",0.017453292519943295]]';
var cs, datum, zone: string;
    cm: single;
begin
  if FIden.EXSP in [1,2] then begin
    datum := Uppercase(FXref[0].HDAT);
    if datum = 'NAS' then datum := 'GEOGCS["GCS_North_American_1927",DATUM["D_North_American_1927",SPHEROID["Clarke_1866",6378206.4,294.9786982]]'
    else if datum = 'NAX' then datum := 'GEOGCS["GCS_North_American_1983",DATUM["North_American_1983",SPHEROID["GRS_1980",6378137,298.257222101]]'
    else if datum = 'WGA' then datum := 'GEOGCS["GCS_WGS_1960",DATUM["WGS_1960",SPHEROID["WGS_1960",6378165,298.3]]'
    else if datum = 'WGB' then datum := 'GEOGCS["GCS_WGS_1966",DATUM["WGS_1966",SPHEROID["WGS_1966",6378145,298.25]]'
    else if datum = 'WGC' then datum := 'GEOGCS["GCS_WGS_1972",DATUM["WGS_1972",SPHEROID["WGS_1972",6378135,298.26]]'
    else if datum = 'WGE' then datum := 'GEOGCS["GCS_WGS_1984",DATUM["WGS_1984",SPHEROID["WGS_1984",6378137,298.257223563]]'
    else datum := '';
  end else datum := '';

  cs := UpperCase(FXref[0].RSNM);
  if cs = 'GEO' then Result := 'GEOGCS["GCS_Geographic",DATUM["Clarke_1866",SPHEROID["Clarke_1866",6378206.4,294.9786982]],' + GREENWICH_DEGREE
  else begin
    zone := FXref[0].ZONE;
    if cs = 'UTM' then begin
      cm := StrToFloat(zone);
      if cm > 0 then zone := FloatToStr(cm) + 'N' else zone := FloatToStr(cm) + 'S';
      cm := 6 * cm - 183;
      Result := Format('PROJCS["UTM_Zone_%s",%s,%s,PROJECTION["Transverse_Mercator"],PARAMETER["False_Easting",500000],PARAMETER["False_Northing",0],PARAMETER["Central_Meridian",%1.0f],PARAMETER["Scale_Factor",0.9996],PARAMETER["Latitude_Of_Origin",0],UNIT["Meter",1]]', [zone, datum, GREENWICH_DEGREE, cm])
    end else if cs = 'UPS' then Result := Format('PROJCS["North_Pole_Stereographic",%s,%s,PROJECTION["Stereographic"],PARAMETER["False_Easting",0],PARAMETER["False_Northing",0],PARAMETER["Central_Meridian",0],PARAMETER["Scale_Factor",1],PARAMETER["Latitude_Of_Origin",90],UNIT["Meter",1]]', [datum, GREENWICH_DEGREE])
    else if cs = 'SPCS' then begin
      // Not yet WKT compliant
      Result := 'State Plane Coordinate System - ' + FXref[0].ZONE;
    end else Result := '';  // will become a null projection
  end;
end;

function TSDTS_RPE.GetRasterSequencing(r: integer): TDSpRasterBandFormat;
var seq: string[2];
begin
  ValidateIndex(0, FNumRasters-1, r);
  seq := UpperCase(FRSDF[r].CSCD);
  case seq[2] of
    'I': Result := dspRasterFlat;              // BSQ
    'J': Result := dspRasterInterleaveLine;    // BIL
    'L': Result := dspRasterInterleavePixel;   // BIP
  else raise Exception.Create('Raster sequencing code invalid');
  end;
end;

function TSDTS_RPE.GetNumLayers(r: integer): integer;
begin
  ValidateIndex(0, FNumRasters-1, r);
  Result := FRsdf[r].NLAY;
end;

function TSDTS_RPE.GetOriginX(r: integer): TDSpFloat64;
begin
  ValidateIndex(0, FNumRasters-1, r);
  Result := FRsdf[r].SADR.X;
end;

function TSDTS_RPE.GetOriginY(r: integer): TDSpFloat64;
begin
  ValidateIndex(0, FNumRasters-1, r);
  Result := FRsdf[r].SADR.Y;
end;

// This is a layer attribute, but the RPE mandates LDEF.NROW to be equal to RSDF.RWXT
function TSDTS_RPE.GetNumRows(r: integer): integer;
begin
  ValidateIndex(0, FNumRasters-1, r);
  Result := FRsdf[r].RWXT;
end;

// This is a layer attribute, but the RPE mandates LDEF.NCOL to be equal to RSDF.CLXT
function TSDTS_RPE.GetNumColumns(r: integer): integer;
begin
  ValidateIndex(0, FNumRasters-1, r);
  Result := FRsdf[r].CLXT;
end;

function TSDTS_RPE.GetLayerType(r, l: integer): TDSpRasterBandType;
begin
  ValidateIndex(0, FNumRasters-1, r);
  ValidateIndex(0, High(FRasters[r]), l);
  if UpCase(FLdef[FRasters[r,l]].CODE[1]) = 'V' then Result := dspRasterBandSurface
  else Result := dspRasterBandOrdinal;
end;

function TSDTS_RPE.GetDataType(r, l: integer): TDSpDataType;
var fmt: string;
begin
  ValidateIndex(0, FNumRasters-1, r);
  ValidateIndex(0, High(FRasters[r]), l);
  fmt := FDdsh[FindDdshIndex(FLdef[FRasters[r,l]].CMNM)].FMT;
  if (fmt = 'I') or (fmt = 'BI32') then Result := dspInt32
  else if fmt = 'R' then Result := dspFloat64
  else if fmt = 'BI8' then Result := dspInt8
  else if fmt = 'BI16' then Result := dspInt16
  else if fmt = 'BUI8' then Result := dspUInt8
  else if fmt = 'BUI16' then Result := dspUInt16
  else if fmt = 'BUI32' then Result := dspUInt32
  else if fmt = 'BFP32' then Result := dspFloat32
  else raise Exception.Create('Cell module has unsupported format for data');
end;

function TSDTS_RPE.GetDataSize(r, l: integer): integer;
var fmt: string;
begin
  ValidateIndex(0, FNumRasters-1, r);
  ValidateIndex(0, High(FRasters[r]), l);
  fmt := FDdsh[FindDdshIndex(FLdef[FRasters[r,l]].CMNM)].FMT;
  if (fmt = 'I') or (fmt = 'BI32') or (fmt = 'BUI32') or (fmt = 'BFP32')then Result := 4
  else if (fmt = 'BI8') or (fmt = 'BUI8') then Result := 1
  else if (fmt = 'BI16') or (fmt = 'BUI16') then Result := 2
  else if fmt = 'R' then Result := 8
  else raise Exception.Create('Cell module has unsupported format for data');
end;

function TSDTS_RPE.ReadRaster(r, l: integer; p: pointer): boolean;
var cellParser: TISO8211;
    cellRecord: TSDTS_CELL;
    size, cols: integer;
    i         : integer;
    pos: pointer;
    nm: string;
begin
  if RasterSequencing[r] <> dspRasterFlat then
    raise Exception.Create('Sorry, but this reader can not handle interleaved rasters.');

  ValidateIndex(0, FNumRasters-1, r);
  ValidateIndex(0, High(FRasters[r]), l);
  size := DataSize[r,l];
  cols := Columns[r];

  // Find the CELL module by name
  nm := FLdef[FRasters[r,l]].CMNM;
  for i := Low(FCatd) to High(FCatd) do
    if FCatd[i].NAME = nm then Break; // CELL module should not be external
  cellParser := ReadCELL(FRoot + FCatd[i].FIL_);

  Result := true;
  try
    cellRecord := ReadCELLRecord(cellParser);
    while cellRecord.MODN <> 'NULL' do begin
      pos := pointer(integer(p) + ((cellRecord.ROWI - 1) * cols + cellRecord.COLI - 1) * size);
      Move(cellRecord.CVLS^, pos^, cellRecord.CVLSsize);
      FreeMem(cellRecord.CVLS);
      cellRecord := ReadCELLRecord(cellParser);
    end;
  except
    Result := false;
  end;
end;

// This is a layer attribute, but the RPE mandates conformity between LDEF and RSDF
function TSDTS_RPE.GetXSize(r: integer): TDSpFloat64;
begin
  ValidateIndex(0, FNumRasters-1, r);
  Result := FIref[r].XHRS;
end;

// This is a layer attribute, but the RPE mandates conformity between LDEF and RSDF
function TSDTS_RPE.GetYSize(r: integer): TDSpFloat64;
begin
  ValidateIndex(0, FNumRasters-1, r);
  Result := FIref[r].YHRS;
end;

procedure TSDTS_RPE.Extent(r: integer; var n, e, s, w: TDSpFloat64);
var v: TDSpFloat64;
begin
  ValidateIndex(0, FNumRasters-1, r);
  v := FIref[FindRCIDIndex(FIref, FRsdf[r].ISID.RCID)].XHRS;
  w := FRsdf[r].SADR.X;
  e := w + FRsdf[r].CLXT * v;
  n := FRsdf[r].SADR.Y;
  s := n - FRsdf[r].RWXT * v;
end;

function TSDTS_RPE.GetNullValue(r: integer): TDSpFloat64;
begin
  Result := FDdom[FindDDOMIndexByDVDF('FILL')].DVAL;
end;

function TSDTS_RPE.GetMissingValue(r: integer): TDSpFloat64;
begin
  Result := FDdom[FindDDOMIndexByDVDF('VOID')].DVAL;
end;

function TSDTS_RPE.GetMinValue(r: integer): TDSpFloat64;
begin
  Result := FDdom[FindDDOMIndexByDVDF('MIN')].DVAL;
end;

function TSDTS_RPE.GetMaxValue(r: integer): TDSpFloat64;
begin
  Result := FDdom[FindDDOMIndexByDVDF('MAX')].DVAL;
end;

end.
