{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpCalibrate.pas. The Original Code is available through}
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements fast calibration of data using the formula:             }
{           Dest := (Source * Scalar) + Offset                                 }
{                                                                              }
{ This formula is often used to compress satellite imagery in a more compact   }
{ data type.                                                                   }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 July 21                                                 }
{                                                                              }
{ Dependencies  : none                                                         }
{                                                                              }
{ Version       : 0.1  03Jul21 Initial coding. Port from an earlier code base  }
{                              developed by the Initial Developer.             }
{                                                                              }
{******************************************************************************}
unit DSpCalibrate;
{$I DSpatial.inc}
interface

  procedure SmallScaleOffsetSingle(Num : cardinal; Src, Dest : pointer; Scale, Off : double); stdcall;

implementation
{$IFDEF BUILD_SSE}
const
  Zero4 : array [0..3] of single = (0, 0, 0, 0);
var
  Scalar4 : array[0..3] of single;
{$ENDIF}

procedure SmallScaleOffsetSingle(Num : cardinal; Src, Dest : pointer; Scale, Off : double); stdcall;
// Dest := (Src * Scalar) + Offset; Src := ^smallint[]; Dest ^single[]
asm
// Load array addresses and scalar into FPU registers
  push ebx
  mov  eax,Num
  lea  ecx,Scale
  fld  qword ptr [ecx]      // 0: Scale
  {$IFDEF BUILD_SSE}        // Move Scale -> scalar4 -> xmm1
  shr  eax,2                // reduce number of iterations
  lea  ecx,Scalar4
  fst  dword ptr [ecx]
  fst  dword ptr [ecx+$04]
  fst  dword ptr [ecx+$08]
  fst  dword ptr [ecx+$0C]
  db   $0F,$10,$09          // movups xmm1,[ecx]: xmm1: Scale x 4
  {$ENDIF}
  lea  edx,Off
  fld  qword ptr [edx]      // 0: Off; 1: Scale
  {$IFDEF BUILD_SSE}
  fst  dword ptr [ecx]      // move Off -> scalar4 -> xmm2
  fst  dword ptr [ecx+$04]
  fst  dword ptr [ecx+$08]
  fst  dword ptr [ecx+$0C]
  db   $0F,$10,$11          // movups xmm2,[ecx]: xmm2: Off x 4
  {$ENDIF}
  mov  edx,Dest
  mov  ebx,Src
@Loop:
  {$IFDEF BUILD_SSE}
  fild word ptr [ebx]       // load 4 Src values
  fild word ptr [ebx+$02]
  fild word ptr [ebx+$04]
  fild word ptr [ebx+$06]
  add  ebx,8

  fstp dword ptr [ecx+$0C]  // move Src -> scalar4 -> xmm0
  fstp dword ptr [ecx+$08]
  fstp dword ptr [ecx+$04]
  fstp dword ptr [ecx]
  db   $0F,$10,$01          // movups xmm0,[ecx]: xmm0: Src x 4

  db   $0F,$59,$C1          // mulps xmm0,xmm1 -> Src * Scale
  db   $0F,$58,$C2          // addps xmm0,xmm2 -> Src * Scale + Off
  db   $0F,$29,$02          // movaps [edx],xmm0
  add  edx,16
  {$ELSE}
  fild word ptr [ebx]       // 0: Src; 1: Off; 2: Scale
  add  ebx,2                // Inc(Src)
  fmul st(0),st(2)          // 0: Src * Scale; 1: Off; 2: Scale
  fadd st(0),st(1)          // 0: Src * Scale + Off; 1: Off; 2: Scale
  fstp dword ptr [edx]      // 0: Off; 1: Scale
  add  edx,4                // Inc(Dest)
  {$ENDIF}
  dec  eax                  // Dec(Num)
  jnz  @Loop

  fstp st(1)
  fstp st(0)                // clear FPU stack
  pop  ebx
end;

end.
 