{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpCoordinateSystem.pas. The Original Code is available }
{ through the home page of DSpatial at http://dspatial.sourceforge.net.        }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit defines coordinate systems in which data set coordinates are       }
{ expressed. A coordinate system consists of a projection to map its geographic}
{ coordinate system to the cartesian coordinate system and a conversion factor }
{ to convert data set units to meters. The projection contains a geographic    }
{ coordinate system establishes a system to express latitude and longitude for }
{ any particular location on Earth, using any particular ellipsoid and local   }
{ datum. The cartesian coordinate system and projection together provide a     }
{ complete solution to converting between cartesian coordinate systems.        }
{                                                                              }
{ DSpatial uses the OpenGIS WKT format for conveying coordinate system         }
{ information to the user and between modules (e.g. PFA reader and data set).  }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 July 27                                                 }
{                                                                              }
{ Dependencies  : DSpProjection                                                }
{                 DSpGlobals                                                   }
{                 SysUtils (Exception, StrToFloat, FloatToString)              }
{                                                                              }
{ Version       : 0.1  03Jul27 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpCoordinateSystem;
{$I DSpatial.inc}
interface
uses DSpGlobals, DSpProjection;
type
  TDSpCoordinateSystem = class
  private
    FRefs: integer;
  protected
    FName: string;
    FToMeter: TDSpUnit;    // convert data set units to meters
    FProj: TDSpProjection; // projection system of the coordinate system
  public
    constructor Create(name: string; proj: TDSpProjection; linearUnit: TDSpUnit);
    destructor  Destroy; override;

    function    AddRef: TDSpCoordinateSystem;
    procedure   Release;

    function    Describe(descr: TDSpDescribe): string;
    function    IsCompatible(cs: TDSpCoordinateSystem): boolean;

    property    Projection: TDSpProjection read FProj;
    property    UOM: TDSpUnit read FToMeter;
  end;

  function CreateCoordinateSystemFromWKT(const wkt: string): TDSpCoordinateSystem;

implementation
uses SysUtils;
////////////////////////////////////////////////////////////////////////////////
// TDSpCoordinateSystem
constructor TDSpCoordinateSystem.Create(name: string; proj: TDSpProjection; linearUnit: TDSpUnit);
begin
  inherited Create;
  FName := name;
  FProj := proj;
  FToMeter := linearUnit;
  FRefs := 1;
end;

destructor TDSpCoordinateSystem.Destroy;
begin
  FToMeter.Free;
  FProj.Free;
  inherited;
end;

function TDSpCoordinateSystem.AddRef: TDSpCoordinateSystem;
begin
  Inc(FRefs);
  Result := self;
end;

procedure TDSpCoordinateSystem.Release;
begin
  Dec(FRefs);
  if FRefs = 0 then Free;
end;

// Describe the coordinate system in friendly format, or as a WKT-compliant string
function TDSpCoordinateSystem.Describe(descr: TDSpDescribe): string;
begin
  if FProj is TDSpProjectionGeographic then Result := FProj.Describe(descr)
  else if FProj is TDSpNullProjection then Result := FProj.Describe(descr)
  else if descr = dspWKT then
    Result := Format('PROJCS["%s",%s,%s]', [FName, FProj.Describe(descr), FToMeter.Describe(descr)])
  else Result := 'Coordinate system: ' + FName + #13#10 + FToMeter.Describe(descr) + 'to meter)'#13#10 +
                 FProj.Describe(descr);
end;

function TDSpCoordinateSystem.IsCompatible(cs: TDSpCoordinateSystem): boolean;
begin
  Result := FProj.IsCompatible(cs.Projection) and FToMeter.IsCompatible(cs.UOM);
end;

////////////////////////////////////////////////////////////////////////////////
// WKT coordinate system parser
function CreateCoordinateSystemFromWKT(const wkt: string): TDSpCoordinateSystem;
type TWKTParameter = record
                       name: string;
                       value: double;
                     end;
var cur, count: integer;
    csName, term, pmname: string;
    toMeter: TDSpUnit;
    pmlng: double;
    param: TWKTParameter;
    proj : TDSpProjection;
    datum: TDSpDatum;

  function ParseLiteralString: string;
  begin
    // Parse to begin of quoted string
    while wkt[cur] <> '"' do Inc(cur);
    Inc(cur);
    // Parse the literal
    count := cur;
    while wkt[count] <> '"' do Inc(count);
    Result := Copy(wkt, cur, count - cur);
    // Update cur, skip over "
    cur := count + 1;
  end;

  function ParseNumber: double;
  begin
    count := cur;
    while wkt[count] in ['0'..'9', '+', '-', '.', 'E', 'e'] do Inc(count);
    Result := StrToFloat(Copy(wkt, cur, count - cur));
    cur := count;
  end;

  procedure ParseAuthority;
  begin
    // Authority is always optional, so keyword starts at cur+1
    if Copy(wkt, cur+1, 9) <> 'AUTHORITY' then Exit;
    Inc(cur, 11);
    ParseLiteralString;
    Inc(cur); // skip over ,
    ParseLiteralString;
    Inc(cur); // skip oner ]
  end;

  function ParseParameter: TWKTParameter;
  begin
    Result.name := ''; // flag termination condition

    // Parameters are always optional, there may be a "," before
    if wkt[cur] = ',' then Inc(cur);
    if Copy(wkt, cur, 9) <> 'PARAMETER' then Exit;
    Inc(cur, 10);
    Result.Name := ParseLiteralString;
    Inc(cur); // skip over ,
    Result.value := ParseNumber;
    Inc(cur); // skip over ]
  end;

  function ParseUnit: TDSpUnit;
  var name: string;
  begin
    if Copy(wkt, cur, 4) <> 'UNIT' then raise Exception.Create('');
    Inc(cur, 5);
    name := ParseLiteralString;
    Inc(cur); // skip over ,
    Result := TDSpUnit.Create(name, ParseNumber);
    ParseAuthority;
    Inc(cur); // skip over ]
  end;

  function ParseDatum: TDSpDatum;
  var name, sphname: string;
      a, invf: double;
  begin
    if Copy(wkt, cur, 5) <> 'DATUM' then raise Exception.Create('');
    Inc(cur, 6);
    name := ParseLiteralString;
    Inc(cur); // skip over ,

    // SPHEROID
    if Copy(wkt, cur, 8) <> 'SPHEROID' then raise Exception.Create('');
    Inc(cur, 9);
    sphname := ParseLiteralString;
    Inc(cur); // skip over ,
    a := ParseNumber;
    Inc(cur);
    invf := ParseNumber;
    ParseAuthority;
    Inc(cur); // skip over ]
    Result := TDSpDatum.Create(name, TDSpSpheroid.Create(sphname, a, invf));

    // TOWGS84
    // Optional, so keyword starts at cur+1
    if Copy(wkt, cur+1, 7) = 'TOWGS84' then begin
      Inc(cur, 9);
      Result.ToWGS84[dspDX] := ParseNumber;
      Inc(cur); // skip over ,
      Result.ToWGS84[dspDY] := ParseNumber;
      Inc(cur); // skip over ,
      Result.ToWGS84[dspDZ] := ParseNumber;
      Inc(cur); // skip over ,
      Result.ToWGS84[dspEX] := ParseNumber;
      Inc(cur); // skip over ,
      Result.ToWGS84[dspEY] := ParseNumber;
      Inc(cur); // skip over ,
      Result.ToWGS84[dspEZ] := ParseNumber;
      Inc(cur); // skip over ,
      Result.ToWGS84[dspPPM] := ParseNumber;
      Inc(cur); // skip over ]
    end;

    ParseAuthority;
    Inc(cur); // skip over ]
  end;

  procedure ParsePrimeMeridian;
  begin
    if Copy(wkt, cur, 6) <> 'PRIMEM' then raise Exception.Create('');
    Inc(cur, 7);
    pmname := ParseLiteralString;
    Inc(cur); // skip over ,
    pmlng := ParseNumber;
    ParseAuthority;
    Inc(cur); // skip oner ]
  end;

  procedure ParseAxes(num: integer);
  // Parses properly, but not yet supported
  var i: integer;
  begin
    for i := 1 to num do begin
      // Axes are always optional, so keyword starts at cur+1
      if Copy(wkt, cur+1, 4) <> 'AXIS' then Exit;
      Inc(cur, 6);
      repeat Inc(cur) until wkt[cur] in [']', ')'];
      if i < num then Inc(cur); // skip over ,
    end;
  end;

  function ParseGEOGCS: TDSpGeographicCoordinateSystem;
  var name: string;
  begin
    if Copy(wkt, cur, 6) <> 'GEOGCS' then raise Exception.Create('');
    Inc(cur, 7);
    name := ParseLiteralString;
    Inc(cur); // skip over ,
    datum := ParseDatum;
    Inc(cur); // skip over ,
    ParsePrimeMeridian;     // sets pmname and pmlong variables
    Inc(cur);
    Result := TDSpGeographicCoordinateSystem.Create(datum, name, pmname, pmlng, ParseUnit);
    ParseAxes(2);
    ParseAuthority;
    Inc(cur); // skip over ]
  end;

  procedure ParseGEOCCS;
  // Parses properly, but not yet supported
  begin
    if Copy(wkt, cur, 6) <> 'GEOCCS' then raise Exception.Create('');
    Inc(cur, 7);
    ParseLiteralString;
    Inc(cur); // skip over ,
    ParseDatum.Free;
    Inc(cur); // skip over ,
    ParsePrimeMeridian;
    Inc(cur);
    ParseUnit.Free;
    ParseAxes(3);
    ParseAuthority;
    Inc(cur); // skip over ]
  end;

  procedure ParseVERTCS;
  // Parses properly, but not yet supported
  begin
    if Copy(wkt, cur, 7) <> 'VERT_CS' then raise Exception.Create('');
    Inc(cur, 8);
    ParseLiteralString;
    Inc(cur); // skip over ,

    // VERT_DATUM
    if Copy(wkt, cur, 10) <> 'VERT_DATUM' then raise Exception.Create('');
    Inc(cur, 11);
    ParseLiteralString;
    Inc(cur); // skip over ,

    ParseNumber;
    ParseAuthority;
    Inc(cur); // skip over ]
    Inc(cur); // skip over ,
    ParseUnit.Free;
    ParseAxes(1);
    ParseAuthority;
    Inc(cur); // skip over ]
  end;

  function ParsePROJCS: string;
  var prjName, parName: string;
      geogcs: TDSpGeographicCoordinateSystem;
      lon0, lat0, fe, fn, k0, lat1, lat2: double;
      f: double;
  begin
    if Copy(wkt, cur, 6) <> 'PROJCS' then raise Exception.Create('');
    Inc(cur, 7);
    Result := ParseLiteralString;
    Inc(cur); // skip over ,
    geogcs := ParseGEOGCS;
    Inc(cur); // skip over ,

    // PROJECTION
    if Copy(wkt, cur, 10) <> 'PROJECTION' then raise Exception.Create('');
    Inc(cur, 11);
    prjName := ParseLiteralString;
    ParseAuthority;
    Inc(cur); // skip over ]

    lon0 := 0.0; fe := 0.0; fn := 0.0; k0 := 1.0; lat1 := 0.0; lat2 := 0.78; lat0 := 0.0;
    f := geogCS.ToRadian;
    repeat
      param := ParseParameter;
      if param.name = '' then Break;
      parName := UpperCase(param.name);
      if (parName = 'CENTRAL_MERIDIAN') or (parName = 'LONGITUDE_OF_CENTER') then lon0 := param.value * f  // all
      else if parName = 'FALSE_EASTING' then fe := param.value              // all
      else if parName = 'FALSE_NORTHING' then fn := param.value             // all
      else if parName = 'SCALE_FACTOR' then k0 := param.value               // STER (std, oblique, polar), TM, MERC
      // EQDC, AZEQ, LAEA, AEAC, LCC, POLY, GNOM, STER, ORTHO:
      else if (parName = 'LATITUDE_OF_CENTER') or (parName = 'LATITUDE_OF_ORIGIN') then lat0 := param.value * f
      else if parName = 'STANDARD_PARALLEL_1' then lat1 := param.value * f  // AEAC, LCC, EQDC
      else if parName = 'STANDARD_PARALLEL_2' then lat2 := param.value * f  // AEAC, LCC, EQDC
      else if (parName = 'FIPSZONE') or (parName = 'ZONE') or (parName = 'AZIMUTH') or
              (parName = 'LONGITUDE_OF_POINT_1') or (parName = 'LATITUDE_OF_POINT_1') or
              (parName = 'LONGITUDE_OF_POINT_2') or (parName = 'LATITUDE_OF_POINT_2') or
              (parName = 'LONGITUDE_OF_POINT_3') or (parName = 'LATITUDE_OF_POINT_3') or
              (parName = 'PERSPECTIVE_POINT_HEIGHT') or (parName = 'LANDSAT_NUMBER') or
              (parName = 'PATH_NUMBER') then begin {disregard} end
      else raise Exception.Create('');
    until false;

    toMeter := ParseUnit;
    ParseAxes(2);
    ParseAuthority;
    Inc(cur); // skip over ]

    // Build the projection object
    if prjName = 'Transverse_Mercator' then proj := TDSpTransverseMercator.Create(geogCS, lon0, fe, fn, lat0, k0)
    else if prjName = 'Cassini_Soldner' then proj := TDSpCassini.Create(geogCS, lon0, fe, fn, lat0)
    else if prjName = 'Cylindrical_Equal_Area' then proj := TDSpCylindricalEqualArea.Create(geogCS, lon0, fe, fn, lat1)
    else if prjName = 'Equirectangular' then proj := TDSpEquirectangular.Create(geogCS, lon0, fe, fn, lat0)
    else if prjName = 'Gall_Stereographic' then proj := TDSpGallStereographic.Create(geogCS, lon0, fe, fn)
    else if prjName = 'Mercator_1SP' then proj := TDSpMercator.Create(geogCS, lon0, fe, fn, lat0, k0)
    else if prjName = 'Mercator_2SP' then proj := TDSpMercator.Create(geogCS, lon0, fe, fn, lat0, k0)
    else if prjName = 'Miller_Cylindrical' then proj := TDSpMiller.Create(geogCS, lon0, fe, fn, lat0)
//Hotine_Oblique_Mercator
//Swiss_Oblique_Cylindrical
//Transverse_Mercator_South_Orientated
//Laborde_Oblique_Mercator
//Oblique_Mercator

    else if prjName = 'Albers_Conic_Equal_Area' then proj := TDSpAlbersEqualAreaConic.Create(geogCS, lon0, fe, fn, lat0, lat1, lat2)
    else if prjName = 'Equidistant_Conic' then proj := TDSpEquidistantConic.Create(geogCS, lon0, fe, fn, lat0, lat1, lat2)
    else if prjName = 'Lambert_Conformal_Conic_1SP' then proj := TDSpLambertConformalConic.Create(geogCS, lon0, fe, fn, lat0)
    else if prjName = 'Lambert_Conformal_Conic_2SP' then proj := TDSpLambertConformalConic.Create(geogCS, lon0, fe, fn, lat0, lat1, lat2)
    else if prjName = 'Polyconic' then proj := TDSpPolyconic.Create(geogCS, lon0, fe, fn, lat0)
//Lambert_Conformal_Conic_2SP_Belgium

    else if prjName = 'Azimuthal_Equidistant' then proj := TDSpAzimuthalEquidistant.Create(geogCS, lon0, fe, fn, lat0)
    else if prjName = 'Gnomonic' then proj := TDSpGnomonic.Create(geogCS, lon0, fe, fn, lat0)
    else if prjName = 'Lambert_Azimuthal_Equal_Area' then proj := TDSpLambertAzimuthalEqualArea.Create(geogCS, lon0, fe, fn, lat0)
    else if prjName = 'Oblique_Stereographic' then proj := TDSpStereographic.Create(geogCS, lon0, fe, fn, lat0, k0)
    else if prjName = 'Orthographic' then proj := TDSpOrthographic.Create(geogCS, lon0, fe, fn, lat0)
    else if prjName = 'Polar_Stereographic' then proj := TDSpStereographic.Create(geogCS, lon0, fe, fn, lat0, k0)
    else if prjName = 'Stereographic' then proj := TDSpStereographic.Create(geogCS, lon0, fe, fn, lat0, k0)

    else if prjName = 'Eckert_IV' then proj := TDSpEckertIV.Create(geogCS, lon0, fe, fn)
    else if prjName = 'Eckert_VI' then proj := TDSpEckertVI.Create(geogCS, lon0, fe, fn)
    else if prjName = 'Mollweide' then proj := TDSpMollweide.Create(geogCS, lon0, fe, fn)
    else if prjName = 'Robinson' then proj := TDSpRobinson.Create(geogCS, lon0, fe, fn)
    else if prjName = 'Sinusoidal' then proj := TDSpSinusoidal.Create(geogCS, lon0, fe, fn)
    else if prjName = 'VanDerGrinten' then proj := TDSpVanDerGrinten.Create(geogCS, lon0, fe, fn)

    else proj := TDSpNullProjection.Create; // raise Exception.Create('');
{
Krovak
New_Zealand_Map_Grid
Tunisia_Mining_Grid
             }
end;

begin
  try
    cur := 1;
    // Read first term, must be either of PROJCS, GEOGCS, or LOCAL_CS. COMPD_CS,
    // GEOCCS, VERT_CS and FITTED_CS not yet supported
    term := Copy(wkt, 1, 4);
    if term = 'PROJ' then csName := ParsePROJCS
    else if term = 'GEOG' then begin
      csName := 'Geographic coordinate system';
      proj := TDSpProjectionGeographic.Create(ParseGEOGCS, 0.0, 0.0, 0.0);
      toMeter := TDSpUnit.Create('Null unit', 1.0);
    end else if term = 'LOCA' then begin
      csName := 'Local coordinate system (meter)';
      proj := TDSpNullProjection.Create;
      toMeter := TDSpUnit.Create('Meter', 1.0);
    end else raise Exception.Create('');
    Result := TDSpCoordinateSystem.Create(csName, proj, toMeter);
  except
//    if Assigned(proj) then proj.Free;
//    if Assigned(toMeter) then toMeter.Free;
    Result := TDSpCoordinateSystem.Create('Local coordinate system (meter)', TDSpNullProjection.Create, TDSpUnit.Create('Meter', 1.0));
  end;
end;

end.
