{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpDataMgr.pas. The Original Code is available through  }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit defines the data manager for DSpatial. A data manager controls     }
{ access to the data sources currently in use in an analysis environment. An   }
{ analysis environment can be a simple view, or a batch process manipulating   }
{ data. A single application can have as many data managers as required. The   }
{ data manager accesses the PFA to open, create, or save data sets. It serves  }
{ as an intermediary between the PFA and the internal representation of data in}
{ DSpatial.                                                                    }
{ The DSpDataMgr class is not interactive, so it can be used for batch         }
{ processing. Beware that it does not enforce compatibility between data sets  }
{ (e.g. projection), it is up to the implementor to check for the suitability  }
{ of data sets for any kind of operation.                                      }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 03                                             }
{                                                                              }
{ Dependencies  : DSpRaster                                                    }
{                 DSpVector                                                    }
{                 DSpDataSet                                                   }
{                 DSpPFA                                                       }
{                 DSpUtils (TDSpQueue)                                         }
{                 DSpGlobals                                                   }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.5  03Dec03 Added AddDataSource                             }
{                 0.4  03Nov20 Added AddDataSet                                }
{                 0.3  03Nov05 PFA overhaul                                    }
{                 0.2  03Sep25 Added support for closing data sets and handlers}
{                 0.1  03Jul18 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpDataMgr;
interface
uses DSpGlobals, DSpUtils, DSpPFA, DSpDataSet;
type
  TDSpDataMgr = class
  protected
    FDataSets   : array of TDSpDataSet;

    function    GetNumDataSets: integer;
    function    GetDataSet(i: integer): TDSpDataSet;
  public
    constructor Create;
    destructor  Destroy; override;

    function    OpenDataSet(h: TDSpSourceHandler; ndx: integer): TDSpDataSet;
    function    AddDataSource(src: TDSpSourceDataSet): TDSpDataSet;
    procedure   AddDataSet(ds: TDSpDataSet);
    procedure   CloseDataSet(ds: TDSpDataSet);
    property    DataSets: integer read GetNumDataSets;
    property    DataSet[i: integer]: TDSpDataSet read GetDataSet;

  end;

implementation
uses SysUtils, DSpRaster, DSpVector;

////////////////////////////////////////////////////////////////////////////////
// TDSpDataMgr
constructor TDSpDataMgr.Create;
begin
  inherited;
end;

destructor TDSpDataMgr.Destroy;
begin
  // for i := 0 to High(FDataSets) do ...
  Finalize(FDataSets);
  inherited;
end;

function TDSpDataMgr.OpenDataSet(h: TDSpSourceHandler; ndx: integer): TDSpDataSet;
var ds: TDSpSourceDataSet;
    dst: TDSpPFAFormat;
begin
  // Get the data set type
  ds := h.ReadLock(ndx);
  try
    dst := ds.DataSetType;
  finally
    ds.ReadUnlock;
  end;

  // Create the data set
  Result := nil; // satify the compiler
  case dst of
    dspPFARaster: Result := TDSpRaster.Open(h, ndx);
    dspPFAVector: Result := TDSpVector.Open(h, ndx);
  end;
  AddDataSet(Result);
end;

function TDSpDataMgr.AddDataSource(src: TDSpSourceDataSet): TDSpDataSet;
begin
  // Create the data set
  Result := nil; // satify the compiler
  case src.DataSetType of                // src has dspRead access
    dspPFARaster: Result := TDSpRaster.CreateFromDataSource(src);
//    dspPFAVector: Result := TDSpVector.CreateFromDataSource(src);
  end;
  AddDataSet(Result);
end;

procedure TDSpDataMgr.AddDataSet(ds: TDSpDataSet);
var i: integer;
begin
  i := Length(FDataSets);
  SetLength(FDataSets, i+1);
  FDataSets[i] := ds;
end;

procedure TDSpDataMgr.CloseDataSet(ds: TDSpDataSet);
var i, ndx: integer;
begin
  // First find the data set in FDataSets
  ndx := -1;
  Assert(Length(FDataSets) > 0);
  for i := 0 to High(FDataSets) do
    if ds = FDataSets[i] then begin
      ndx := i;
      Break;
    end;
  if ndx = -1 then Exit;

  // Now zap the object and rearrange FDataSets if needed
  ds.Free;
  if ndx < High(FDataSets) then
    for i := ndx to High(FDataSets) - 1 do FDataSets[i] := FDataSets[i+1];
  SetLength(FDataSets, High(FDataSets));
end;

function TDSpDataMgr.GetNumDataSets: integer;
begin
  Result := Length(FDataSets);
end;

function TDSpDataMgr.GetDataSet(i: integer): TDSpDataSet;
begin
  Assert(i in [0..High(FDataSets)]);
  Result := FDataSets[i];
end;

end.
