{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpDataSet.pas. The Original Code is available through  }
{  the home page of DSpatial at http://dspatial.sourceforge.net.               }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit defines the ancestor class of all spatial data sets. It contains   }
{ abstract methods and should never be instantiated. Use descendant classes    }
{ instead.                                                                     }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 05                                             }
{                                                                              }
{ Dependencies  : DSpGlobals                                                   }
{                 DSpAttributes                                                }
{                 DSpCoordinateSystem                                          }
{                 DSpPFA                                                       }
{                 SysUtils (IntToStr)                                          }
{                                                                              }
{ Version       : 0.1  03Nov05 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpDataSet;
interface
uses DSpGlobals, DSpPFA, DSpCoordinateSystem;
type
  // The Mother-of-all-DataSets. This class provides the basic access to the
  // different kinds of geospatial data sets managed by the PFA, including
  // common meta-information such as a unique identifier, projection parameters,
  // extent, unit. Otherwise, this class is a mostly abstract ancestor to more
  // useful classes.
  TDSpDataSet = class
  protected
    FLocalName: string;
    FHandler  : TDSpSourceHandler;    // referenced, not owned
    FIndex    : integer;

    function    GetSource: TDSpSourceDataSet;
    function    GetID: integer;
    function    GetName: string;
    procedure   SetName(name: string);
    function    GetCoordSys: TDSpCoordinateSystem;
  public
    constructor create(src: TDSpSourceHandler; ndx: integer);
    destructor  Destroy; override;

    property    Source: TDSpSourceDataSet read GetSource;
    property    SourceIndex: integer read FIndex write FIndex;
    property    Handler: TDSpSourceHandler read FHandler;
    property    ID: integer read GetID;
    property    Name: string read GetName write SetName;
    property    CoordinateSystem: TDSpCoordinateSystem read GetCoordSys;
    function    DescribeCS: string;
    function    Extent: TDSpExtent;
  end;

implementation
uses SysUtils, SyncObjs;

////////////////////////////////////////////////////////////////////////////////
// TDSpDataSet
constructor TDSpDataSet.Create(src: TDSpSourceHandler; ndx: integer);
begin
  inherited Create;
  FLocalName := '';
  FHandler := src;
  FIndex := ndx;
  FHandler.AddDataSetRef(FIndex);
end;

destructor TDSpDataSet.Destroy;
begin
  FHandler.CloseDataSetRef(FIndex);
  inherited;
end;

// This is a dangerous method to call. Don't call it unless you really have to.
// Be sure to always unlock the source data set as soon as possible. Failure to
// do so will effectively make the data source read-only (by disallowing write
// access because of pending reads), and your application will not unload
// properly because the lock can not be freed.
function TDSpDataSet.GetSource: TDSpSourceDataSet;
begin
  Result := FHandler.ReadLock(FIndex);
end;

function TDSpDataSet.GetID: integer;
var ds: TDSpSourceDataSet;
begin
  ds := FHandler.ReadLock(FIndex);
  try
    Result := ds.Id;
  finally
    ds.ReadUnlock;
  end;
end;

function TDSpDataSet.GetName: string;
var ds: TDSpSourceDataSet;
begin
  if FLocalName <> '' then Result := FLocalName
  else begin
    ds := FHandler.ReadLock(FIndex);
    try
      Result := ds.Name;
    finally
      ds.ReadUnlock;
    end;
  end;
end;

procedure TDSpDataSet.SetName(name: string);
begin
  FLocalName := name;
{ if mode = write then begin
    FInfo.BeginWrite;
    FInfo.DataSetName[FIndex] := name;
    FInfo.EndWrite;
  end;}
end;

function TDSpDataSet.GetCoordSys: TDSpCoordinateSystem;
var ds: TDSpSourceDataSet;
begin
  ds := FHandler.ReadLock(FIndex);
  try
    Result := ds.CoordinateSystem;
  finally
    ds.ReadUnlock;
  end;
end;

function TDSpDataSet.DescribeCS: string;
var ds: TDSpSourceDataSet;
begin
  ds := FHandler.ReadLock(FIndex);
  try
    Result := ds.DescribeCS(dspFriendly);
  finally
    ds.ReadUnlock;
  end;
end;

function TDSpDataSet.Extent: TDSpExtent;
var ds: TDSpSourceDataSet;
begin
  ds := FHandler.ReadLock(FIndex);
  try
    Result := ds.Extent;
  finally
    ds.ReadUnlock;
  end;
end;

end.
