{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpGlobals.pas. The Original Code is available through  }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit defines global variables, constants, types, etc. for use with      }
{ DSpatial. Most importantly, this unit contains the TDSpatial class, and its  }
{ DSpatial instance, which controls the global behaviour of any application    }
{ based on DSpatial. This unit should be included in virtually every other unit}
{ that is used in DSpatial and it should be the first listed DSpatial unit in  }
{ the uses clause of the project file (*.DPR).                                 }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 19                                             }
{                                                                              }
{ Dependencies  : DSpPFA (access to PFA instance)                              }
{                 DSpUtils (FloatsSame)                                        }
{                 Types (TRect)                                                }
{                 SysUtils (Format)                                            }
{                 Math (Ceil, Floor)                                           }
{                                                                              }
{ Version       : 0.7  03Dec19 Conditional compilation based on platform and   }
{                              compiler. TDSpRenderMap has selectable bit depth}
{                              of the map. TDSpExtent can handle nil arguments.}
{                              TDSpExtent.IsInside overloaded for point data.  }
{                 0.6  03Nov27 TDSpatial now supports idle processing.         }
{                 0.5  03Nov05 Message IDs changed. Changes to TDSpExtent and  }
{                              TDSpRenderMap.                                  }
{                 0.4  03Aug16 Added Rows and Columns property to render map   }
{                 0.3  03Aug14 Conversion of TDSpUnit to a class               }
{                              Addition of TDSpRasterRenderMap                 }
{                 0.2  03Jul18 Addition of TDSpatial                           }
{                              Changes/Additions to TDSpExtent                 }
{                 0.1  03Apr26 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpGlobals;
interface
{$IFDEF WIN32}
uses DSpWin32;
{$ENDIF}
const
  // Global constants
  DSP_MAXDATASIZE         = $60000000;  // 1.5GB, maximum allocation of memory
                                        // for data sets
  DSP_FOREVER             = $0FFFFFFF;  // wait a very long time

  // Message IDs for message passing and thread management
  DSPMSG_DSPBASE          = 0;
  DSPMSG_NULL             = DSPMSG_DSPBASE;
  DSPMSG_TERMINATE        = DSPMSG_DSPBASE + 1;
  DSPMSG_OK               = DSPMSG_DSPBASE + 2;
  DSPMSG_DONE             = DSPMSG_DSPBASE + 3;
  DSPMSG_ERROR            = DSPMSG_DSPBASE + 4;

  DSPMSG_PFABASE          = DSPMSG_DSPBASE + 200;
  DSPMSG_SOURCEINIT       = DSPMSG_PFABASE + 1;
  DSPMSG_SOURCEOPEN       = DSPMSG_PFABASE + 2;
  DSPMSG_SOURCESAVE       = DSPMSG_PFABASE + 3;
  DSPMSG_SOURCECLOSE      = DSPMSG_PFABASE + 4;
  DSPMSG_RASTERADDBAND    = DSPMSG_PFABASE + 11;
  DSPMSG_RASTERDELETEBAND = DSPMSG_PFABASE + 12;
  DSPMSG_BANDREADCOORDS   = DSPMSG_PFABASE + 31;
  DSPMSG_BANDREADAREA     = DSPMSG_PFABASE + 32;
  DSPMSG_BANDWRITECOORDS  = DSPMSG_PFABASE + 41;
  DSPMSG_VECTORINFO       = DSPMSG_PFABASE + 51;
  DSPMSG_RENDERRASTERMAP  = DSPMSG_PFABASE + 81;

type
  TDSpExtent = class;

  // Data types. For spatial data these identifiers must be used. For any other
  // persistent data these types are highly recommended.
  TDSpUInt8    = byte;
  TDSpUInt16   = word;
  TDSpUInt32   = cardinal;
  TDSpInt8     = shortint;
  TDSpInt16    = smallint;
  TDSpInt32    = integer;
  TDSpInt64    = Int64;
  TDSpFloat32  = single;
  TDSpFloat64  = double;

  PDSpUInt8    = ^TDSpUInt8;
  PDSpUInt16   = ^TDSpUInt16;
  PDSpUInt32   = ^TDSpUInt32;
  PDSpInt8     = ^TDSpInt8;
  PDSpInt16    = ^TDSpInt16;
  PDSpInt32    = ^TDSpInt32;
  PDSpInt64    = ^TDSpInt64;
  PDSpFloat32  = ^TDSpFloat32;
  PDSpFloat64  = ^TDSpFloat64;

  TDSpDataType = (dspNil, dspUInt8, dspUInt16, dspUInt32, dspInt8, dspInt16,
                  dspInt32, dspInt64, dspFloat32, dspFloat64);     // Don't change this order

  // Object states or results. These can be used by many different objects.
  TDSpState = (dspOk, dspError, dspIgnore, dspWarning, dspFail, dspPending,
               dspIdle, dspBusy, dspReading, dspWriting, dspProcessing,
               dspRendering, dspDeleting);
  TDSpStates = set of TDSpState;
  
  // PFA types
  TDSpPFACap = (dspReadRaster, dspWriteRaster, dspReadVector, dspWriteVector,
                dspReadAttribute, dspWriteAttribute);
  TDSpPFACaps = set of TDSpPFACap;

  TDSpPFAFormat = (dspDontKnow, dspPFARaster, dspPFAVector);

  TDSpPFAAccess = (dspNoAccess, dspRead, dspWrite, dspExcl);

  TDSpRasterBandType = (dspRasterBandNil, dspRasterBandSurface, dspRasterBandOrdinal, dspRasterBandImage);

  TDSpRasterBandFormat = (dspRasterFlat, dspRasterInterleaveLine, dspRasterInterleavePixel,
                          dspRasterTiled);

  // Data set types
  TDSpDatasetInfo = record
                      Feature : TDSpPFAFormat;
                      ID      : integer;
                      Extent  : TDSpExtent;
                      Name    : string;
                      CoordSys: string;
                    end;
  PDSpDatasetInfo = ^TDSpDatasetInfo;

  TDSpRasterBandInfo = record
                         BandIndex     : integer;       // index of raster band in file
                         DataType      : TDSpDataType;
                         BandType      : TDSpRasterBandType;
                         MissingValue  : TDSpFloat64;
                         Min, Max      : TDSpFloat64;
                       end;
  PDSpRasterBandInfo = ^TDSpRasterBandInfo;
  ADSpRasterBandInfo = array of TDSpRasterBandInfo;

  TDSpRasterInfo = record
                     BandFormat    : TDSpRasterBandFormat;
                     Bands         : integer;
                     Band          : ADSpRasterBandInfo;
                     Rows, Columns : integer;
                     PixelSize     : TDSpFloat64;
                   end;
  PDSpRasterInfo = ^TDSpRasterInfo;

  // Message parameter structures
  TDSpMsgReadRasterBandCoords = record
                                  Band   : integer;
                                  Data   : pointer;
                                  Coords : TRect;
                                end;
  PDSpMsgReadRasterBandCoords = ^TDSpMsgReadRasterBandCoords;


  // Other stuff
  TParameterEvent = procedure (Sender: TObject; p: double) of object;

  TDSpDescribe  = (dspFriendly, dspWKT);
  TDSpDirection = (dspHorz, dspVert);

  // TDSpatial
  // Never instantiate this class. Use the global DSpatial variable for access
  TDSpatial = class
  private
  protected
    constructor Create;
  public
    destructor  Destroy; override;

    procedure   IdleProcessing(Sender: TObject; var Done: boolean);
    function    NewDataSetID: TDSpUInt32;
    function    TempDir: string;
  end;

  // Extent management and manipulation
  // This class will not function properly with lat/long coordinate systems
  // that span the date line.
  TDSpExtent = class
  protected
    FNorth, FEast, FSouth, FWest : TDSpFloat64;
  private
    function GetArea: TDSpFloat64;
  public
    constructor Create(north, east, south, west: TDSpFloat64);

    function    Clone: TDSpExtent;
    procedure   CopyFrom(ex: TDSpExtent);
    procedure   UnionWith(ex: TDSpExtent);
    function    Intersects(ex: TDSpExtent): boolean; overload;
    function    Intersects(n, e, s, w: TDSpFloat64): boolean; overload;
    function    Intersection(ex: TDSpExtent): TDSpExtent;
    function    IsInside(ex: TDSpExtent): boolean; overload;
    function    IsInside(x, y: TDSpFloat64): boolean; overload;
    function    PixelExtent(const view: TDSpExtent; const area: TRect; pixel: TDSpFloat64): TRect;
    property    Area: TDSpFloat64 read GetArea;
    property    West: TDSpFloat64 read FWest write FWest;
    property    East: TDSpFloat64 read FEast write FEast;
    property    North: TDSpFloat64 read FNorth write FNorth;
    property    South: TDSpFloat64 read FSouth write FSouth;
  end;

  // Units of measure
  // There are two fundamental units: meter for linear features, and radian for
  // angular features. This class can describe any linear of angular unit and
  // it has a conversion factor to one of the fundamental units.
  TDSpUnit  = class
  protected
    FName   : string;
    FAbbrev : string;
    FConvert: double;
  public
    constructor Create(name: string; convert: double);

    function    Describe(descr: TDSpDescribe): string;
    function    IsCompatible(u: TDSpUnit): boolean;
    property    Name: string read FName write FName;
    property    Convert: double read FConvert;
    property    Short: string read FAbbrev write FAbbrev;
  end;

  // Raster render map
  TDSpRenderMap = class
  private
    FSize: integer;
  protected
    FMap: pointer;
    FRows, FColumns: integer;
    FDepth: integer;              // bit depth of the map

    function    GetRow(r: integer): pointer;
    function    GetMaxValue: cardinal;
  public
    constructor Create(r, c, depth: integer);
    destructor  Destroy; override;

    procedure   Resize(r, c: integer);
    procedure   Pack;

    property    Row[i: integer]: pointer read GetRow;
    property    Rows: integer read FRows;
    property    Columns: integer read FColumns;
    property    BitDepth: integer read FDepth;
    property    MaxValue: cardinal read GetMaxValue;
  end;

  TDSpRenderRasterMap = record
                          Band    : integer;
                          Map     : TDSpRenderMap;
                          MapArea : TRect;
                          Area    : TDSpExtent;
                          Pixel,
                          Min, Max: TDSpFloat64;
                        end;
  PDSpRenderRasterMap = ^TDSpRenderRasterMap;

  function DataSize(t: TDSpDataType): TDSpInt32;

var
  DSpatial: TDSpatial;

implementation
// Load the necessary resources
{$R DSpatial.res}
uses SysUtils, Math, DSpUtils, DSpPFA;

////////////////////////////////////////////////////////////////////////////////
// TDSpatial
constructor TDSpatial.Create;
begin
  inherited;
end;

destructor TDSpatial.Destroy;
begin
  inherited;
end;

procedure TDSpatial.IdleProcessing(Sender: TObject; var Done: boolean);
begin
  PFA.IdleProcessing;
  Done := true;
end;

function TDSpatial.NewDataSetID: TDSpUInt32;
begin
  Result := 0; // This obviously requires some work {;-)
end;

function TDSpatial.TempDir: string;
begin
  Result := GetCurrentDir; // This obviously requires some work {;-)
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpExtent
constructor TDSpExtent.Create(north, east, south, west: TDSpFloat64);
begin
  FEast  := east;
  FWest  := west;
  FNorth := north;
  FSouth := south;
end;

function TDSpExtent.Clone: TDSpExtent;
begin
  Result := TDSpExtent.Create(FNorth, FEast, FSouth, FWest);
end;

procedure TDSpExtent.CopyFrom(ex: TDSpExtent);
begin
  if ex = nil then Exit;
  FEast := ex.FEast;
  FWest := ex.FWest;
  FNorth := ex.FNorth;
  FSouth := ex.FSouth;
end;

procedure TDSpExtent.UnionWith(ex: TDSpExtent);
begin
  if ex = nil then Exit;
  if ex.FEast > FEast then FEast := ex.FEast;
  if ex.FWest < FWest then FWest := ex.FWest;
  if ex.FNorth > FNorth then FNorth := ex.FNorth;
  if ex.FSouth < FSouth then FSouth := ex.FSouth;
end;

function TDSpExtent.Intersects(ex: TDSpExtent): boolean;
begin
  if ex = nil then Result := false
  else Result := (FNorth > ex.FSouth) and (FSouth < ex.FNorth) and (FEast > ex.FWest) and (FWest < ex.FEast);
end;

function TDSpExtent.Intersects(n, e, s, w: TDSpFloat64): boolean;
begin
  Result := (FNorth > s) and (FSouth < n) and (FEast > w) and (FWest < e);
end;

function TDSpExtent.GetArea: TDSpFloat64;
begin
  Result := (FNorth - FSouth) * (FEast - FWest);
end;

function TDSpExtent.Intersection(ex: TDSpExtent): TDSpExtent;
var n, e, s, w: TDSpFloat64;
begin
  if ex = nil then Result := nil
  else begin
    if FWest < ex.West then w := ex.West else w := FWest;
    if FEast > ex.East then e := ex.East else e := FEast;
    if FSouth < ex.South then s := ex.South else s := FSouth;
    if FNorth > ex.North then n := ex.North else n := FNorth;
    if (w >= e) or (s >= n) then Result := nil
    else Result := TDSpExtent.Create(n, e, s, w);
  end;
end;

function TDSpExtent.IsInside(ex: TDSpExtent): boolean;
begin
  if ex = nil then Result := false
  else Result := (ex.West >= FWest) and (ex.East <= FEast) and
                 (ex.North <= FNorth) and (ex.South >= FSouth);
end;

function TDSpExtent.IsInside(x, y: TDSpFloat64): boolean;
begin
  Result := (x >= FWest) and (x <= FEast) and (y <= FNorth) and (y >= FSouth);
end;

// The extent of self in pixels relative to some viewing geometry
function TDSpExtent.PixelExtent(const view: TDSpExtent; const area: TRect; pixel: TDSpFloat64): TRect;
var invPixel: TDSpFloat64;
    overlap: TDSpExtent;
begin
  overlap := Intersection(view);
  if overlap = nil then begin
    Result.Left := 0;
    Result.Right := -1;
  end else begin
    invPixel := 1 / pixel;
    Result.Top := Floor((view.North - overlap.North) * invPixel);
    Result.Bottom := Ceil((view.North - overlap.South) * invPixel);
    Result.Right := Ceil((overlap.East - view.West) * invPixel);
    Result.Left := Floor((overlap.West - view.West) * invPixel);
    overlap.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpUnit
constructor TDSpUnit.Create(name: string; convert: double);
begin
  inherited Create;
  FName := name;
  FConvert := convert;
end;

function TDSpUnit.Describe(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then Result := Format('UNIT["%s",%15.13f]', [FName, FConvert])
  else Result := '  Unit: ' + FName + Format(' (%15.13f ', [FConvert]);
end;

function TDSpUnit.IsCompatible(u: TDSpUnit): boolean;
begin
  Result := FloatsSame(FConvert, u.Convert);
end;

////////////////////////////////////////////////////////////////////////////////
//
constructor TDSpRenderMap.Create(r, c, depth: integer);
begin
  Assert(depth in [8, 16, 32]);
  inherited Create;
  FRows := r;
  FColumns := c;
  FDepth := depth;
  FSize := r * c * FDepth div 8;
  if FSize = 0 then FMap := nil
  else FMap := AllocMem(FSize);
end;

destructor TDSpRenderMap.Destroy;
begin
  if FMap <> nil then FreeMem(FMap);
  inherited;
end;

procedure TDSpRenderMap.Resize(r, c: integer);
begin
  FRows := r;
  FColumns := c;
  if (FRows = 0) or (FColumns = 0) then begin
    if FMap <> nil then FreeMem(FMap);
    FSize := 0;
    Exit;
  end;
  if FRows * FColumns > FSize then begin
    FSize := FRows * FColumns * FDepth div 8;
    FMap := AllocMem(FSize * 2);
  end else FillChar(FMap^, FSize, 0);
end;

procedure TDSpRenderMap.Pack;
begin
  if FRows * FColumns < FSize then begin
    FSize := FRows * FColumns;
    ReallocMem(FMap, FSize);
  end;
end;

function TDSpRenderMap.GetRow(r: integer): pointer;
begin
  if FMap = nil then Result := nil
  else Result := pointer(integer(FMap) + r * FColumns * FDepth div 8);
end;

function TDSpRenderMap.GetMaxValue: cardinal;
begin
  case FDepth of
     8: Result := 255;
    16: Result := 65535;
    32: Result := 4294967295;
  else Result := 0; // to satify the compiler
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Global functions
function DataSize(t: TDSpDataType): TDSpInt32;
begin
  case t of
    dspUInt8, dspInt8               : Result := 1;
    dspUInt16, dspInt16             : Result := 2;
    dspUInt32, dspInt32, dspFloat32 : Result := 4;
    dspInt64, dspFloat64            : Result := 8;
  else Result := 0;            // satisfy the compiler
  end;
end;

initialization
  Randomize;                   // enable random number generation
  DSpatial := TDSpatial.Create;
finalization
  DSpatial.Free;
end.
