{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpHistogram.pas. The Original Code is available        }
{ through the home page of DSpatial at http://dspatial.sourceforge.net.        }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements histogram analysis functions. Histograms can be         }
{ constructed from any kind of data that can be presented serially (as an      }
{ array). The histogram can be analyzed programmatically or visually using the }
{ TDSpHistogram component. Note that the TDSpHisto class has abstract methods  }
{ and you should instantiate one of its descendants, which are specific to a   }
{ particular data type. When you want to display a histogram, you first create }
{ the TDSpHisto descendant and add the data for analysis, after which you      }
{ assign the histogram to the TDSpHistogram component for display.             }
{                                                                              }
{ Data can be added in chunks using the AddData method. Thus, if you have a    }
{ huge data set, you can call AddData several times with the different parts of}
{ the data set and update the histogram incrementally. Currently, when you add }
{ a raster band for analysis, the MissingValue value is not considered. As long}
{ as that value is outside of the range [FMin..FMax], any pixels having a value}
{ of MissingValue will end up in the FUnder or FOver count and thus not be used}
{ in the histogram. The total count FCount of the TDSpHisto instance does not  }
{ include those out-of-range pixels. If MissingValue is within the range you   }
{ are out of luck in the current version.                                      }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 21                                             }
{                                                                              }
{ Dependencies  : DSpGlobals                                                   }
{                 Math                                                         }
{                 DSpWin32 (TRect)                                             }
{                 SysUtils                                                     }
{                 Classes                                                      }
{                 Graphics                                                     }
{                 Controls                                                     }
{                                                                              }
{ Version       : 0.1  03Nov21 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpHistogram;
{$I DSpatial.inc}
interface
uses {$IFDEF WIN32} Controls, Classes, Graphics, DSpWin32, {$ENDIF} DSpGlobals;
type
  TDSpHistoData = array of cardinal;
  
  // Abstract ancestor of all histogram classes. Descendants implement processing
  // of data of a specific type.
  TDSpHisto = class
  private
    FHisto    : TDSpHistoData;
    FCount,                           // total number of elements in [FMin..FMax]
    FUnder,                           // number of elements < FMin
    FOver     : cardinal;             // number of elements > FMax
    FMin,
    FMax      : TDSpFloat64;
    FBins     : integer;
  protected
    function    GetCount(bin: integer): cardinal;
  public
    constructor Create(min, max: TDSpFloat64; bins: integer = 512);
    destructor  Destroy; override;

    function    AddData(const data: pointer; num: integer): cardinal; virtual; abstract;
    function    DetermineBestMaxValue(roundUp: boolean = true): cardinal;

    property    Min: TDSpFloat64 read FMin;
    property    Max: TDSpFloat64 read FMax;
    property    Bins: integer read FBins;
    property    Count[bin: integer]: cardinal read GetCount;
    property    Total: cardinal read FCount;
    property    Histogram: TDSpHistoData read FHisto;
  end;

  TDSpHistoUInt8 = class(TDSpHisto)
  public
    constructor Create(min: TDSpUInt8 = 0; max: TDSpUInt8 = 255; bins: integer = 256);
    function    AddData(const data: pointer; num: integer): cardinal; override;
  end;

  TDSpHistoInt8 = class(TDSpHisto)
  public
    constructor Create(min: TDSpInt8 = -128; max: TDSpInt8 = 127; bins: integer = 256);
    function    AddData(const data: pointer; num: integer): cardinal; override;
  end;

  TDSpHistoUInt16 = class(TDSpHisto)
  public
    constructor Create(min: TDSpUInt16 = 0; max: TDSpUInt16 = 65535; bins: integer = 512);
    function    AddData(const data: pointer; num: integer): cardinal; override;
  end;

  TDSpHistoInt16 = class(TDSpHisto)
  public
    constructor Create(min: TDSpInt16 = -32768; max: TDSpInt16 = 32767; bins: integer = 512);
    function    AddData(const data: pointer; num: integer): cardinal; override;
  end;

  TDSpHistoUInt32 = class(TDSpHisto)
  public
    constructor Create(min, max: TDSpUInt32; bins: integer = 512);
    function    AddData(const data: pointer; num: integer): cardinal; override;
  end;

  TDSpHistoInt32 = class(TDSpHisto)
  public
    constructor Create(min, max: TDSpInt32; bins: integer = 512);
    function    AddData(const data: pointer; num: integer): cardinal; override;
  end;

  TDSpHistoInt64 = class(TDSpHisto)
  public
    constructor Create(min, max: TDSpInt64; bins: integer = 512);
    function    AddData(const data: pointer; num: integer): cardinal; override;
  end;

  TDSpHistoFloat32 = class(TDSpHisto)
  public
    constructor Create(min, max: TDSpFloat32; bins: integer = 512);
    function    AddData(const data: pointer; num: integer): cardinal; override;
  end;

  TDSpHistoFloat64 = class(TDSpHisto)
  public
    constructor Create(min, max: TDSpFloat64; bins: integer = 512);
    function    AddData(const data: pointer; num: integer): cardinal; override;
  end;

  // The histogram component
  TDSpHistogram = class(TCustomControl)
  private
    FHisto      : TDSpHisto;
    FMax        : cardinal;    // max number of elements
    FUserMax    : cardinal;    // User set maximum
    FDataArea   : TRect;       // area where data is drawn, rest is for axes and labels
    FExclMin    : boolean;     // scale histogram without considering count of minimum value
    FExclMax    : boolean;     // same as above for maximum value
    FBinLabels  : boolean;
    FValueLabels: boolean;
    FValueLabelsInside: boolean;// if true, value label is written inside the data area
    FAxisColor,
    FDataColor  : TColor;
    FMajorBinTicks,            // number of major ticks, labeled if F*Labels
    FMinorBinTicks,            // number of minor ticks between major ticks
    FMajorValueTicks,
    FMinorValueTicks : integer;
  protected
    procedure   Paint; override;
    procedure   Resize; override;
    procedure   MouseMove(Shift: TShiftState; X, Y: Integer); override;

    procedure   SetMax(max: cardinal);
    procedure   SetTick(Index: integer; Value: integer);
  public
    constructor Create(AOwner: TComponent); override;

    procedure   AssignHisto(h: TDSpHisto);
  published
    // Properties
    property    Align;
    property    Color;
    property    Font;
    property    ShowHint;
    property    ExcludeMin: boolean read FExclMin write FExclMin default false;
    property    ExcludeMax: boolean read FExclMax write FExclMax default false;
    property    MaxValue: cardinal read FUserMax write SetMax;
    property    LabelBins: boolean read FBinLabels write FBinLabels default false;
    property    LabelValues: boolean read FValueLabels write FValueLabels default false;
    property    AxisColor: TColor read FAxisColor write FAxisColor default clRed;
    property    DataColor: TColor read FDataColor write FDataColor default clBlue;
    property    MajorBinTicks: integer index 1 read FMajorBinTicks write SetTick default 3;
    property    MinorBinTicks: integer index 2 read FMinorBinTicks write SetTick default 0;
    property    MajorValueTicks: integer index 3 read FMajorValueTicks write SetTick default 3;
    property    MinorValueTicks: integer index 4 read FMinorValueTicks write SetTick default 0;
    property    ValueLabelsInside: boolean read FValueLabelsInside write FValueLabelsInside default false;

    // Events
    property    OnClick;
    property    OnDblClick;
    property    OnMouseDown;
    property    OnMouseMove;
    property    OnMouseUp;
    property    OnEnter;
    property    OnExit;
  end;

  procedure Register;

implementation
uses Math, SysUtils;

////////////////////////////////////////////////////////////////////////////////
// TDSpHistogram
constructor TDSpHisto.Create(min, max: TDSpFloat64; bins: integer = 512);
var i: integer;
begin
  inherited Create;
  Assert((min < max) and (bins > 0) and (bins <= 4096));
  FMin := min;
  FMax := max;

  // Initialize the bins
  FBins := bins;
  SetLength(FHisto, bins);
  for i := 0 to bins - 1 do FHisto[i] := 0;
end;

destructor TDSpHisto.Destroy;
begin
  Finalize(FHisto);
  inherited;
end;

function TDSpHisto.GetCount(bin: integer): cardinal;
begin
  Assert((bin >= 0) and (bin < Length(FHisto)));
  Result := FHisto[bin];
end;

function TDSpHisto.DetermineBestMaxValue(roundUp: boolean = true): cardinal;
var i, scale, temp: cardinal;
begin
  // Get the max value without including the first and last bin
  Result := 0;
  for i := 1 to FBins - 2 do
    if FHisto[i] > Result then Result := FHisto[i];

  // Check if lowest or highest bins should be included
  if (FHisto[0] > Result) and (FHisto[0] / Result < 1.2) then temp := FHisto[0] else temp := Result;
  if (FHisto[FBins-1] > Result) and (FHisto[FBins-1] / Result < 1.2) then Result := FHisto[FBins-1];
  if temp > Result then Result := temp;

  // Round up if required
  if roundUp then begin
    scale := Floor(Log10(Result));
    if scale > 1 then Dec(scale);
    scale := Round(IntPower(10, scale));
    if Result mod scale > 0 then Result := (Result div scale + 1) * scale;
  end;
end;


////////////////////////////////////////////////////////////////////////////////
// Fast 8 bit full histogram
procedure _Histogram8BitFull(const data: pointer; num: integer; histo: pointer);
{var i: integer;
begin
  for i := 1 to num do begin
    Inc(histo[data^]);
    Inc(data);
  end;   }
asm
  push   ebx                      // save ebx on the stack
  xor    ebx,ebx                  // ebx := 0

@loop:
  mov    bl,[eax]                 // move byte from data into bl
  inc    dword ptr [ecx+ebx*4]    // Inc(histo[bl])
  inc    eax                      // move to the next byte in data
  dec    edx                      // Dec(num)
  jnz    @loop                    // if num > 0 then loop

  pop    ebx                      // restore ebx from the stack
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpHistoUInt8
constructor TDSpHistoUInt8.Create(min: TDSpUInt8 = 0; max: TDSpUInt8 = 255; bins: integer = 256);
begin
  inherited Create(min, max, bins);
end;

function TDSpHistoUInt8.AddData(const data: pointer; num: integer): cardinal;
var i: integer;
    bp: PDSpUInt8;
    min: TDSpUInt8;
    scale: TDSpFloat64;
begin
  if FBins = 256 then begin
    _Histogram8BitFull(data, num, @FHisto[0]);
    Inc(FCount, num);
  end else begin
    bp := data;
    min := Round(FMin);
    scale := FBins / (FMax - FMin);
    for i := 1 to num do begin
      if bp^ < min then Inc(FUnder)
      else if bp^ > max then Inc(FOver)
      else begin
        Inc(FHisto[Floor((bp^ - min) * scale)]);
        Inc(FCount);
      end;
      Inc(bp);
    end;
  end;
  Result := FCount;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpHistoInt8
constructor TDSpHistoInt8.Create(min: TDSpInt8 = -128; max: TDSpInt8 = 127; bins: integer = 256);
begin
  inherited Create(min, max, bins);
end;

function TDSpHistoInt8.AddData(const data: pointer; num: integer): cardinal;
var i: integer;
    shp: PDSpInt8;
    min: TDSpInt8;
    scale: TDSpFloat64;
begin
  if FBins = 256 then begin
    _Histogram8BitFull(data, num, @FHisto[0]);
    Inc(FCount, num);
  end else begin
    shp := data;
    min := Round(FMin);
    scale := FBins / (FMax - FMin);
    for i := 1 to num do begin
      if shp^ < min then Inc(FUnder)
      else if shp^ > max then Inc(FOver)
      else begin
        Inc(FHisto[Floor((shp^ - min) * scale)]);
        Inc(FCount);
      end;
      Inc(shp);
    end;
  end;
  Result := FCount;
end;

constructor TDSpHistoUInt16.Create(min: TDSpUInt16 = 0; max: TDSpUInt16 = 65535; bins: integer = 512);
begin
  inherited Create(min, max, bins);
end;

function TDSpHistoUInt16.AddData(const data: pointer; num: integer): cardinal;
var i: integer;
    wp: PDSpUInt16;
    min: TDSpUInt16;
    scale: TDSpFloat64;
begin
  wp := data;
  min := Round(FMin);
  scale := FBins / (FMax - FMin);
  for i := 1 to num do begin
    if wp^ < min then Inc(FUnder)
    else if wp^ > max then Inc(FOver)
    else begin
      Inc(FHisto[Floor((wp^ - min) * scale)]);
      Inc(FCount);
    end;
    Inc(wp);
  end;
  Result := FCount;
end;

constructor TDSpHistoInt16.Create(min: TDSpInt16 = -32768; max: TDSpInt16 = 32767; bins: integer = 512);
begin
  inherited Create(min, max, bins);
end;

function TDSpHistoInt16.AddData(const data: pointer; num: integer): cardinal;
var i: integer;
    smp: PDSpInt16;
    min: TDSpInt16;
    scale: TDSpFloat64;
begin
  smp := data;
  min := Round(FMin);
  scale := FBins / (FMax - FMin);
  for i := 1 to num do begin
    if smp^ < min then Inc(FUnder)
    else if smp^ > max then Inc(FOver)
    else begin
      Inc(FHisto[Floor((smp^ - min) * scale)]);
      Inc(FCount);
    end;
    Inc(smp);
  end;
  Result := FCount;
end;

constructor TDSpHistoUInt32.Create(min, max: TDSpUInt32; bins: integer = 512);
begin
  inherited Create(min, max, bins);
end;

function TDSpHistoUInt32.AddData(const data: pointer; num: integer): cardinal;
var i: integer;
    cp: PDSpUInt32;
    min: TDSpUInt32;
    scale: TDSpFloat64;
begin
  cp := data;
  min := Round(FMin);
  scale := FBins / (FMax - FMin);
  for i := 1 to num do begin
    if cp^ < min then Inc(FUnder)
    else if cp^ > max then Inc(FOver)
    else begin
      Inc(FHisto[Floor((cp^ - min) * scale)]);
      Inc(FCount);
    end;
    Inc(cp);
  end;
  Result := FCount;
end;

constructor TDSpHistoInt32.Create(min, max: TDSpInt32; bins: integer = 512);
begin
  inherited Create(min, max, bins);
end;

function TDSpHistoInt32.AddData(const data: pointer; num: integer): cardinal;
var i: integer;
    ip: PDSpInt32;
    min: TDSpInt32;
    scale: TDSpFloat64;
begin
  ip := data;
  min := Round(FMin);
  scale := FBins / (FMax - FMin);
  for i := 1 to num do begin
    if ip^ < min then Inc(FUnder)
    else if ip^ > max then Inc(FOver)
    else begin
      Inc(FHisto[Floor((ip^ - min) * scale)]);
      Inc(FCount);
    end;
    Inc(ip);
  end;
  Result := FCount;
end;

constructor TDSpHistoInt64.Create(min, max: TDSpInt64; bins: integer = 512);
begin
  inherited Create(min, max, bins);
end;

function TDSpHistoInt64.AddData(const data: pointer; num: integer): cardinal;
var i: integer;
    ip: PDSpInt64;
    min: TDSpInt64;
    scale: TDSpFloat64;
begin
  ip := data;
  min := Round(FMin);
  scale := FBins / (FMax - FMin);
  for i := 1 to num do begin
    if ip^ < min then Inc(FUnder)
    else if ip^ > max then Inc(FOver)
    else begin
      Inc(FHisto[Floor((ip^ - min) * scale)]);
      Inc(FCount);
    end;
    Inc(ip);
  end;
  Result := FCount;
end;

constructor TDSpHistoFloat32.Create(min, max: TDSpFloat32; bins: integer = 512);
begin
  inherited Create(min, max, bins);
end;

function TDSpHistoFloat32.AddData(const data: pointer; num: integer): cardinal;
var i: integer;
    sp: PDSpFloat32;
    scale: TDSpFloat64;
begin
  sp := data;
  scale := FBins / (FMax - FMin);
  for i := 1 to num do begin
    if sp^ < FMin then Inc(FUnder)
    else if sp^ > max then Inc(FOver)
    else begin
      Inc(FHisto[Floor((sp^ - FMin) * scale)]);
      Inc(FCount);
    end;
    Inc(sp);
  end;
  Result := FCount;
end;

constructor TDSpHistoFloat64.Create(min, max: TDSpFloat64; bins: integer = 512);
begin
  inherited Create(min, max, bins);
end;

function TDSpHistoFloat64.AddData(const data: pointer; num: integer): cardinal;
var i: integer;
    dp: PDSpFloat64;
    scale: TDSpFloat64;
begin
  dp := data;
  scale := FBins / (FMax - FMin);
  for i := 1 to num do begin
    if dp^ < FMin then Inc(FUnder)
    else if dp^ > max then Inc(FOver)
    else begin
      Inc(FHisto[Floor((dp^ - FMin) * scale)]);
      Inc(FCount);
    end;
    Inc(dp);
  end;
  Result := FCount;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpHistogram
constructor TDSpHistogram.Create(AOwner: TComponent);
begin
  inherited;
  FBinLabels := false;
  FValueLabels := false;
  FValueLabelsInside := false;
  FAxisColor := clRed;
  FDataColor := clBlue;
  FMajorBinTicks := 3;
  FMinorBinTicks := 0;
  FMajorValueTicks := 3;
  FMinorValueTicks := 0;
  FHisto := nil;
end;

procedure TDSpHistogram.AssignHisto(h: TDSpHisto);
var i, start, stop: integer;
begin
  FHisto := h;
  if FExclMin then start := 1 else start := 0;
  if FExclMax then stop := FHisto.Bins - 2 else stop := FHisto.Bins - 1;
  if FUserMax = 0 then begin
    FMax := FHisto.Count[start];
    for i := start + 1 to stop do
      if FMax < FHisto.Count[i] then FMax := FHisto.Count[i];
  end else FMax := FUserMax;
  Resize;
  Repaint;
end;

procedure TDSpHistogram.Paint;
var i, j, bins, line, len, len2: integer;
    hscale, vscale: double;
    str: string;
begin
  inherited;
  // Erase the background
  Canvas.Brush.Color := Color;
  Canvas.FillRect(ClientRect);

  Canvas.Pen.Color := FAxisColor;

  // Need some data and space to work with, otherwise exit
  if (FHisto = nil) or (ClientWidth < 10) or (ClientHeight < 10) then begin
    Canvas.MoveTo(2, ClientHeight - 5);
    Canvas.LineTo(ClientWidth - 2, ClientHeight - 5);
    Canvas.MoveTo(4, 2);
    Canvas.LineTo(4, ClientHeight - 2);
    Exit;
  end;

  // Data scaling
  bins := FHisto.Bins;
  hscale := (FDataArea.Right - FDataArea.Left + 1) / bins;
  vscale := (FDataArea.Bottom - FDataArea.Top + 1) / FMax;

  // Draw the axes and the ticks
  Canvas.MoveTo(FDataArea.Left - 1, FDataArea.Top);
  Canvas.LineTo(FDataArea.Left - 1, FDataArea.Bottom + 5);
  if (FMajorValueTicks > 0) and (FMajorValueTicks < (FDataArea.Bottom - FDataArea.Top)) then begin
    len := Round((FDataArea.Bottom - FDataArea.Top + 1) / FMajorValueTicks);
    for i := 0 to FMajorValueTicks - 1 do begin
      Canvas.MoveTo(FDataArea.Left - 5, FDataArea.Top + len * i);
      Canvas.LineTo(FDataArea.Left - 1, FDataArea.Top + len * i);
      if FMinorValueTicks > 0 then begin
        len2 := Round(len / (FMinorValueTicks + 1));
        for j := 1 to FMinorValueTicks do begin
          Canvas.MoveTo(FDataArea.Left - 3, FDataArea.Top + len * i + len2 * j);
          Canvas.LineTo(FDataArea.Left - 1, FDataArea.Top + len * i + len2 * j);
        end;
      end;
    end;
  end;
  // Bin axis
  Canvas.MoveTo(FDataArea.Left - 5, FDataArea.Bottom + 1);
  Canvas.LineTo(FDataArea.Right, FDataArea.Bottom + 1);
  if (FMajorBinTicks > 0) and (FMajorBinTicks < FHisto.Bins) and
     (FMajorBinTicks < (FDataArea.Right - FDataArea.Left)) then begin
    len := Round((FDataArea.Right - FDataArea.Left + 1) / FMajorBinTicks);
    for i := 0 to FMajorBinTicks - 1 do begin
      Canvas.MoveTo(FDataArea.Right - len * i, FDataArea.Bottom + 1);
      Canvas.LineTo(FDataArea.Right - len * i, FDataArea.Bottom + 5);
      if FMinorBinTicks > 0 then begin
        len2 := Round(len / (FMinorBinTicks + 1));
        for j := 1 to FMinorBinTicks do begin
          Canvas.MoveTo(FDataArea.Right - len * i - len2 * j, FDataArea.Bottom + 1);
          Canvas.LineTo(FDataArea.Right - len * i - len2 * j, FDataArea.Bottom + 3);
        end;
      end;
    end;
  end;

  // Draw the data
  Canvas.Pen.Color := FDataColor;
  for i := 0 to bins - 1 do begin
    line := FDataArea.Left + Trunc(i * hscale);
    if FHisto.Count[i] >= FMax then len := FDataArea.Bottom - FDataArea.Top + 1
    else len := Round(FHisto.Count[i] * vscale);
    Canvas.MoveTo(line, FDataArea.Bottom);
    Canvas.LineTo(line, FDataArea.Bottom - len);
  end;

  // Label the axes and the ticks
  Canvas.Font := Font;
  j := Abs(Font.Height) div 2;
  if FValueLabels and (FMajorValueTicks > 0) then begin
    len2 := Round((FDataArea.Bottom - FDataArea.Top + 1) / FMajorValueTicks);
    for i := 1 to FMajorValueTicks do begin
      str := IntToStr(Round(FMax * i / FMajorValueTicks));
      if FValueLabelsInside then len := FDataArea.Left + 1
      else len := FDataArea.Left - Canvas.TextWidth(str) - 6;
      Canvas.TextOut(len, FDataArea.Bottom - len2 * i - j, str);
    end;
  end;
  if FBinLabels and (FMajorBinTicks > 0) then begin
    len := Canvas.TextWidth(FloatToStrF(FHisto.Min, ffGeneral, 12, 1));
    Canvas.TextOut(FDataArea.Left - len div 2, FDataArea.Bottom + 7, FloatToStrF(FHisto.Min, ffGeneral, 12, 1));
    len2 := Round((FDataArea.Right - FDataArea.Left + 1) / FMajorBinTicks);
    for i := 1 to FMajorBinTicks do begin
      str := FloatToStrF(FHisto.Min + (FHisto.Max - FHisto.Min) * i / FMajorBinTicks, ffGeneral, 12, 1);
      len := FDataArea.Left + len2 * i - Canvas.TextWidth(str) div 2;
      Canvas.TextOut(len, FDataArea.Bottom + 7, str);
    end;
  end;
end;

procedure TDSpHistogram.Resize;
var h, len: integer;
begin
  inherited;
  if (ClientWidth < 10) or (ClientHeight < 10) then
    FDataArea := Rect(0, 0, 0, 0)
  else begin
    // Top & Bottom
    h := Abs(Font.Height);
    if FValueLabels then begin
      FDataArea.Top := 3 + h div 2;
      if FBinLabels then FDataArea.Bottom := ClientHeight - 9 - h
      else FDataArea.Bottom := ClientHeight - Max(9, h div 2);
    end else begin
      FDataArea.Top := 3;
      if FBinLabels then FDataArea.Bottom := ClientHeight - 9 - h
      else FDataArea.Bottom := ClientHeight - 9;
    end;

    // Left & Right
    Canvas.Font := Font;
    len := Canvas.TextWidth(FloatToStrF(FMax, ffGeneral, 12, 1));
    if FBinLabels then begin
      if Assigned(FHisto) then begin
        if FValueLabels and not FValueLabelsInside then
          FDataArea.Left := 6 + Max(len, Canvas.TextWidth(FloatToStrF(FHisto.Min, ffGeneral, 12, 1)) div 2)
        else FDataArea.Left := 6 + Canvas.TextWidth(FloatToStrF(FHisto.Min, ffGeneral, 12, 1)) div 2;
        FDataArea.Right := ClientWidth - 6 - len div 2;
      end else begin
        FDataArea.Left := 6 + Canvas.TextWidth('0') div 2;
        FDataArea.Right := ClientWidth - 6 - Canvas.TextWidth('4096') div 2;
      end;
    end else begin
      FDataArea.Left := 6;
      FDataArea.Right := ClientWidth - 6;
    end;
  end;
end;

procedure TDSpHistogram.MouseMove(Shift: TShiftState; X, Y: Integer);
var ndx: integer;
begin
  inherited;
  if Assigned(FHisto) and (X >= FDataArea.Left) and (X <= FDataArea.Right) and
     (Y >= FDataArea.Top) and (Y <= FDataArea.Bottom) then begin
    ndx := Min(Round((X - FDataArea.Left) * FHisto.Bins / (FDataArea.Right - FDataArea.Left + 1)), FHisto.Bins - 1);
    Hint := Format('Bin: %d'#13#10'Value: %d', [ndx, FHisto.Count[ndx]]);
  end else Hint := '';
end;

procedure TDSpHistogram.SetMax(max: cardinal);
begin
  FUserMax := max;
  FMax := max;
  Resize;
  Repaint;
end;

procedure TDSpHistogram.SetTick(Index: integer; Value: integer);
begin
  case Index of
    1: FMajorBinTicks := Max(0, Value);
    2: FMinorBinTicks := Min(Max(0, Value), 9);
    3: FMajorValueTicks := Max(0, Value);
    4: FMinorValueTicks := Min(Max(0, Value), 9);
  end;
end;

procedure Register;
begin
  RegisterComponents('DSpatial', [TDSpHistogram]);
end;

end.
