{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRaster.pas. The Original Code is available through   }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit provides basic data access services for DSpatial modules. In a     }
{ typical situation, a user process will first read data, getting the data in  }
{ a format specific to the feature (for instance, a raster band is delivered as}
{ a block of memory), process it, and then create a new data set from the      }
{ result of the processing. This works well for smaller data sets, but huge    }
{ data sets require more sophisticated methods.                                }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 20                                             }
{                                                                              }
{ Dependencies  : DSpDataSet                                                   }
{                 DSpGlobals                                                   }
{                 DSpPFA                                                       }
{                 DSpRasterSource                                              }
{                                                                              }
{ Version       : 0.1  03Nov20 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpModuleBase;
{$I DSpatial.inc}
interface
uses DSpGlobals, DSpRaster;

  function DSpReadRasterBandData(const r: TDSpRaster; band: integer): pointer;
  function DSpWriteRasterBandData(const r: TDSpRaster; band: integer; data: pointer): boolean;
//  function DSpCreateRasterFromData(const r: TDSpRaster; dt: TDSpDataType; const p: pointer): TDSpRaster;

implementation
uses SysUtils, SyncObjs, DSpPFA, DSpRasterSource;

function DSpReadRasterBandData(const r: TDSpRaster; band: integer): pointer;
var rows, cols: integer;
    msg: TDSpThreadMessage;
    params: PDSpMsgReadRasterBandCoords;
begin
  rows := r.Rows;
  cols := r.Columns;

  // Set up the message structure
  New(params);
  params^.Band := band;
  params^.Coords.Left := 0;
  params^.Coords.Top := 0;
  params^.Coords.Right := cols - 1;
  params^.Coords.Bottom := rows - 1;
  params^.Data := AllocMem(rows * cols * DataSize(r.BandDataType[band]));

  New(msg);
  msg^.DataSourceIndex := r.SourceIndex;
  msg^.Event := TEvent.Create(nil, false, false, '');
  msg^.Msg := DSPMSG_BANDREADCOORDS;
  msg^.Parameter := params;

  r.Handler.QueueMessage(msg);
  if msg^.Event.WaitFor(60000) <> wrSignaled then begin
    Result := nil;
    Freemem(params^.Data);
  end else Result := params^.Data;

  msg^.Event.Free;
  Dispose(msg);
end;

function DSpWriteRasterBandData(const r: TDSpRaster; band: integer; data: pointer): boolean;
var rows, cols: integer;
    msg: TDSpThreadMessage;
    params: PDSpMsgReadRasterBandCoords;
begin
  rows := r.Rows;
  cols := r.Columns;

  // Set up the message structure
  New(params);
  params^.Band := band;
  params^.Coords.Left := 0;
  params^.Coords.Top := 0;
  params^.Coords.Right := cols - 1;
  params^.Coords.Bottom := rows - 1;
  params^.Data := data;

  New(msg);
  msg^.DataSourceIndex := r.SourceIndex;
  msg^.Event := TEvent.Create(nil, false, false, '');
  msg^.Msg := DSPMSG_BANDWRITECOORDS;
  msg^.Parameter := params;

  r.Handler.QueueMessage(msg);
  Result := msg^.Event.WaitFor(60000) = wrSignaled;

  msg^.Event.Free;
  Dispose(msg);
end;

    {
function DSpCreateRasterFromData(const r: TDSpRaster; dt: TDSpDataType; const p: pointer): TDSpRaster;
var h: TDSpSourceHandler;
    rasterSource: TDSpRasterSource;
begin
  // Create the temporary raster that will hold the result
  rasterSource := r.Source as TDSpRasterSource;        // the input source
  try
    h := PFA.CreateTempRaster(rasterSource);
  finally
    rasterSource.ReadUnlock;
  end;

  // These operations can take place in the main thread for two reasons:
  // 1: The data set is temporary and just created, so no other threads are
  //    accessing it.
  // 2: No disk or network access is necessary, because the memory is already
  //    allocated and initialized.
  rasterSource := h.WriteLock(0) as TDSpRasterSource;  // the output source
  try
    rasterSource.AddBand(dt, dspRasterBandImage).SetBandData(p);
  finally
    rasterSource.WriteUnlock;
  end;

  // Finally, create the result
  Result := TDSpRaster.Open(h, 0);
end;
     }
end.
