{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpPFA.pas. The Original Code is available through      }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit defines the Pluggable Format Architecture (PFA). The PFA interfaces}
{ between drivers that read and write the different formats of spatial data and}
{ the client application. Thus the application has a uniform view of disparate }
{ data access methods. The PFA has subclasses specific to the type of data     }
{ being accessed: vector, raster, tabular. In every application there should   }
{ be exactly one instance of the TDSpPFA class with global scope. That instance}
{ is created in the initialization section of this unit as the PFA variable. It}
{ is automatically destroyed in the finalization section upon closing of the   }
{ application. This unit must be listed in the uses clause of the project      }
{ before any of the data set classes and statically loaded data brokers, as    }
{ they might reference the PFA variable in their initialization clauses.       }
{                                                                              }
{ The PFA manages multiple data sources simultaneously, such that true multi-  }
{ processing can be achieved on multi-processor systems (every handler operates}
{ as a separate thread), but even on single processor systems the latency of   }
{ slower components such as hard disks and network access is offset through the}
{ separation by threads of slower code.                                        }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 19                                             }
{                                                                              }
{ Dependencies  : DSpDataMgr                                                   }
{                 DSpRasterSource                                              }
{                 DSpVectorSource                                              }
{                 DSpTempRaster                                                }
{                 DSpCoordinateSystem                                          }
{                 DSpUtils (ValidateIndex)                                     }
{                 DSpGlobals                                                   }
{                 SyncObjs (TEvent)                                            }
{                 SysUtils (ExtractFileExt, Exception,                         }
{                           TMultiReadExclusiveWriteSynchronizer)              }
{                 Classes (TThread)                                            }
{                                                                              }
{ Version       : 0.7  03Dec19 TDSpSourceDataSet now uses access specifier.    }
{                              IdentifyBroker uses full source string. Opening }
{                              a source is now asynchronous. Full support for  }
{                              creating new handlers and data sets from        }
{                              existing ones. Support for adding and deleting  }
{                              raster bands. Data sources can now be found by  }
{                              name (TDSpSourceHandler.Read/WriteLock).        }
{                 0.6  03Nov27 DSpPFA class uses synchronization lock. Added   }
{                              support for reading source data into memory.    }
{                              Source deletion now happens during idle time.   }
{                 0.5  03Nov07 PFA model overhaul                              }
{                 0.4  03Sep25 TDSpPFAObject now supports state flag and data  }
{                              set removal.                                    }
{                 0.3  03Aug14 Many changes to raster calls                    }
{                 0.2  03Jul18 Addition of TDSpPFAObject                       }
{                              Changes to TDSpPFAHandler                       }
{                 0.1  03Apr26 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpPFA;
{$I DSpatial.inc}
interface
uses Classes, SysUtils, SyncObjs, DSpGlobals, DSpUtils, DSpCoordinateSystem;

type
  TDSpBaseDriver    = class;
  TDSpSourceHandler = class;

  TDSpPFAMessage = record
                     DataSourceIndex : integer;
                     Event           : TEvent;
                     Msg             : cardinal;            // DSPMSG_*
                     Parameter       : pointer;
                   end;
  TDSpThreadMessage = ^TDSpPFAMessage;

  TDSpDataSetInfo = record
                      ID             : integer;       // identifier for use by the driver
                      DescriptiveName: string;
                      CoordSys       : string;
                      Extent         : TDSpExtent;
                      DataSetType    : TDSpPFAFormat;
                      DataSetData    : pointer;       // depends on data set type
                    end;

  // The Mother-of-all-Data sources. The class provides the basic structure of the
  // different kinds of geospatial data sources, including common meta-information
  // such as a unique identifier, projection parameters, extent, unit. Otherwise
  // this class is a mostly abstract ancestor to more useful classes.
  TDSpSourceDataSet = class
  private
    FReads      : integer;              // number of read accesses
    FLock       : TMultiReadExclusiveWriteSynchronizer;
  protected
    FId         : TDSpUInt32;           // unique identifier
    FName       : string;               // descriptive name of the data set
    FRefs       : integer;              // number of references to this data set
    FAccess     : TDSpPFAAccess;        // access mode

    FIndex      : integer;
    FDriver     : TDSpBaseDriver;
    FHandler    : TDSpSourceHandler;
    FDataSetType: TDSpPFAFormat;

    FExtent     : TDSpExtent;           // extreme coordinates of the data set
    FCoordSys   : TDSpCoordinateSystem; // coordinate system; never free a TDSpCoordinateSystem: reference counted

    procedure   ReadLock;
    procedure   WriteLock;
    procedure   StartWriting; virtual;
    procedure   EndWriting; virtual;

    procedure   PurgeData; virtual; abstract;
    procedure   SetExtent(ext: TDSpExtent);
  public
//    constructor Create(const drv: TDSpBaseDriver; dataSet: integer; const h: TDSpSourceHandler);
    constructor Open(const drv: TDSpBaseDriver; dataSet: integer; const h: TDSpSourceHandler);
    destructor  Destroy; override;
    procedure   Close; virtual; abstract;

    function    Access: TDSpPFAAccess;
    function    GetAccess(acc: TDSpPFAAccess): boolean;
    procedure   DropAccess(oldAccess: TDSpPFAAccess);

    procedure   AddRef;
    function    CloseRef: integer;
    procedure   ReadUnlock;
    procedure   WriteUnlock;

    procedure   SetCoordinateSystem(cs: TDSpCoordinateSystem); overload;
    procedure   SetCoordinateSystem(const wkt: string); overload;
    function    DescribeCS(descr: TDSpDescribe): string;
    function    CoordinateSystem: TDSpCoordinateSystem;

    property    Id: TDSpUInt32 read FId;
    property    Name: string read FName write FName;
    property    DataSetType: TDSpPFAFormat read FDataSetType;
    property    DataSetIndex: integer read FIndex;
    property    Extent: TDSpExtent read FExtent write SetExtent;
    property    Driver: TDSpBaseDriver read FDriver;
    property    Handler: TDSpSourceHandler read FHandler;
  end;

  //////////////////////////////////////////////////////////////////////////////
  // PFA objects: TDSpBaseDriver, TDSpBaseBroker, TDSpSourceHandler, and TDSpPFA

  // TDSpBaseDriver is the ancestor of all the drivers available through the PFA
  TDSpBaseDriver = class
  protected
    function    Description: string; virtual; abstract;
    function    DataSets: integer; virtual; abstract;

    function    DataSetID(i: integer): integer; virtual; abstract;
    function    DataSetDescription(i: integer): string; virtual; abstract;
    function    DataSetCoordSys(i: integer): string; virtual; abstract;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); virtual; abstract;
    function    DataSetType(i: integer): TDSpPFAFormat; virtual; abstract;
  public
    function    CreateDataSet(ds: PDSpDataSetInfo; p: pointer): integer; virtual; abstract;
    function    AddDataSet(const ds: TDSpSourceDataSet): integer; virtual; abstract;
    function    CloneDataSet(const ds: TDSpSourceDataSet): integer; virtual; abstract;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); virtual; abstract;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); virtual; abstract;
    procedure   AddRasterBand(i: integer; info: PDSpRasterbandInfo); virtual; abstract;
    procedure   DeleteRasterBand(i: integer); virtual; abstract;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); virtual; abstract;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); virtual; abstract;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); virtual; abstract;
  end;

  TDSpBaseBroker = class
  protected
    FDescription: string;
    FFilter     : string;    // filter for file dialogs
    FDrivers    : TList;     // of TDSpBaseDriver;
    FCaps       : TDSpPFACaps;
  public
    constructor Create;
    destructor  Destroy; override;

    function    Filter: string;
    function    SupportsSource(const source: string): boolean; virtual; abstract;
    function    CanSave(const ds: TDSpSourceDataSet): string; virtual; abstract;

    function    CreateSource(const fn: string): TDSpSourceHandler; virtual; abstract;
    function    OpenSource(const fn: string): TDSpSourceHandler; virtual; abstract;
    function    CloseDriver(r: TDSpBaseDriver): boolean; virtual;

    property    Capabilities: TDSpPFACaps read FCaps;
  end;

  // A handler to a data source. The PFA issues this handler when the application
  // opens a source. The handler runs in a separate thread such that multiple
  // data sources can be accessed simultaneously by multiple handlers. Instances
  // of this class are owned by the PFA and clients will receive a reference to
  // the instance; when the handler is no longer required, the client should
  // close the reference using PFA.CloseSource, NOT delete the instance.
  TDSpSourceHandler = class(TThread)
  private
    FQueue      : TDSpThreadQueue;             // queue to receive messages in
    FLock       : TMultiReadExclusiveWriteSynchronizer; // protect access to local members
    FStatus     : TDSpStates;                  // status of the reader
  protected
    FConnect    : string;                      // connection string of the source, typically a file name
    FDescription: string;                      // general description of the source
    FRefs       : integer;                     // number of references to this source
    FDriver     : TDSpBaseDriver;              // the driver accessing the file
    FDataSets   : array of TDSpSourceDataSet;  // the data sources themselves

    procedure   Execute; override;

    procedure   OpenSource;

    procedure   AddRasterBand(msg: TDSpThreadMessage);
    procedure   DeleteRasterBand(msg: TDSpThreadMessage);
    procedure   RenderRasterMap(msg: TDSpThreadMessage);
    procedure   ReadRasterBandDataByCoords(msg: TDSpThreadMessage);
    procedure   WriteRasterBandDataByCoords(msg: TDSpThreadMessage);

    function    GetSourceString: string;
    function    GetNumDataSets: integer;
    procedure   AddRef;
    function    TotalRefs: integer;
    function    FindDataSetByName(const name: string): integer;
  public
    constructor Create(r: TDSpBaseDriver; const source: string);
    destructor  Destroy; override;
    procedure   Close;

    procedure   QueueMessage(msg: TDSpThreadMessage);
    function    ReadLock(ndx: integer): TDSpSourceDataSet; overload;
    function    ReadLock(const name: string): TDSpSourceDataSet; overload;
    function    WriteLock(ndx: integer): TDSpSourceDataSet; overload;
    function    WriteLock(const name: string): TDSpSourceDataSet; overload;
    property    Status: TDSpStates read FStatus;
    property    Source: string read GetSourceString;
    property    Driver: TDSpBaseDriver read FDriver;
    property    DataSets: integer read GetNumDataSets;
    function    CreateDataSet(ds: PDSpDataSetInfo; p: pointer): TDSpSourceDataSet;
    function    AddDataSet(ds: TDSpSourceDataSet): TDSpSourceDataSet;
    function    CloneDataSetFrom(ds: TDSpSourceDataSet): TDSpSourceDataSet;

    procedure   AddDataSetRef(ndx: integer);
    procedure   CloseDataSetRef(ndx: integer);
  end;

  // The data broker of the PFA. Do not instantiate this class, but access it
  // through the PFA global variable.
  TDSpPFA = class
  private
    FStaticBrokers   : array of TDSpBaseBroker;
//    FDynamicBrokers : array of TDSpBaseBroker;
    FSources         : array of TDSpSourceHandler;
    FLock            : TMultiReadExclusiveWriteSynchronizer; // protect access to local members

    function    IdentifyBroker(const source: string): TDSpBaseBroker;
    procedure   InsertHandler(s: TDSpSourceHandler);
  protected
    constructor Create;
  public
    destructor  Destroy; override;

    procedure   RegisterBroker(d: TDSpBaseBroker);
    function    FilterString(cap: TDSpPFACap): string;
    function    SaveAsFilter(const ds: TDSpSourceDataSet): string;

    function    CreateSource(const source: string): TDSpSourceHandler;
    function    OpenSource(const source: string): TDSpSourceHandler;
    function    CreateTempRaster(ds: PDSpDatasetInfo; r: PDSpRasterInfo): TDSpSourceHandler; overload;
    function    CreateTempRaster(r: TDSpSourceDataSet): TDSpSourceHandler; overload;

    procedure   IdleProcessing;
  end;

  function DSpKernelReadRasterBand(const ds: TDSpSourceDataSet; band: integer): pointer;

var
  PFA : TDSpPFA;

implementation
uses {$IFDEF WIN32} DSpWin32, {$ELSE} DSpLinux, {$ENDIF}
     DSpRasterSource, DSpVectorSource, DSpTempRaster;

////////////////////////////////////////////////////////////////////////////////
// Utility functions
function DSpKernelReadRasterBand(const ds: TDSpSourceDataSet; band: integer): pointer;
var r, c: integer;
    msg: TDSpThreadMessage;
    params: PDSpMsgReadRasterBandCoords;
begin
  Assert(ds is TDSpRasterSource);
  r := TDSpRasterSource(ds).Rows;
  c := TDSpRasterSource(ds).Columns;

  // Set up the message structure
  New(params);
  params^.Band := band;
  params^.Coords.Left := 0;
  params^.Coords.Top := 0;
  params^.Coords.Right := c - 1;
  params^.Coords.Bottom := r - 1;
  params^.Data := AllocMem(r * c * DataSize(TDSpRasterSource(ds).Band[band].DataType));

  New(msg);
  msg^.DataSourceIndex := ds.DataSetIndex;
  msg^.Event := TEvent.Create(nil, false, false, '');
  msg^.Msg := DSPMSG_BANDREADCOORDS;
  msg^.Parameter := params;

  ds.Handler.QueueMessage(msg);
  if msg^.Event.WaitFor($FFFFFFFF) <> wrSignaled then
    Assert(false, 'Wait event unexpectedly failed');
  Result := params^.Data;
  
  msg^.Event.Free;
  Dispose(msg);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpSourceDataSet
{constructor TDSpSourceDataSet.Create(const drv: TDSpBaseDriver; dataSet: integer; const h: TDSpSourceHandler);
begin
  inherited Create;
  FId := 0;
  FRefs := 0;
  FAccess := dspExcl;
  FLock := TMultiReadExclusiveWriteSynchronizer.Create;
  FDriver := drv;
  FIndex := dataSet;
  FHandler := h;
  // Other parameters (see Open) must be set before the data set is fully defined.
end;
     }
constructor TDSpSourceDataSet.Open(const drv: TDSpBaseDriver; dataSet: integer; const h: TDSpSourceHandler);
begin
  inherited Create;
  FRefs := 0;
  FLock := TMultiReadExclusiveWriteSynchronizer.Create;
  FAccess := dspNoAccess;
  FDriver := drv;
  FIndex := dataSet;
  FHandler := h;

  FID := FDriver.DataSetID(FIndex);
  FName := FDriver.DataSetDescription(FIndex);
  FCoordSys := CreateCoordinateSystemFromWKT(FDriver.DataSetCoordSys(FIndex));
  FExtent := TDSpExtent.Create(0, 0, 0, 0);
  FDriver.DataSetExtent(FIndex, FExtent);
end;

destructor TDSpSourceDataSet.Destroy;
begin
  if Assigned(FCoordSys) then FCoordSys.Release; // never free a TDSpCoordinateSystem: reference counted
  if Assigned(FExtent) then FExtent.Free;
  FLock.Free;
  inherited;
end;

function TDSpSourceDataSet.Access: TDSpPFAAccess;
begin
  Result := FAccess;
end;

function TDSpSourceDataSet.GetAccess(acc: TDSpPFAAccess): boolean;
begin
  if (FAccess = dspExcl) or ((FAccess = dspWrite) and (acc in [dspWrite, dspExcl])) or
     ((acc = dspExcl) and (FRefs > 1)) then
    Result := false
  else begin
    Result := true;
    if acc in [dspWrite, dspExcl] then FAccess := acc
    else if (acc = dspRead) and (FAccess = dspNoAccess) then FAccess := dspRead;
    if acc = dspRead then Inc(FReads);
  end;
end;

procedure TDSpSourceDataSet.DropAccess(oldAccess: TDSpPFAAccess);
begin
  if oldAccess = dspRead then Dec(FReads);
  if FReads = 0 then FAccess := dspNoAccess else FAccess := dspRead;
end;

procedure TDSpSourceDataSet.AddRef;
begin
  FLock.BeginWrite;
  Inc(FRefs);
  FLock.EndWrite;
end;

function TDSpSourceDataSet.CloseRef: integer;
begin
  FLock.BeginWrite;
  try
    if FRefs > 0 then Dec(FRefs);
    if FRefs = 0 then PurgeData;
    Result := FRefs;
  finally
    FLock.EndWrite;
  end;
end;

procedure TDSpSourceDataSet.ReadLock;
begin
  FLock.BeginRead;
end;
                         
procedure TDSpSourceDataSet.ReadUnlock;
begin
  FLock.EndRead;
end;

procedure TDSpSourceDataSet.WriteLock;
begin
  FLock.BeginWrite;
  StartWriting;
end;

procedure TDSpSourceDataSet.WriteUnlock;
begin
  EndWriting;
  FLock.EndWrite;
end;

procedure TDSpSourceDataSet.StartWriting;
begin
end;

procedure TDSpSourceDataSet.EndWriting;
begin
end;

procedure TDSpSourceDataSet.SetCoordinateSystem(cs: TDSpCoordinateSystem);
begin
  if FAccess = dspExcl then begin
    if Assigned(FCoordSys) then FCoordSys.Release;
    FCoordSys := cs;
  end;
end;

procedure TDSpSourceDataSet.SetCoordinateSystem(const wkt: string);
begin
  if FAccess = dspExcl then begin
    if Assigned(FCoordSys) then FCoordSys.Release;
    FCoordSys := CreateCoordinateSystemFromWKT(wkt);
  end;
end;

function TDSpSourceDataSet.DescribeCS(descr: TDSpDescribe): string;
begin
  if Assigned(FCoordSys) then Result := FCoordSys.Describe(descr)
  else Result := '(No coordinate system defined.)';
end;

function TDSpSourceDataSet.CoordinateSystem: TDSpCoordinateSystem;
begin
  Result := FCoordSys;
end;

procedure TDSpSourceDataSet.SetExtent(ext: TDSpExtent);
begin
  if FAccess = dspExcl then
    FExtent := ext.Clone;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpSourceHandler
constructor TDSpSourceHandler.Create(r: TDSpBaseDriver; const source: string);
begin
  inherited Create(true);
  FLock := TMultiReadExclusiveWriteSynchronizer.Create;
  FLock.BeginWrite;
  try
    FRefs := 1;
    FDriver := r;
    FStatus := [dspOk, dspReading, dspBusy];
    FQueue := TDSpThreadQueue.Create;
  finally
    FLock.EndWrite;
  end;
end;

destructor TDSpSourceHandler.Destroy;
var i: integer;
begin
  FLock.BeginWrite;
  try
    if Length(FDataSets) > 0 then
      for i := 0 to High(FDataSets) do
        if Assigned(FDataSets[i]) then FDataSets[i].Free;
    Finalize(FDataSets);
    FDriver.Free;
    FQueue.Free;
  finally
    FLock.EndWrite;
  end;
  inherited;
end;

procedure TDSpSourceHandler.Close;
begin
  FLock.BeginWrite;
  try
    if FRefs > 0 then Dec(FRefs);
  finally
    FLock.EndWrite;
  end;
end;

procedure TDSpSourceHandler.Execute;
var msg: TDSpThreadMessage;
begin
  while not Terminated do begin
    msg := TDSpThreadMessage(FQueue.Get);

    if msg = nil then begin
      FStatus := [dspOk, dspIdle];  // no lock required here as FStatus is private
      Suspend; 
      FStatus := [dspOk, dspBusy];
      Continue;                     // after waking up, start by getting a new message
    end;

    try
      case msg^.Msg of
        DSPMSG_NULL             : ; // dummy message, just set the event
        DSPMSG_TERMINATE        : Terminate;
        DSPMSG_SOURCEOPEN       : OpenSource;
        DSPMSG_RASTERADDBAND    : AddRasterBand(msg);
        DSPMSG_RASTERDELETEBAND : DeleteRasterBand(msg);
        DSPMSG_RENDERRASTERMAP  : RenderRasterMap(msg);
        DSPMSG_BANDREADCOORDS   : ReadRasterBandDataByCoords(msg);
        DSPMSG_BANDWRITECOORDS  : WriteRasterBandDataByCoords(msg);
      else Assert(false, 'Unhandled message id in TDSpSourceHandler.Execute');
      end;
    except
      msg^.Msg := DSPMSG_ERROR;
      FStatus := [dspError];
    end;

    if msg^.Event = nil then Dispose(msg)
    else msg^.Event.SetEvent;
  end;
end;

procedure TDSpSourceHandler.OpenSource;
var i, num: integer;
    dst: TDSpPFAFormat;
begin
  FStatus := FStatus + [dspReading];
  FLock.BeginWrite;
  try
    // Get global data source info
    FDescription := FDriver.Description;
    num := FDriver.DataSets;
    SetLength(FDataSets, num);
    if num = 0 then Exit;

    // Create the data sets
    for i := 0 to num - 1 do begin
      dst := FDriver.DataSetType(i);
      case dst of
        dspPFARaster: FDataSets[i] := TDSpRasterSource.Open(FDriver, i, self);
        dspPFAVector: ;
      end;
    end;
  finally
    FLock.EndWrite;     // Reset the lock so others can access the handler
    FStatus := [dspOk, dspIdle];
  end;
end;

function TDSpSourceHandler.CreateDataSet(ds: PDSpDataSetInfo; p: pointer): TDSpSourceDataSet;
var ndx, drvndx: integer;
begin
  FStatus := [dspBusy];
  FLock.BeginWrite;
  try
    ndx := Length(FDataSets);                // Make some room in the list of data sets
    SetLength(FDataSets, ndx+1);
    drvndx := FDriver.CreateDataSet(ds, p);  // Create the data set in the driver
    if drvndx = -1 then Result := nil        // Data set could not be created by the driver
    else begin
      case ds^.Feature of                    // Open the result data set in self
        dspPFARaster: FDataSets[ndx] := TDSpRasterSource.Open(FDriver, drvndx, self);
        dspPFAVector: ;
      end;
      Result := FDataSets[ndx];
    end;
  finally
    FLock.EndWrite;
  end;
  FStatus := [dspIdle];
end;

// AddDataSet adds an existing data set to the handler. All the information,
// data, etc. is copied from ds into a new data set in this handler
function TDSpSourceHandler.AddDataSet(ds: TDSpSourceDataSet): TDSpSourceDataSet;
var ndx, drvndx: integer;
begin
  FStatus := [dspBusy];
  FLock.BeginWrite;
  try
    ndx := Length(FDataSets);           // Make some room in the list of data sets
    SetLength(FDataSets, ndx+1);
    drvndx := FDriver.AddDataSet(ds);   // Add the data set to the driver
    if drvndx = -1 then Result := nil   // Data set could not be added
    else begin
      case ds.DataSetType of            // Create the result data set in self
        dspPFARaster: FDataSets[ndx] := TDSpRasterSource.Open(FDriver, drvndx, self);
        dspPFAVector: ;
      end;
      Result := FDataSets[ndx];
    end;
  finally
    FLock.EndWrite;
  end;
  FStatus := [dspIdle];
end;

// CloneDataSetFrom is used to create a data set that is compatible with an
// existing data set: same extent, etc. For rasters this means same rows,
// columns, and pixel size.
function TDSpSourceHandler.CloneDataSetFrom(ds: TDSpSourceDataSet): TDSpSourceDataSet;
var ndx, drvndx: integer;
begin
  FStatus := [dspBusy];
  FLock.BeginWrite;
  try
    ndx := Length(FDataSets);           // Make some room in the list of data sets
    SetLength(FDataSets, ndx+1);
    drvndx := FDriver.CloneDataSet(ds); // Clone the data set to the driver, no data added only structure
    if drvndx = -1 then Result := nil   // Data set could not be added
    else begin
      case ds.DataSetType of            // Create the result data set in self
        dspPFARaster: FDataSets[ndx] := TDSpRasterSource.Open(FDriver, drvndx, self);
        dspPFAVector: ;
      end;
      Result := FDataSets[ndx];
    end;
  finally
    FLock.EndWrite;
  end;
  FStatus := [dspIdle];
end;

procedure TDSpSourceHandler.AddRasterBand(msg: TDSpThreadMessage);
begin
  Assert(FDataSets[msg^.DataSourceIndex] is TDSpRasterSource);
  FLock.BeginRead;
  try
    TDSpRasterSource(FDataSets[msg^.DataSourceIndex]).AddBand(msg^.Parameter);
  finally
    FLock.EndRead;
  end;
end;

procedure TDSpSourceHandler.DeleteRasterBand(msg: TDSpThreadMessage);
begin
  Assert(FDataSets[msg^.DataSourceIndex] is TDSpRasterSource);
  FLock.BeginRead;
  try
    TDSpRasterSource(FDataSets[msg^.DataSourceIndex]).DeleteBand(integer(msg^.Parameter));
  finally
    FLock.EndRead;
  end;
end;

procedure TDSpSourceHandler.RenderRasterMap(msg: TDSpThreadMessage);
begin
  Assert(FDataSets[msg^.DataSourceIndex] is TDSpRasterSource);
  FLock.BeginRead;
  try
    TDSpRasterSource(FDataSets[msg^.DataSourceIndex]).RenderBand(msg^.Parameter);
  finally
    FLock.EndRead;
  end;
end;

procedure TDSpSourceHandler.ReadRasterBandDataByCoords(msg: TDSpThreadMessage);
begin
  Assert(FDataSets[msg^.DataSourceIndex] is TDSpRasterSource);
  FLock.BeginRead;
  try
    TDSpRasterSource(FDataSets[msg^.DataSourceIndex]).ReadBandByCoords(msg^.Parameter);
  finally
    FLock.EndRead;
  end;
end;

procedure TDSpSourceHandler.WriteRasterBandDataByCoords(msg: TDSpThreadMessage);
begin
  Assert(FDataSets[msg^.DataSourceIndex] is TDSpRasterSource);
  FLock.BeginRead;
  try
    TDSpRasterSource(FDataSets[msg^.DataSourceIndex]).WriteBandByCoords(msg^.Parameter);
  finally
    FLock.EndRead;
  end;
end;

procedure TDSpSourceHandler.QueueMessage(msg: TDSpThreadMessage);
begin
  FQueue.Put(msg); // don't need lock, because the queue is thread-safe
  Resume;
end;

function TDSpSourceHandler.ReadLock(ndx: integer): TDSpSourceDataSet;
begin
  FLock.BeginRead;
  try
    Result := FDataSets[ndx];
    Result.ReadLock;
  finally
    FLock.EndRead;
  end;
end;

function TDSpSourceHandler.ReadLock(const name: string): TDSpSourceDataSet;
var ndx: integer;
begin
  FLock.BeginRead;
  try
    ndx := FindDataSetByName(name);
    if ndx > -1 then begin
      Result := FDataSets[ndx];
      Result.ReadLock;
    end else Result := nil;
  finally
    FLock.EndRead;
  end;
end;

function TDSpSourceHandler.WriteLock(ndx: integer): TDSpSourceDataSet;
begin
  FLock.BeginRead;
  try
    Result := FDataSets[ndx];
    Result.WriteLock;
  finally
    FLock.EndRead;
  end;
end;

function TDSpSourceHandler.WriteLock(const name: string): TDSpSourceDataSet;
var ndx: integer;
begin
  FLock.BeginRead;
  try
    ndx := FindDataSetByName(name);
    if ndx > -1 then begin
      Result := FDataSets[ndx];
      Result.WriteLock;
    end else Result := nil;
  finally
    FLock.EndRead;
  end;
end;

function TDSpSourceHandler.FindDataSetByName(const name: string): integer;
var i: integer;
    ds: TDSpSourceDataSet;
    up: string;
begin
  Result := -1;    // flag failure
  if Length(FDataSets) > 0 then begin
    up := UpperCase(name);
    for i := 0 to High(FDataSets) do begin
      ds := ReadLock(i);
      try
        if UpperCase(ds.Name) = up then begin
          Result := i;
          Break;
        end;
      finally
        ds.ReadUnlock;
      end;
    end;
  end;
end;

function TDSpSourceHandler.GetSourceString: string;
begin
  Result := FConnect;
end;

function TDSpSourceHandler.GetNumDataSets: integer;
begin
  FLock.BeginRead;
  try
    Result := Length(FDataSets);
  finally
    FLock.EndRead;
  end;
end;

procedure TDSpSourceHandler.AddRef;
begin
  Inc(FRefs);
end;

procedure TDSpSourceHandler.AddDataSetRef(ndx: integer);
begin
  Assert(ndx in [0..High(FDataSets)]);
  FLock.BeginRead;
  try
    FDataSets[ndx].AddRef;
  finally
    FLock.EndRead;
  end;
end;

procedure TDSpSourceHandler.CloseDataSetRef(ndx: integer);
begin
  Assert(ndx in [0..High(FDataSets)]);
  FLock.BeginRead;
  try
    FDataSets[ndx].CloseRef;
  finally
    FLock.EndRead;
  end;
end;

function TDSpSourceHandler.TotalRefs: integer;
var i: integer;
    ds: TDSpSourceDataSet;
begin
  FLock.BeginRead;
  try
    Result := FRefs;
  finally
    FLock.EndRead;
  end;

  if Length(FDataSets) > 0 then
    for i := 0 to High(FDataSets) do begin
      ds := ReadLock(i);
      try
        Inc(Result, ds.FRefs);
      finally
        ds.ReadUnlock;
      end;
    end;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpBaseBroker - ancestor of all PFA drivers
constructor TDSpBaseBroker.Create;
begin
  // descendants should set the FDescription and FExtension members in their constructors
  inherited;
  FDrivers := TList.Create;
end;

destructor TDSpBaseBroker.Destroy;
begin
  // NOTE: descendants must free any open readers before invoking this destructor
  FDrivers.Clear;
  FDrivers.Free;
  inherited;
end;

function TDSpBaseBroker.Filter: string;
begin
  Result := FDescription + '|' + FFilter;
end;

// Basic termination of the driver thread. Override this virtual method for
// broker specific action, but ALWAYS call this method from the descendant
// method.
function TDSpBaseBroker.CloseDriver(r: TDSpBaseDriver): boolean;
begin
  Result := true;
  Free;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpPFA
constructor TDSpPFA.Create;
begin
  inherited;
  FLock := TMultiReadExclusiveWriteSynchronizer.Create;
  SetLength(FStaticBrokers, 1);
  FStaticBrokers[0] := TDSpTempRasterBroker.Create;
end;

destructor TDSpPFA.Destroy;
var i: integer;
begin
  FLock.BeginWrite;
  try
    if Length(FSources) > 0 then
      for i := 0 to High(FSources) do FSources[i].Free;
    Finalize(FSources);

    for i := High(FStaticBrokers) downto 0 do FStaticBrokers[i].Free;
    Finalize(FStaticBrokers);
  finally
    FLock.EndWrite;
  end;
  FLock.Free;
  inherited;
end;

// Register a static broker with the PFA. Static brokers cannot be removed from
// the PFA (hence the name). Use dynamic brokers in DLLs for resource
// conservation (mostly useful for less used data formats).
procedure TDSpPFA.RegisterBroker(d: TDSpBaseBroker);
begin
  // Lock is not required here because this function will only be called upon
  // initialization.
  SetLength(FStaticBrokers, Length(FStaticBrokers)+1);
  FStaticBrokers[High(FStaticBrokers)] := d;
end;

function TDSpPFA.FilterString(cap: TDSpPFACap): string;
var i: integer;
begin
  FLock.BeginRead;
  try
    Result := '';
    for i := 0 to High(FStaticBrokers) do
      if cap in FStaticBrokers[i].Capabilities then begin
        if Result <> '' then Result := Result + '|';
        Result := Result + FStaticBrokers[i].Filter;
      end;
  finally
    FLock.EndRead;
  end;
end;

function TDSpPFA.SaveAsFilter(const ds: TDSpSourceDataSet): string;
var i: integer;
    cap: TDSpPFACap;
    filter: string;
begin
  FLock.BeginRead;
  try
    Result := '';
    if ds is TDSpRasterSource then cap := dspWriteRaster
    else {if ds is TDSpVectorSource then} cap := dspWriteVector;

    for i := 0 to High(FStaticBrokers) do
      if cap in FStaticBrokers[i].Capabilities then begin
        filter := FStaticBrokers[i].CanSave(ds);
        if filter <> '' then begin
          if Result <> '' then Result := Result + '|';
          Result := Result + filter;
        end;
      end;
  finally
    FLock.EndRead;
  end;
end;

procedure TDSpPFA.InsertHandler(s: TDSpSourceHandler);
begin
  FLock.BeginWrite;
  try
    SetLength(FSources, Length(FSources) + 1);
    FSources[High(FSources)] := s;
  finally
    FLock.EndWrite;
  end;
end;

// Create a new data source
function TDSpPFA.CreateSource(const source: string): TDSpSourceHandler;
var broker: TDSpBaseBroker;
begin
  Result := nil;
  FLock.BeginRead;
  try
    // Find the right broker
    broker := IdentifyBroker(source);
    if broker = nil then Exit;

    // Create the handler
    Result := broker.CreateSource(source);
    Yield;
  finally
    FLock.EndRead;
  end;

  // Register the handler with the PFA
  if Result <> nil then InsertHandler(Result);
end;

// Create a handler to an existing data source
function TDSpPFA.OpenSource(const source: string): TDSpSourceHandler;
var broker: TDSpBaseBroker;
    i: integer;
    msg: TDSpThreadMessage;
begin
  Result := nil;
  FLock.BeginRead;
  try
    // First check if the source is already handled. If found, return that
    // handler and increase its ref count.
    if Length(FSources) > 0 then
      for i := 0 to High(FSources) do
        if source = FSources[i].Source then begin
          Result := FSources[i];
          Result.AddRef;
          Exit;
        end;

    // Find the right broker
    broker := IdentifyBroker(source);
    if broker = nil then Exit;

    // Create the handler and register it with the PFA
    Result := broker.OpenSource(source);
    New(msg);
    msg^.Event := nil;
    msg^.Msg := DSPMSG_SOURCEOPEN;
    msg^.Parameter := nil;
    Result.QueueMessage(msg);
    Yield;                      // Give the handler thread a chance to process
  finally                       // the message
    FLock.EndRead;
  end;
  if Result <> nil then InsertHandler(Result);
end;

// Create a temporary raster
function TDSpPFA.CreateTempRaster(ds: PDSpDatasetInfo; r: PDSpRasterInfo): TDSpSourceHandler;
begin
  Result := FStaticBrokers[0].CreateSource('');
  if Result.CreateDataSet(ds, r) = nil then FreeAndNil(Result)
  else InsertHandler(Result);
end;

function TDSpPFA.CreateTempRaster(r: TDSpSourceDataSet): TDSpSourceHandler;
begin
  Result := FStaticBrokers[0].CreateSource('');
  Result.CloneDataSetFrom(r);
  InsertHandler(Result);
end;

function TDSpPFA.IdentifyBroker(const source: string): TDSpBaseBroker;
var i: integer;
begin
  FLock.BeginRead;
  try
    for i := 0 to High(FStaticBrokers) do
      if FStaticBrokers[i].SupportsSource(source) then begin
        Result := FStaticBrokers[i];
        Exit;
      end;
    Result := nil;
  finally
    FLock.EndRead;
  end;
end;

// This method is called from TDSpatial.IdleProcessing when the application is
// idle. The list of sources is scanned to find any sources to which all
// references have been dropped. These sources are then deleted and the list
// updated.
procedure TDSpPFA.IdleProcessing;
var i, j, max, del: integer;
begin
  if Length(FSources) = 0 then Exit;

  FLock.BeginWrite;
  try
    max := High(FSources);
    del := 0;
    for i := 0 to max do  // First delete any unused sources...
      if FSources[i].TotalRefs = 0 then begin
        FSources[i].Free;
        FSources[i] := nil;
        Inc(del);
      end;
    if del > 0 then begin // ... then update the list if needed
      if max > 0 then begin
        j := max;
        while FSources[j] = nil do Dec(j); // j points to the highest index source
        for i := 0 to max - del do
          if FSources[i] = nil then begin
            FSources[i] := FSources[j];
            repeat Dec(j) until FSources[j] <> nil;
          end;
      end;
      SetLength(FSources, max - del + 1);
    end;
  finally
    FLock.EndWrite;
  end;
end;

initialization
  PFA := TDSpPFA.Create;
finalization
  if Assigned(PFA) then PFA.Free;
end.
