{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpProjection.pas. The Original Code is available       }
{ through the home page of DSpatial at http://dspatial.sourceforge.net.        }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit provides support for projecting coordinates between a number of    }
{ coordinate systems. Every data set has an instance of TDSpCoordinateSystem,  }
{ which has a member of TDSpProjection type which completely defines the       }
{ coordinate system of the data set. DSpatial natively supports a small number }
{ of projections. Other projections are supported through the PROJ library,    }
{ which is maintained by Frank Warmerdam (http://www.remotesensing.org/proj/)  }
{ and for which Alex Weidauer developed a Delphi wrapper to access the DLL.    }
{                                                                              }
{ DSpatial uses the OpenGIS WKT format for conveying projection information to }
{ the user and between modules (e.g. PFA reader and data set).                 }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 August 14                                               }
{                                                                              }
{ Dependencies  : DSpUtils (FloatsSame)                                        }
{                 DSpGlobals                                                   }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.1 03Aug14 Full coordinate system support, compatible with  }
{                             ISO/TC211 and OGC.                               }
{                 0.2 03May19 Initial coding.                                  }
{                                                                              }
{******************************************************************************}
unit DSpProjection;
{$I DSpatial.inc}
interface
uses DSpGlobals;
type
  TDSpSpheroid = class
  protected
    FName: string;
    FSemiMajorAxis: double;
    FInvFlattening: double;
  public
    constructor Create(name: string; a, invf: double);

    function    Describe(descr: TDSpDescribe): string;
    function    IsCompatible(s: TDSpSpheroid): boolean;
    property    Name: string read FName write FName;
    property    SemiMajorAxis: double read FSemiMajorAxis write FSemiMajorAxis;
    property    InvFlattening: double read FInvFlattening write FInvFlattening;
  end;

  TDSpDatumToWGS84 = (dspDX, dspDY, dspDZ, dspEX, dspEY, dspEZ, dspPPM);

  TDSpDatum = class
  protected
    FName: string;
    FSpheroid: TDSpSpheroid;
    FToWGS84: array[TDSpDatumToWGS84] of double; // conversion to WGS1984

    procedure   SetSpheroid(spheroid: TDSpSpheroid);
    function    GetConvert(i: TDSpDatumToWGS84): double;
    procedure   SetConvert(i: TDSpDatumToWGS84; value: double);
  public
    constructor Create(name: string; spheroid: TDSpSpheroid);
    destructor  Destroy; override;

    function    Describe(descr: TDSpDescribe): string;
    function    IsCompatible(d: TDSpDatum): boolean;
    property    Name: string read FName write FName;
    property    Spheroid: TDSpSpheroid read FSpheroid write SetSpheroid;
    property    ToWGS84[i: TDSpDatumToWGS84]: double read GetConvert write SetConvert;
  end;

  TDSpGeographicCoordinateSystem = class
  protected
    FName: string;
    FDatum: TDSpDatum;
    FPrimeMeridian: double;
    FPMName: string;
    FUnit: TDSpUnit;

    function    GetRadianConvert: double;
  public
    constructor Create(datum: TDSpDatum; name, pmName: string; pmLong: double; angularUnit: TDSpUnit);
    destructor  Destroy; override;

    function    Describe(descr: TDSpDescribe): string;
    function    IsCompatible(cs: TDSpGeographicCoordinateSystem): boolean;
    property    ToRadian: double read GetRadianConvert;
    property    Datum: TDSpDatum read FDatum;
    property    UOM: TDSpUnit read FUnit;
    property    PrimeMeridian: double read FPrimeMeridian;
  end;

  //////////////////////////////////////////////////////////////////////////////
  TDSpProjection = class
  public
    function    Describe(descr: TDSpDescribe): string; virtual; abstract;
    function    IsCompatible(proj: TDSpProjection): boolean; virtual; abstract;
    function    PROJstring: string; virtual; abstract;
  end;

  TDSpNullProjection = class(TDSpProjection)
  public
    function    Describe(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;
    function    PROJstring: string; override;
  end;

  TDSpBaseProjection = class(TDSpProjection)
  protected
    FGeogCS : TDSpGeographicCoordinateSystem;
    FEasting,
    FNorthing: double;         // in coordinate system units
    FCentralMeridian: double;  // in radians
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn: double);
    destructor  Destroy; override;

    function    Describe(descr: TDSpDescribe): string; override;
    function    WKTName: string; virtual; abstract;
    function    WKTParameters(descr: TDSpDescribe): string; virtual;
    function    IsCompatible(proj: TDSpProjection): boolean; override;
    function    PROJellps: string;

    property    GeographicCS: TDSpGeographicCoordinateSystem read FGeogCS;
    property    CentralMeridian: double read FCentralMeridian;
    property    FalseEasting: double read FEasting;
    property    FalseNorthing: double read FNorthing;
  end;

  TDSpProjectionGeographic = class(TDSpBaseProjection)
  public
    function    Describe(descr: TDSpDescribe): string; override;
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  //////////////////////////////////////////////////////////////////////////////
  TDSpProjectionCylindrical = class(TDSpBaseProjection)
  protected
    FLatitudeOfOrigin: double; // in radians
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin: double);

    function    WKTParameters(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;

    property    LatitudeOfOrigin: double read FLatitudeOfOrigin;
  end;

  TDSpCassini = class(TDSpProjectionCylindrical)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpCylindricalEqualArea = class(TDSpProjectionCylindrical)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpMercator = class(TDSpProjectionCylindrical)
  protected
    FScaleFactor: double; // dimensionless
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, k0: double);

    function    WKTName: string; override;
    function    WKTParameters(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;
    function    PROJstring: string; override;

    property    ScaleFactor: double read FScaleFactor;
  end;

  TDSpTransverseMercator = class(TDSpProjectionCylindrical)
  protected
    FScaleFactor: double; // dimensionless
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, k0: double);

    function    WKTName: string; override;
    function    WKTParameters(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;
    function    PROJstring: string; override;

    property    ScaleFactor: double read FScaleFactor;
  end;

  TDSpMiller = class(TDSpProjectionCylindrical)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

{  TDSpHotineObliqueMercator = class(TDSpProjectionCylindrical)
  public
    function    WKTName: string; override;
  end;
 }
  TDSpEquirectangular = class(TDSpProjectionCylindrical)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  //////////////////////////////////////////////////////////////////////////////
  TDSpProjectionConic = class(TDSpBaseProjection)
  protected
    FLatitudeOfOrigin: double; // in radians
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin: double);

    function    WKTParameters(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;
    property    LatitudeOfOrigin: double read FLatitudeOfOrigin;
  end;

  TDSpAlbersEqualAreaConic = class(TDSpProjectionConic)
  protected
    FFirstParallel, FSecondParallel: double; // in radians
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, lat1, lat2: double);

    function    WKTName: string; override;
    function    WKTParameters(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;
    function    PROJstring: string; override;

    property    FirstParallel: double read FFirstParallel;
    property    SecondParallel: double read FSecondParallel;
  end;

  TDSpEquidistantConic = class(TDSpProjectionConic)
  protected
    FFirstParallel, FSecondParallel: double; // in radians
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, lat1, lat2: double);

    function    WKTName: string; override;
    function    WKTParameters(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;
    function    PROJstring: string; override;

    property    FirstParallel: double read FFirstParallel;
    property    SecondParallel: double read FSecondParallel;
  end;

  TDSpLambertConformalConic = class(TDSpProjectionConic)
  protected
    FFirstParallel, FSecondParallel: double; // in radians
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin: double; lat1: double = -10; lat2: double = -10);

    function    WKTName: string; override;
    function    WKTParameters(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;
    function    PROJstring: string; override;

    property    FirstParallel: double read FFirstParallel;
    property    SecondParallel: double read FSecondParallel;
  end;

  TDSpPolyconic = class(TDSpProjectionConic)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  //////////////////////////////////////////////////////////////////////////////
  TDSpProjectionAzimuthal = class(TDSpBaseProjection)
  protected
    FLatitudeOfOrigin: double; // in radians
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin: double);

    function    WKTParameters(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;

    property    LatitudeOfOrigin: double read FLatitudeOfOrigin;
  end;

  TDSpAzimuthalEquidistant = class(TDSpProjectionAzimuthal)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpGnomonic = class(TDSpProjectionAzimuthal)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpLambertAzimuthalEqualArea = class(TDSpProjectionAzimuthal)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpOrthographic = class(TDSpProjectionAzimuthal)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpStereographic = class(TDSpProjectionAzimuthal)
  protected
    FScaleFactor: double; // dimensionless
  public
    constructor Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, k0: double);

    function    WKTName: string; override;
    function    WKTParameters(descr: TDSpDescribe): string; override;
    function    IsCompatible(proj: TDSpProjection): boolean; override;
    function    PROJstring: string; override;
    property    ScaleFactor: double read FScaleFactor;
  end;

////////////////////////////////////////////////////////////////////////////////
  TDSpGallStereographic = class(TDSpBaseProjection)   // cylindrical, but no lat0
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpEckertIV = class(TDSpBaseProjection)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpEckertVI = class(TDSpBaseProjection)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpMollweide = class(TDSpBaseProjection)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpRobinson = class(TDSpBaseProjection)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpSinusoidal = class(TDSpBaseProjection)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

  TDSpVanDerGrinten = class(TDSpBaseProjection)
  public
    function    WKTName: string; override;
    function    PROJstring: string; override;
  end;

implementation
uses SysUtils, DSpUtils;

////////////////////////////////////////////////////////////////////////////////
// TDSpSpheroid
constructor TDSpSpheroid.Create(name: string; a, invf: double);
begin
  inherited Create;
  FName := name;
  FSemiMajorAxis := a;
  FInvFlattening := invf;
end;

function TDSpSpheroid.Describe(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('SPHEROID["%s",%11.3f,%15.12f]', [FName, FSemiMajorAxis, FInvFlattening])
  else Result := '      Spheroid:'#9#9 + FName + #13#10 +
                 Format('        Semi-major axis:'#9'%11.3f', [FSemiMajorAxis]) + ' meter'#13#10 +
                 Format('        Inverse flattening:'#9'%15.12f', [FInvFlattening]) + #13#10;
end;

function TDSpSpheroid.IsCompatible(s: TDSpSpheroid): boolean;
begin
  Result := FloatsSame(FSemiMajorAxis, s.SemiMajorAxis) and
            FloatsSame(FInvFlattening, s.InvFlattening);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpDatum
constructor TDSpDatum.Create(name: string; spheroid: TDSpSpheroid);
begin
  FName := name;
  Assert(Assigned(spheroid));
  FSpheroid := spheroid;
end;

destructor TDSpDatum.Destroy;
begin
  FSpheroid.Free;
  inherited;
end;

procedure TDSpDatum.SetSpheroid(spheroid: TDSpSpheroid);
begin
  FSpheroid.Free;
  Assert(Assigned(spheroid));
  FSpheroid := spheroid;
end;

function TDSpDatum.Describe(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then begin
    Result := Format('DATUM["%s",%s', [FName, FSpheroid.Describe(descr)]);
    if FToWGS84[dspDX] > 0.0 then
      Result := Format('%s,TOWGS84[%8.6f,%8.6f,%8.6f,%8.6f,%8.6f,%8.6f,%8.6f]]',
                       [Result, FToWGS84[dspDX], FToWGS84[dspDY], FToWGS84[dspDZ],
                        FToWGS84[dspEX], FToWGS84[dspEY], FToWGS84[dspEZ], FToWGS84[dspPPM]])
    else Result := Result + ']';
  end else begin
    Result := '    Datum: ' + FName + #13#10 + FSpheroid.Describe(descr);
    if FToWGS84[dspDX] > 0.0 then
      Result := Result + Format('      Convert to WGS84:'#9'%8.6f, %8.6f, %8.6f, %8.6f, %8.6f, %8.6f, %8.6f',
                       [FToWGS84[dspDX], FToWGS84[dspDY], FToWGS84[dspDZ],
                        FToWGS84[dspEX], FToWGS84[dspEY], FToWGS84[dspEZ], FToWGS84[dspPPM]]);

  end;
end;

function TDSpDatum.IsCompatible(d: TDSpDatum): boolean;
begin
  Result := FSpheroid.IsCompatible(d.Spheroid);
  if Result and (FToWGS84[dspDX] > 0.0) then
    Result := FloatsSame(FToWGS84[dspDX], d.ToWGS84[dspDX]) and
              FloatsSame(FToWGS84[dspDY], d.ToWGS84[dspDY]) and
              FloatsSame(FToWGS84[dspDZ], d.ToWGS84[dspDZ]) and
              FloatsSame(FToWGS84[dspEX], d.ToWGS84[dspEX]) and
              FloatsSame(FToWGS84[dspEY], d.ToWGS84[dspEY]) and
              FloatsSame(FToWGS84[dspEZ], d.ToWGS84[dspEZ]) and
              FloatsSame(FToWGS84[dspPPM], d.ToWGS84[dspPPM]);
end;

function TDSpDatum.GetConvert(i: TDSpDatumToWGS84): double;
begin
  Result := FToWGS84[i];
end;

procedure TDSpDatum.SetConvert(i: TDSpDatumToWGS84; value: double);
begin
  FToWGS84[i] := value;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpGeographicCoordinateSystem
constructor TDSpGeographicCoordinateSystem.Create(datum: TDSpDatum; name, pmName: string; pmLong: double; angularUnit: TDSpUnit);
begin
  inherited Create;
  FName := name;
  FDatum := datum;
  FPMName := pmName;
  FUnit := angularUnit;
  FPrimeMeridian := pmLong * FUnit.Convert;
end;

destructor TDSpGeographicCoordinateSystem.Destroy;
begin
  FUnit.Free;
  FDatum.Free;
  inherited;
end;

function TDSpGeographicCoordinateSystem.GetRadianConvert: double;
begin
  Result := FUnit.Convert;
end;

function TDSpGeographicCoordinateSystem.Describe(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('GEOGCS["%s",%s,PRIMEM["%s",%15.13f],%s]',
                     [FName, FDatum.Describe(descr), FPMName, FPrimeMeridian / FUnit.Convert, FUnit.Describe(descr)])
  else Result := '  Earth system: ' + FName + #13#10'    Prime meridian: ' + FPMName + ' (' +
                 FloatToStrF(FPrimeMeridian / FUnit.Convert, ffGeneral, 15, 10) + ')'#13#10 +
                 FDatum.Describe(descr) + '  ' + FUnit.Describe(descr) + 'to radian)'#13#10;
end;

function TDSpGeographicCoordinateSystem.IsCompatible(cs: TDSpGeographicCoordinateSystem): boolean;
begin
  Result := FDatum.IsCompatible(cs.Datum) and FUnit.IsCompatible(cs.UOM) and FloatsSame(FPrimeMeridian, cs.PrimeMeridian);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpNullProjection
function TDSpNullProjection.Describe(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := 'LOCAL_CS["Local coordinate system (meter)",LOCAL_DATUM["Local datum",0],UNIT["Meter",1.0],AXIS["X",EAST],AXIS["Y",NORTH]]'
  else Result := 'Local coordinate system (or unknown projected coordinate system)';
end;

function TDSpNullProjection.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := proj is TDSpNullProjection;
end;

function TDSpNullProjection.PROJstring: string;
begin
  Result := '+proj=';
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpBaseProjection
constructor TDSpBaseProjection.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn: double);
begin
  inherited Create;
  FGeogCS := geogCS;
  FCentralMeridian := cm;
  FEasting := fe;
  FNorthing := fn;
end;

destructor TDSpBaseProjection.Destroy;
begin
  FGeogCS.Free;
  inherited;
end;

function TDSpBaseProjection.Describe(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('%s,PROJECTION["%s",%s', [FGeogCS.Describe(descr), WKTName, WKTParameters(descr)])
  else Result := FGeogCS.Describe(descr) + '  Projection:'#9#9 + WKTName + #13#10 + WKTParameters(descr);
end;

function TDSpBaseProjection.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('PARAMETER["Central_Meridian",%15.13f],PARAMETER["False_Easting",%5.3f],PARAMETER["False_Northing",%5.3f]',
                     [FCentralMeridian / FGeogCS.ToRadian, FEasting, FNorthing])
  else Result := '    Central meridian:'#9#9 + FloatToStrF(FCentralMeridian / FGeogCS.ToRadian, ffGeneral, 15, 13) +
                 #13#10'    False easting:'#9#9 + FloatToStrF(FEasting, ffGeneral, 12, 3) +
                 #13#10'    False northing:'#9#9 + FloatToStrF(FNorthing, ffGeneral, 12, 3) + #13#10;
end;

function TDSpBaseProjection.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpBaseProjection) and
            FGeogCS.IsCompatible((proj as TDSpBaseProjection).GeographicCS) and
            FloatsSame(FCentralMeridian, (proj as TDSpBaseProjection).CentralMeridian) and
            FloatsSame(FEasting, (proj as TDSpBaseProjection).FalseEasting) and
            FloatsSame(FNorthing, (proj as TDSpBaseProjection).FalseNorthing);
end;

function TDSpBaseProjection.PROJellps: string;
begin
  with FGeogCS.Datum.Spheroid do
    Result := Format('+a=%11.3f +rf=%11.7f +lon_0=%7.5f +x_0=%5.3f +y_0=%5.3f', [SemiMajorAxis, InvFlattening, FCentralMeridian / FGeogCS.ToRadian, FalseEasting, FalseNorthing]);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpProjectionGeographic
function TDSpProjectionGeographic.Describe(descr: TDSpDescribe): string;
begin
  Result := FGeogCS.Describe(descr);
end;

function TDSpProjectionGeographic.WKTName: string;
begin
  Result := 'Geographic';
end;

function TDSpProjectionGeographic.PROJstring: string;
begin
  Result := '+proj=latlong';
end;

////////////////////////////////////////////////////////////////////////////////
// Cylindrical projections
constructor TDSpProjectionCylindrical.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin: double);
begin
  inherited Create(geogCS, cm, fe, fn);
  FLatitudeOfOrigin := latOrigin;
end;

function TDSpProjectionCylindrical.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('%s,PARAMETER["Latitude_Of_Origin",%15.13f]',
              [inherited WKTParameters(descr), FLatitudeOfOrigin / FGeogCS.ToRadian])
  else Result := inherited WKTParameters(descr) + '    Latitude of origin:'#9 +
                 FloatToStrF(FLatitudeOfOrigin / FGeogCS.ToRadian, ffGeneral, 15, 13) + #13#10;
end;

function TDSpProjectionCylindrical.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpProjectionCylindrical) and
            inherited IsCompatible(proj) and
            FloatsSame(FLatitudeOfOrigin, (proj as TDSpProjectionCylindrical).LatitudeOfOrigin);
end;

function TDSpCassini.WKTName: string;
begin
  Result := 'Cassini_Soldner';
end;

function TDSpCassini.PROJstring: string;
begin
  Result := Format('+proj=cass %s +lat_0=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

function TDSpCylindricalEqualArea.WKTName: string;
begin
  Result := 'Cylindrical_Equal_Area';
end;

function TDSpCylindricalEqualArea.PROJstring: string;
begin
  Result := Format('+proj=cea %s +lat_ts=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

function TDSpGallStereographic.WKTName: string;
begin
  Result := 'Gall_Stereographic';
end;

function TDSpGallStereographic.PROJstring: string;
begin
  Result := Format('+proj=gall %s', [PROJellps]);
end;

constructor TDSpMercator.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, k0: double);
begin
  inherited Create(geogCS, cm, fe, fn, latOrigin);
  FScaleFactor := k0;
end;

function TDSpMercator.WKTName: string;
begin
  Result := 'Mercator_1SP';
end;

function TDSpMercator.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('%s,PARAMETER["Scale_Factor",%10.8f]', [inherited WKTParameters(descr), FScaleFactor])
  else Result := inherited WKTParameters(descr) + '    Scale factor:'#9#9 + FloatToStrF(FScaleFactor, ffGeneral, 10, 8) + #13#10;
end;

function TDSpMercator.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpMercator) and
            inherited IsCompatible(proj) and
            FloatsSame(FSCaleFactor, (proj as TDSpMercator).ScaleFactor);
end;

function TDSpMercator.PROJstring: string;
begin
  Result := Format('+proj=merc %s +lat_ts=%7.5f +k_0=%7.5f', [PROJellps, FLatitudeOfOrigin, FScaleFactor]);
end;

constructor TDSpTransverseMercator.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, k0: double);
begin
  inherited Create(geogCS, cm, fe, fn, latOrigin);
  FScaleFactor := k0;
end;

function TDSpTransverseMercator.WKTName: string;
begin
  Result := 'Transverse_Mercator';
end;

function TDSpTransverseMercator.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('%s,PARAMETER["Scale_Factor",%10.8f]', [inherited WKTParameters(descr), FScaleFactor])
  else Result := inherited WKTParameters(descr) + '    Scale factor:'#9#9 + FloatToStrF(FScaleFactor, ffGeneral, 10, 8) + #13#10;
end;

function TDSpTransverseMercator.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpTransverseMercator) and
            inherited IsCompatible(proj) and
            FloatsSame(FScaleFactor, (proj as TDSpTransverseMercator).ScaleFactor);
end;

function TDSpTransverseMercator.PROJstring: string;
begin
  Result := Format('+proj=tmerc %s +lat_0=%7.5f +k_0=%7.5f', [PROJellps, FLatitudeOfOrigin, FScaleFactor]);
end;

function TDSpMiller.WKTName: string;
begin
  Result := 'Miller_Cylindrical';
end;

function TDSpMiller.PROJstring: string;
begin
  Result := Format('+proj=mill %s +lat_0=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

function TDSpEquirectangular.WKTName: string;
begin
  Result := 'Equirectangular';
end;

function TDSpEquirectangular.PROJstring: string;
begin
  Result := Format('+proj=eqc %s +lat_ts=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

////////////////////////////////////////////////////////////////////////////////
// Conic projections
constructor TDSpProjectionConic.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin: double);
begin
  inherited Create(geogCS, cm, fe, fn);
  FLatitudeOfOrigin := latOrigin;
end;

function TDSpProjectionConic.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('%s,PARAMETER["Latitude_Of_Origin",%15.13f]',
              [inherited WKTParameters(descr), FLatitudeOfOrigin / FGeogCS.ToRadian])
  else Result := inherited WKTParameters(descr) + '    Latitude of origin:'#9 +
                 FloatToStrF(FLatitudeOfOrigin / FGeogCS.ToRadian, ffGeneral, 15, 13) + #13#10;
end;

function TDSpProjectionConic.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpProjectionConic) and
            inherited IsCompatible(proj) and
            FloatsSame(FLatitudeOfOrigin, (proj as TDSpProjectionConic).LatitudeOfOrigin);
end;

constructor TDSpAlbersEqualAreaConic.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, lat1, lat2: double);
begin
  inherited Create(geogCS, cm, fe, fn, latOrigin);
  FFirstParallel := lat1;
  FSecondParallel := lat2;
end;

function TDSpAlbersEqualAreaConic.WKTName: string;
begin
  Result := 'Albers_Conic_Equal_Area';
end;

function TDSpAlbersEqualAreaConic.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('%s,PARAMETER["Standard_Parallel_1",%15.13f],PARAMETER["Standard_Parallel_2",%15.13f]',
                     [inherited WKTParameters(descr), FFirstParallel / FGeogCS.ToRadian, FSecondParallel / FGeogCS.ToRadian])
  else Result := inherited WKTParameters(descr) +
                 '    Standard parallel 1:'#9 + FloatToStrF(FFirstParallel / FGeogCS.ToRadian, ffGeneral, 15, 13) +
                 #13#10'    Standard parallel 2:'#9 + FloatToStrF(FSecondParallel / FGeogCS.ToRadian, ffGeneral, 15, 13) + #13#10;
end;

function TDSpAlbersEqualAreaConic.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpAlbersEqualAreaConic) and
            inherited IsCompatible(proj) and
            FloatsSame(FFirstParallel, (proj as TDSpAlbersEqualAreaConic).FirstParallel) and
            FloatsSame(FSecondParallel, (proj as TDSpAlbersEqualAreaConic).SecondParallel);
end;

function TDSpAlbersEqualAreaConic.PROJstring: string;
begin
  Result := Format('+proj=aea %s +lat_1=%7.5f +lat_2=%7.5f +lat_0=%7.5f', [PROJellps, FFirstParallel, FSecondParallel, FLatitudeOfOrigin]);
end;

constructor TDSpEquidistantConic.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, lat1, lat2: double);
begin
  inherited Create(geogCS, cm, fe, fn, latOrigin);
  FFirstParallel := lat1;
  FSecondParallel := lat2;
end;

function TDSpEquidistantConic.WKTName: string;
begin
  Result := 'Equidistant_Conic';
end;

function TDSpEquidistantConic.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('%s,PARAMETER["Standard_Parallel_1",%15.13f],PARAMETER["Standard_Parallel_2",%15.13f]',
                     [inherited WKTParameters(descr), FFirstParallel / FGeogCS.ToRadian, FSecondParallel / FGeogCS.ToRadian])
  else Result := inherited WKTParameters(descr) +
                 '    Standard parallel 1:'#9 + FloatToStrF(FFirstParallel / FGeogCS.ToRadian, ffGeneral, 15, 13) +
                 #13#10'    Standard parallel 2:'#9 + FloatToStrF(FSecondParallel / FGeogCS.ToRadian, ffGeneral, 15, 13) + #13#10;
end;

function TDSpEquidistantConic.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpEquidistantConic) and
            inherited IsCompatible(proj) and
            FloatsSame(FFirstParallel, (proj as TDSpEquidistantConic).FirstParallel) and
            FloatsSame(FSecondParallel, (proj as TDSpEquidistantConic).SecondParallel);
end;

function TDSpEquidistantConic.PROJstring: string;
begin
  Result := Format('+proj=eqdc %s +lat_1=%7.5f +lat_2=%7.5f +lat_0=%7.5f', [PROJellps, FFirstParallel, FSecondParallel, FLatitudeOfOrigin]);
end;

constructor TDSpLambertConformalConic.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin: double; lat1: double = -10; lat2: double = -10);
// The default values for lat1 and lat2 indicate use of single parallel latOrigin
begin
  inherited Create(geogCS, cm, fe, fn, latOrigin);
  FFirstParallel := lat1;
  FSecondParallel := lat2;
end;

function TDSpLambertConformalConic.WKTName: string;
begin
  if FFirstParallel > -1 then Result := 'Lambert_Conformal_Conic_2SP'
  else Result := 'Lambert_Conformal_Conic_1SP';
end;

function TDSpLambertConformalConic.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then begin
    if FFirstParallel > -1 then
      Result := Format('%s,PARAMETER["Standard_Parallel_1",%15.13f],PARAMETER["Standard_Parallel_2",%15.13f]',
                     [inherited WKTParameters(descr), FFirstParallel / FGeogCS.ToRadian, FSecondParallel / FGeogCS.ToRadian])
    else Result := inherited WKTParameters(descr);
  end else begin
    Result := inherited WKTParameters(descr);
    if FFirstParallel > -1 then
      Result := Result + '    Standard parallel 1:'#9 + FloatToStrF(FFirstParallel / FGeogCS.ToRadian, ffGeneral, 15, 13) +
                   #13#10'    Standard parallel 2:'#9 + FloatToStrF(FSecondParallel / FGeogCS.ToRadian, ffGeneral, 15, 13) + #13#10;
  end;
end;

function TDSpLambertConformalConic.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpLambertConformalConic) and inherited IsCompatible(proj);
  if Result and (FFirstParallel > -1) then
    Result := FloatsSame(FFirstParallel, (proj as TDSpLambertConformalConic).FirstParallel) and
              FloatsSame(FSecondParallel, (proj as TDSpLambertConformalConic).SecondParallel);
end;

function TDSpLambertConformalConic.PROJstring: string;
begin
  if FFirstParallel > -1 then
    Result := Format('+proj=lcc %s +lat_1=%7.5f +lat_2=%7.5f +lat_0=%7.5f', [PROJellps, FFirstParallel, FSecondParallel, FLatitudeOfOrigin])
  else
    Result := Format('+proj=lcc %s +lat_1=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

function TDSpPolyconic.WKTName: string;
begin
  Result := 'Polyconic';
end;

function TDSpPolyconic.PROJstring: string;
begin
  Result := Format('+proj=poly %s +lat_0=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

////////////////////////////////////////////////////////////////////////////////
// Azimuthal projections
constructor TDSpProjectionAzimuthal.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin: double);
begin
  inherited Create(geogCS, cm, fe, fn);
  FLatitudeOfOrigin := latOrigin;
end;

function TDSpProjectionAzimuthal.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('%s,PARAMETER["Latitude_Of_Origin",%15.13f]',
              [inherited WKTParameters(descr), FLatitudeOfOrigin / FGeogCS.ToRadian])
  else Result := inherited WKTParameters(descr) + '    Latitude of origin:'#9 +
                 FloatToStrF(FLatitudeOfOrigin / FGeogCS.ToRadian, ffGeneral, 15, 13) + #13#10;
end;

function TDSpProjectionAzimuthal.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpProjectionAzimuthal) and
            inherited IsCompatible(proj) and
            FloatsSame(FLatitudeOfOrigin, (proj as TDSpProjectionAzimuthal).LatitudeOfOrigin);
end;

function TDSpAzimuthalEquidistant.WKTName: string;
begin
  Result := 'Azimuthal_Equidistant';
end;

function TDSpAzimuthalEquidistant.PROJstring: string;
begin
  Result := Format('+proj=aeqd %s +lat_0=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

function TDSpGnomonic.WKTName: string;
begin
  Result := 'Gnomonic';
end;

function TDSpGnomonic.PROJstring: string;
begin
  Result := Format('+proj=gnom %s +lat_0=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

function TDSpLambertAzimuthalEqualArea.WKTName: string;
begin
  Result := 'Lambert_Azimuthal_Equal_Area';
end;

function TDSpLambertAzimuthalEqualArea.PROJstring: string;
begin
  Result := Format('+proj=laea %s +lat_0=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

function TDSpOrthographic.WKTName: string;
begin
  Result := 'Orthographic';
end;

function TDSpOrthographic.PROJstring: string;
begin
  Result := Format('+proj=ortho %s +lat_0=%7.5f', [PROJellps, FLatitudeOfOrigin]);
end;

constructor TDSpStereographic.Create(geogCS: TDSpGeographicCoordinateSystem; cm, fe, fn, latOrigin, k0: double);
begin
  inherited Create(geogCS, cm, fe, fn, latOrigin);
  FScaleFactor := k0;
end;

function TDSpStereographic.WKTName: string;
begin
  if FloatsSame(Abs(FLatitudeOfOrigin), Pi/2) then Result := 'Polar_Stereographic'
  else Result := 'Oblique_Stereographic';
end;

function TDSpStereographic.WKTParameters(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then
    Result := Format('%s,PARAMETER["Scale_Factor",%10.8f]', [inherited WKTParameters(descr), FScaleFactor])
  else Result := inherited WKTParameters(descr) + '    Scale factor:'#9#9 + FloatToStrF(FScaleFactor, ffGeneral, 10, 8) + #13#10;
end;

function TDSpStereographic.IsCompatible(proj: TDSpProjection): boolean;
begin
  Result := (proj is TDSpStereographic) and
            inherited IsCompatible(proj) and
            FloatsSame(FSCaleFactor, (proj as TDSpStereographic).ScaleFactor);
end;

function TDSpStereographic.PROJstring: string;
begin
  Result := Format('+proj=stere %s +lat_0=%7.5f +k_0=%7.5f', [PROJellps, FLatitudeOfOrigin, FSCaleFactor]);
end;

////////////////////////////////////////////////////////////////////////////////
// Miscellaneous projections
function TDSpEckertIV.WKTName: string;
begin
  Result := 'Eckert_IV';
end;

function TDSpEckertIV.PROJstring: string;
begin
  Result := Format('+proj=eck4 %s', [PROJellps]);
end;

function TDSpEckertVI.WKTName: string;
begin
  Result := 'Eckert_VI';
end;

function TDSpEckertVI.PROJstring: string;
begin
  Result := Format('+proj=eck6 %s', [PROJellps]);
end;

function TDSpMollweide.WKTName: string;
begin
  Result := 'Mollweide';
end;

function TDSpMollweide.PROJstring: string;
begin
  Result := Format('+proj=moll %s', [PROJellps]);
end;

function TDSpRobinson.WKTName: string;
begin
  Result := 'Robinson';
end;

function TDSpRobinson.PROJstring: string;
begin
  Result := Format('+proj=robin %s', [PROJellps]);
end;

function TDSpSinusoidal.WKTName: string;
begin
  Result := 'Sinusoidal';
end;

function TDSpSinusoidal.PROJstring: string;
begin
  Result := Format('+proj=sinu %s', [PROJellps]);
end;

function TDSpVanDerGrinten.WKTName: string;
begin
  Result := 'VanDerGrinten';
end;

function TDSpVanDerGrinten.PROJstring: string;
begin
  Result := Format('+proj=vandg %s', [PROJellps]);
end;

end.
