{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpProjector.pas. The Original Code is available through}
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Alexander Weidauer             }
{ (alexweidauer@users.sourceforge.net). Portions created by the Initial        }
{ Developer are Copyright (C) of the Initial Developer. Changes to the Original}
{ Code available through DSpatial are Copyright (C) of the Contributing        }
{ Authors.                                                                     }
{                                                                              }
{******************************************************************************}
{ This unit provides an object wrapper called TDSpProjector around the         }
{ application interface to Frank Warmerdam's geographic projection toolkit     }
{ proj, version 4.4.7, for use with DSpatial. This unit is specific to the     }
{ Win32 environment.                                                           }
{                                                                              }
{ The PROJ library is open source software, originally developed by Gerald     }
{ Everden and maintained by Frank Warmerdam. PROJ can be found at              }
{ http://www.remotesensing.org/proj/.                                          }
{                                                                              }
{ Unit owner    : Alexander Weidauer                                           }
{ Last modified : 2003 August 01                                               }
{                                                                              }
{ Dependencies  : DSpGlobals                                                   }
{                 proj447.dll (the PROJ library)                               }
{                 SysUtils (Exception)                                         }
{                 Math                                                         }
{                 Windows (library access)                                     }
{                                                                              }
{ Version       : 0.1  03Aug01 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpProjector;
{$I DSpatial.inc}
interface
uses DSpGlobals;
type
  // Point record and analoga
  TProjUV = record
              U, V: TDSpFloat64;
            end;
  TProjXY = TProjUV;
  TProjLP = TProjUV;

  PProjPJ = Pointer;

  TDSpProjector = class
  private
    FPJ       : PProjPJ;   // Pointer to the projection internal structure.
    FProjStr  : string;    // Current parameter string for initialization.
    FCurrentLL: TProjUV;   // Latitude / longitude buffer
    FCurrentXY: TProjXY;   // Plane XY data buffer
  public
    constructor Create;
    destructor  Destroy; override;

    function    SetProjection(Parameter: string): boolean;
    procedure   ClearProjection;
    procedure   LonLatToPlane(Lon, Lat: TDSpFloat64; var X, Y: TDSpFloat64);
    function    LonLatToPlaneCheck(Lon, Lat: TDSpFloat64; var X, Y: TDSpFloat64): boolean;
    procedure   PlaneToLonLat(X, Y: TDSpFloat64; var Lon, Lat: TDSpFloat64);

    property    Parameter: string read FProjStr;
    property    CurrentLonLat: TProjUV read FCurrentLL;
    property    CurrentPlane : TProjXY read FCurrentXY;
  end;

implementation
uses Windows, Math, SysUtils;
const
  { Name of the DLL }
  cProjLibName = 'proj447.dll';
  { Scaling value radiant to decimal degree }
  cRAD_TO_DEG   =	57.29577951308232;
  { Scaling value decimal degree to radiant }
  cDEG_TO_RAD   =	0.0174532925199432958;

  // Errorstrings
  cPjErrorOpenDLL = 'Library "%s" not found!';
  cPjErrorOpenProc = 'Procedure "%s" in library "%s" not found!';
  cPjErrorInitPath = 'Initialization file "proj_def.dat" in path "%s" not found!';
  cPjErrorInitLibrary = 'Initialization of projection toolkit failed!';
  cPjErrorNotInitLibrary = 'The projection toolkit is not initilized!';
  cPjErrorForward = 'Fatal error in projection toolkit while forward projection!';
  cPjErrorInverse = 'Fatal error in projection toolkit while inverse projection!';
  cPjDefaultProjection = '+proj=merc +ellps=WGS84 +lon_0=0';

  // Mathmatical constants
  Infinity    = 1/0;       // this is tricky
  NaN         = 0/0;       // this is tricky
  NegInfinity = -Infinity;

  // Function prototype names in DLLL
  cPjInitialize = '_pj_init_plus_path';
  cPjForward    = '_pj_fwd';
  cPjInverse    = '_pj_inv';
  cPjDispose    = '_pj_free';

type
  // Function prototypes for DLL access
  TPjInitialize = function(const Args: PChar; const Path: PChar): PProjPJ; cdecl;
  TPjForward    = function(ProjLP: TProjLP; projPJ: PProjPJ): TProjXY; cdecl;
  TPjInverse    = function(ProjXY: TProjXY; projPJ: PProjPJ): TProjLP; cdecl;
  TPjDispose    = function(projPJ: PProjPJ): TDSpInt32; cdecl;

var
 PjInitialize: TPjInitialize;
 PjForward   : TPjForward;
 PjInverse   : TPjInverse;
 PjDispose   : TPjDispose;
 DllHandle   : THandle;

constructor TDSpProjector.Create;
begin
  inherited;
  // Initialize the projection
  FProjStr := cPjDefaultProjection;
  FPJ := PjInitialize(PChar(FProjStr), nil);
  if FPJ = nil then raise Exception.Create(cPjErrorInitLibrary);
end;

destructor TDSpProjector.Destroy;
begin
  ClearProjection;
  inherited;
end;

// Clear the current projection
procedure TDSpProjector.ClearProjection;
begin
  if FPJ <> nil then PjDispose(FPJ);
  FPJ := nil;
end;

// Set a certain projection
function TDSpProjector.SetProjection(Parameter: string): boolean;
begin
  Result := false;

  // Dismiss the current projection
  ClearProjection;

  // try to set up a new one
  FPJ := PjInitialize(PChar(Parameter), nil);
  if FPJ = nil then raise Exception.Create(cPjErrorInitLibrary)
  else Result := true;
end;

// Calculate the forward projection. Input is in radians, output is in meter
procedure TDSpProjector.LonLatToPlane(Lon, Lat: TDSpFloat64; Var X, Y: TDSpFloat64);
begin
  // If there is no projection available, throw an exception.
  if FPJ = nil then raise Exception.Create(cPjErrorNotInitLibrary);

  // Set the data
  FCurrentLL.U := Lon;
  FCurrentLL.V := Lat;
  try
    FCurrentXY := PjForward(FCurrentLL, FPJ);
  except
    raise Exception.Create(cPjErrorForward);
  end;

  // Set the output values
  X := FCurrentXY.U;
  Y := FCurrentXY.V;
end;

function TDSpProjector.LonLatToPlaneCheck(Lon, Lat: TDSpFloat64; var X, Y: TDSpFloat64): boolean;
begin
  LonLatToPlane(Lon, Lat, X, Y);
  Result := not((X = Infinity) or (Y = Infinity) or (X = -Infinity) or (Y = -Infinity));
end;

// Calculate the inverse projection. Input is in meter, output is in radians
procedure TDSpProjector.PlaneToLonLat(X, Y: TDSpFloat64; var Lon, Lat: TDSpFloat64);
begin
  // If there is no projection available, throw an exception
  if FPJ = nil then raise Exception.Create(cPjErrorNotInitLibrary);

  // Set the data
  FCurrentXY.U := X;
  FCurrentXY.V := Y;
  try
    FCurrentLL := PjInverse(FCurrentXY, FPJ);
  except
    raise Exception.Create(cPjErrorInverse);
  end;

  // Set the output values
  Lon := FCurrentLL.U;
  Lat := FCurrentLL.V;
end;

initialization
  // Initialize the DLL and set up the basic projection routines
  DLLHandle := LoadLibrary(cProjLibName);
  if DLLHandle = 0 then
    raise Exception.Create(Format(cPjErrorOpenDLL, [cProjLibName]));

  // Load the routine addresses from the DLL
  @PjInitialize := GetProcAddress(DLLHandle, cPjInitialize);
  if @PjInitialize = nil then
    raise Exception.Create(Format(cPjErrorOpenProc, [cPjInitialize, cProjLibName]));

  @PjForward := GetProcAddress(DLLHandle, cPjForward);
  if @PjForward = nil then
    raise Exception.Create(Format(cPjErrorOpenProc, [cPjForward]));

  @PjInverse:=GetProcAddress(DLLHandle, cPjInverse);
  if @PjInverse = nil then
    raise Exception.Create(Format(cPjErrorOpenProc, [cPjInverse]));

  @PjDispose:=GetProcAddress(DLLHandle, cPjDispose);
  if @PjDispose = nil then
    raise Exception.Create(Format(cPjErrorOpenProc, [cPjDispose]));

finalization
  // Free the library
  FreeLibrary(DLLHandle);
  DLLHandle := 0;
End.

