{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRRT.pas. The Original Code is available through the  }
{ home page of DSpatial at http://dspatial.sourceforge.net.                    }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements the so-called raster renderer transforms, to convert    }
{ from raster band values to colours or colour bands. The RRT is owned by the  }
{ renderer and it is passed to its associated raster objects for rendering of  }
{ a bitmap of an invalidated screen area.                                      }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 16                                             }
{                                                                              }
{ Dependencies  : DSpGlobals                                                   }
{                 Math (Ceil)                                                  }
{                 Classes (Rect)                                               }
{                 Graphics (TBitmap)                                           }
{                 Windows (RGBTriple)                                          }
{                                                                              }
{ Version       : 0.4  03Nov16 Raster band transformation uses part of render  }
{                              map instead of entire map. TRect replaced by    }
{                              TDSpRect.                                       }
{                 0.3  03Nov07 Bitmaps filled with transparent colour before   }
{                              rendering.                                      }
{                 0.2  03Aug14 Complete overhaul of previous version. Addition }
{                              of TDSpRRTSmoothRamp.                           }
{                 0.1  03Jul18 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpRRT;
{$I DSpatial.inc}
interface
{$IFDEF WIN32}
uses Windows, Graphics, DSpWin32,
{$ENDIF}
     DSpGlobals;
type
  TDSpFloat64Array = array of TDSpFloat64;
  TDSpWordArray   = array of word;
  TDSpColorArray  = array of TColor;

  TDSpRRTBand = class
  end;

  TDSpRRT = class
  public
    procedure Transform(renderMap: TDSpRenderMap; area: TRect; bmp: TBitmap); virtual;
    procedure Ramp(bmp: TBitMap; dir: TDSpDirection); virtual; abstract;
    function  Transparent: TColor; virtual; abstract;
  end;

  TDSpRRTSingleBand = class(TDSpRRT)
  end;

  TDSpRRT3Band = class(TDSpRRT)
  protected
    FRed, FGreen, FBlue: TDSpRRTBand;
  end;

  TDSpRRTGrayScale = class(TDSpRRTSingleBand)
  public
    procedure Transform(renderMap: TDSpRenderMap; area: TRect; bmp: TBitmap); override;
    procedure Ramp(bmp: TBitMap; dir: TDSpDirection); override;
    function  Transparent: TColor; override;
  end;

  TDSpRRTHotToCold = class(TDSpRRTSingleBand)
  public
    procedure Transform(renderMap: TDSpRenderMap; area: TRect; bmp: TBitmap); override;
    procedure Ramp(bmp: TBitMap; dir: TDSpDirection); override;
    function  Transparent: TColor; override;
  end;

  TDSpRampData = record
                   Value  : TDSpFloat64;
                   Red,
                   Green,
                   Blue,
                   Alpha  : byte;
                   Break  : word;
                   Diff   : word;
                 end;

  TDSpRRTSmoothRamp = class(TDSpRRTSingleBand)
  protected
    FData     : array of TDSpRampData;

    function    GetBreaks: integer;
    procedure   SetValue(ndx: integer; val: TDSpFloat64);
    function    GetValue(ndx: integer): TDSpFloat64;
    procedure   SetColour(ndx: integer; colour: TColor);
    function    GetColour(ndx: integer): TColor;
    procedure   RecalcIntervals;
  public
    constructor Create;
    destructor  Destroy; override;

    function    Insert(ndx: integer): integer;
    function    Delete(ndx: integer): boolean;
    procedure   Invert;
    procedure   Transform(renderMap: TDSpRenderMap; area: TRect; bmp: TBitmap); override;
    procedure   Ramp(bmp: TBitMap; dir: TDSpDirection); override;
    function    Transparent: TColor; override;

    property    Intervals: integer read GetBreaks;
    property    Value[ndx: integer]: TDSpFloat64 read GetValue write SetValue;
    property    Colour[ndx: integer]: TColor read GetColour write SetColour;
  end;

implementation
uses Classes, Math;

procedure TDSpRRT.TransForm(renderMap: TDSpRenderMap; area: TRect; bmp: TBitmap);
begin
  bmp.TransparentColor := Transparent;
  bmp.Canvas.Brush.Color := Transparent;
  bmp.Canvas.FillRect(Rect(0, 0, bmp.Width, bmp.Height));
end;

procedure TDSpRRTHotToCold.TransForm(renderMap: TDSpRenderMap; area: TRect; bmp: TBitmap);
var val: PWord;
    r, c: integer;
    sl: PRGBTriple;
begin
  inherited;
  Assert(bmp <> nil);
  for r := 0 to bmp.Height - 1 do begin
    sl := bmp.ScanLine[r];
    val := PWord(integer(renderMap.Row[r+area.Top]) + area.Left * 2);
    for c := 0 to bmp.Width - 1 do begin
      if val^ > 0 then
        if val^ < 16384 then begin
          sl^.rgbtRed := 0;
          sl^.rgbtGreen := val^ shr 6;
          sl^.rgbtBlue := 255;
        end else if val^ < 32768 then begin
          sl^.rgbtRed := 0;
          sl^.rgbtGreen := 255;
          sl^.rgbtBlue := (32767 - val^) shr 6;
        end else if val^ < 49152 then begin
          sl^.rgbtRed := (val^ - 32768) shr 6;
          sl^.rgbtGreen := 255;
          sl^.rgbtBlue := 0;
        end else begin
          sl^.rgbtRed := 255;
          sl^.rgbtGreen := (65535 - val^) shr 6;
          sl^.rgbtBlue := 0;
        end;
      Inc(sl);
      Inc(val);
    end;
  end;
end;

procedure TDSpRRTHotToCold.Ramp(bmp: TBitMap; dir: TDSpDirection);
var run, r, c: integer;
    sl: PRGBTriple;
    val: word;
    red, green, blue: byte;
begin
  if dir = dspVert then begin
    run := bmp.Height - 1;
    for r := 0 to run do begin
      sl := bmp.ScanLine[r];
      val := (run - r) * 65534 div run + 1;
      if val < 16384 then begin
        red := 0;
        green := val shr 6;
        blue := 255;
      end else if val < 32768 then begin
        red := 0;
        green := 255;
        blue := (32767 - val) shr 6;
      end else if val < 49152 then begin
        red := (val - 32768) shr 6;
        green := 255;
        blue := 0;
      end else begin
        red := 255;
        green := (65535 - val) shr 6;
        blue := 0;
      end;
      for c := 0 to bmp.Width - 1 do begin
        sl^.rgbtRed := red;
        sl^.rgbtGreen := green;
        sl^.rgbtBlue := blue;
        Inc(sl);
      end;
    end;
  end else if dir = dspHorz then begin
    // No use for this just yet
  end;
end;

function TDSpRRTHotToCold.Transparent: TColor;
begin
  Result := $00FFFFFF;
end;

procedure TDSpRRTGrayScale.TransForm(renderMap: TDSpRenderMap; area: TRect; bmp: TBitmap);
var val: PWord;
    r, c: integer;
    sl: PRGBTriple;
begin
  inherited;
  for r := 0 to bmp.Height - 1 do begin
    sl := bmp.ScanLine[r];
    val := PWord(integer(renderMap.Row[r+area.Top]) + area.Left * 2);
    for c := 0 to bmp.Width - 1 do begin
      if val^ > 0 then begin
        sl^.rgbtRed   := val^ shr 8;
        sl^.rgbtGreen := sl^.rgbtRed;
        sl^.rgbtBlue  := sl^.rgbtRed;
      end;
      Inc(sl);
      Inc(val);
    end;
  end;
end;

procedure TDSpRRTGrayScale.Ramp(bmp: TBitMap; dir: TDSpDirection);
var run, r, c: integer;
    sl : PRGBTriple;
    col: TRGBTriple;
    val: byte;
begin
  if dir = dspVert then begin
    run := bmp.Height - 1;
    for r := 0 to run do begin
      sl := bmp.ScanLine[r];
      val := (run - r) * 254 div run + 1;
      col.rgbtRed := val;
      col.rgbtGreen := val;
      col.rgbtBlue := val;
      for c := 0 to bmp.Width - 1 do begin
        sl^ := col;
        Inc(sl);
      end;
    end;
  end else if dir = dspHorz then begin

  end;
end;

function TDSpRRTGrayScale.Transparent: TColor;
begin
  Result := $0000000F;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRRTSmoothRamp
constructor TDSpRRTSmoothRamp.Create;
begin
  SetLength(FData, 2);
  FData[0].Value := 0.0;
  FData[1].Value := 1000.0;

  FData[0].Red   := $76;
  FData[0].Green := $68;
  FData[0].Blue  := $49;   // brownish
  FData[1].Red   := $BC;
  FData[1].Green := $D3;
  FData[1].Blue  := $49;   // brownish

  RecalcIntervals;
end;

destructor TDSpRRTSmoothRamp.Destroy;
begin
  Finalize(FData);
  inherited;
end;

function TDSpRRTSmoothRamp.GetBreaks: integer;
begin
  Result := Length(FData);
end;

procedure TDSpRRTSmoothRamp.RecalcIntervals;
var i, max: integer;
    min, scale: TDSpFloat64;
begin
  max := High(FData);
  FData[0].Break := 1;
  FData[max].Break := 65535;
  if max > 1 then begin
    min := FData[0].Value;
    scale := 65535 / (FData[max].Value - min);
    for i := 1 to max - 1 do
      FData[i].Break := Ceil((FData[i].Value - min) * scale);
  end;
  for i := 1 to max do
    FData[i].Diff := FData[i].Break - FData[i-1].Break;
end;

procedure TDSpRRTSmoothRamp.SetValue(ndx: integer; val: TDSpFloat64);
begin
  Assert((ndx >= 0) and (ndx <= High(FData)));
  if (ndx > 0) and (val < FData[ndx-1].Value) then Exit;
  if (ndx < High(FData)) and (val > FData[ndx+1].Value) then Exit;
  FData[ndx].Value := val;
  RecalcIntervals;
end;

function TDSpRRTSmoothRamp.GetValue(ndx: integer): TDSpFloat64;
begin
  Assert((ndx >= 0) and (ndx <= High(FData)));
  Result := FData[ndx].Value;
end;

procedure TDSpRRTSmoothRamp.SetColour(ndx: integer; colour: TColor);
begin
  Assert((ndx >= 0) and (ndx <= High(FData)));
  FData[ndx].Red := colour and $FF;
  FData[ndx].Green := (colour shr 8) and $FF;
  FData[ndx].Blue := (colour shr 16) and $FF;
end;

function TDSpRRTSmoothRamp.GetColour(ndx: integer): TColor;
begin
  Assert((ndx >= 0) and (ndx <= High(FData)));
  Result := (FData[ndx].Blue shl 16) + (FData[ndx].Green shl 8) + FData[ndx].Red;
end;

function TDSpRRTSmoothRamp.Insert(ndx: integer): integer;
var i, max: integer;
begin
  max := Length(FData);
  SetLength(FData, max+1);
  if ndx <= 0 then begin
    Result := 0;
    for i := max downto 1 do begin
      FData[i].Value := FData[i-1].Value;
      FData[i].Red := FData[i-1].Red;
      FData[i].Green := FData[i-1].Green;
      FData[i].Blue := FData[i-1].Blue;
    end;
    FData[0].Value := 2 * FData[1].Value - FData[2].Value;
  end else if ndx >= max then begin
    Result := max;
    FData[max].Value := 2 * FData[max-1].Value - FData[max-2].Value;
  end else begin
    Result := ndx;
    for i := max downto ndx + 1 do begin
      FData[i].Value := FData[i-1].Value;
      FData[i].Red := FData[i-1].Red;
      FData[i].Green := FData[i-1].Green;
      FData[i].Blue := FData[i-1].Blue;
    end;
    FData[ndx].Value := (FData[ndx-1].Value + FData[ndx+1].Value) * 0.5;
  end;
  FData[Result].Red := Random(255);
  FData[Result].Green := Random(255);
  FData[Result].Blue := Random(255);
  RecalcIntervals;
end;

function TDSpRRTSmoothRamp.Delete(ndx: integer): boolean;
var i: integer;
begin
  Result := (ndx >= 0) and (ndx <= High(FData)) and (Length(FData) > 2);
  if not Result then Exit;
  if ndx < High(FData) then
    for i := ndx to High(FData) - 1 do begin
      FData[i].Value := FData[i+1].Value;
      FData[i].Red := FData[i+1].Red;
      FData[i].Green := FData[i+1].Green;
      FData[i].Blue := FData[i+1].Blue;
    end;
  SetLength(FData, High(FData));
  RecalcIntervals;
end;

procedure TDSpRRTSmoothRamp.Invert;
var i, max: integer;
    temp: byte;
begin
  i := 0;
  max := High(FData);
  while i < max - i do begin
    temp := FData[i].Red;
    FData[i].Red := FData[max-i].Red;
    FData[max-i].Red := temp;
    temp := FData[i].Green;
    FData[i].Green := FData[max-i].Green;
    FData[max-i].Green := temp;
    temp := FData[i].Blue;
    FData[i].Blue := FData[max-i].Blue;
    FData[max-i].Blue := temp;
    Inc(i);
  end;
end;

procedure TDSpRRTSmoothRamp.TransForm(renderMap: TDSpRenderMap; area: TRect; bmp: TBitmap);
var val: PWord;
    r, c, i: integer;
    sl: PRGBTriple;
begin
  inherited;
  for r := 0 to bmp.Height - 1 do begin
    sl := bmp.ScanLine[r];
    val := PWord(integer(renderMap.Row[r+area.Top]) + area.Left * 2);
    for c := 0 to bmp.Width - 1 do begin
      if val^ > 0 then begin
        i := 0;
        while FData[i].Break < val^ do Inc(i);
        if val^ = FData[i].Break then begin
          sl^.rgbtRed   := FData[i].Red;
          sl^.rgbtGreen := FData[i].Green;
          sl^.rgbtBlue  := FData[i].Blue;
        end else begin
          sl^.rgbtRed   := FData[i-1].Red + (FData[i].Red - FData[i-1].Red) * (val^ - FData[i-1].Break) div FData[i].Diff;
          sl^.rgbtGreen := FData[i-1].Green + (FData[i].Green - FData[i-1].Green) * (val^ - FData[i-1].Break) div FData[i].Diff;
          sl^.rgbtBlue  := FData[i-1].Blue + (FData[i].Blue - FData[i-1].Blue) * (val^ - FData[i-1].Break) div FData[i].Diff;
        end;
      end;
      Inc(sl);
      Inc(val);
    end;
  end;
end;

procedure TDSpRRTSmoothRamp.Ramp(bmp: TBitMap; dir: TDSpDirection);
var run, r, c, i: integer;
    sl: PRGBTriple;
    val: word;
    red, green, blue: byte;
begin
  if dir = dspVert then begin
    run := bmp.Height - 1;
    i := High(FData);
    for r := 0 to run do begin
      sl := bmp.ScanLine[r];
      val := (run - r) * 65534 div run + 1;
      if val <= FData[i-1].Break then Dec(i);
      if val = FData[i].Break then begin
        red   := FData[i].Red;
        green := FData[i].Green;
        blue  := FData[i].Blue;
      end else begin
        red   := FData[i-1].Red + (FData[i].Red - FData[i-1].Red) * (val - FData[i-1].Break) div FData[i].Diff;
        green := FData[i-1].Green + (FData[i].Green - FData[i-1].Green) * (val - FData[i-1].Break) div FData[i].Diff;
        blue  := FData[i-1].Blue + (FData[i].Blue - FData[i-1].Blue) * (val - FData[i-1].Break) div FData[i].Diff;
      end;
      for c := 0 to bmp.Width - 1 do begin
        sl^.rgbtRed := red;
        sl^.rgbtGreen := green;
        sl^.rgbtBlue := blue;
        Inc(sl);
      end;
    end;
  end else if dir = dspHorz then begin
    // need some code here
  end;
end;

function TDSpRRTSmoothRamp.Transparent: TColor;
begin
  Result := $00000000;      // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
end;

end.
