{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRaster.pas. The Original Code is available through   }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit defines the user view of a raster and its bands. The TDSpRaster    }
{ class mostly serves as a conduit of data passing between a user process and  }
{ the PFA.                                                                     }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 18                                             }
{                                                                              }
{ Dependencies  : DSpDataSet                                                   }
{                 DSpGlobals                                                   }
{                 DSpPFA                                                       }
{                 DSpRasterSource                                              }
{                 DSpUtils (FloatsSameInv)                                     }
{                 SyncObjs (TEvent)                                            }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.3  03Dec18 Added CreateFromDataSource, AddBand, and        }
{                              DeleteBand. TDSpRaster methods now check for nil}
{                              raster bands. TDSpRasters can now be opened by  }
{                              name. TDSpRasterBand can be named.              }
{                 0.2  03Nov20 Updated message passing, raster bands don't send}
{                              messages anymore. Rendering moved to DSpRenderer}
{                              Added TDSpRasterBandImage.                      }
{                 0.1  03Nov07 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpRaster;
interface
{$I DSpatial.inc}
uses DSpGlobals, DSpPFA, DSpDataSet;
type
  TDSpRaster = class;

  TDSpRasterBand = class
  protected
    FRaster   : TDSpRaster;
    FBand     : integer;
    FName     : string;

    function    GetDataType: TDSpDataType;
    function    GetMinValue: TDSpFloat64;
    function    GetMaxValue: TDSpFloat64;
    function    GetMissingValue: TDSpFloat64;
  public
    constructor Create(raster: TDSpRaster; band: integer);
    destructor  Destroy; override;

    property    Raster: TDSpRaster read FRaster;
    property    BandIndex: integer read FBand;
    property    Name: string read FName write FName;
    property    DataType: TDSpDataType read GetDataType;
    property    Min: TDSpFloat64 read GetMinValue;
    property    Max: TDSpFloat64 read GetMaxValue;
    property    MissingValue: TDSpFloat64 read GetMissingValue;
  end;

  TDSpRasterBandOrdinal = class(TDSpRasterBand)
  public
    constructor Create(raster: TDSpRaster; band: integer);
  end;

  TDSpRasterBandSurface = class(TDSpRasterBand)
  public
    constructor Create(raster: TDSpRaster; band: integer);
  end;

  TDSpRasterBandImage = class(TDSpRasterBand)
  public
    constructor Create(raster: TDSpRaster; band: integer);
  end;

  TDSpRaster = class(TDSpDataSet)
  private
    procedure   DefineBands(src: TDSpSourceDataSet);
  protected
    FBands    : array of TDSpRasterBand;

    function    GetRows: integer;
    function    GetColumns: integer;
    function    GetPixelSize: TDSpFloat64;
    function    GetBand(b: integer): TDSpRasterBand;
    function    GetBands: integer;
    function    GetBandType(b: integer): TDSpRasterBandType;
    function    GetBandDataType(b: integer): TDSpDataType;
    function    GetBandMin(b: integer): TDSpFloat64;
    function    GetBandMax(b: integer): TDSpFloat64;
    function    GetBandMissingValue(b: integer): TDSpFloat64;
  public
    constructor Open(src: TDSpSourceHandler; ndx: integer); overload;
    constructor Open(src: TDSpSourceHandler; const name: string); overload;
    constructor CreateFromDataSource(src: TDSpSourceDataSet);
    destructor  Destroy; override;

    property    Rows: integer read GetRows;
    property    Columns: integer read GetColumns;
    property    PixelSize: TDSpFloat64 read GetPixelSize;

    property    Band[b: integer]: TDSpRasterBand read GetBand;
    property    Bands: integer read GetBands;
    function    AddBand(const name: string; bt: TDSpRasterBandType; dt: TDSpDataType; null: TDSpFloat64): TDSpRasterBand;
    procedure   DeleteBand(b: integer);
    property    BandType[b: integer]: TDSpRasterBandType read GetBandType;
    property    BandDataType[b: integer]: TDSpDataType read GetBandDataType;
  end;

implementation
uses SysUtils, SyncObjs, DSpUtils, DSpRasterSource;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterBand
constructor TDSpRasterBand.Create(raster: TDSpRaster; band: integer);
begin
  inherited Create;
  FRaster := raster;
  FBand := band;
end;

destructor TDSpRasterBand.Destroy;
begin
  inherited;
end;

function TDSpRasterBand.GetDataType: TDSpDataType;
begin
  Result := FRaster.GetBandDataType(FBand);
end;

function TDSpRasterBand.GetMinValue: TDSpFloat64;
begin
  Result := FRaster.GetBandMin(FBand);
end;

function TDSpRasterBand.GetMaxValue: TDSpFloat64;
begin
  Result := FRaster.GetBandMax(FBand);
end;

function TDSpRasterBand.GetMissingValue: TDSpFloat64;
begin
  Result := FRaster.GetBandMissingValue(FBand);
end;

////////////////////////////////////////////////////////////////////////////////
// Raster band types
constructor TDSpRasterBandOrdinal.Create(raster: TDSpRaster; band: integer);
begin
  inherited;
  FName := Format('Ordinal band (%d)', [band]);
end;

constructor TDSpRasterBandSurface.Create(raster: TDSpRaster; band: integer);
begin
  inherited;
  FName := Format('Surface band (%d)', [band]);
end;

constructor TDSpRasterBandImage.Create(raster: TDSpRaster; band: integer);
begin
  inherited;
  FName := Format('Image band (%d)', [band]);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRaster
constructor TDSpRaster.Open(src: TDSpSourceHandler; ndx: integer);
var source: TDSpRasterSource;
begin
  source := TDSpRasterSource(src.ReadLock(ndx));
  if source = nil then
    raise Exception.Create('Could not create raster');

  inherited Create(src, ndx);
  try
    DefineBands(source);
  finally
    source.ReadUnlock;
  end;
end;

constructor TDSpRaster.Open(src: TDSpSourceHandler; const name: string);
var source: TDSpRasterSource;
begin
  source := TDSpRasterSource(src.ReadLock(name));
  if source = nil then
    raise Exception.Create('Could not create raster');

  inherited Create(src, source.DataSetIndex);
  try
    DefineBands(source);
  finally
    source.ReadUnlock;
  end;
end;

constructor TDSpRaster.CreateFromDataSource(src: TDSpSourceDataSet);
begin
  inherited Create(src.Handler, src.DataSetIndex);
  DefineBands(src);
end;

destructor TDSpRaster.Destroy;
var i: integer;
begin
  if Length(FBands) > 0 then
    for i := 0 to High(FBands) do
      if FBands <> nil then FBands[i].Free;
  Finalize(FBands);
  inherited;
end;

// Define the bands in this raster from the source object. The src parameter
// must have dspRead access or there could be thread synchronization problems.
procedure TDSpRaster.DefineBands(src: TDSpSourceDataSet);
var i: integer;
begin
  with src as TDSpRasterSource do
    if Bands > 0 then begin
      SetLength(FBands, Bands);
      for i := 0 to Bands - 1 do begin
        case Band[i].BandType of
          dspRasterBandSurface: FBands[i] := TDSpRasterBandSurface.Create(self, i);
          dspRasterBandOrdinal: FBands[i] := TDSpRasterBandOrdinal.Create(self, i);
          dsprasterBandImage  : FBands[i] := TDSpRasterBandImage.Create(self, i);
        end;
      end;
    end;
end;

function TDSpRaster.GetBand(b: integer): TDSpRasterBand;
begin
  Assert(b in [0..High(FBands)]);
  Result := FBands[b];
end;

function TDSpRaster.GetBands: integer;
begin
  Result := Length(FBands);
end;

function TDSpRaster.GetRows: integer;
var raster: TDSpRasterSource;
begin
  raster := TDSpRasterSource(FHandler.ReadLock(FIndex));
  try
    Result := raster.Rows;
  finally
    raster.ReadUnlock;
  end;
end;

function TDSpRaster.GetColumns: integer;
var raster: TDSpRasterSource;
begin
  raster := TDSpRasterSource(FHandler.ReadLock(FIndex));
  try
    Result := raster.Columns;
  finally
    raster.ReadUnlock;
  end;
end;

function TDSpRaster.GetPixelSize: TDSpFloat64;
var raster: TDSpRasterSource;
begin
  raster := TDSpRasterSource(FHandler.ReadLock(FIndex));
  try
    Result := raster.PixelSize;
  finally
    raster.ReadUnlock;
  end;
end;

function TDSpRaster.GetBandType(b: integer): TDSpRasterBandType;
var raster: TDSpRasterSource;
begin
  if FBands[b] <> nil then begin
    raster := TDSpRasterSource(FHandler.ReadLock(FIndex));
    try
      Result := raster.Band[b].BandType;
    finally
      raster.ReadUnlock;
    end;
  end else Result := dspRasterBandNil;
end;

function TDSpRaster.GetBandDataType(b: integer): TDSpDataType;
var raster: TDSpRasterSource;
begin
  if FBands[b] <> nil then begin
    raster := TDSpRasterSource(FHandler.ReadLock(FIndex));
    try
      Result := raster.Band[b].DataType;
    finally
      raster.ReadUnlock;
    end;
  end else Result := dspNil;
end;

function TDSpRaster.GetBandMin(b: integer): TDSpFloat64;
var raster: TDSpRasterSource;
begin
  if FBands[b] <> nil then begin
    raster := TDSpRasterSource(FHandler.ReadLock(FIndex));
    try
      Result := raster.MinValue[b];
    finally
      raster.ReadUnlock;
    end;
  end else Result := -9E200;
end;

function TDSpRaster.GetBandMax(b: integer): TDSpFloat64;
var raster: TDSpRasterSource;
begin
  if FBands[b] <> nil then begin
    raster := TDSpRasterSource(FHandler.ReadLock(FIndex));
    try
      Result := raster.MaxValue[b];
    finally
      raster.ReadUnlock;
    end;
  end else Result := -9E200;
end;

function TDSpRaster.GetBandMissingValue(b: integer): TDSpFloat64;
var raster: TDSpRasterSource;
begin
  if FBands[b] <> nil then begin
    raster := TDSpRasterSource(FHandler.ReadLock(FIndex));
    try
      Result := raster.MissingValue[b];
    finally
      raster.ReadUnlock;
    end;
  end else Result := -9E200;
end;

function TDSpRaster.AddBand(const name: string; bt: TDSpRasterBandType; dt: TDSpDataType; null: TDSpFloat64): TDSpRasterBand;
var msg: TDSpThreadMessage;
    bandinfo: PDSpRasterBandInfo;
    ndx: integer;
begin
  // Collect band data
  New(bandinfo);
  bandinfo^.DataType := dt;
  bandinfo^.BandType := bt;
  bandinfo^.MissingValue := null;
  bandinfo^.Min := 0;
  bandinfo^.Max := -1;

  // Fire off the request
  New(msg);
  msg^.DataSourceIndex := FIndex;
  msg^.Event := TEvent.Create(nil, false, false, '');
  msg^.Msg := DSPMSG_RASTERADDBAND;
  msg^.Parameter := bandinfo;
  FHandler.QueueMessage(msg);

  // Wait for the request to complete. Add the band if successful, else return nil
  if (msg^.Event.WaitFor($FFFFFFFF) = wrSignaled) and (bandinfo^.BandIndex > -1) then begin
    ndx := Length(FBands);
    SetLength(FBands, ndx+1);
    case bt of
      dspRasterBandSurface: FBands[ndx] := TDSpRasterBandSurface.Create(self, ndx);
      dspRasterBandOrdinal: FBands[ndx] := TDSpRasterBandOrdinal.Create(self, ndx);
      dsprasterBandImage  : FBands[ndx] := TDSpRasterBandImage.Create(self, ndx);
    end;
    FBands[ndx].Name := name;
    Result := FBands[ndx];
  end else Result := nil;

  // Clean up
  msg^.Event.Free;
  Dispose(msg);
  Dispose(bandinfo);
end;

procedure TDSpRaster.DeleteBand(b: integer);
var msg: TDSpThreadMessage;
begin
  // Fire off the request
  New(msg);
  msg^.DataSourceIndex := FIndex;
  msg^.Event := TEvent.Create(nil, false, false, '');
  msg^.Msg := DSPMSG_RASTERDELETEBAND;
  msg^.Parameter := pointer(b);
  FHandler.QueueMessage(msg);

  // Wait for the request to complete. Delete the band if successful.
  if (msg^.Event.WaitFor($FFFFFFFF) = wrSignaled) and (integer(msg^.Parameter) > -1) then begin
    FBands[b].Free;
    FBands[b] := nil;
  end;

  // Clean up
  msg^.Event.Free;
  Dispose(msg);
end;

end.
