{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRasterSource.pas. The Original Code is available     }
{ through the home page of DSpatial at http://dspatial.sourceforge.net.        }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit defines the raster source class, and its associated bands. The     }
{ raster source class is referenced by the source handler when raster data sets}
{ are being accessed, where this class manages data access.                    }
{                                                                              }
{ This unit is mostly a repackaging of the old DSpRaster unit, with the new PFA}
{ architecture of data access.                                                 }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 29                                             }
{                                                                              }
{ Dependencies  : DSpGlobals                                                   }
{                 DSpDataMgr                                                   }
{                 DSpRRT (rendering)                                           }
{                 DSpPFA                                                       }
{                 DSpCoordinateSystem                                          }
{                 DSpRasterData                                                }
{                 DSpWin32/DSpLinux (Yield)                                    }
{                 DSpUtils (FloatsSameInv)                                     }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.3  03Dec29 Methods use access mode of the data set. Open   }
{                              and create constructor pair.                    }
{                 0.2  03Nov16 Added support for purging unused data. Added    }
{                              support for reading data into user memory.      }
{                 0.1  03Nov05 Initial coding. Repacking of old DSpRaster unit.}
{                                                                              }
{******************************************************************************}
unit DSpRasterSource;
interface
{$I DSpatial.inc}
uses DSpGlobals, DSpPFA, DSpRRT, DSpCoordinateSystem, DSpRasterData;
type
  TDSpRasterSource = class;

  TDSpSourceRasterBand = class
  protected
    FRaster   : TDSpRasterSource;
    FBandIndex: integer;
    FDataType : TDSpDataType;
    FBandType : TDSpRasterBandType;
    FData     : TDSpRasterData;

    procedure   PurgeData;
  public
    constructor Create(const raster: TDSpRasterSource; b: integer);
    destructor  Destroy; override;

    property    Raster: TDSpRasterSource read FRaster;
    property    Band: integer read FBandIndex;
    property    DataType: TDSpDataType read FDataType;
    property    BandType: TDSpRasterBandType read FBandType;
    procedure   SetMinMax(var min, max: TDSpFloat64);

    procedure   RenderMap(params: PDSpRenderRasterMap);
    procedure   ReadDataByCoords(params: PDSpMsgReadRasterBandCoords);
    procedure   WriteDataByCoords(params: PDSpMsgReadRasterBandCoords);
  end;

  TDSpRasterSource = class(TDSpSourceDataSet)
  protected
    FBands    : array of TDSpSourceRasterBand;
    FInfo     : TDSpRasterInfo;

    procedure   PurgeData; override;

    function    GetNumBands: integer;
    function    GetBand(b: integer): TDSpSourceRasterBand;
    function    GetRows: integer;
    function    GetColumns: integer;
    function    GetPixelSize: TDSpFloat64;
    function    GetMissingValue(b: integer): TDSpFloat64;
    function    GetMinValue(b: integer): TDSpFloat64;
    function    GetMaxValue(b: integer): TDSpFloat64;
  public
//    constructor Create(const driver: TDSpBaseDriver; dataSet: integer; const h: TDSpSourceHandler);
    constructor Open(const driver: TDSpBaseDriver; dataSet: integer; const h: TDSpSourceHandler);
    destructor  Destroy; override;
    procedure   Close; override;

    procedure   SetRasterStructure(rows, cols, bands: integer; pixel: TDSpFloat64);

    property    Bands: integer read GetNumBands;
    property    Band[b: integer]: TDSpSourceRasterBand read GetBand;
    procedure   SetBandStructure(band: integer; bandType: TDSpRasterBandType; dataType: TDSpDataType);
    function    AddBand(info: PDSpRasterBandInfo): TDSpSourceRasterBand;
    procedure   DeleteBand(b: integer);

    property    Rows: integer read GetRows;
    property    Columns: integer read GetColumns;
    property    PixelSize: TDSpFloat64 read GetPixelSize;
    property    MissingValue[b: integer]: TDSpFloat64 read GetMissingValue;
    property    MinValue[b: integer]: TDSpFloat64 read GetMinValue;
    property    MaxValue[b: integer]: TDSpFloat64 read GetMaxValue;


    procedure   RenderBand(params: PDSpRenderRasterMap);
    procedure   ReadBandByCoords(params: PDSpMsgReadRasterBandCoords);
    procedure   WriteBandByCoords(params: PDSpMsgReadRasterBandCoords);
  end;

implementation
uses SysUtils, {$IFDEF WIN32} DSpWin32, {$ELSE} DSpLinux, {$ENDIF} DSpUtils;
////////////////////////////////////////////////////////////////////////////////
// TDSpSourceRasterBand
constructor TDSpSourceRasterBand.Create(const raster: TDSpRasterSource; b: integer);
begin
  inherited Create;
  FRaster := raster;
  FBandIndex := b;
  with FRaster.FInfo.Band[FBandIndex] do begin
    FBandType := BandType;
    FDataType := DataType;
    FData := CreateStorage(FRaster, FBandIndex, FDataType);
  end;
end;

destructor TDSpSourceRasterBand.Destroy;
begin
  FData.Free;
  inherited;
end;

procedure TDSpSourceRasterBand.SetMinMax(var min, max: TDSpFloat64);
begin
  FData.MinMax(min, max);
end;

procedure TDSpSourceRasterBand.RenderMap(params: PDSpRenderRasterMap);
begin
  FData.Render(params);
end;

procedure TDSpSourceRasterBand.PurgeData;
begin
  FData.PurgeData;
end;

procedure TDSpSourceRasterBand.ReadDataByCoords(params: PDSpMsgReadRasterBandCoords);
begin
  FData.ReadDataByCoords(params);
end;

procedure TDSpSourceRasterBand.WriteDataByCoords(params: PDSpMsgReadRasterBandCoords);
begin
  FData.WriteDataByCoords(params);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterSource
//constructor TDSpRasterSource.Create(const driver: TDSpBaseDriver; dataSet: integer; const h: TDSpSourceHandler);
//begin
//  inherited;
//  FDataSetType := dspPFARaster;
//end;

constructor TDSpRasterSource.Open(const driver: TDSpBaseDriver; dataSet: integer; const h: TDSpSourceHandler);
var b: integer;
begin
  inherited;                           // access is dspNoAccess
  FDataSetType := dspPFARaster;
  FDriver.GetRasterInfo(FIndex, @FInfo);
  SetLength(FInfo.Band, FInfo.Bands);

  if FInfo.Bands > 0 then begin
    FDriver.GetRasterBandInfo(FIndex, @FInfo);
    SetLength(FBands, FInfo.Bands);
    for b := 0 to FInfo.Bands - 1 do
      FBands[b] := TDSpSourceRasterBand.Create(self, b);
  end;
end;

destructor TDSpRasterSource.Destroy;
var i: integer;
begin
  if Length(FBands) > 0 then
    for i := 0 to High(FBands) do FBands[i].Free;
  Finalize(FBands);
  inherited;
end;

procedure TDSpRasterSource.Close;
begin
  Destroy;
end;

procedure TDSpRasterSource.PurgeData;
var i: integer;
begin
  if FInfo.Bands > 0 then
    for i := 0 to High(FBands) do FBands[i].PurgeData;
end;

procedure TDSpRasterSource.SetRasterStructure(rows, cols, bands: integer; pixel: TDSpFloat64);
begin
  if FAccess = dspExcl then begin
    FInfo.Rows := rows;
    FInfo.Columns := cols;
    FInfo.Bands := bands;
    FInfo.PixelSize := pixel;
    SetLength(FInfo.Band, bands);
  end;
end;

procedure TDSpRasterSource.SetBandStructure(band: integer; bandType: TDSpRasterBandType; dataType: TDSpDataType);
begin
  Assert((band >= 0) and (band < Length(FInfo.Band)));
  if FAccess = dspExcl then begin
    with FInfo.Band[band] do begin
      BandType := bandType;
      DataType := DataType;
      MissingValue := 9E200;
      Min := 0;
      Max := -1;
    end;
    FBands[band] := TDSpSourceRasterBand.Create(self, band);
  end;
end;

function TDSpRasterSource.GetNumBands: integer;
begin
  Result := Length(FBands);
end;

function TDSpRasterSource.GetBand(b: integer): TDSpSourceRasterBand;
begin
  Assert((b >= 0) and (b < Length(FBands)));
  Result := FBands[b];
end;

function TDSpRasterSource.GetRows: integer;
begin
  Result := FInfo.Rows;
end;

function TDSpRasterSource.GetColumns: integer;
begin
  Result := FInfo.Columns;
end;

function TDSpRasterSource.GetPixelSize: TDSpFloat64;
begin
  Result := FInfo.PixelSize;
end;

function TDSpRasterSource.GetMissingValue(b: integer): TDSpFloat64;
begin
  Result := FInfo.Band[b].MissingValue;
end;

function TDSpRasterSource.GetMinValue(b: integer): TDSpFloat64;
var min, max: TDSpFloat64;
begin
  if FInfo.Band[b].Min > FInfo.Band[b].Max then begin
    FBands[b].SetMinMax(min, max);
    FInfo.Band[b].Min := min;
    FInfo.Band[b].Max := max;
  end;
  Result := FInfo.Band[b].Min;
end;

function TDSpRasterSource.GetMaxValue(b: integer): TDSpFloat64;
var min, max: TDSpFloat64;
begin
  if FInfo.Band[b].Min > FInfo.Band[b].Max then begin
    FBands[b].SetMinMax(min, max);
    FInfo.Band[b].Min := min;
    FInfo.Band[b].Max := max;
  end;
  Result := FInfo.Band[b].Max;
end;

procedure TDSpRasterSource.RenderBand(params: PDSpRenderRasterMap);
begin
  Assert(params^.Band in [0..High(FBands)]);
  if FAccess <> dspExcl then
    FBands[params^.Band].RenderMap(params);
end;

procedure TDSpRasterSource.ReadBandByCoords(params: PDSpMsgReadRasterBandCoords);
begin
  Assert(params^.Band in [0..High(FBands)]);
  if FAccess <> dspExcl then
    FBands[params^.Band].ReadDataByCoords(params);
end;

procedure TDSpRasterSource.WriteBandByCoords(params: PDSpMsgReadRasterBandCoords);
begin
  Assert(params^.Band in [0..High(FBands)]);
  repeat Yield until GetAccess(dspWrite);       // tricky!!!

  if FAccess = dspWrite then begin
    FBands[params^.Band].WriteDataByCoords(params);
    FInfo.Band[params^.Band].Min := 0;             // reset stats
    FInfo.Band[params^.Band].Max := -1;
  end;
  DropAccess(dspWrite);
end;

function TDSpRasterSource.AddBand(info: PDSpRasterBandInfo): TDSpSourceRasterBand;
var max: integer;
begin
  Result := nil;
  info^.BandIndex := -1; // flag failure

  // First try to gain exclusive access. Try hard, but give up eventually.
  for max := 1 to 5 do begin
    if GetAccess(dspExcl) then Break;
    Yield;
  end;
  if FAccess <> dspExcl then Exit;

  // Check if the driver will take the band. If so, add the band to FBands
  FDriver.AddRasterBand(FIndex, info);
  if info^.BandIndex > -1 then begin
    max := Length(FBands);
    SetLength(FBands, max+1);
    SetLength(FInfo.Band, max+1);
    FInfo.Band[max] := info^;
    Result := TDSpSourceRasterBand.Create(self, max);
    FBands[max] := Result;
  end;

  // Drop the access
  DropAccess(dspexcl);
end;

procedure TDSpRasterSource.DeleteBand(b: integer);
begin
end;

end.
