{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRenderer.pas. The Original Code is available through }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements the different kinds of renderers for data sets inside a }
{ DSpViewer. Depending on the characteristics of the data set a particular     }
{ renderer is used. Renderers can be organized into a DSpLegend component for  }
{ display next to a DSpViewer. The raster renderer makes use of a helper class }
{ which defines the actual transformation from raster band values to colours.  }
{ These so-called raster renderer transforms (RRT) are located in the unit     }
{ DSpRRT.                                                                      }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 12                                             }
{                                                                              }
{ Dependencies  : DSpRaster                                                    }
{                 DSpVector                                                    }
{                 DSpDataSet                                                   }
{                 DSpRRT                                                       }
{                 DSpViewer                                                    }
{                 DSpPFA (TDSpThreadMessage)                                   }
{                 DSpUtils (FloatsSame)                                        }
{                 DSpGlobals                                                   }
{                 Graphics (TBitmap)                                           }
{                 SysUtils (FreeAndNil)                                        }
{                 SyncObjs (TEvent)                                            }
{                 Classes (Rect function)                                      }
{                 Windows (TRect, IntersectRect)                               }
{                 Math (Ceil)                                                  }
{                                                                              }
{ Version       : 0.6  03Dec12 TDSpRasterRenderer3Channel class added.         }
{                              TDSpBandRendererData class added to handle      }
{                              raster bands for both the simple and the 3      }
{                              channel raster renderer.                        }
{                 0.5  03Nov14 Min/Max value of raster bands supported.        }
{                              Rendering of raster bands done locally.         }
{                 0.4  03Nov07 PFA overhaul                                    }
{                 0.3  03Sep25 Added SaveToImage                               }
{                 0.2  03Aug14 Rendering engine overhauled, many changes.      }
{                 0.1  03Jul18 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpRenderer;
{$I DSpatial.inc}
interface
uses Windows, Graphics, SyncObjs, DSpGlobals, DSpPFA, DSpDataSet, DSpRaster,
     DSpVector, DSpRRT;
type
  TDSpRenderer = class
  protected
    FLabel    : string;
    FViewer   : TObject;       // = TDSpViewer
    FTotalArea: TDSpExtent;    // total extent of bands in renderer
    FAOI      : TDSpExtent;    // area-of-interest
    FViewArea : TRect;         // area of the renderer in pixel space

    function    GetExtent: TDSpExtent;
    procedure   SetExtent(ext: TDSpExtent);
  public
    constructor Create(const viewer: TObject; lab: string);
    destructor  Destroy; override;

    function    CloseDataSet(const ds: TDSpDataSet): boolean; virtual; abstract;
    procedure   DissociateFromViewer;
    procedure   ResetArea; virtual; abstract;
    procedure   SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64); virtual;
    procedure   Paint; virtual; abstract;
    function    SaveImage(const fn: string): TBitmap; virtual; abstract;
    property    DisplayArea: TDSpExtent read GetExtent write SetExtent;
    property    Name: string read FLabel write FLabel;
  end;

  TDSpRasterRenderer = class(TDSpRenderer)
  protected
    FBitmap   : TBitmap;

    procedure   DoRender(const clip: TRect); virtual; abstract;
  public
    constructor Create(const viewer: TObject; lab: string);
    destructor  Destroy; override;

    procedure   Paint; override;
  end;

  TDSpBandRenderer = record
                       Band  : TDSpRasterBand;
                       Msg   : TDSpThreadMessage;
                       Status: TDSpState;
                     end;

  TDSpBandRendererData = class
  protected
    FBands    : array of TDSpBandRenderer;
    FMin, FMax: TDSpFloat64;
    FMap      : TDSpRenderMap;
    FDepth    : integer;

    procedure   AddRasterBand(band: TDSpRasterBand);
    procedure   DeleteRasterBand(band: TDSpRasterBand);

    procedure   SetMapArea(view: TDSpExtent; area, renderArea: TRect; pixel: TDSpFloat64);
    procedure   RenderBest(renderArea: TDSpExtent);

    procedure   SetMin(val: TDSpFloat64);
    function    GetMin: TDSpFloat64;
    procedure   SetMax(val: TDSpFloat64);
    function    GetMax: TDSpFloat64;
    function    GetNumBands: integer;
    function    GetBand(b: integer): TDSpRasterBand;

    procedure   WaitFor(ndx: integer);
    function    GetMap: TDSpRenderMap;
    property    Map: TDSpRenderMap read GetMap;
  public
    constructor Create(bitDepth: integer);
    destructor  Destroy; override;

    procedure   Range(min, max: TDSpFloat64);
    property    Min: TDSpFloat64 read GetMin write SetMin;
    property    Max: TDSpFloat64 read GetMax write SetMax;
    property    Bands: integer read GetNumBands;
    property    Band[b: integer]: TDSpRasterBand read GetBand;
  end;

  TDSpRasterRendererSimple = class(TDSpRasterRenderer)
  // A single raster band determines the entire RGB-triplet
  protected
    FData       : TDSpBandRendererData;
    FTrans      : TDSpRRTSingleBand;

    procedure   SetTrans(t: TDSpRRTSingleBand);
    function    GetTrans: TDSpRRTSingleBand;
    procedure   SetMin(val: TDSpFloat64);
    function    GetMin: TDSpFloat64;
    procedure   SetMax(val: TDSpFloat64);
    function    GetMax: TDSpFloat64;
    function    GetNumBands: integer;
    function    GetBand(b: integer): TDSpRasterBand;
  public
    constructor Create(const viewer: TObject; lab: string);
    destructor  Destroy; override;

    function    CloseDataSet(const ds: TDSpDataSet): boolean; override;
    procedure   AddRasterBand(band: TDSpRasterBand);
    procedure   DeleteRasterBand(band: TDSpRasterBand);
    procedure   DoRender(const clip: TRect); override;
    function    SaveImage(const fn: string): TBitmap; override;
    procedure   SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64); override;
    procedure   ResetArea; override;
    procedure   Range(min, max: TDSpFloat64);

    property    Transformation: TDSpRRTSingleBand read GetTrans write SetTrans;
    property    Min: TDSpFloat64 read GetMin write SetMin;
    property    Max: TDSpFloat64 read GetMax write SetMax;
    property    Bands: integer read GetNumBands;
    property    Band[b: integer]: TDSpRasterBand read GetBand;
  end;

  TDSpRasterRenderer3Channel = class(TDSpRasterRenderer)
  // The RGB-triplet is composed of bands in each of the three channels 0..2.
  // Channel 3 can be used for image fusion
  protected
    FChannel   : array[0..3] of TDSpBandRendererData;

    procedure   SetMin(channel: integer; val: TDSpFloat64);
    function    GetMin(channel: integer): TDSpFloat64;
    procedure   SetMax(channel: integer; val: TDSpFloat64);
    function    GetMax(channel: integer): TDSpFloat64;
    function    GetChannel(channel: integer): TDSpBandRendererData;
    function    GetNumBands: integer;
    function    GetBand(channel: integer; b: integer): TDSpRasterBand;

    function    Sharpen(r, g, b, pan: byte): TRGBTriple;
  public
    constructor Create(const viewer: TObject; lab: string);
    destructor  Destroy; override;

    function    CloseDataSet(const ds: TDSpDataSet): boolean; override;
    procedure   AddRasterBand(channel: integer; band: TDSpRasterBand);
    procedure   DeleteRasterBand(channel: integer; band: TDSpRasterBand);
    procedure   DoRender(const clip: TRect); override;
    function    SaveImage(const fn: string): TBitmap; override;
    procedure   SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64); override;
    procedure   ResetArea; override;
    procedure   Range(channel: integer; min, max: TDSpFloat64);

    property    Min[channel: integer]: TDSpFloat64 read GetMin write SetMin;
    property    Max[channel: integer]: TDSpFloat64 read GetMax write SetMax;
    property    Bands: integer read GetNumBands;
    property    Band[channel: integer; b: integer]: TDSpRasterBand read GetBand;
    property    Channel[channel: integer]: TDSpBandRendererData read GetChannel;
  end;

  TDSpFeatureRenderer = class(TDSpRenderer)
  protected
    FDataSet : TDSpVector;

  public
    constructor Create(const viewer: TObject; lab: string);

    procedure   Paint; override;
    function    SaveImage(const fn: string): TBitmap; override;
  end;

implementation
uses Math, Classes, SysUtils, DSpViewer, DSpUtils;
////////////////////////////////////////////////////////////////////////////////
// Utility routines
function ViewAreaToRendererArea(view, renderer: TRect): TRect;
begin
  Result.Top    := view.Top - renderer.Top;
  Result.Bottom := view.Bottom - renderer.Top;
  Result.Left   := view.Left - renderer.Left;
  Result.Right  := view.Right - renderer.Left;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRenderer
constructor TDSpRenderer.Create(const viewer: TObject; lab: string);
begin
  Assert(viewer is TDSpViewer);
  FViewer := viewer;
  FLabel := lab;
end;

destructor TDSpRenderer.Destroy;
begin
  if Assigned(FTotalArea) then FTotalArea.Free;
  if Assigned(FAOI) then FAOI.Free;
  inherited;
end;

procedure TDSpRenderer.DissociateFromViewer;
begin
  (FViewer as TDSpViewer).DeleteRenderer(self);
end;

procedure TDSpRenderer.SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64);
var ext: TDSpExtent;
begin
  if not Assigned(FTotalArea) then Exit;
  if Assigned(FAOI) then ext := FAOI else ext := FTotalArea;
  FViewArea := ext.PixelExtent(view, area, pixel);
end;

function TDSpRenderer.GetExtent: TDSpExtent;
begin
  if Assigned(FAOI) then Result := FAOI else Result := FTotalArea;
end;

procedure TDSpRenderer.SetExtent(ext: TDSpExtent);
begin
  if Assigned(FAOI) then FAOI.Free;
  FAOI := ext.Clone;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterRenderer
constructor TDSpRasterRenderer.Create(const viewer: TObject; lab: string);
begin
  inherited;
  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat := pf24bit;
  FBitmap.Transparent := true;
end;

destructor TDSpRasterRenderer.Destroy;
begin
  FBitmap.Free;
  inherited;
end;

procedure TDSpRasterRenderer.Paint;
var clip: TRect;
begin
  if not Assigned(FTotalArea) then Exit;
  if not IntersectRect(clip, (FViewer as TDSpViewer).Canvas.ClipRect, FViewArea) then Exit;

  // Create the bitmap to fill the "dirty" area
  FBitmap.Width := clip.Right - clip.Left + 1;
  FBitmap.Height := clip.Bottom - clip.Top + 1;

  // Dispatch to the appropriate sub-class
  DoRender(clip);
  (FViewer as TDSpViewer).Canvas.Draw(clip.Left, clip.Top, FBitmap);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpBandRendererData
constructor TDSpBandRendererData.Create(bitDepth: integer);
begin
  FDepth := bitDepth;
  FMap := TDSpRenderMap.Create(0, 0, FDepth);
  FMin := 0;
  if FDepth = 8 then FMax := 255 else FMax := 1000;
end;

destructor TDSpBandRendererData.Destroy;
var i: integer;
begin
  if Length(FBands) > 0 then
    for i := 0 to High(FBands) do begin
      Dispose(PDSpRenderRasterMap(FBands[i].Msg^.Parameter));
      FBands[i].Msg^.Event.Free;
      Dispose(FBands[i].Msg);
    end;
  Finalize(FBands);
  FMap.Free;
  inherited;
end;

procedure TDSpBandRendererData.AddRasterBand(band: TDSpRasterBand);
var min, max: TDSpFloat64;
    i: integer;
begin
  i := Length(FBands);
  SetLength(FBands, i+1);
  FBands[i].Band := band;
  New(FBands[i].Msg);
  FBands[i].Msg^.DataSourceIndex := band.Raster.SourceIndex;
  FBands[i].Msg^.Event := TEvent.Create(nil, false, false, '');
  FBands[i].Msg^.Msg := DSPMSG_RENDERRASTERMAP;;
  New(PDSpRenderRasterMap(FBands[i].Msg^.Parameter));
  PDSpRenderRasterMap(FBands[i].Msg^.Parameter)^.Band := band.BandIndex;
  PDSpRenderRasterMap(FBands[i].Msg^.Parameter)^.Map := FMap;
  FBands[i].Status := dspIdle;

  min := band.Min;
  max := band.Max;
  if Length(FBands) = 1 then begin
    if min < max then begin
      FMin := min;
      FMax := max;
    end;
  end else begin
    if min < max then begin
      if min < FMin then FMin := min;
      if max > FMax then FMax := max;
    end;
  end;
end;

procedure TDSpBandRendererData.DeleteRasterBand(band: TDSpRasterBand);
var i, max: integer;
begin
  if Length(FBands) = 0 then Exit;
  max := High(FBands);
  for i := 0 to max do
    if FBands[i].Band = band then begin
      Dispose(PDSpRenderRasterMap(FBands[i].Msg^.Parameter));
      FBands[i].Msg^.Event.Free;
      Dispose(FBands[i].Msg);
      if i < max then FBands[i] := FBands[max];
      SetLength(FBands, max);
      Break;
    end;
end;

function TDSpBandRendererData.GetNumBands: integer;
begin
  Result := Length(FBands);
end;

function TDSpBandRendererData.GetBand(b: integer): TDSpRasterBand;
begin
  Assert((b >= 0) and (b < Length(FBands)));
  Result := FBands[b].Band;
end;

procedure TDSpBandRendererData.SetMapArea(view: TDSpExtent; area, renderArea: TRect; pixel: TDSpFloat64);
var i: integer;
    overlap : TDSpExtent;
begin
  if renderArea.Left > renderArea.Right then FMap.Resize(0, 0)
  else FMap.Resize(renderArea.Bottom - renderArea.Top + 1, renderArea.Right - renderArea.Left + 1);

  for i := 0 to High(FBands) do begin
    overlap := FBands[i].Band.Raster.Extent.Intersection(view);
    if overlap <> nil then begin
      PDSpRenderRasterMap(FBands[i].Msg^.Parameter)^.MapArea := overlap.PixelExtent(view, area, pixel);
      with PDSpRenderRasterMap(FBands[i].Msg^.Parameter)^ do begin
        if MapArea.Left > MapArea.Right then Continue; // band not in view
        MapArea := ViewAreaToRendererArea(MapArea, renderArea);
      end;
      PDSpRenderRasterMap(FBands[i].Msg^.Parameter)^.Area := view;
      PDSpRenderRasterMap(FBands[i].Msg^.Parameter)^.Pixel := pixel;
      PDSpRenderRasterMap(FBands[i].Msg^.Parameter)^.Min := FMin;
      PDSpRenderRasterMap(FBands[i].Msg^.Parameter)^.Max := FMax;
      overlap.Free;
      FBands[i].Status := dspPending;
      FBands[i].Band.Raster.Handler.QueueMessage(FBands[i].Msg);
    end;
  end;
end;

procedure TDSpBandRendererData.RenderBest(renderArea: TDSpExtent);
var r, c, b: integer;
    pxl: TDSpFloat64;
    a: TRect;
begin
  if Length(FBands) = 1 then begin
    r := FBands[0].Band.Raster.Rows;
    c := FBands[0].Band.Raster.Columns;
    FMap.Resize(r, c);
    PDSpRenderRasterMap(FBands[0].Msg^.Parameter)^.Min := FMin;
    PDSpRenderRasterMap(FBands[0].Msg^.Parameter)^.Max := FMax;
    PDSpRenderRasterMap(FBands[0].Msg^.Parameter)^.Pixel := -1; // flag full resolution
    FBands[0].Status := dspPending;
    FBands[0].Band.Raster.Handler.QueueMessage(FBands[0].Msg);
  end else begin
    // Establish pixel size
    pxl := 0.0;
    for b := 0 to High(FBands) do
      if pxl < FBands[b].Band.Raster.PixelSize then pxl := FBands[b].Band.Raster.PixelSize;

    // Establish r & c, based on renderArea
    c := Ceil((renderArea.East - renderArea.West) / pxl);
    r := Ceil((renderArea.North - renderArea.South) / pxl);
    FMap.Resize(r, c);

    // Instruct the bands to fetch the data
    a := Rect(0, 0, c-1, r-1);
    for b := 0 to High(FBands) do with PDSpRenderRasterMap(FBands[b].Msg^.Parameter)^ do begin
      MapArea := FBands[b].Band.Raster.Extent.PixelExtent(renderArea, a, pxl);
      Area := renderArea;
      Pixel := pxl;
      Min := FMin;
      Max := FMax;
      FBands[b].Status := dspPending;
      FBands[b].Band.Raster.Handler.QueueMessage(FBands[b].Msg);
    end;
  end;
end;

function TDSpBandRendererData.GetMap: TDSpRenderMap;
var i: integer;
begin
  // Make sure all data is loaded, then return the map
  for i := 0 to High(FBands) do WaitFor(i);
  Result := FMap;
end;

procedure TDSpBandRendererData.WaitFor(ndx: integer);
begin
  if FBands[ndx].Status = dspPending then
    if FBands[ndx].Msg^.Event.WaitFor($FFFFFFFF) = wrSignaled then
      FBands[ndx].Status := dspIdle
    else Assert(false, 'Wait event in TDSpBandRendererData unexpectedly failed');
end;

procedure TDSpBandRendererData.Range(min, max: TDSpFloat64);
begin
  FMin := min;
  FMax := max;
end;

procedure TDSpBandRendererData.SetMin(val: TDSpFloat64);
begin
  FMin := val;
end;

function TDSpBandRendererData.GetMin: TDSpFloat64;
begin
  Result := FMin;
end;

procedure TDSpBandRendererData.SetMax(val: TDSpFloat64);
begin
  FMax := val;
end;

function TDSpBandRendererData.GetMax: TDSpFloat64;
begin
  Result := FMax;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterRendererSimple
constructor TDSpRasterRendererSimple.Create(const viewer: TObject; lab: string);
begin
  inherited;
  FTrans := TDSpRRTHotToCold.Create;
  FData := TDSpBandRendererData.Create(16);
end;

destructor TDSpRasterRendererSimple.Destroy;
begin
  FTrans.Free;
  FData.Free;
  inherited;
end;

procedure TDSpRasterRendererSimple.AddRasterBand(band: TDSpRasterBand);
begin
  FData.AddRasterBand(band);

  if FData.Bands = 1 then FTotalArea := band.Raster.Extent.Clone
  else FTotalArea.UnionWith(band.Raster.Extent);
  (FViewer as TDSpViewer).MergeExtent(FTotalArea);
end;

function TDSpRasterRendererSimple.CloseDataSet(const ds: TDSpDataSet): boolean;
var i, num: integer;
begin
  Result := false;
  if (not (ds is TDSpRaster)) or (TDSpRaster(ds).Bands = 0) then Exit;

  num := GetNumBands;
  for i := 0 to TDSpRaster(ds).Bands - 1 do
    FData.DeleteRasterBand(TDSpRaster(ds).Band[i]);
  Result := GetNumBands < num;
end;

procedure TDSpRasterRendererSimple.DeleteRasterBand(band: TDSpRasterBand);
begin
  FData.DeleteRasterBand(band);
  ResetArea;
end;

procedure TDSpRasterRendererSimple.SetTrans(t: TDSpRRTSingleBand);
begin
  FTrans.Free;
  FTrans := t;
end;

function TDSpRasterRendererSimple.GetTrans: TDSpRRTSingleBand;
begin
  Result := FTrans;
end;

function TDSpRasterRendererSimple.GetNumBands: integer;
begin
  Result := FData.Bands;
end;

function TDSpRasterRendererSimple.GetBand(b: integer): TDSpRasterBand;
begin
  Result := FData.Band[b];
end;

procedure TDSpRasterRendererSimple.SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64);
begin
  inherited; // Set FViewArea
  FData.SetMapArea(view, area, FViewArea, pixel);
end;

procedure TDSpRasterRendererSimple.DoRender(const clip: TRect);
var area: TRect;
begin
  if FData.Bands = 0 then Exit;

  // Prep the bitmap
  FBitmap.TransparentColor := FTrans.Transparent;
  FBitMap.Canvas.Brush.Color := FBitmap.TransparentColor;
  FBitMap.Canvas.FillRect(Rect(0, 0, FBitMap.Width, FBitMap.Height));

  // Set the dirty area in FData.Map coordinates and make the transformation
  area := ViewAreaToRendererArea(clip, FViewArea);
  FTrans.Transform(FData.Map, area, FBitmap);
end;

function TDSpRasterRendererSimple.SaveImage(const fn: string): TBitmap;
var r, c: integer;
begin
  Result := TBitmap.Create;
  Result.PixelFormat := pf24Bit;
  if FData.Bands = 0 then Exit;

  FData.RenderBest(FTotalArea);
  r := FData.Map.Rows;
  c := FData.Map.Columns;

  // Render the map to the result bitmap
  Result.TransparentColor := FTrans.Transparent;
  Result.Canvas.Brush.Color := FBitmap.TransparentColor;
  Result.Width := c;
  Result.Height := r;
  Result.Canvas.FillRect(Rect(0, 0, c-1, r-1));
  FTrans.Transform(FData.Map, Rect(0, 0, c-1, r-1), Result);

  // Restore the viewer geometry
  TDSpViewer(FViewer).SetRendererMap;
end;

procedure TDSpRasterRendererSimple.ResetArea;
var i: integer;
begin
  if FData.Bands = 0 then Exit;
  FTotalArea.CopyFrom(FData.Band[0].Raster.Extent);
  if FData.Bands > 1 then
    for i := 1 to FData.Bands - 1 do
      FTotalArea.UnionWith(FData.Band[i].Raster.Extent);
  TDSpViewer(FViewer).ResetExtent;
end;

procedure TDSpRasterRendererSimple.Range(min, max: TDSpFloat64);
begin
  FData.Range(min, max);
end;

procedure TDSpRasterRendererSimple.SetMin(val: TDSpFloat64);
begin
  FData.Min := val;
end;

function TDSpRasterRendererSimple.GetMin: TDSpFloat64;
begin
  Result := FData.Min;
end;

procedure TDSpRasterRendererSimple.SetMax(val: TDSpFloat64);
begin
  FData.Max := val;
end;

function TDSpRasterRendererSimple.GetMax: TDSpFloat64;
begin
  Result := FData.Max;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterRenderer3Channel
constructor TDSpRasterRenderer3Channel.Create(const viewer: TObject; lab: string);
begin
  inherited;
  FChannel[0] := TDSpBandRendererData.Create(8);
  FChannel[1] := TDSpBandRendererData.Create(8);
  FChannel[2] := TDSpBandRendererData.Create(8);
  FChannel[3] := TDSpBandRendererData.Create(8);
end;

destructor TDSpRasterRenderer3Channel.Destroy;
begin
  FChannel[0].Free;
  FChannel[1].Free;
  FChannel[2].Free;
  FChannel[3].Free;              
  inherited;
end;

procedure TDSpRasterRenderer3Channel.AddRasterBand(channel: integer; band: TDSpRasterBand);
begin
  FChannel[channel].AddRasterBand(band);

  if GetNumBands = 1 then FTotalArea := band.Raster.Extent.Clone
  else FTotalArea.UnionWith(band.Raster.Extent);
  (FViewer as TDSpViewer).MergeExtent(FTotalArea);
end;

function TDSpRasterRenderer3Channel.CloseDataSet(const ds: TDSpDataSet): boolean;
var i, c, num: integer;
begin
  Result := false;
  if (not (ds is TDSpRaster)) or (TDSpRaster(ds).Bands = 0) then Exit;

  num := GetNumBands;
  for i := 0 to TDSpRaster(ds).Bands - 1 do
    for c := 0 to 3 do FChannel[c].DeleteRasterBand(TDSpRaster(ds).Band[i]);
  Result := GetNumBands < num;
end;

procedure TDSpRasterRenderer3Channel.DeleteRasterBand(channel: integer; band: TDSpRasterBand);
begin
  FChannel[channel].DeleteRasterBand(band);
  ResetArea;
end;

function TDSpRasterRenderer3Channel.GetChannel(channel: integer): TDSpBandRendererData;
begin
  Result := FChannel[channel];
end;

function TDSpRasterRenderer3Channel.GetNumBands: integer;
begin
  Result := FChannel[0].Bands + FChannel[1].Bands + FChannel[2].Bands + FChannel[3].Bands;
end;

function TDSpRasterRenderer3Channel.GetBand(channel: integer; b: integer): TDSpRasterBand;
begin
  Result := FChannel[channel].Band[b];
end;

procedure TDSpRasterRenderer3Channel.SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64);
var c: integer;
begin
  inherited; // Set FViewArea
  for c := 0 to 3 do FChannel[c].SetMapArea(view, area, FViewArea, pixel);
end;

procedure TDSpRasterRenderer3Channel.DoRender(const clip: TRect);
var area: TRect;
    r, c: integer;
    sl: PRGBTriple;
    red, green, blue, pan: ^byte;
begin
  if GetNumBands = 0 then Exit;

  // Prep the bitmap
  FBitmap.TransparentColor := $02000000;
  FBitMap.Canvas.Brush.Color := FBitmap.TransparentColor;
  FBitMap.Canvas.FillRect(Rect(0, 0, FBitMap.Width, FBitMap.Height));

  // Set the dirty area in map coordinates and blend the RGB colour maps. If
  // channel 3 has bands assigned to it, do the pan-sharpening.
  area := ViewAreaToRendererArea(clip, FViewArea);
  if FChannel[3].Bands = 0 then
    for r := 0 to FBitMap.Height - 1 do begin
      sl := FBitmap.ScanLine[r];
      red := pointer(integer(FChannel[0].Map.Row[r+area.Top]) + area.Left);
      green := pointer(integer(FChannel[1].Map.Row[r+area.Top]) + area.Left);
      blue := pointer(integer(FChannel[2].Map.Row[r+area.Top]) + area.Left);
      for c := 0 to FBitmap.Width - 1 do begin
        sl^.rgbtRed := red^;
        sl^.rgbtGreen := green^;
        sl^.rgbtBlue := blue^;
        Inc(sl);
        Inc(red);
        Inc(green);
        Inc(blue);
      end;
    end
  else
    // Use pan-sharpening
    for r := 0 to FBitMap.Height - 1 do begin
      sl := FBitmap.ScanLine[r];
      red := pointer(integer(FChannel[0].Map.Row[r+area.Top]) + area.Left);
      green := pointer(integer(FChannel[1].Map.Row[r+area.Top]) + area.Left);
      blue := pointer(integer(FChannel[2].Map.Row[r+area.Top]) + area.Left);
      pan := pointer(integer(FChannel[3].Map.Row[r+area.Top]) + area.Left);
      for c := 0 to FBitmap.Width - 1 do begin
        sl^ := Sharpen(red^, green^, blue^, pan^);
        Inc(sl);
        Inc(red);
        Inc(green);
        Inc(blue);
        Inc(pan);
      end;
    end
end;

function TDSpRasterRenderer3Channel.SaveImage(const fn: string): TBitmap;
var r, c: integer;
begin
  Result := TBitmap.Create;
  Result.PixelFormat := pf24Bit;
  if GetNumBands = 0 then Exit;

//  FData.RenderBest(FTotalArea);
  r := FChannel[0].Map.Rows;
  c := FChannel[0].Map.Columns;

  // Render the map to the bitmap
  Result.TransparentColor := $02000000;
  Result.Canvas.Brush.Color := Result.TransparentColor;
  Result.Width := c;
  Result.Height := r;
  Result.Canvas.FillRect(Rect(0, 0, c-1, r-1));

  // Restore the viewer geometry
  TDSpViewer(FViewer).SetRendererMap;
end;

procedure TDSpRasterRenderer3Channel.ResetArea;
var i, c: integer;
    ext: TDSpExtent;
begin
  if GetNumBands = 0 then Exit;

  FreeAndNil(FTotalArea);
  for c := 0 to 3 do
    if FChannel[c].Bands > 0 then
      for i := 0 to FChannel[c].Bands - 1 do begin
        ext := FChannel[c].Band[i].Raster.Extent;
        if Assigned(FTotalArea) then FTotalArea.UnionWith(ext)
        else FTotalArea := ext.Clone;
      end;

  TDSpViewer(FViewer).ResetExtent;
end;

procedure TDSpRasterRenderer3Channel.Range(channel: integer; min, max: TDSpFloat64);
begin
  FChannel[channel].Range(min, max);
end;

procedure TDSpRasterRenderer3Channel.SetMin(channel: integer; val: TDSpFloat64);
begin
  FChannel[channel].Min := val;
end;

function TDSpRasterRenderer3Channel.GetMin(channel: integer): TDSpFloat64;
begin
  Result := FChannel[channel].Min;
end;

procedure TDSpRasterRenderer3Channel.SetMax(channel: integer; val: TDSpFloat64);
begin
  FChannel[channel].Max := val;
end;

function TDSpRasterRenderer3Channel.GetMax(channel: integer): TDSpFloat64;
begin
  Result := FChannel[channel].Max;
end;

function TDSpRasterRenderer3Channel.Sharpen(r, g, b, pan: byte): TRGBTriple;
var m1, m2, s: byte;
    h, sum, diff: smallint;

  function HueToRGB(hue: smallint): byte;
  begin
    if hue < 0 then Inc(hue, 255)
    else if hue > 255 then Dec(hue, 255);

    if hue < 43 then Result := m1 + (diff * hue + 85) div 43
    else if hue < 128 then Result := m2
    else if hue < 171 then Result := m1 + (diff * (171 - hue) + 85) div 43
    else Result := m1;
  end;

begin
  // Calculate intermediate values
  m1 := r;
  m2 := r;
  if g > m1 then m1 := g else m2 := g;
  if b > m1 then m1 := b
  else if b < m2 then m2 := b;
  sum := m1 + m2;
  diff := m1 - m2;

  // Compute H and S
  if m1 = m2 then begin           // r=g=b: achromatic
    s := 0;                       // saturation
    h := 0;                       // hue, undefined
  end else begin
    if sum < 255 then
      s := (diff * 255 + sum div 2) div sum
    else s := (diff * 255 + (510 - sum) div 2) div (510 - sum);

    if r = m1 then h := ((g - b) * 43) div diff
    else if g = m1 then h := 85 + ((b - r) * 43) div diff
    else h := 171 + ((r - g) * 43) div diff;
  end;

  // Now convert HLS back to RGB, using pan for L
  if s = 0 then begin            // achromatic
    Result.rgbtRed   := pan;
    Result.rgbtGreen := pan;
    Result.rgbtBlue  := pan;
  end else begin                 //chromatic
    // set up magic numbers
    if pan < 128 then m2 := (pan * (255 + s) + 128) div 255
    else m2 := pan + s - (pan * s + 128) div 255;
    m1 := 2 * pan - m2;

    Result.rgbtRed   := HueToRGB(h + 85);// * 255 + 127) div 255;
    Result.rgbtGreen := HueToRGB(h);//      * 255 + 127) div 255;
    Result.rgbtBlue  := HueToRGB(h - 85);// * 255 + 127) div 255;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Feature renderer
constructor TDSpFeatureRenderer.Create(const viewer: TObject; lab: string);
begin
  Assert(viewer is TDSpVector);
  FViewer := viewer;
  FLabel := lab;
end;

procedure TDSpFeatureRenderer.Paint;
begin
end;

function TDSpFeatureRenderer.SaveImage(const fn: string): TBitmap;
begin
  Result := TBitmap.Create;
  Result.PixelFormat := pf24Bit;
end;

end.
