{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpUtils.pas. The Original Code is available through    }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit contains general purpose conversion routines and testing for       }
{ special values, as well as synchronization utilities.                        }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 05                                             }
{                                                                              }
{ Dependencies  : DSpGlobals                                                   }
{                 DSpWin32 (TMutex, Yield)                                     }
{                 SysUtils (ExtractFileExt, Exception)                         }
{                                                                              }
{ Version       : 0.7  03Dec05 Added DSpTempFileName                           }
{                 0.6  03Nov22 Moved SaveAsImage to DSpGraphics.               }
{                 0.5  03Nov05 TFileMapping & TFileView removed,               }
{                              TDSpQueue virtualised, TDSpThreadQueue added.   }
{                 0.4  03Sep24 Added SaveAsImage                               }
{                 0.3  03Jul18 Added TDSpQueue, FloatsSameInv                  }
{                 0.2  03May09 Added support for mutexes and file mapping.     }
{                 0.1  03Apr17 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpUtils;
interface
uses DSpWin32;

type
  // TDSpQueue
  PQueueItem = ^TQueueItem;
  TQueueItem = record
    Item: pointer;
    Next: PQueueItem;
  end;

  TDSpQueue = class
  protected
    FHead, FTail: PQueueItem;
    FLength     : integer;
  public
    constructor Create;
    destructor  Destroy; override;

    procedure   Put(x: pointer); virtual;
    function    Get: pointer; virtual;
    function    Peek: pointer; virtual;
    procedure   HeadToTail; virtual;
    function    Empty: boolean; virtual;
    property    Length: integer read FLength;
  end;

  TDSpThreadQueue = class(TDSpQueue)
  private
    FMutex    : TDSpMutex;
  public
    constructor Create;
    destructor  Destroy; override;

    procedure   Put(x: pointer); override;
    function    Get: pointer; override;
    function    Peek: pointer; override;
    procedure   HeadToTail; override;
  end;

  function  DSpTempFileName: string;

  procedure ValidateIndex(const low, high, ndx: integer);
  function  FloatsSame(const a, b: double): boolean;
  function  FloatsSameInv(const a, binv: double): boolean;

  function IntMax(a, b: integer) : integer;
  function IntMin(a, b: integer) : integer;
  function CardMax(a, b: cardinal) : cardinal;
  function CardMin(a, b: cardinal) : cardinal;

  function NaNextended : extended;
  function NaNdouble : double;
  function NaNsingle : single;

  function IsNan(v : extended) : boolean; overload;
  function IsNaN(v : double) : boolean; overload;
  function IsNaN(v : single) : boolean; overload;

  function EndianSwap(n : integer) : integer; overload;
  function EndianSwap(n : cardinal) : cardinal; overload;
  function EndianSwap(n : word)    : word; overload;

  procedure Swap2Bytes(const x; n : integer); register;
  procedure Swap4Bytes(const x; n : integer); register;
  procedure SwapBytes(var x : cardinal); overload;
  procedure SwapBytes(var x : word); overload;
  procedure SwapBytes(var x : smallint); overload;

implementation
uses SysUtils, DSpGlobals;

////////////////////////////////////////////////////////////////////////////////
// TDSpQueue
constructor TDSpQueue.Create;
begin
  inherited;
  FHead := nil;
end;

destructor TDSpQueue.Destroy;
var i : PQueueItem;
begin
  while FHead <> nil do begin   // the items themselves are not deleted
    i := FHead;
    FHead := FHead^.Next;
    Dispose(i);
  end;
end;

procedure TDSpQueue.Put(x: pointer);
var i : PQueueItem;
begin
  New(i);
  i^.Next := nil;
  i^.Item := x;
  if FTail = nil then FHead := i
  else FTail^.Next := i;
  FTail := i;
  Inc(FLength);
end;

function TDSpQueue.Get: pointer;
var i : PQueueItem;
begin
  if FHead = nil then Result := nil
  else begin
    Result := FHead^.Item;
    i := FHead;
    FHead := FHead^.Next;
    if FTail = i then FTail := nil;
    Dispose(i);
    Dec(FLength);
  end;
end;

function TDSpQueue.Peek: pointer;
begin
  if FHead = nil then Result := nil
  else Result := FHead^.Item;
end;

procedure TDSpQueue.HeadToTail;
var i: PQueueItem;
begin
  if FLength < 2 then Exit;
  i := FHead;
  FHead := FHead^.Next;
  i^.Next := nil;
  FTail^.Next := i;
  FTail := i;
end;

function TDSpQueue.Empty : boolean;
begin
  Result := FLength = 0;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpThreadQueue
constructor TDSpThreadQueue.Create;
begin
  FMutex := TDSpMutex.Create(3000);
  inherited;
end;

destructor TDSpThreadQueue.Destroy;
begin
  FMutex.Get(MaxInt); // make sure the mutex is available, wait forever if needs be
  inherited;
  FMutex.Release;
  FMutex.Free;
end;

procedure TDSpThreadQueue.Put(x: pointer);
begin
  FMutex.Get(MaxInt); // make sure the mutex is available, wait forever if needs be
  inherited;
  FMutex.Release;
end;

function TDSpThreadQueue.Get: pointer;
begin
  if not FMutex.Get then Result := nil
  else begin
    Result := inherited Get;
    FMutex.Release;
  end;
end;

function TDSpThreadQueue.Peek: pointer;
begin
  if not FMutex.Get then Result := nil
  else begin
    Result := inherited Peek;
    FMutex.Release;
  end;
end;

procedure TDSpThreadQueue.HeadToTail;
begin
  FMutex.Get(MaxInt); // make sure the mutex is available, wait forever if needs be
  inherited;
  FMutex.Release;
end;

////////////////////////////////////////////////////////////////////////////////
// utility routines
function DSpTempFileName: string;
var dt: double;
    i: Int64;
begin
  repeat
    dt := Now;
    Move(dt, i, 8);
    Result := Format('%s\Tmp%s%d.dsp', [DSpatial.TempDir, IntToHex(i, 8), Random(1000)]);
  until not FileExists(Result);
end;

procedure ValidateIndex(const low, high, ndx: integer);
begin
  if not ndx in [low..high] then
    raise Exception.Create('Invalid index');
end;

function FloatsSame(const a, b: double): boolean;
begin
  if b = 0.0 then Result := a = 0.0
  else Result := Abs(a/b - 1) <= 0.000001;
end;

function FloatsSameInv(const a, binv: double): boolean;
begin
  Result := Abs((a * binv) - 1) <= 0.000001;
end;

function EndianSwap(n : integer) : integer;
// Swaps 32bit integer value in big endian to little endian and vice versa
asm
  BSWAP EAX
end;

function EndianSwap(n : cardinal) : cardinal;
// Swaps 32bit cardinal value in big endian to little endian and vice versa
asm
  BSWAP EAX
end;

function EndianSwap(n : word) : word;
// Swaps 16bit integer value in big endian to little endian and vice versa
asm
  XCHG  AL, AH
end;

// x contains the address to n things of 2 bytes long
procedure Swap2Bytes(const x; n : integer); register;
asm
@NotDone:
  MOV   CX,[EAX]
  XCHG  CH,CL
  MOV   [EAX],CX
  ADD   EAX,2
  DEC   EDX
  JNZ   @NotDone
end;

// x contains the address to n things of 4 bytes long
procedure Swap4Bytes(const x; n : integer); register;
asm
@NotDone:
  MOV   ECX,[EAX]
  BSWAP ECX
  MOV   [EAX],ECX
  ADD   EAX,4
  DEC   EDX
  JNZ   @NotDone
end;

procedure SwapBytes(var x : cardinal);
asm
  BSWAP EAX
end;

procedure SwapBytes(var x : word);
asm
  XCHG  AL, AH
end;

procedure SwapBytes(var x : smallint);
asm
  XCHG  AL, AH
end;

function IntMax(a, b: integer) : integer;
begin
  if a > b then IntMax := a else IntMax := b;
end;

function IntMin(a, b: integer) : integer;
begin
  if a < b then IntMin := a else IntMin := b;
end;

function CardMax(a, b: cardinal) : cardinal;
begin
  if a > b then CardMax := a else CardMax := b;
end;

function CardMin(a, b: cardinal) : cardinal;
begin
  if a < b then CardMin := a else CardMin := b;
end;

function NaNextended : extended;
const b : array [1..10] of byte = ($FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF);
begin
  Move(b, Result, 10);
end;

function NaNdouble : double;
const b : array [1..8] of byte = ($7F, $FF, $FF, $FF, $FF, $FF, $FF, $FF);
begin
  Move(b, Result, 8);
end;

function NaNsingle : single;
const b : array [1..4] of byte = ($7F, $FF, $FF, $FF);
begin
  Move(b, Result, 4);
end;

function IsNan(v : extended) : boolean;
var b : array [1..10] of byte;
    i : integer;
begin
  Move(v, b, 10);
  Result := ((b[1] shl 1) = $FE);
  i := 2;
  while Result and (i <= 10) do begin
    Result := (b[i] = $FF);
    Inc(i);
  end;
end;

function IsNaN(v : double) : boolean;
var i : comp;
begin
  Move(v, i, 8);
  Result := ((i = $7FFFFFFFFFFFFFFF) or (i = $FFFFFFFFFFFFFFFF));
    // No bit-wise operations on comp type possible
end;

function IsNaN(v : single) : boolean;
var i : cardinal;
begin
  Move(v, i, 4);
  Result := ((i shl 1) = $FFFFFFFE); // Discard sign
end;

end.
