{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpViewer.pas. The Original Code is available through   }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements the 2D viewer of spatial data for DSpatial. The viewer  }
{ controls a set of data sets through its FDataMgr member. Several helper      }
{ classes can attach to the viewer for specific functionality (e.g. for        }
{ displaying legends, tool bars, etc.). The viewer is the major interactive    }
{ access into DSpatial.                                                        }
{ The viewer is implemented as a control, descending from TCustomControl. The  }
{ viewer will be installed on the DSpatial component palette page.             }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 20                                             }
{                                                                              }
{ Dependencies  : DSpRenderer                                                  }
{                 DSpDataMgr                                                   }
{                 DSpUtils (FloatsSame)                                        }
{                 DSpGlobals                                                   }
{                 Controls (TCustomControl)                                    }
{                 Forms (TScreen)                                              }
{                 Classes (TControl, TList)                                    }
{                 Graphics (TPen, TColor)                                      }
{                 Windows (TPoint)                                             }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.3  03Nov20 Support for panning added. Zoom and pan modes   }
{                              are now persistent. Cursor changes during Paint }
{                              and SetRendererMap.                             }
{                 0.2  03Nov05 SetRendererMap in response to size and scale    }
{                              changes.                                        }
{                 0.1  03Jul18 Initial coding. Does not yet register on a      }
{                              palette in the IDE.                             }
{                                                                              }
{******************************************************************************}
unit DSpViewer;
{$I DSpatial.inc}
interface
uses Windows, SysUtils, Classes, Controls, Graphics, DSpGlobals, DSpRenderer;
type
  TDSpViewAction = (dspViewPoint, dspViewPan, dspViewZoom, dspViewSelect);

  TDSpViewer = class(TCustomControl)
  private
    FAffineFactor: double;             // pixel space to world space
    FCurrExtent  : TDSpExtent;         // extent of current display
    FOffsetX,
    FOffsetY     : double;             // world-coordinate of lower-left corner
    FScale       : double;             // the denominator of the scale: 1:FScale
    FActionState : TDSpViewAction;     // context of current operation
    FTracking    : boolean;            // flag to indicate mouse tracking mode
    FOrigin,                           // starting point of mouse action
    FCurrent     : TPoint;             // current point of mouse action

    FOnScaleChange : TParameterEvent;
  protected
    FRenderers   : TList;              // list of renderers
    FTotalExtent : TDSpExtent;         // union of extents of all data sets

    procedure   Paint; override;
    procedure   Resize; override;

    procedure   SetScale(const s: double);
    function    GetProjectedCoordX(x: integer): double;
    function    GetProjectedCoordY(y: integer): double;

  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

    property    Renderers: TList read FRenderers;
    procedure   AddRenderer(rend: TDSpRenderer);
    procedure   DeleteRenderer(rend: TDSpRenderer);
    procedure   SwapRenderers(r1, r2: integer);
    function    NumRenderers: integer;
    procedure   SetRendererMap;

    procedure   SetViewExtent(const north, east, south, west: double); overload;
    procedure   SetViewExtent(const ext: TDSpExtent); overload;
    procedure   MergeExtent(const ex: TDSpExtent);
    procedure   ResetExtent;
    procedure   StartZoom;
    procedure   StartPan;
    procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure   ZoomToAll;

    property    CurrentExtent: TDSpExtent read FCurrExtent;
    property    OriginX: double read FOffsetX;
    property    OriginY: double read FOffsetY;
    property    AffineScalar: double read FAffineFactor;
    property    ProjectedX[x: integer]: double read GetProjectedCoordX;
    property    ProjectedY[y: integer]: double read GetProjectedCoordY;

    property    Canvas;        // expose inherited property
  published
    property    Align;         // expose inherited properties
    property    Color;
    property    OnMouseMove;
    property    OnClick;
    property    OnDblClick;
    property    Scale: double read FScale write SetScale;

    property    OnScaleChange: TParameterEvent read FOnScaleChange write FOnScaleChange;
  end;

  procedure Register;

implementation
uses Forms, DSpUtils;

const
  dspCursorZoomIn = 21;

constructor TDSpViewer.Create(AOwner: TComponent);
begin
  inherited;
  FScale := 1.0;
  FAffineFactor := 0.0254 / Screen.PixelsPerInch;
  Color := clBlack;
  FRenderers := TList.Create;
  Screen.Cursors[dspCursorZoomIn] := LoadCursor(HInstance, 'ZOOMIN');
end;

destructor TDSpViewer.Destroy;
var i: integer;
begin
  if Assigned(FCurrExtent) then FCurrExtent.Free;
  if Assigned(FTotalExtent) then FTotalExtent.Free;
  if FRenderers.Count > 0 then
    for i := 0 to FRenderers.Count - 1 do
      TDSpRenderer(FRenderers[i]).Free;
  FRenderers.Free;
  inherited;
end;

procedure TDSpViewer.Paint;
var i: integer;
    oldCursor: TCursor;
begin
  if FRenderers.Count = 0 then Exit;

  oldCursor := Cursor;
  Cursor := crHourGlass;

  // Get the renderers going, in rverse order
  for i := FRenderers.Count - 1 downto 0 do begin
    TDSpRenderer(FRenderers[i]).Paint;
//    Application.ProcessMessages;
  end;

  Cursor := oldCursor;
end;

procedure TDSpViewer.Resize;
begin
  inherited;
  if Assigned(FCurrExtent) then
    with FCurrExtent do
      SetViewExtent(North, East, South, West);
end;

procedure TDSpViewer.AddRenderer(rend: TDSpRenderer);
var ext: TDSpExtent;
begin
  FRenderers.Add(rend);
  ext := rend.DisplayArea;
  if ext = nil then Exit;
  
  if not Assigned(FTotalExtent) then FTotalExtent := ext.Clone
  else FTotalExtent.UnionWith(ext);
  if not Assigned(FCurrExtent) then begin
    FCurrExtent := FTotalExtent.Clone;
    SetViewExtent(FCurrExtent.North, FCurrExtent.East, FCurrExtent.South, FCurrExtent.West);
  end;
end;
      
procedure TDSpViewer.DeleteRenderer(rend: TDSpRenderer);
begin
  FRenderers.Remove(rend);
  ResetExtent;
end;

procedure TDSpViewer.SwapRenderers(r1, r2: integer);
begin
  FRenderers.Exchange(r1, r2);
end;

function TDSpViewer.NumRenderers: integer;
begin
  Result := FRenderers.Count;
end;

procedure TDSpViewer.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  FCurrent := Point(X, Y);
  if FActionState in [dspViewZoom, dspViewPan] then begin
    FTracking := true;
    Canvas.Pen.Color := clRed;
    FOrigin := Point(X, Y);
    Canvas.MoveTo(X, Y);
  end;
end;

procedure TDSpViewer.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var n, e, s, w, dx, dy: double;
begin
  inherited;
  if FTracking then begin
    Canvas.Pen.Mode := pmNot;
    Canvas.MoveTo(FOrigin.X, FOrigin.Y);
    case FActionState of
      dspViewZoom: begin
        Canvas.LineTo(FOrigin.X, FCurrent.Y);
        Canvas.LineTo(FCurrent.X, FCurrent.Y);
        Canvas.LineTo(FCurrent.X, FOrigin.Y);
        Canvas.LineTo(FOrigin.X, FOrigin.Y);
      end;
      dspViewPan : Canvas.LineTo(FCurrent.X, FCurrent.Y);
    end;
    Canvas.Pen.Mode := pmCopy;
    FTracking := false;
  end;
  FCurrent := Point(X, Y);
  case FActionState of
    dspViewZoom:
      if (FCurrent.Y <> FOrigin.Y) and (FCurrent.X <> FOrigin.X) then begin
        if FCurrent.Y > FOrigin.Y then begin
          n := GetProjectedCoordY(FCurrent.Y);
          s := GetProjectedCoordY(FOrigin.Y);
        end else begin
          s := GetProjectedCoordY(FCurrent.Y);
          n := GetProjectedCoordY(FOrigin.Y);
        end;
        if FCurrent.X > FOrigin.X then begin
          e := GetProjectedCoordX(FCurrent.X);
          w := GetProjectedCoordX(FOrigin.X);
        end else begin
          w := GetProjectedCoordX(FCurrent.X);
          e := GetProjectedCoordX(FOrigin.X);
        end;
        SetViewExtent(n, e, s, w);
      end;
    dspViewPan:
      if (FCurrent.Y <> FOrigin.Y) or (FCurrent.X <> FOrigin.X) then begin
        dx := GetProjectedCoordX(FCurrent.X) - GetProjectedCoordX(FOrigin.X);
        dy := GetProjectedCoordY(FCurrent.Y) - GetProjectedCoordY(FOrigin.Y);
        n := FCurrExtent.North - dy;
        s := FCurrExtent.South - dy;
        e := FCurrExtent.East - dx;
        w := FCurrExtent.West - dx;
        SetViewExtent(n, e, s, w);
      end;
  end;
//  FActionState := dspViewPoint;
//  Cursor := crDefault;
end;

procedure TDSpViewer.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if FTracking then begin
    Canvas.Pen.Mode := pmNot;
    Canvas.MoveTo(FOrigin.X, FOrigin.Y);
    case FActionState of
      dspViewZoom: begin
        Canvas.LineTo(FOrigin.X, FCurrent.Y);
        Canvas.LineTo(FCurrent.X, FCurrent.Y);
        Canvas.LineTo(FCurrent.X, FOrigin.Y);
        Canvas.LineTo(FOrigin.X, FOrigin.Y);
        Canvas.LineTo(FOrigin.X, Y);
        Canvas.LineTo(X, Y);
        Canvas.LineTo(X, FOrigin.Y);
        Canvas.LineTo(FOrigin.X, FOrigin.Y);
      end;
      dspViewPan : begin
        Canvas.LineTo(FCurrent.X, FCurrent.Y);
        Canvas.MoveTo(FOrigin.X, FOrigin.Y);
        Canvas.LineTo(X, Y);
      end;
    end;
    Canvas.Pen.Mode := pmCopy;
  end;
  FCurrent := Point(X, Y);
end;

procedure TDSpViewer.ZoomToAll;
begin
  if not Assigned(FTotalExtent) then Exit;
  ResetExtent;
  with FTotalExtent do
    SetViewExtent(North, East, South, West);
end;

procedure TDSpViewer.SetViewExtent(const north, east, south, west: double);
var c, d: double;
begin
  if not Assigned(FCurrExtent) then FCurrExtent := TDSpExtent.Create(0, 0, 0, 0);
  
  if Width * (north - south) > Height * (east - west) then begin
    // Adjust east/west
    FCurrExtent.North := north;
    FCurrExtent.South := south;
    c := west + (east - west) * 0.5;
    d := 0.5 * (north - south) * Width / Height;
    FCurrExtent.West := c - d;
    FCurrExtent.East := c + d;
  end else begin
    // Adjust north/south
    FCurrExtent.West := west;
    FCurrExtent.East := east;
    c := south + (north - south) * 0.5;
    d := 0.5 * (east - west) * Height / Width;
    FCurrExtent.South := c - d;
    FCurrExtent.North := c + d;
  end;

  // Calculate scale and set parameters
  FScale := Screen.PixelsPerInch * (FCurrExtent.East - FCurrExtent.West) / (Width * 0.0254);
  FAffineFactor := 0.0254 * FScale / Screen.PixelsPerInch;
  FOffsetX := FCurrExtent.West;
  FOffsetY := FCurrExtent.South;
  SetRendererMap;
  if Assigned(OnScaleChange) then OnScaleChange(self, FScale);
  Invalidate;
end;

procedure TDSpViewer.SetViewExtent(const ext: TDSpExtent);
begin
  SetViewExtent(ext.North, ext.East, ext.South, ext.West);
end;

procedure TDSpViewer.MergeExtent(const ex: TDSpExtent);
begin
  if Assigned(FTotalExtent) then FTotalExtent.UnionWith(ex)
  else FTotalExtent := ex.Clone;
//  if FCurrExtent = nil then FCurrExtent := FTotalExtent.Clone;
end;

procedure TDSpViewer.StartZoom;
begin
  if Assigned(FCurrExtent) then begin
    FActionState := dspViewZoom;
    Cursor := dspCursorZoomIn;
  end;
end;

procedure TDSpViewer.StartPan;
begin
  if Assigned(FCurrExtent) then begin
    FActionState := dspViewPan;
    Cursor := crHandPoint;
  end;
end;

procedure TDSpViewer.SetScale(const s: double);
var c, d: double;    // c = center; d = distance from center to edge
begin
  if (s <= 0.0) or FloatsSame(s, FScale) then Exit;
  if Assigned(FCurrExtent) then begin
    c := FCurrExtent.West + (FCurrExtent.East - FCurrExtent.West) * 0.5;
    d := (c - FCurrExtent.West) * s / FScale;
    FCurrExtent.West := c - d;
    FCurrExtent.East := c + d;
    c := FCurrExtent.South + (FCurrExtent.North - FCurrExtent.South) * 0.5;
    d := (c - FCurrExtent.South) * s / FScale;
    FCurrExtent.South := c - d;
    FCurrExtent.North := c + d;
    FScale := s;
    FAffineFactor := 0.0254 * FScale / Screen.PixelsPerInch;
    FOffsetX := FCurrExtent.West;
    FOffsetY := FCurrExtent.South;
    SetRendererMap;
  end else FScale := s;
  if Assigned(OnScaleChange) then OnScaleChange(self, FScale);
  Invalidate;
end;

procedure TDSpViewer.SetRendererMap;
var i: integer;
    oldCursor: TCursor;
begin
  if FRenderers.Count > 0 then begin

    oldCursor := Cursor;
    Cursor := crHourGlass;

    for i := 0 to FRenderers.Count - 1 do
      TDSpRenderer(FRenderers[i]).SetMapArea(FCurrExtent, ClientRect, FAffineFactor);

    Cursor := oldCursor;
  end;
  Invalidate;
end;

function TDSpViewer.GetProjectedCoordX(x: integer): double;
begin
  if FScale <= 0.0 then Result := 0.0
  else Result := x * FAffineFactor + FOffsetX;   
end;

function TDSpViewer.GetProjectedCoordY(y: integer): double;
begin
  if FScale <= 0.0 then Result := 0.0
  else Result := (Height - y) * FAffineFactor + FOffsetY;
end;

procedure TDSpViewer.ResetExtent;
var i: integer;
begin
  if FRenderers.Count = 0 then Exit;
  FTotalExtent.CopyFrom(TDSpRenderer(FRenderers[0]).DisplayArea);
  if FRenderers.Count = 1 then Exit;
  for i := 1 to FRenderers.Count - 1 do
    FTotalExtent.UnionWith(TDSpRenderer(FRenderers[i]).DisplayArea);
end;

procedure Register;
begin
  RegisterComponents('DSpatial', [TDSpViewer]);
end;

end.
